CokeScanner : Object ( characterTable characterEscapes file char tokenType token matchingToken )

CokeScanner withFile: aFile
[
    self := self new.
    file := aFile.
    self nextChar.
    self nextToken.
]

CokeScanner initialize
[
    super initialize.
    tokenType := nil.
    token := WriteStream on: (String new: 32).
    matchingToken := nil.
    characterTable := Array new: 256.
    characterTable atAllPut: #xIllegal.
    33 to: 126   do: [:i | characterTable at: 1 + i       put: #xLetter].
    ' \t'        do: [:c | characterTable at: 1 + c value put: #xSpace].
    '\n'         do: [:c | characterTable at: 1 + c value put: #xLineFeed].
    '\r'         do: [:c | characterTable at: 1 + c value put: #xCarriageReturn].
    ';'          do: [:c | characterTable at: 1 + c value put: #xComment].
    '"'          do: [:c | characterTable at: 1 + c value put: #xDoubleQuote].
    ''''         do: [:c | characterTable at: 1 + c value put: #xSingleQuote].
    '0123456789' do: [:c | characterTable at: 1 + c value put: #xDigit].
    '('          do: [:c | characterTable at: 1 + c value put: #xParenthesis].
    '['          do: [:c | characterTable at: 1 + c value put: #xBracket].
    '{'          do: [:c | characterTable at: 1 + c value put: #xBrace].
    ')]}'        do: [:c | characterTable at: 1 + c value put: #xClose].
    (characterEscapes := IdentityDictionary new)
	at: $a  put: $\a; "bel"
	at: $b  put: $\b; "bs"
	at: $e  put: $\e; "esc"
	at: $f  put: $\f; "ff"
	at: $n  put: $\n; "nl"
	at: $r  put: $\r; "cr"
	at: $t  put: $\t; "ht"
	at: $v  put: $\v. "vt"
]

CokeScanner charType	[ ^characterTable at: 1 + char value ]
CokeScanner nextChar	[ ^char := file next ]

CokeScanner tokenType	[ ^tokenType ]
CokeScanner token	[ ^token contents ]

CokeScanner nextToken
[
    token resetToStart.
    tokenType := nil.
    [tokenType isNil]
	whileTrue:
	   [char isNil
		ifTrue:  [tokenType := #eof]
		ifFalse: [self perform: (characterTable at: 1 + char value)]].
    ^tokenType
]

CokeScanner xIllegal
[
    self error: ('illegal character: ', char printString)
]

CokeScanner xSpace
[
    self nextChar
]

CokeScanner xLineFeed
[
    [self nextChar notNil and: [char == $\r]] whileTrue.
]

CokeScanner xCarriageReturn
[
    [self nextChar notNil and: [char == $\n]] whileTrue.
]

CokeScanner xLetter
[
    token nextPut: char.
    self nextChar.
    [self charType == #xLetter]
	whileTrue:
	   [token nextPut: char.
	    self nextChar].
    ^tokenType := #identifier
]

CokeScanner xDigit
[
    token nextPut: char.
    self nextChar.
    [self charType == #xDigit]
	whileTrue:
	   [token nextPut: char.
	    self nextChar].
    ^tokenType := #integer
]

CokeScanner xParenthesis
[
    token nextPut: char.
    matchingToken := ')'.
    self nextChar.
    ^tokenType := #open
]

CokeScanner xBracket
[
    token nextPut: char.
    matchingToken := ']'.
    self nextChar.
    ^tokenType := #open
]

CokeScanner xBrace
[
    token nextPut: char.
    matchingToken := '}'.
    self nextChar.
    ^tokenType := #open
]

CokeScanner xClose
[
    token nextPut: char.
    self nextChar.
    ^tokenType := #close
]

CokeScanner xSingleQuote
[
    self nextChar.
    [char notNil and: [char ~~ $']]
	whileTrue:
	   [token nextPut: char.
	    self nextChar].
    char == $' ifFalse: [self error: 'EOF in literal'].
    self nextChar.
    ^tokenType := #singleQuote
]

CokeScanner xDoubleQuote
[
    self nextChar.
    [char notNil and: [char ~~ $"]] whileTrue: [self scanCharacter].
    char isNil ifTrue: [self error: 'EOF in literal'].
    self nextChar.
    ^tokenType := #doubleQuote
]

CokeScanner scanCharacter
[
    | d |
    d := char.
    self nextChar.
    d == $\\ ifTrue: [self scanEscape] ifFalse: [token nextPut: d]
]

CokeScanner scanEscape
[
    | d |
    char isNil ifTrue: [self error: 'EOF while scanning escaped character'].
    char == $\n ifTrue: [^self nextChar].	"ignore escaped EOL"
    d := characterEscapes at: char ifAbsent: [(char between: $0 and: $7) ifTrue: [self scanOctal] ifFalse: [char]].
    self nextChar.
    token nextPut: d
]

CokeScanner scanOctal
[
    | d |
    d := char digitValue.
    2 timesRepeat:
	[self nextChar.
	 char isNil ifTrue: [self error: 'EOF while scanning octal character escape'].
	 (char between: $0 and: $7) ifFalse: [self error: 'illegal digit in octal character escape'].
	 d := d * 8 + char digitValue].
    ^Character value: d
]
