{ import: st80 }

{ input: File }
{ input: CType }
{ input: CNode }
{ input: CokeScanner }
{ input: CokeParser }


File print: anObject
[
    self write: anObject printString
]


Stack : OrderedCollection ()

Stack push: anObject	[ ^self add: anObject ]
Stack pop		[ ^self removeLast ]


CTransformer : Object ()

CTransformer eval: expr		[ ^expr eval: self ]

CNode    eval: eval	[ ^self subclassResponsibility: #eval: ]
CInteger eval: eval	[ ^eval __integer: self ]
CString  eval: eval	[ ^eval __string: self ]
CSymbol  eval: eval	[ ^eval __symbol: self ]
CList    eval: eval	[ ^eval __list: self ]
CList    evalArgs: eval	[ ^value collect: [:arg | arg eval: eval] startingAt: 2 ]

CTransformer __integer: cInteger	[ ^self subclassResponsibility: #__integer: ]
CTransformer __string: cString		[ ^self subclassResponsibility: #__string:  ]
CTransformer __symbol: cSymbol		[ ^self subclassResponsibility: #__symbol:  ]
CTransformer __syntax: anObject		[ ^nil ]
CTransformer __operator: anObject	[ ^nil ]

CTransformer __list: cForm
[
    | head selector |
    cForm value isEmpty ifTrue: [^cForm].
    (head := cForm value first) isCSymbol
	ifTrue:
	   [(selector := self __syntax:   head) notNil ifTrue: [^self perform: selector with:  cForm].
	    (selector := self __operator: head) notNil ifTrue: [^self perform: selector with: (cForm evalArgs: self)]].
    ^self __apply: cForm
]

CTransformer __apply: cForm
[
    | form |
    form := cForm value.
    ^self __apply: (form first eval: self)
	       to: (form collect: [:arg | arg eval: self] startingAt: 2)
]

CTransformer __apply: cObject to: args	[ ^self subclassResponsibility: #__apply:to: ]

CTransformer __add: cForm	[ ^self subclassResponsibility: #__add: ]
CTransformer __sub: cForm	[ ^self subclassResponsibility: #__sub: ]
CTransformer __mul: cForm	[ ^self subclassResponsibility: #__mul: ]
CTransformer __div: cForm	[ ^self subclassResponsibility: #__div: ]
CTransformer __mod: cForm	[ ^self subclassResponsibility: #__mod: ]


CEnvironment : IdentityDictionary ( parent )

CEnvironment withParent: cEnvironment
[
    self := self new.
    parent := cEnvironment.
]

CEnvironment at: key ifAbsent: aBlock
[
    ^super at: key
	   ifAbsent: [parent notNil
			  ifTrue:  [parent at: key ifAbsent: aBlock]
			  ifFalse: [aBlock value]]
]

CEnvironment push	[ ^self withParent: self ]
CEnvironment pop	[ ^parent ]


CEvaluator : CTransformer ( syntax operators environment )

CEvaluator initialize
[
    super initialize.
    (syntax := IdentityDictionary new)
	at: #if		put: #__if:;
	at: #let	put: #__let:;
	at: #define	put: #__define:;
	at: #lambda	put: #__lambda:.
    (operators := IdentityDictionary new)
	at: #+		put: #__add:;
	at: #-		put: #__sub:;
	at: #*		put: #__mul:;
	at: #/		put: #__div:;
	at: #%		put: #__mod:;
	at: #<		put: #__lt:;
	at: #<=		put: #__le:;
	at: #=		put: #__eq:;
	at: #!=		put: #__ne:;
	at: #>=		put: #__ge:;
	at: #>		put: #__gt:;
	at: #println	put: #__println:.
    environment := CEnvironment new.
]

CEvaluator __syntax: cObject	[ ^cObject isCSymbol ifTrue: [syntax    at: cObject value ifAbsent: []] ]
CEvaluator __operator: cObject	[ ^cObject isCSymbol ifTrue: [operators at: cObject value ifAbsent: []] ]

CEvaluator __integer: cInteger	[ ^cInteger value ]
CEvaluator __string: cString	[ ^cString  value ]

CEvaluator __symbol: cSymbol
[
    ^environment at: cSymbol value ifAbsent: [self error: cSymbol value, ' is undefined']
]

CEvaluator __apply: cObject to: arguments
[
    | form formals result |
    cObject isCLambda ifFalse: [^self errorCannotApply].
    form := cObject value.
    self pushEnvironment.
    formals := form second value.
    formals size == arguments size ifFalse: [^self errorArgumentCount].
    formals with: arguments do: [:cSymbol :value | environment at: cSymbol value put: value].
    3 to: form size do: [:index | result := (form at: index) eval: self].
    self popEnvironment.
    ^result
]

CEvaluator __add: args		[ ^args injectFirstInto: [:sum  :arg | sum  + arg] ]
CEvaluator __sub: args		[ ^args injectFirstInto: [:diff :arg | diff - arg] ]
CEvaluator __mul: args		[ ^args injectFirstInto: [:prod :arg | prod * arg] ]
CEvaluator __div: args		[ ^args injectFirstInto: [:quo  :arg | quo  / arg] ]
CEvaluator __mod: args		[ ^args injectFirstInto: [:rem  :arg | rem  % arg] ]

CEvaluator __lt: args		[ ^args first <  args second ]
CEvaluator __le: args		[ ^args first <= args second ]
CEvaluator __eq: args		[ ^args first == args second ]
CEvaluator __ne: args		[ ^args first ~~ args second ]
CEvaluator __ge: args		[ ^args first >= args second ]
CEvaluator __gt: args		[ ^args first >  args second ]

CEvaluator __println: args	[ ^args do: [:arg | arg printString putln] ]

CEvaluator pushEnvironment	[ environment := environment push ]
CEvaluator popEnvironment	[ environment := environment pop ]

CEvaluator __if: cForm
[
    | form formals |
    form := cForm value.
    (form size between: 3 and: 4) ifFalse: [self errorSyntax].
    ^(form second eval: self)
	ifTrue:  [form third eval: self]
	ifFalse: [form size == 4 ifTrue: [form fourth eval: self]]
]

CEvaluator __let: cForm
[
    | form vars result |
    form := cForm value.
    form size >= 2 ifFalse: [self errorSyntax].
    (vars := form second) isCList ifFalse: [self errorSyntax].
    self pushEnvironment.
    vars value do: [:init |
	init isCList ifFalse: [self errorSyntax].
	init value size == 2 ifFalse: [self errorSyntax].
	init value first isCSymbol ifFalse: [self errorSyntax].
	environment at: init value first value put: (init value second eval: self)].
    3 to: form size do: [:i | result := (form at: i) eval: self].
    self popEnvironment.
    ^result
]

CEvaluator __define: cForm
[
    | form |
    form := cForm value.
    form size == 3 ifFalse: [self errorArgumentCount].
    form second isCSymbol ifFalse: [self errorSyntax].
    ^environment at: form second value put: (form third eval: self).
]

CEvaluator __lambda: cForm
[
    | form formals |
    form := cForm value.
    form size >= 2 ifFalse: [self errorSyntax].
    form second isCList ifFalse: [self errorSyntax].
    form second value do: [:arg | arg isCSymbol ifFalse: [self errorSyntax]].
    ^cForm
]


Foo : Object ()
Foo a				[ ('Foo a') putln ]
Foo a: a			[ ('Foo a: ', a printString) putln ]
Foo a: a b: b			[ ('Foo a: ', a printString, ' b: ', b printString) putln ]
Foo a: a b: b c: c		[ ('Foo a: ', a printString, ' b: ', b printString, ' c: ', c printString) putln ]
Foo a: a b: b c: c d: d		[ ('Foo a: ', a printString, ' b: ', b printString, ' c: ', c printString, ' d: ', d printString) putln ]
Foo a: a b: b c: c d: d e: e	[ ('Foo a: ', a printString, ' b: ', b printString, ' c: ', c printString, ' d: ', d printString, ' e: ', e printString) putln ]

[
    Foo perform: #a withArguments: #().
    Foo perform: #a: withArguments: #(1).
    Foo perform: #a:b: withArguments: #(1 2).
    Foo perform: #a:b:c: withArguments: #(1 2 foo).
    Foo perform: #a:b:c:d: withArguments: #(1 2 foo bar).
    Foo perform: #a:b:c:d:e: withArguments: #(1 2 foo bar (42 666)).
    Smalltalk quit.
]



[
    | parser env expr |
    parser := CokeParser withFile: StandardInput.
    env := CEvaluator new.
    [(expr := parser parse) notNil]
	whileTrue:
	   [env eval: expr].
]
