{ import: Object }
{ import: Assembler }	"for labels"
{ import: CokeType }
{ import: CokeExpression }
{ import: CokeTransform }

Object freezeTypes []

Expression freezeTypes
[
    self first == #par ifTrue: [^self freezePar].
    self first == #gen ifTrue: [^self second].
    self freezeType; from: 1 doWithIndex: [:e :i | self at: i put: e freezeTypes].
]

Expression freezeType
[
    type isNil ifTrue: [type := Type vType].
    self first: (self first asString , type fullSuffix) asSymbol.
]

Expression freezePar
[
    | common |
    self from: 1 doWithIndex: [:e :i | (self at: i put: e freezeTypes) type isVoidType ifFalse: [common := e type]].
    self from: 1 do: [:e | e type isVoidType ifFalse: [common := common combine: e type for: #par]].
    self from: 1 doWithIndex: [:e :i | e type isVoidType ifFalse: [self at: i put: (e type convert: e to: common)]].
    type := common.
]

Object makeTypes []

Expression makeTypes
[
    type isNil ifTrue: [type := Type fromOperator: self first].
    self from: 1 do: [:e | e makeTypes].
]

"----------------------------------------------------------------"

CokeCompiler : CokeTransform ()

CokeCompiler new	[ ^self withDefault: #apply ]

CokeCompiler newLabel	[ ^AssemblerLabel new ]

CokeCompiler compile: expr
[
    listing nextPutAll: 'compile: '; print: expr; cr.
    expr := self transform: expr asExpression.
    expr := (Expression with: #ret with: expr freezeTypes) type: expr type; freezeType.
    listing nextPutAll: ' --> '; print: expr; cr.
    ^expr
]

CokeCompiler integer: anInteger		[ ^(Expression with: #cnst with: anInteger) type: (SignedType  size: 4) ]
CokeCompiler string:  aString		[ ^(Expression with: #cnst with: aString)   type: Type pcType ]
CokeCompiler symbol:  aSymbol		[ ^(Expression with: #addr with: aSymbol)   type: Type pvType ]

CokeCompiler + expr			[ ^self binary: #+ generic: #add on: expr ]
CokeCompiler - expr			[ ^self binary: #- generic: #sub on: expr ]
CokeCompiler * expr			[ ^self binary: #* generic: #mul on: expr ]
CokeCompiler / expr			[ ^self binary: #/ generic: #div on: expr ]
CokeCompiler % expr			[ ^self binary: #% generic: #mod on: expr ]

CokeCompiler binary: operator generic: opcode on: expr					"xxx POINTERS xxx"
[
    | type |
    expr := self begin :expr.	"transform all arguments"
    type := expr second type.
    2 to: expr size - 1 do: [:i | type := type combine: (expr at: i) type for: operator].
    1 to: expr size - 1 do: [:i | expr at: i put: ((expr at: i) type convert: (expr at: i) to: type)].
    expr first: opcode; type: type.
    expr size > 3 ifTrue: [expr := self associate: opcode in: expr].
    ^expr
]

CokeCompiler associate: opcode in: expr
[
    | root |
    root := (Expression with: expr first with: expr second with: expr third) type: expr type.
    3 to: expr size - 1 do: [:i | root := (Expression with: expr first with: root with: (expr at: i)) type: expr type].
    StdOut nextPutAll: 'associate: '; print: expr; nextPutAll: ' -> '; print: root; cr.
    ^root
]

CokeCompiler < 	expr			[ ^self relation: #<  generic: #lt on: expr ]
CokeCompiler <= expr			[ ^self relation: #<= generic: #le on: expr ]
CokeCompiler == expr			[ ^self relation: #== generic: #eq on: expr ]
CokeCompiler != expr			[ ^self relation: #!= generic: #ne on: expr ]
CokeCompiler >= expr			[ ^self relation: #>= generic: #ge on: expr ]
CokeCompiler >  expr			[ ^self relation: #>  generic: #gt on: expr ]

CokeCompiler relation: operator generic: opcode on: expr					"xxx POINTERS xxx"
[
    | type |
    expr := self begin :expr.	"transform all arguments"
    type := expr second type.
    2 to: expr size - 1 do: [:i | type := type combine: (expr at: i) type for: operator].
    1 to: expr size - 1 do: [:i | expr at: i put: ((expr at: i) type convert: (expr at: i) to: type)].
    expr first: opcode; type: (SignedType size: 4).
    ^expr
]

CokeCompiler if :expr
[
    expr := self begin :expr.	"transform all operands"
    expr size == 3 ifTrue: [^self if2: expr].
    expr size == 4 ifTrue: [^self if3: expr].
    self syntaxError: expr
]

CokeCompiler if2: expr
[
    ^#(par (jumpe :0 :3) :1 (jump :4) (label :3) :2 (label :4))
	unquoting: (Array with: expr second with: expr third
			  with: ((Expression with: #cnst with: 0) type: Type iType)
			  with: self newLabel with: self newLabel)
]

CokeCompiler if3: expr
[
    ^#(par (jumpe :0 :3) :1 (jump :4) (label :3) :2 (label :4))
	unquoting: (Array with: expr second with: expr third with: expr fourth
			  with: self newLabel with: self newLabel)
]

CokeCompiler gen :expr
[
    ^expr type: expr second makeTypes type.
]

CokeCompiler apply :expr		"(fn args ...) -> (call (fn) (arg) ...)"
[
    | call |
    StdErr nextPutAll: 'apply: '; print: expr; cr.
    self begin :expr.
    call := Expression new: expr size + 1.
    expr size downTo: 2 do: [:i | call at: i put: (self makeArgument: (expr at: i - 1) index: i - 2)].
    call second: (self makeFunction: expr first).
    call first: #call; type: Type iType.					"xxx default: i4 xxx"
    StdErr nextPutAll: 'apply: '; print: expr; nextPutAll: ' -> '; print: call; cr.
    ^call
]

CokeCompiler makeArgument: expr index: index
[
    ^(Expression with: #arg with: expr) type: expr type
]

CokeCompiler makeFunction: expr
[
    ^expr isSymbol
	ifTrue:  [^(Expression with: #addrg with: expr) type: Type pvType]
	ifFalse: [^self transform: expr]
]

{ import: PowerPC }

CokeCompiler evaluate: expr
[
    ^PowerPC new evaluate: (self compile: expr)
]

Expression compile
[
    ^CokeCompiler new compile: self
]
