{ import: Object }
{ import: Scanner }
{ import: CokeExpression }

CokeScanner : Scanner ( characterEscapes tokenType tokenValue )

CokeScanner on: aStream
[
    self := super on: aStream.
]

CokeScanner initialize
[
    super initialize.
    characterTable
	atAll: '(['			put: #xOpen;
	atAll: '])'			put: #xClose;
	atAll: '!$%&*+,-./:<=>?@\^_|~'	put: #xLetter;
	from:  $a to: $z		put: #xLetter;
	from:  $A to: $Z		put: #xLetter;
	from:  $0 to: $9		put: #xDigit;
	from:   0 to: 32		put: #xSpace;
	at:    $"			put: #xString;
	at:    $;			put: #xComment.
    (characterEscapes := IdentityDictionary new)
	at: $a  put: $\a; "bel"
	at: $b  put: $\b; "bs"
	at: $e  put: $\e; "esc"
	at: $f  put: $\f; "ff"
	at: $n  put: $\n; "nl"
	at: $r  put: $\r; "cr"
	at: $t  put: $\t; "ht"
	at: $v  put: $\v. "vt"
]

CokeScanner nextCharWithEscape
[
    nextChar == $\\ ifFalse: [^nextChar].
    self nextChar isNil ifTrue: [^self error: 'incomplete character escape'].
    ^characterEscapes at: nextChar ifAbsent: [nextChar]
]


CokeScanner xSpace
[
    [self nextChar notNil and: [(characterTable at: nextChar) == #xSpace]] whileTrue.
    ^nil
]

CokeScanner xComment
[
    [self nextChar notNil and: [nextChar ~~ $\n and: [nextChar ~~ $\r]]] whileTrue.
    ^nil
]

CokeScanner xString
[
    tokenValue := WriteStream on: (String new: 8).
    [self nextChar notNil and: [nextChar ~~ $"]]
	whileTrue: [tokenValue nextPut: self nextCharWithEscape].
    nextChar isNil ifTrue: [self error: 'EOF in string literal'].
    self nextChar.
    tokenValue := tokenValue contents.
    ^#string
]

CokeScanner xDigit
[
    tokenValue := nextChar digitValue.
    [(characterTable at: self nextChar) == #xDigit]
	whileTrue:
	   [tokenValue := tokenValue * 10 + nextChar digitValue].
    ^#number
]

CokeScanner xLetter
[
    tokenValue := WriteStream on: (String new: 8).
    tokenValue nextPut: nextChar.
    [(characterTable at: self nextChar) == #xLetter or: [(characterTable at: nextChar) == #xDigit]]
	whileTrue:
	   [tokenValue nextPut: nextChar].
    tokenValue := tokenValue contents.
    ^#name
]

CokeScanner xOpen
[
    self nextChar.
    ^#open
]

CokeScanner xClose
[
    self nextChar.
    ^#close
]

CokeScanner next
[
    ^(tokenType := super nextToken) notNil ifTrue: [self parse]
]

CokeScanner parse
[
    ^self perform: tokenType
]

CokeScanner number
[
    ^tokenValue.
]

CokeScanner string
[
    ^tokenValue.
]

CokeScanner name
[
    ^tokenValue asSymbol.
]

CokeScanner open
[
    | list |
    list := WriteStream on: (Expression new: 8).
    [(tokenType := self nextToken) notNil and: [tokenType ~~ #close]]
	whileTrue: [list nextPut: self parse].
    ^list contents
]
