{ import: Object }
{ include "tag.h" }

[ Object := _object ]	"alias"
[ Symbol := _selector ]	"alias"

SmallInteger value_: _value		{ return _O(v__value); }
SmallInteger _integerValue		{ return (oop)_I(v_self); }

ByteArray size_: _size value_: _value
[
    self := self new.
    size := SmallInteger value_: _size.
    _bytes := _value.
]

[ UndefinedObject _beNilType ]
[ SmallInteger    _beTagType ]

Object new			[ ^self _vtable _alloc: self _sizeof ]

Object yourself			[ "^self" ]

Object identityHash		{ return (oop)((long)self >> 3 | 1); }

Object _palloc: size		{ return (oop)_palloc(sizeof(oop) * _I(v_size)); }
Object _balloc: size		{ return (oop)_balloc(              _I(v_size)); }

UndefinedObject _nil		{ return 0; }

[ nil   := UndefinedObject _nil ]		{ pragma: type nil   nil   }
[ true  := Object ]				{ pragma: type true  true  }
[ false := nil ]				{ pragma: type false false }

UndefinedObject not		[ ^true ]
Object not			[ ^false ]

UndefinedObject isNil		[ ^true ]
Object isNil			[ ^false ]

UndefinedObject notNil		[ ^false ]
Object notNil			[ ^true ]

Object debugName			[ ^String value_: self _debugName ]
Object printOn: aStream			[ ^aStream nextPutAll: self debugName ]
UndefinedObject printOn: aStream	[ ^aStream nextPutAll: 'nil' ]

Object ifNil: nilBlock			[ ]
UndefinedObject ifNil: nilBlock		[ ^nilBlock value ]

Object ifTrue: aBlock			[ ^aBlock value ]
UndefinedObject ifTrue: aBlock		[ ^nil ]

Object ifFalse: aBlock			[ ^nil ]
UndefinedObject ifFalse: aBlock		[ ^aBlock value ]

Object abort		{ abort(); }
Object exit		{ exit(0); }
Object halt		[ StdErr cr; nextPutAll: 'halt'; cr.  { exit(1); } ]

Object perform: s
{
    struct __closure *c= _local_bind(v_s, v_self);
    return (c->method)((oop)c, v_self);
}

Object perform: s with: a
{
    struct __closure *c= _local_bind(v_s, v_self);
    return (c->method)((oop)c, v_self, v_a);
}

Object perform: s with: a with: b
{
    struct __closure *c= _local_bind(v_s, v_self);
    return (c->method)((oop)c, v_self, v_a, v_b);
}

Object error: reason
[
    StdErr cr; nextPutAll: reason; cr.
    { abort(); }.
]

Object shouldNotImplement: aSelector
[
    StdErr cr; nextPutAll: self debugName; nextPutAll: ' should not implement '; print: aSelector; cr.
    { abort(); }.
]

Object errorOutOfBounds: anIndex
[
    StdErr cr;
	print: anIndex; nextPutAll: ' is out of bounds in ';
	nextPutAll: self debugName; nextPut: $(; print: self size; nextPut: $);
	cr.
    { abort(); }.
]

"----------------------------------------------------------------"

SmallInteger identityHash	[ ^self * 17 ]

SmallInteger // arg
[
    { if ((long)v_self & (long)v_arg & 1) return _O(_I(v_self) / _I(v_arg)); }.
    ^self primitiveFailed
]

SmallInteger \\ arg
[
    { if ((long)v_self & (long)v_arg & 1) return _O(_I(v_self) % _I(v_arg)); }.
    ^self primitiveFailed
]

SmallInteger printOn: aStream
[
    ^self printOn: aStream base: 10
]

SmallInteger printOn: aStream base: base
[
    self < 0 ifTrue: [^aStream nextPut: $-; print: 0 - self base: base].
    self >= base ifTrue: [self // base printOn: aStream base: base].
    aStream nextPut: (self \\ base) asDigit.
]

SmallInteger asDigit
[
    ^self + (self > 9 ifTrue: [$a - 10] ifFalse: [$0])
]

SmallInteger digitValue
[
    (self between: $A and: $Z) ifTrue: [^self - $A + 10].
    (self between: $a and: $z) ifTrue: [^self - $a + 10].
    (self between: $0 and: $9) ifTrue: [^self - $0     ].
    ^nil
]

SmallInteger max: anInteger
[
    ^self < anInteger ifTrue: [anInteger] ifFalse: [self]
]

SmallInteger between: low and: high
[
    ^low <= self and: [self <= high]
]

SmallInteger to: limit do: unaryBlock
[
    [self <= limit]
	whileTrue:
	   [unaryBlock value: self.
	    self := self + 1]
]

SmallInteger downTo: limit do: unaryBlock
[
    [self >= limit]
	whileTrue:
	   [unaryBlock value: self.
	    self := self - 1]
]

"----------------------------------------------------------------"

String asString []

String value_: _value
[
    self := self new.
    _bytes := _value.
    size := SmallInteger value_: self _size.
]

String _size
{
    return (oop)strlen((char *)self->v__bytes);
}

String _cString
{
    char *string= (char *)malloc((size_t)(long)self->v_size >> 1);
    memcpy(string, (void *)self->v__bytes, ((size_t)(long)self->v_size >> 1) + 1);
    string[(size_t)(long)self->v_size >> 1]= '\0';
    return (oop)string;
}

String beginsWith: aString
[
    self size < aString size ifTrue: [^false].
    aString doWithIndex: [:e :i | (self at: i) == e ifFalse: [^false]].
    ^true
]

String asSymbol		[ ^Symbol intern_: self _bytes size_: size _integerValue ]

String printOn: aStream	[ ^aStream nextPut: $'; nextPutAll: self; nextPut: $' ]


"----------------------------------------------------------------"

Object isSymbol	[ ^false ]
Symbol isSymbol	[ ^true ]

Symbol intern_: _string size_: _stringSize
{
    char *buf= (char *)alloca((int)v__stringSize + 1);
    memcpy((void *)buf, (void *)v__string, (int)v__stringSize);
    buf[(int)v__stringSize]= '\0';
    return _selector(buf);
}

Symbol intern: aString
[
    ^self intern_: aString _cString size_: aString size _integerValue
]

Symbol _bytes	[ ^_elements ]
Symbol size	[ ^SmallInteger value_: _size ]

Symbol _cString	[ ^_elements ]

Symbol asString	[ ^String value_: _elements ]

Symbol beginsWith: aString	[ ^self asString beginsWith: aString ]

Symbol printOn: aStream	[ ^aStream nextPut: $#; nextPutAll: self asString ]

"----------------------------------------------------------------"

StaticBlockClosure function_: _implementationAddress arity_: _argumentCount
[
    { v_self= _alloc(v_self, sizeof(*self)); }.
    _arity    := _argumentCount.
    _function := _implementationAddress.
]

StaticBlockClosure arity	[ ^SmallInteger value_: _arity ]

"evaluating"

StaticBlockClosure value
[
    { if ((int)self->v__arity == 0) return ((_imp_t)(self->v__function))(v__closure, v_self); }.
    ^self errorArgumentCount: 0
]

StaticBlockClosure value: a
[
    { if ((int)self->v__arity == 1) return ((_imp_t)(self->v__function))(v__closure, v_self, v_a); }.
    ^self errorArgumentCount: 1
]

StaticBlockClosure value: a value: b
[
    { if ((int)self->v__arity == 2) return ((_imp_t)(self->v__function))(v__closure, v_self, v_a, v_b); }.
    ^self errorArgumentCount: 1
]

StaticBlockClosure value: a value: b value: c
[
    { if ((int)self->v__arity == 3) return ((_imp_t)(self->v__function))(v__closure, v_self, v_a, v_b, v_c); }.
    ^self errorArgumentCount: 1
]

StaticBlockClosure value: a value: b value: c value: d
[
    { if ((int)self->v__arity == 4) return ((_imp_t)(self->v__function))(v__closure, v_self, v_a, v_b, v_c, v_d); }.
    ^self errorArgumentCount: 1
]

StaticBlockClosure value: a value: b value: c value: d value: e
[
    { if ((int)self->v__arity == 5) return ((_imp_t)(self->v__function))(v__closure, v_self, v_a, v_b, v_c, v_d, v_e); }.
    ^self errorArgumentCount: 1
]

"controling"

StaticBlockClosure whileTrue
[
    [self value] whileTrue: []
]

StaticBlockClosure whileTrue: aBlock
[
    self value
	ifTrue:
	   [aBlock value.
	    self whileTrue: aBlock].
]

StaticBlockClosure whileFalse
[
    [self value] whileFalse: []
]

StaticBlockClosure whileFalse: aBlock
[
    self value
	ifFalse:
	   [aBlock value.
	    self whileFalse: aBlock].
]

StaticBlockClosure repeat
[
    [true] whileTrue: [self value]
]


BlockClosure function_: _implementationAddress arity_: _argumentCount
	     outer: outerBlock state: variableArray nlr_: _dynamicEnvironment
[
    { v_self= _alloc(v_self, sizeof(*self)); }.
    _function := _implementationAddress.
    _arity    := _argumentCount.
    outer     := outerBlock.
    state     := variableArray.
    _nlr      := _dynamicEnvironment.
]

"----------------------------------------------------------------"

"{ import: Collection }"

"{ import: Object }"

Collection isEmpty
[
    ^self size == 0
]

Collection addAll: aCollection
[
    aCollection do: [:elt | self add: elt]
]

Collection printOn: aStream
[
    aStream nextPutAll: self debugName.
    self printContentsOn: aStream.
]

Collection printContentsOn: aStream
[
    aStream nextPut: $(.
    self do: [:element | element printOn: aStream] separatedBy: [aStream space].
    aStream nextPut: $).
]

Collection do: elementBlock separatedBy: betweenBlock
[
    | first |
    first := true.
    self do: [:element |
	first ifTrue: [first := false] ifFalse: [betweenBlock value].
	elementBlock value: element].
]

SequenceableCollection first		[ ^self at: 0			]
SequenceableCollection first: anObject	[ ^self at: 0 put: anObject	]
SequenceableCollection second		[ ^self at: 1			]
SequenceableCollection second: anObject	[ ^self at: 1 put: anObject	]
SequenceableCollection third		[ ^self at: 2			]
SequenceableCollection third: anObject	[ ^self at: 2 put: anObject	]
SequenceableCollection fourth		[ ^self at: 3			]
SequenceableCollection last		[ ^self at: self size - 1	]

SequenceableCollection new: size withAll: value
[
    ^(self new: size) atAllPut: value; yourself
]

SequenceableCollection atAllPut: value
[
    0 to: self size - 1 do: [:key | self at: key put: value]
]

SequenceableCollection atAll: keyCollection put: value
[
    keyCollection do: [:key | self at: key put: value]
]

SequenceableCollection from: firstKey to: lastKey put: value
[
    firstKey to: lastKey do: [:key | self at: key put: value]
]

SequenceableCollection do: elementBlock			[ self from: 0 do: elementBlock ]
SequenceableCollection reverseDo: elementBlock		[ self from: 0 reverseDo: elementBlock ]
SequenceableCollection doWithIndex: elementIndexBlock	[ self from: 0 doWithIndex: elementIndexBlock ]
SequenceableCollection reverseDoWithIndex: eiBlock	[ self from: 0 reverseDoWithIndex: eiBlock ]
SequenceableCollection do: aBlock separatedBy: bBlock	[ self from: 0 do: aBlock separatedBy: bBlock ]

SequenceableCollection from: start do: elementBlock
[
    start to: self size - 1 do: [:offset | elementBlock value: (self at: offset)].
]

SequenceableCollection from: start reverseDo: elementBlock
[
    self size - 1 downTo: start do: [:offset | elementBlock value: (self at: offset)].
]

SequenceableCollection from: start doWithIndex: elementIndexBlock
[
    start to: self size - 1 do: [:offset | elementIndexBlock value: (self at: offset) value: offset].
]

SequenceableCollection from: start reverseDoWithIndex: elementIndexBlock
[
    self size - 1 downTo: start do: [:offset | elementIndexBlock value: (self at: offset) value: offset].
]

SequenceableCollection from: start do: elementBlock separatedBy: betweenBlock
[
    | first |
    first := true.
    self from: start do: [:e |
	first ifTrue: [first := false] ifFalse: [betweenBlock value].
	elementBlock value: e].
]

SequenceableCollection with: aCollection do: elementBlock
[
    self size == aCollection size ifFalse: [self error: 'collection sizes differ'].
    0 to: self size - 1 do: [:offset | elementBlock value: (self at: offset) value: (aCollection at: offset)].
]

SequenceableCollection with: aCollection doWithIndex: elementsIndexBlock
[
    self size == aCollection size ifFalse: [self error: 'collection sizes differ'].
    0 to: self size - 1 do: [:offset | elementsIndexBlock value: (self at: offset) value: (aCollection at: offset) value: offset].
]

SequenceableCollection with: aCollection replace: elementsBlock
[
    self with: aCollection doWithIndex: [:a :b :i | self at: i put: (elementsBlock value: a value: b)]
]

SequenceableCollection replaceFrom: first to: last with: aCollection
[
    ^self replaceFrom: first to: last with: aCollection startingAt: 0
]

SequenceableCollection replaceFrom: first to: last with: aCollection startingAt: collectionStart
[
    0 to: last - first do: [:offset | self at: first + offset put: (aCollection at: collectionStart + offset)].
]

SequenceableCollection copy
[
    ^(self new: self size) replaceFrom: 0 to: self size - 1 with: self startingAt: 0
]

SequenceableCollection copyFrom: first to: last
[
    | size |
    size := 0 max: last - first + 1.
    ^(self new: size) replaceFrom: 0 to: size - 1 with: self startingAt: 0
]

SequenceableCollection , aCollection
[
    ^(self new: self size + aCollection size)
	replaceFrom: 0 to: self size - 1 with: self;
	replaceFrom: self size to: self size + aCollection size - 1 with: aCollection
]

"----------------------------------------------------------------"

ArrayedCollection with: a
[
    ^(self new: 1)
	at: 0 put: a;
	yourself
]

ArrayedCollection with: a with: b
[
    ^(self new: 2)
	at: 0 put: a;
	at: 1 put: b;
	yourself
]

ArrayedCollection with: a with: b with: c
[
    ^(self new: 3)
	at: 0 put: a;
	at: 1 put: b;
	at: 2 put: c;
	yourself
]

ArrayedCollection with: a with: b with: c with: d
[
    ^(self new: 4)
	at: 0 put: a;
	at: 1 put: b;
	at: 2 put: c;
	at: 3 put: d;
	yourself
]

ArrayedCollection with: a with: b with: c with: d with: e
[
    ^(self new: 5)
	at: 0 put: a;
	at: 1 put: b;
	at: 2 put: c;
	at: 3 put: d;
	at: 4 put: e;
	yourself
]

ArrayedCollection new: capacity
[
    self := super new.
    size := capacity.
]

ArrayedCollection isEmpty
[
    ^size == 0.
]

ArrayedCollection size
[
    ^size.
]

"----------------------------------------------------------------"

_object isArray	[ ^false ]
Array isArray	[ ^true ]

Array size_: _size value_: _value
[
    self := self new: (SmallInteger value_: _size).
    _oops := _value.
    {
      int i;
      for (i= 0;  i < (int)v__size;  ++i)
        ((oop *)self->v__oops)[i]= *((oop **)v__value)[i];
    }.
]

Array new: capacity
[
    self := super new.
    size := capacity.
    _oops := self _palloc: size
]

Array _elements	[ ^_oops ]

Array at: offset
[
    ^(0 <= offset and: [offset < size])
	ifTrue:  [self at_: offset _integerValue]
	ifFalse: [self errorOutOfBounds: offset]
]

Array at_: _offset
{
    return ((oop *)self->v__oops)[(int)v__offset];
}

Array at: offset put: anObject
[
    ^(0 <= offset and: [offset < size])
	ifTrue:  [self at_: offset _integerValue put: anObject]
	ifFalse: [self errorOutOfBounds: offset]
]

Array at_: _offset put: anObject
{
    return ((oop *)self->v__oops)[(int)v__offset]= v_anObject;
}

Array  argumentVectorSize	[ ^SmallInteger value_: self _argumentVectorSize ]
Array _argumentVectorSize	{ return (oop)_param(0); }
Array  argumentVectorAt: offset	[ ^String value_: (self _argumentVectorAt_: offset _integerValue) ]
Array _argumentVectorAt_: _off	{ return (oop)(((char **)_param(1))[(int)v__off]); }

Array argumentVector
[
    self := self new: self argumentVectorSize.
    0 to: size - 1 do: [:offset | self at: offset put: (self argumentVectorAt: offset)].
]

Array printOn: aStream
[
    aStream nextPut: $#.
    self printContentsOn: aStream.
]

"----------------------------------------------------------------"

ByteArray : ArrayedCollection ( _bytes )

ByteArray new: capacity
[
    self := super new: capacity.
    _bytes := self _balloc: capacity.
]

ByteArray _bytes	[ ^_bytes ]
ByteArray _elements	[ ^_bytes ]

ByteArray at: offset
[
    ^(0 <= offset and: [offset < size])
	ifTrue:  [SmallInteger value_: (self _at_: offset _integerValue)]
	ifFalse: [self errorOutOfBounds: offset]
]

ByteArray _at_: _offset
{
    return (oop)((long)(((unsigned char *)self->v__bytes)[(int)v__offset]));
}

ByteArray at: offset put: anInteger
[
    (0 <= offset and: [offset < size and: [0 <= anInteger and: [anInteger < 256]]])
	ifTrue:  [self at_: offset _integerValue put_: anInteger _integerValue]
	ifFalse: [self errorOutOfBounds: offset].
    ^anInteger.
]

ByteArray at_: _offset put_: _integer
{
    ((char *)self->v__bytes)[(int)v__offset]= (int)v__integer;
}


IdentitySet : Collection ( tally lists )

IdentitySet lists
[
    ^lists
]

IdentitySet new
[
    ^self new: 1.
]

IdentitySet new: width
[
    self := super new.
    tally := 0.
    lists := Array new: width.
]

IdentitySet size	[ ^tally ]

IdentitySet add: anObject
[
    | listOffset list |
    listOffset := anObject identityHash \\ lists size.
    list := lists at: listOffset.
    list isNil
	ifTrue:
	   [list := lists at: listOffset put: self newList.
	    tally := tally + 1.
	    ^list at: 0 put: anObject].
    0 to: list size - 1 do: [:offset |
	| elt |
	elt := list at: offset.
	elt == nil
	    ifTrue:
	       [tally := tally + 1.
		^list at: offset put: anObject].
	(self compare: elt with: anObject)
	    ifTrue: [^elt]].
    ^self
	grow: list at: listOffset;
	add: anObject
]

IdentitySet newList	[ ^Array new: 4 ]

IdentitySet grow: list at: listOffset
[
    list size * 4 > lists size
	ifTrue:  [self widen]
	ifFalse: [self deepen: list at: listOffset]
]

IdentitySet findElement: anObject
[
    | listOffset list |
    listOffset := anObject identityHash \\ lists size.
    list := lists at: listOffset.
    list isNil ifTrue: [^nil].
    0 to: list size - 1 do: [:offset |
	| elt |
	elt := list at: offset.
	elt == nil ifTrue: [^nil].
	(self compare: elt with: anObject) ifTrue: [^elt]].
    ^nil
]

IdentitySet compare: anEntry with: anObject
[
    ^anEntry == anObject
]


IdentitySet widen
[
    "StdOut nextPutAll: 'WIDEN '; print: self size; space; print: lists size + 1 * 2 - 1; cr."
    lists := (self new: lists size + 1 * 2 - 1) addAll: self; lists.
]

IdentitySet deepen: list at: listOffset
[
    | newList |
    "StdOut nextPutAll: 'DEEPEN '; print: self size; space; print: self lists size; space; print: list size * 2; cr."
    newList := list new: list size * 2.
    newList replaceFrom: 0 to: list size - 1 with: list startingAt: 0.
    lists at: listOffset put: newList.
]

IdentitySet do: unaryBlock
[
    lists do: [:list | list notNil ifTrue: [list do: [:elt | elt notNil ifTrue: [unaryBlock value: elt]]]]
]

IdentitySet includes: anObject
[
    ^(self findElement: anObject ifAbsent: []) notNil
]


FastIdentitySet : IdentitySet ()

FastIdentitySet newList	[ ^Array new: 1 ]

FastIdentitySet grow: list at: listOffset
[
    list size >= 4
	ifTrue:  [self widen]
	ifFalse: [self deepen: list at: listOffset]
]

"----------------------------------------------------------------"

Association : Object ( key value )

Association key		[ ^key ]
Association value	[ ^value ]

Object -> anObject
[
    ^Association withKey: self value: anObject
]

Association withKey: aKey value: aValue
[
    self := self new.
    key := aKey.
    value := aValue.
]

Association setValue: aValue
[
    ^value := aValue.
]

Association identityHash
[
    ^key identityHash
]

Association printOn: aStream
[
    aStream nextPut: $(; print: key; nextPutAll: ' -> '; print: value; nextPut: $)
]

"----------------------------------------------------------------"

IdentityDictionary compare: anEntry with: anObject
[
    ^anEntry key == anObject key
]

IdentityDictionary at: aKey
[
    ^self at: aKey ifAbsent: [self errorKeyNotFound]
]

IdentityDictionary at: aKey ifAbsent: failBlock
[
    | assoc |
    ^(assoc := self findElement: aKey -> nil) notNil
	ifTrue:  [assoc value]
	ifFalse: [failBlock value]
]

IdentityDictionary at: aKey put: anObject
[
    ^(self add: aKey -> anObject) setValue: anObject
]


FastIdentityDictionary : IdentityDictionary ()

FastIdentityDictionary newList	[ ^Array new: 1 ]

FastIdentityDictionary grow: list at: listOffset
[
    list size >= 4
	ifTrue:  [self widen]
	ifFalse: [self deepen: list at: listOffset]
]

"----------------------------------------------------------------"

OrderedCollection : SequenceableCollection ( firstIndex lastIndex array )

OrderedCollection new	[ ^(self new: 4) removeAll ]

OrderedCollection new: size
[
    self := super new.
    array := Array new: size.
    firstIndex := 0.
    lastIndex := size.
]

OrderedCollection size			[ ^lastIndex - firstIndex ]

OrderedCollection add: anObject		[ ^self addLast: anObject ]

OrderedCollection addFirst: anObject
[
    firstIndex == 0 ifTrue: [self makeRoomAtFirst].
    firstIndex := firstIndex - 1.
    array at: firstIndex put: anObject.
    ^anObject
]

OrderedCollection addLast: anObject
[
    lastIndex == array size ifTrue: [self makeRoomAtLast].
    array at: lastIndex put: anObject.
    lastIndex := lastIndex + 1.
    ^anObject
]

OrderedCollection makeRoomAtFirst
[
    | delta newArray |
    delta := array size.
    (self size == 0 and: [delta > 0]) ifTrue: [^self resetToEnd].
    (self size * 2 < delta) ifTrue: [^self shiftToEnd].
    newArray := (array new: self size + delta) replaceFrom: firstIndex + delta to: lastIndex + delta - 1 with: array startingAt: firstIndex.
    array := newArray.
    firstIndex := firstIndex + delta.
    lastIndex := lastIndex + delta.
]

OrderedCollection makeRoomAtLast
[
    (self size == 0 and: [array size > 0]) ifTrue: [^self resetToBeginning].
    (self size * 2 < array size) ifTrue: [^self shiftToBeginning].
    array := (array new: self size + array size)
	replaceFrom: firstIndex to: lastIndex - 1 with: array startingAt: firstIndex;
	yourself.
]

OrderedCollection resetToBegining	[ firstIndex := lastIndex := 0]
OrderedCollection resetToEnd		[ firstIndex := lastIndex := array size ]

OrderedCollection removeFirst
[
    | first |
    self emptyCheck.
    first := array at: firstIndex.
    firstIndex := firstIndex + 1.
    ^first
]

OrderedCollection removeLast
[
    | last |
    self emptyCheck.
    lastIndex := lastIndex - 1.
    last := array at: lastIndex.
    ^last
]

OrderedCollection removeAll
[
    firstIndex := lastIndex := 0.
]

OrderedCollection at: offset
[
    | arrayOffset |
    arrayOffset := firstIndex + offset.
    ^(offset >= 0 and: [arrayOffset < lastIndex])
	ifTrue:  [array at: arrayOffset]
	ifFalse: [self errorOutOfBounds: offset]
]

OrderedCollection do: unaryBlock
[
    firstIndex to: lastIndex - 1 do: [:index | unaryBlock value: (array at: index)]
]


"----------------------------------------------------------------"

File : Object ( _fd )

File withFd: fd
[
    self := self new.
    _fd := fd _integerValue.
]

File print: anObject		[ anObject printOn: self ]
File print: aNumber base: base	[ aNumber printOn: self base: base ]
File cr				[ self nextPut: $\n ]
File space			[ self nextPut: $   ]
File tab			[ self nextPut: $\t ]

File nextPut: aByte
[
    { if ((long)v_aByte & 1) { char c= _I(v_aByte);  return _O(write((int)self->v__fd, (void *)&c, 1)); } }.
    ^self primitiveFailed
]

File nextPutAll: aString
[
    | _bytes _size |
    _bytes := aString _bytes.
    _size  := aString size _integerValue.
    { return _O(write((int)self->v__fd, (void *)v__bytes, (size_t)v__size)); }.
]

File  read: aCollection			[ ^self read: aCollection size: aCollection size ]
File  read: aCollection size: size	[ ^SmallInteger value_: (self _read_: aCollection _elements size_: size _integerValue) ]
File _read_: _bytes size_: _size	{ return (oop)read((int)self->v__fd, (void *)v__bytes, (size_t)v__size); }

File isInteractive			{ return (oop)isatty((int)self->v__fd); }

"----------------------------------------------------------------"

SinkStream : Object ()

SinkStream cr			[]
SinkStream tab			[]
SinkStream space		[]
SinkStream print: o		[]
SinkStream print: o base: b	[]
SinkStream nextPut: o		[]
SinkStream nextPutAll: o	[]

"----------------------------------------------------------------"

ReadStream : Object ( collection position readLimit )

ReadStream collection	[ ^collection ]

ReadStream on: aCollection
[
    self := self new.
    collection := aCollection.
    position := 0.
    readLimit := collection size.
]

ReadStream peek
[
    ^self atEnd ifFalse: [collection at: position]
]

ReadStream next
[
    ^self atEnd ifFalse: [collection at: (position := position + 1) - 1]
]

ReadStream atEnd
[
    ^position >= readLimit and: [self refill not]
]

ReadStream refill
[
    ^false
]

"----------------------------------------------------------------"

WriteStream : ReadStream ( writeLimit )

WriteStream on: aCollection
[
    self := super on: aCollection.
    writeLimit := collection size.
    readLimit := 0.
]

WriteStream nextPutAll: aCollection
[
    aCollection do: [:anObject | self nextPut: anObject].
    ^aCollection
]

WriteStream nextPut: anObject
[
    position >= writeLimit ifTrue: [self grow].
    collection at: position put: anObject.
    readLimit := (position := position + 1) max: readLimit.
    ^anObject
]

WriteStream grow
[
    | newCollection |
    newCollection := collection new: (writeLimit := collection size * 2).
    newCollection replaceFrom: 0 to: collection size - 1 with: collection startingAt: 0.
    collection := newCollection.
]

WriteStream contents
[
    ^collection copyFrom: 0 to: position - 1
]



WriteStream dump
[
    StdOut nextPutAll: self debugName;
	nextPutAll: '(collection='; nextPutAll: collection debugName; nextPut: $[; print: collection size; nextPut: $];
	nextPutAll: ' position=';   print: position;
	nextPutAll: ' readLimit=';  print: readLimit;
	nextPutAll: ' writeLimit='; print: writeLimit;
	nextPut: $); cr.
]


"----------------------------------------------------------------"

FileStream : ReadStream ( file )

FileStream refill
[
    | n |
    (n := file read: collection) > 0 ifFalse: [^false].
    position := 0.
    readLimit := n.
]

FileStream on: aFile
[
    self := super on: (ByteArray new: 8192).
    file := aFile.
    position := collection size.
]

"----------------------------------------------------------------"

ConsoleFileStream : FileStream ( prompt )

ConsoleFileStream on: aFile
[
    ^self on: aFile prompt: (aFile isInteractive ifTrue: ['> '] ifFalse: ['']).
]

ConsoleFileStream on: aFile prompt: aString
[
    self := super on: aFile.
    prompt := aString.
]

ConsoleFileStream refill
[
    StdOut nextPutAll: prompt.
    ^super refill
]

"----------------------------------------------------------------"

[ StdIn  := File withFd: 0 ]
[ StdOut := File withFd: 1 ]
[ StdErr := File withFd: 2 ]

File readStream
[
    ^(self isInteractive ifTrue: [ConsoleFileStream] ifFalse: [FileStream]) on: self
]

"
Object properties
[
    ^properties
]

Object get: propertyName
[
    ^properties notNil ifTrue: [properties at: propertyName ifAbsent: []]
]

Object put: propertyName value: propertyValue
[
    properties isNil ifTrue: [properties := IdentityDictionary new].
    ^properties at: propertyName put: propertyValue
]
"
