#ifndef __vpu_CPU_h
#define __vpu_CPU_h

#include "Heap.h"
#include "Insns.h"

#include "md.h"


class CPU;
class Insn;
class VPU;

enum {
  IREG= 0,
  FREG= 1,
  NREG= 2,
};

enum {
  IVAR= 0,
  FVAR= 1,
  NVAR= 2,
};


struct InsnMode
{
  unsigned      mask;
  Insn::Emitter emit;
  unsigned	clobbers[NREG];
};


class CPU : public Bulk
{
public:

  VPU     *vpu;
  int      asm_pass;
  insn    *asm_pc;
  int      codeSize;
  void    *code;

  unsigned ccri;
  unsigned csri;
  unsigned ccrf;
  unsigned csrf;

  bool	   canSaveArgs;
  bool	   reverseArgs;

  void initialiseFrame(VPU *vpu);

  CPU(void)
  {
    vpu=      	 0;
    asm_pass= 	 0;
    asm_pc=   	 0;
    codeSize=  	 0;
    code=     	 0;
    ccri= csri=  0;
    ccrf= csrf=  0;
    canSaveArgs= false;
    reverseArgs= false;
  }

  void  initialise(VPU *vpu);
  void  initialiseArgs(VPU *vpu);

  void  constrainCall(Call *call);

  void  pass(VPU *vpu, int n);
  void *allocateCode(void);
  void  flush(void);
  int   disassemble(insn *addr);

# define IMODE(WHO)	void emit##WHO(WHO *insn, VPU *vpu);
# define PMODE(WHO)	static InsnMode *modes##WHO(void);
# include "modes.h"
# undef	 IMODE
# undef	 PMODE

# define INSN(WHO)
# define MODE(I,O,L,R,CI,CF)	void emit##I##_##O##_##L##_##R(I *insn, VPU *vpu);
# define ENDINSN(WHO)
# include "md-cpu.h"
# undef	 INSN
# undef	 MODE
# undef	 ENDINSN

  void noEmit(Insn *i, VPU *vpu);
};


#endif // __vpu_CPU_h
