#ifndef __vpu_Insn_h
#define __vpu_Insn_h

#include "VPU.h"
#include "Heap.h"
#include "List.h"
#include "Stack.h"

#include "types.h"

class CPU;
class Loc;

struct InsnMode;


class Insn : public Bulk
{
public:
  // Deque
  Insn	   *prev;
  Insn	   *next;

  typedef void (CPU::*Emitter)(Insn *insn, VPU *vpu);

  Loc      *loc;
  Stack     in;
  Stack     out;
  TempStack temps;
  Emitter   emit;
  unsigned  clobbers[2];	//xxx should be [NREG]
  bool	    spilled;
  bool	    reloaded;

  Insn(void)
  {
    prev= next=  0;
    loc=      	 0;
    emit=     	 0;
    clobbers[0]= 0;
    clobbers[1]= 0;
    spilled=  	 false;
    reloaded= 	 false;
  }

  virtual ~Insn(void) {}

  void follow(VPU *vpu, Insn *insn);
  void append(Insn *insn);
  void prepend(Insn *insn);

  struct StackEffect
  {
    unsigned read : 8;
    unsigned drop : 8;
    unsigned push : 8;
    unsigned copy : 8;
    StackEffect(unsigned r, unsigned d, unsigned p, unsigned c)
    {
      read= r;  drop= d;  push= p;  copy= c;
    }
  };

  virtual StackEffect stackEffect(void) { return StackEffect(0,0,0,0); }

# define declare(N,R,D,P,C)							\
  virtual const char *name(void) { return #N; }					\
  virtual StackEffect stackEffect(void) { return StackEffect(R,D,P,C); }	\
  N

  virtual void initStack(VPU *vpu);
  virtual void initFlow(VPU *vpu, Insn *prev);
  virtual void setOutputLoc(Loc *loc);
  virtual void initOutput(VPU *vpu);
  virtual void initCall(VPU *vpu);
  virtual void typeCheck(VPU *vpu);
  virtual void initMode(VPU *vpu);
  virtual void selectInsn(VPU *vpu);
  virtual void constrainRegisters(VPU *vpu);
  virtual void targetRegisters(VPU *vpu);
  virtual bool allocateRegisters(VPU *vpu);
  virtual bool isSavedArg(void) { return false; }
  virtual void reload(VPU *vpu);
  virtual void moveRegs(void);
  virtual bool hasMovableReg(void);
  virtual void moveReg(void);
          void allocateFrame(VPU *vpu);
  inline  void generate(VPU *vpu) { assert(emit); (vpu->cpu->*emit)(this, vpu); }

	  bool findEmitter(InsnMode *modes, int out, int lhs, int rhs);
  virtual bool hasMode(VPU *vpu)= 0;
  virtual InsnMode *modes(void);

  virtual bool isCmp(void);
  virtual bool isJump(void);
  virtual bool isTemp(void);
  virtual bool isArgOut(void);

  virtual const char *name(void)= 0;

          int  printType(void);
  virtual int  printArgs(void);
	  int  print(void);
	  void println(void);
	  int  programSize(void);
};


#endif // __vpu_Insn_h
