#ifndef __vpu_Insns_h
#define __vpu_Insns_h

#include "Insn.h"

class Loc;
class Reg;
class Var;


class LocalLabel;


class Gdef : public Insn
{
public:
  Label *label;
  declare(Gdef,0,0,0,0)(Label *l) { label= l; }
  virtual void initMode(VPU *vpu);
  virtual bool hasMode(VPU *vpu);
  virtual int  printArgs(void);
};


class Ldef : public Insn
{
public:
  int         index;
  LocalLabel *label;
  bool	      reached;
  declare(Ldef,0,0,0,0)(int i, LocalLabel *l);
  virtual void initStack(VPU *vpu);
  virtual void initFlow(VPU *vpu, Insn *prev);
  virtual void initMode(VPU *vpu);
  virtual bool hasMode(VPU *vpu);
  virtual bool allocateRegisters(VPU *vpu);
  virtual int  printArgs(void);
};


class IndexedInsn : public Insn
{
public:
  int index;
  IndexedInsn(int i) { index= i; }
  virtual int printArgs(void);
};

class Begin : public IndexedInsn
{
public:
  declare(Begin,0,0,0,0)(int i) : IndexedInsn(i) {}
  virtual void initMode(VPU *vpu);
  virtual bool hasMode(VPU *vpu);
  virtual int  printArgs(void);
};

class End : public IndexedInsn
{
public:
  declare(End,0,0,0,0)(int i) : IndexedInsn(i) {}
  virtual void initMode(VPU *vpu);
  virtual bool hasMode(VPU *vpu);
  virtual int  printArgs(void);
};


class Entry : public Insn
{
public:
  int returnType;
  declare(Entry,0,0,0,0)(int t) { returnType= t; }
  void inReg(VPU *vpu, int mode, Reg *r, Reg *q= 0);
  void onStack(VPU *vpu, int mode, Var *v);
  virtual void initMode(VPU *vpu);
  virtual bool allocateRegisters(VPU *vpu);
  virtual bool hasMode(VPU *vpu);
  virtual InsnMode *modes(void);
};


class Move : public Insn
{
public:
  declare(Move,0,0,0,0)(Loc *l= 0) { loc= l; }
  virtual Loc *src(void) { return in[0]->loc; }
  virtual Loc *dst(void) { return loc; }
  virtual void selectInsn(VPU *vpu);
  virtual bool allocateRegisters(VPU *vpu);
  virtual void reload(VPU *vpu);
  virtual bool hasMode(VPU *vpu);
  virtual InsnMode *modes(void);
  inline  Loc *lhs(void) { return src(); }
  virtual int  printArgs(void);
};


class Arg : public Move
{
public:
  int   refType;
  Loc  *ref;
  bool  addressed;
  declare(Arg,0,0,0,0)(int t) { refType= t; ref= 0; addressed= false; }
  virtual Loc *src(void) { return ref; }
  void inReg(VPU *vpu, int clazz, Reg *r, Reg *q= 0);
  void onStack(VPU *vpu, int clazz, Var *v);
  virtual void initStack(VPU *vpu);
  virtual void typeCheck(VPU *vpu);
  virtual void initMode(VPU *vpu);
  virtual bool allocateRegisters(VPU *vpu);
  virtual void reload(VPU *vpu);
  virtual InsnMode *modes(void);
  virtual bool hasMode(VPU *vpu);
  virtual bool isTemp(void);
  virtual int  printArgs(void);
};


class Dup : public Move
{
public:
  int index;
  declare(Dup,index,0,1,0)(int n) { index= n; }
  virtual Loc *src(void) { return in[index]->loc; }
  virtual void setOutputLoc(Loc *l);
  virtual void typeCheck(VPU *vpu);
  virtual void initMode(VPU *vpu);
  virtual bool hasMovableReg(void);
  virtual void moveReg(void);
  virtual int  printArgs(void);
};


class Temp : public Insn
{
public:
  int   refType;
  bool  addressed;
  declare(Temp,0,0,0,0)(int t) { refType= t; addressed= false; }
  virtual void initStack(VPU *vpu);
  virtual void typeCheck(VPU *vpu);
  virtual void initMode(VPU *vpu);
  virtual bool allocateRegisters(VPU *vpu);
  virtual bool hasMode(VPU *vpu);
  virtual bool isTemp(void);
};

class LdArg : public Move
{
public:
  int   index;
  Arg  *arg;
  declare(LdArg,0,0,1,0)(int i) { index= i; arg= 0; }
  virtual Loc *src(void) { return arg->loc; }
  virtual void initStack(VPU *vpu);
  virtual void setOutputLoc(Loc *l);
  virtual void typeCheck(VPU *vpu);
  virtual void reload(VPU *vpu);
  virtual bool hasMovableReg(void);
  virtual void moveReg(void);
  virtual int  printArgs(void);
};

class LdTemp : public Move
{
public:
  int   index;
  Temp *temp;
  declare(LdTemp,0,0,1,0)(int i) { index= i; temp= 0; }
  virtual Loc *src(void) { return temp->loc; }
  virtual void initStack(VPU *vpu);
  virtual void setOutputLoc(Loc *l);
  virtual void typeCheck(VPU *vpu);
  virtual void reload(VPU *vpu);
  virtual bool hasMovableReg(void);
  virtual void moveReg(void);
  virtual int  printArgs(void);
};

class StArg : public Move
{
public:
  int   index;
  Arg  *arg;
  declare(StArg,1,1,1,1)(int i) { index= i; arg= 0; }
  virtual Loc *src(void) { return in[0]->loc; }
  virtual Loc *dst(void) { return arg->loc; }
  virtual void initStack(VPU *vpu);
  virtual void setOutputLoc(Loc *l);
  virtual void initOutput(VPU *vpu);
  virtual void typeCheck(VPU *vpu);
  virtual void initMode(VPU *vpu);
  inline  Loc *lhs(void) { return in.first->elt->loc; }
  virtual int  printArgs(void);
};

class StTemp : public Move
{
public:
  int   index;
  Temp *temp;
  declare(StTemp,1,1,1,1)(int i) { index= i; temp= 0; }
  virtual Loc *src(void) { return in[0]->loc; }
  virtual Loc *dst(void) { return temp->loc; }
  virtual void initStack(VPU *vpu);
  virtual void setOutputLoc(Loc *l);
  virtual void initOutput(VPU *vpu);
  virtual void typeCheck(VPU *vpu);
  virtual void initMode(VPU *vpu);
  inline  Loc *lhs(void) { return in.first->elt->loc; }
  virtual int  printArgs(void);
};

class Put : public Move
{
public:
  int   index;
  declare(Put,1,1,1,1)(int i) { index= i; }
  virtual Loc *src(void) { return in[0]->loc; }
  virtual Loc *dst(void) { return in[index]->loc; }
  virtual void setOutputLoc(Loc *l);
  virtual void initOutput(VPU *vpu);
  virtual void typeCheck(VPU *vpu);
  virtual void initMode(VPU *vpu);
  inline  Loc *lhs(void) { return in.first->elt->loc; }
  virtual int  printArgs(void);
};

class AddrArg: public Insn
{
public:
  Arg *var;
  int  index;
  declare(AddrArg,0,0,1,0)(int n) { var= 0; index= n; }
  virtual void initStack(VPU *vpu);
  virtual void setOutputLoc(Loc *l);
  virtual void typeCheck(VPU *vpu);
  virtual void selectInsn(VPU *vpu);
  virtual bool allocateRegisters(VPU *vpu);
  virtual void reload(VPU *vpu);
  virtual bool hasMode(VPU *vpu);
  virtual InsnMode *modes(void);
};

class AddrTemp: public Insn
{
public:
  Temp *var;
  int   index;
  declare(AddrTemp,0,0,1,0)(int n) { var= 0; index= n; }
  virtual void initStack(VPU *vpu);
  virtual void setOutputLoc(Loc *l);
  virtual void typeCheck(VPU *vpu);
  virtual void selectInsn(VPU *vpu);
  virtual bool allocateRegisters(VPU *vpu);
  virtual void reload(VPU *vpu);
  virtual bool hasMode(VPU *vpu);
  virtual InsnMode *modes(void);
};

class DropTemp : public Insn
{
public:
  Temp *temp;
  declare(DropTemp,0,0,0,0)(void) { temp= 0; }
  virtual void initStack(VPU *vpu);
  virtual void initMode(VPU *vpu);
  virtual bool allocateRegisters(VPU *vpu);
  virtual bool hasMode(VPU *vpu);
  virtual int  printArgs(void);
};


class LdFP: public Insn
{
public:
  declare(LdFP,0,0,1,0)(void) {}
  virtual void setOutputLoc(Loc *l);
  virtual void typeCheck(VPU *vpu);
  virtual void selectInsn(VPU *vpu);
  virtual bool allocateRegisters(VPU *vpu);
  virtual void reload(VPU *vpu);
  virtual bool hasMode(VPU *vpu);
  virtual InsnMode *modes(void);
};


class Ld : public Insn
{
public:
  int  refType;
  long long value;
  char *string;
  declare(Ld,0,0,1,0)(int t, long long i, char *s= 0)
  {
    refType= t;
    value=   i;
    string=  s;
  }
  virtual void setOutputLoc(Loc *l);
  virtual void typeCheck(VPU *vpu);
  virtual void initMode(VPU *vpu);
  virtual void selectInsn(VPU *vpu);
  virtual bool allocateRegisters(VPU *vpu);
  virtual void reload(VPU *vpu);
  virtual bool hasMode(VPU *vpu);
  virtual InsnMode *modes(void);
  virtual int  printArgs(void);
};


class Addr : public Ld
{
public:
  LocalLabel *label;
  declare(Addr,0,0,1,0)(LocalLabel *l) : Ld(LI4, 0xDeadBeef, 0) { label= l; }
  virtual int printArgs(void);
};

class AddrList : public List<Addr> {};
typedef Link<Addr> AddrLink;


class Drop : public IndexedInsn
{
public:
  declare(Drop,0,index,0,0)(int i) : IndexedInsn(i) {}
  virtual void initMode(VPU *vpu);
  virtual void selectInsn(VPU *vpu);
  virtual bool allocateRegisters(VPU *vpu);
  virtual void reload(VPU *vpu);
  virtual bool hasMode(VPU *vpu);
  inline  Loc *lhs(void) { return in.first->elt->loc; }
};


class Unary : public Insn
{
public:
  virtual void setOutputLoc(Loc *l);
  virtual void initOutput(VPU *vpu);
  virtual void typeCheck(VPU *vpu);
  virtual void selectInsn(VPU *vpu);
  virtual bool allocateRegisters(VPU *vpu);
  virtual void reload(VPU *vpu);
  virtual bool hasMode(VPU *vpu);
  virtual int  printArgs(void);
  inline  Loc *lhs(void) { return in.first->elt->loc; }
};


class Cvt : public Unary
{
public:
  int outType;
  declare(Cvt,1,1,1,0)(int t, Loc *l= 0) { loc= l; outType= t; }
  virtual void setOutputLoc(Loc *l);
  virtual void typeCheck(VPU *vpu);
  virtual void selectInsn(VPU *vpu);
  virtual bool allocateRegisters(VPU *vpu);
  virtual void reload(VPU *vpu);
  virtual bool hasMode(VPU *vpu);
  virtual InsnMode *modes(void);
  inline  Loc *lhs(void) { return in.first->elt->loc; }
  virtual int  printArgs(void);
};


class Binary : public Insn
{
public:
  virtual void setOutputLoc(Loc *l);
  virtual void initOutput(VPU *vpu);
  virtual void typeCheck(VPU *vpu);
  virtual void selectInsn(VPU *vpu);
  virtual void targetRegisters(VPU *vpu);
  virtual bool allocateRegisters(VPU *vpu);
  virtual void reload(VPU *vpu);
  virtual void moveRegs(void);
  virtual bool hasMode(VPU *vpu);
  inline  Loc *rhs(void) { return in.first->elt->loc; }
  inline  Loc *lhs(void) { return in.first->next->elt->loc; }
  virtual int  printArgs(void);
};

#define unary(WHO)				\
class WHO : public Unary			\
{						\
public:						\
  declare(WHO,1,1,1,1)(void) {}			\
  virtual InsnMode *modes(void);		\
}

#define binop(WHO)				\
class WHO : public Binary			\
{						\
public:						\
  declare(WHO,2,2,1,2)(void) {}			\
  virtual InsnMode *modes(void);		\
}

unary(SaveArgs);
unary(Alloc);

unary(Not);
unary(Neg);

binop(Add);
binop(Sub);
binop(Mul);
binop(Div);
binop(Mod);
binop(And);
binop(Or );
binop(Xor);
binop(Asl);
binop(Asr);
binop(Lsl);
binop(Lsr);
binop(Wrb);
binop(Wrh);
binop(Wrw);

binop(Rsub);	// subtract from
binop(Rdiv);	// divide into
binop(Rmod);	// modulo into
binop(Rasl);	// reversed shift left
binop(Rasr);	// reversed shift right
binop(Rlsl);	// reversed shift left
binop(Rlsr);	// reversed shift right

#undef unary
#undef binop

class Read : public Insn
{
public:
  virtual void setOutputLoc(Loc *l);
  virtual void typeCheck(VPU *vpu);
  virtual void selectInsn(VPU *vpu);
  virtual bool allocateRegisters(VPU *vpu);
  virtual void reload(VPU *vpu);
  virtual bool hasMode(VPU *vpu);
  virtual int  printArgs(void);
  inline  Loc *lhs(void) { return in.first->elt->loc; }
};

#define read(WHO)				\
class WHO : public Read				\
{						\
public:						\
  declare(WHO,1,1,1,1)(void) {}			\
  virtual InsnMode *modes(void);		\
}

read(Rdb);
read(Rdh);
read(Rdw);

#undef read

class Cmp : public Binary
{
public:
  enum { Lt, Le, Eq, Ne, Ge, Gt };
  static int reflect(int cc);
  static int invert(int cc);

  int op;
  declare(Cmp,2,2,1,0)(int o) { op= o; }
  inline  int  reflected(void) { return reflect(op); }
  inline  int  inverted(void) { return invert(op); }
  virtual void setOutputLoc(Loc *l);
  virtual void initOutput(VPU *vpu);
  virtual void typeCheck(VPU *vpu);
  virtual void targetRegisters(VPU *vpu);
  virtual void reload(VPU *vpu);
  virtual InsnMode *modes(void);
  virtual bool isCmp(void);
  inline  Loc *rhs(void) { return in.first->elt->loc; }
  inline  Loc *lhs(void) { return in.first->next->elt->loc; }
  virtual int  printArgs(void);
};


class BranchInsn : public IndexedInsn
{
public:
  LocalLabel *label;
  BranchInsn(int i, LocalLabel *l) : IndexedInsn(i) { label= l; }
  virtual void initStack(VPU *vpu);
  virtual void initFlow(VPU *vpu, Insn *prev);
  virtual int  printArgs(void);
};

class Br : public BranchInsn
{
public:
  declare(Br,0,0,0,0)(int i, LocalLabel *l) : BranchInsn(i, l) {}
  virtual void initMode(VPU *vpu);
  virtual bool hasMode(VPU *vpu);
  virtual InsnMode *modes(void);
  virtual bool isJump(void);
};

class Bc : public BranchInsn
{
public:
  enum { F, T };
  int cond;
  declare(Bc,1,1,0,0)(int c, int i, LocalLabel *l) : BranchInsn(i, l) { cond= c; }
  virtual void typeCheck(VPU *vpu);
  virtual void initMode(VPU *vpu);
  virtual void selectInsn(VPU *vpu);
  virtual bool allocateRegisters(VPU *vpu);
  virtual void reload(VPU *vpu);
  virtual bool hasMode(VPU *vpu);
  virtual InsnMode *modes(void);
  inline  Loc *lhs(void) { return in.first->elt->loc; }
  virtual int  printArgs(void);
};

class Bi : public Insn
{
public:
  declare(Bi,1,1,0,0)(void) {}
  virtual void typeCheck(VPU *vpu);
  virtual void initMode(VPU *vpu);
  virtual void selectInsn(VPU *vpu);
  virtual bool allocateRegisters(VPU *vpu);
  virtual void reload(VPU *vpu);
  virtual bool hasMode(VPU *vpu);
  virtual InsnMode *modes(void);
  inline  Loc *lhs(void) { return in.first->elt->loc; }
  virtual bool isJump(void);
};


class ArgOut : public Insn
{
public:
  bool  savesArg;
  int   mode;
  Reg  *reg[2];	// argument register(s)
  int   offset;	// frame offset
  declare(ArgOut,1,0,0,1)(Loc *l)
  {
    loc=            l;
    savesArg=       false;
    mode=           0;
    reg[0]= reg[1]= 0;
    offset=         0;
  }
  virtual void typeCheck(VPU *vpu);
  virtual void initMode(VPU *vpu);
  virtual bool allocateRegisters(VPU *vpu);
  virtual bool isSavedArg(void) { return savesArg; }
  virtual void reload(VPU *vpu);
  virtual void moveRegs(void);
  virtual bool hasMode(VPU *vpu);
  virtual InsnMode *modes(void);
  inline  Loc *lhs(void) { return in.first->elt->loc; }
  virtual bool isArgOut(void);
  virtual int  printArgs(void);
};


class Call : public Insn
{
public:
  int    resultType;
  int    nArgs;
  int	 argsSize;
  void  *func;
  Label *label;
  declare(Call,nArgs,nArgs,1,0)(int t, int n, void *f)
  {
    resultType= t;  nArgs= n;  argsSize= 0;  func= f;  label= 0;
  }
  Call(int t, int n, Label *l)
  {
    resultType= t;  nArgs= n;  argsSize= 0;  func= 0;  label= l;
  }
  virtual bool isIndirect(void) { return false; }
  virtual void initStack(VPU *vpu);
  virtual void setOutputLoc(Loc *l);
  virtual void initCall(VPU *vpu);
  virtual void typeCheck(VPU *vpu);
  virtual void selectInsn(VPU *vpu);
  virtual void constrainRegisters(VPU *vpu);
  virtual bool allocateRegisters(VPU *vpu);
  virtual void reload(VPU *vpu);
  virtual bool hasMode(VPU *vpu);
  virtual InsnMode *modes(void);
  virtual int  printArgs(void);
};


class Calli : public Call
{
public:
  declare(Calli,nArgs+1,nArgs+1,1,0)(int t, int n) : Call(t, n, (void *)0) {}
  virtual bool isIndirect(void) { return true; }
  virtual void selectInsn(VPU *vpu);
  virtual void reload(VPU *vpu);
  virtual bool hasMode(VPU *vpu);
  virtual InsnMode *modes(void);
  inline  Loc *lhs(void) { return in.first->elt->loc; }
};


class Ret : public Insn
{
public:
  declare(Ret,1,1,0,0)(void) {}
  virtual void typeCheck(VPU *vpu);
  virtual void initMode(VPU *vpu);
  virtual void selectInsn(VPU *vpu);
  virtual void targetRegisters(VPU *vpu);
  virtual bool allocateRegisters(VPU *vpu);
  virtual void reload(VPU *vpu);
  virtual bool isJump(void);
  virtual bool hasMode(VPU *vpu);
  virtual InsnMode *modes(void);
  inline  Loc *lhs(void) { return in.first->elt->loc; }
  virtual int  printArgs(void);
};


#endif // __vpu_Insns_h
