#ifndef __vpu_List_h
#define __vpu_List_h

#include <assert.h>
#include <stdio.h>

#include "Heap.h"

#include "error.h"


template <class Elt_t>
class Link : public Bulk
{
public:
  Elt_t		*elt;
  Link<Elt_t>	*next;

  Link(Elt_t *e) { elt= e; next= 0; }
};


template <class Elt_t>
class List : public Bulk
{
public:
  Link<Elt_t> *first;
  Link<Elt_t> *last;

  List(void) { first= last= 0; }

  inline bool isEmpty(void)
  {
    return first == 0;
  }

  inline int size(void)
  {
    int n= 0;
    for (Link<Elt_t> *l= first;  l;  l= l->next) ++n;
    return n;
  }

  inline Elt_t *at(int i)
  {
    Link<Elt_t> *l= first;
    while (i-- && (l= l->next));
    if (!l) fatal("index out of range");
    return l->elt;
  }

  inline Elt_t *operator [](int i) { return at(i); }

  Elt_t *addFirst(Heap *heap, Elt_t *elt)
  {
    Link<Elt_t> *link= new(heap) Link<Elt_t>(elt);
    if (first == 0)
      {
	assert(last == 0);
	first= last= link;
      }
    else
      {
	assert(last != 0);
	link->next= first;
	first= link;
      }
    return elt;
  }

  Elt_t *addLast(Heap *heap, Elt_t *elt)
  {
    Link<Elt_t> *link= new(heap) Link<Elt_t>(elt);
    if (first == 0)
      {
	assert(last == 0);
	first= last= link;
      }
    else
      {
	assert(last != 0);
	last->next= link;
	last= link;
      }
    return elt;
  }

  Elt_t *removeFirst(void)
  {
    assert(!isEmpty());
    Link<Elt_t> *link= first;
    first= first->next;
    if (first == 0)
      last= 0;
    return link->elt;
  }
};


#endif // __vpu_List_h
