#ifndef __vpu_Loc_h
#define __vpu_Loc_h

#include "Heap.h"

#include "types.h"

class VPU;
class Reg;
class Var;


class Loc : public Bulk
{
  static int nextSeq;
  int seq;

public:
  int	    type;	// this is ONLY used during typecheck
  int	    mode;
  bool	    shared;
  bool	    aliased;
  unsigned  clobbered[2];
  Reg	   *targeted[2];
  union {
    Reg      *reg[2];
    Var      *var;
    long long value;
  };

  Loc(int m= BOT, Reg *r= 0, Reg *q= 0)
  {
    seq=     	  nextSeq++;
    type=    	  BOT;
    mode=    	  m;
    shared=  	  false;
    aliased=  	  false;
    clobbered[0]= 0;
    clobbered[1]= 0;
    targeted[0]=  0;
    targeted[1]=  0;
    reg[0]=	  r;
    reg[1]=	  q;
  }

  Loc(int m, Var *v)
  {
    seq=     	  nextSeq++;
    type=    	  BOT;
    mode=    	  m;
    shared=  	  false;
    aliased=  	  false;
    clobbered[0]= 0;
    clobbered[1]= 0;
    targeted[0]=  0;
    targeted[1]=  0;
    var=	  v;
  }

  inline Loc &operator =(Loc &l)
  {
    seq=     	  nextSeq++;
    type=    	  l.type;
    mode=    	  l.mode;
    shared=  	  l.shared;
    aliased=  	  l.aliased;
    clobbered[0]= l.clobbered[0];
    clobbered[1]= l.clobbered[1];
    targeted[0]=  l.targeted[0];
    targeted[1]=  l.targeted[1];
    reg[0]=	  l.reg[0];
    reg[1]=	  l.reg[1];
    return *this;
  }

  inline Loc(Loc &l) { *this= l; }

  void     beRegister(Reg *r= 0, Reg *q= 0);
  void     beVariable(Var *v= 0);

  int      varLimit(int varClass);
  int      regClass(void);
  int      regSize(void);
  unsigned regMask(void);

  inline unsigned regMask(int clazz)
  {
    return (isReg() && (regClass() == clazz)) ? regMask() : 0;
  }

  void clobber(int clazz, unsigned mask, VPU *vpu);
  void target(Reg *r, Reg *q= 0);

  inline bool isLiteral(void) { return (modeStor(mode) == StorLit); }
  inline bool isBool(void)    { return (mode == CCR); }
  inline bool isReg(void)     { return (modeStor(mode) == StorReg) && !isBool(); }
  inline bool isVar(void)     { return (modeStor(mode) == StorVar); }

  static char *modeName(int t);
  inline char *modeName(void) { return modeName(mode); }
  static char *typeName(int t);
  inline char *typeName(void) { return typeName(type); }

  int  print(void);
  int  printSeq(void);
  int  printType(void);
  static int printMode(unsigned mode);
         int printMode(void);
  void println(void);
};


#endif // __vpu_Loc_h
