#ifndef __vpu_Reg_h
#define __vpu_Reg_h

#include "Heap.h"


class RegSet;

class Reg : public Bulk
{
public:
  int       number;	// ccg's register number
  unsigned  mask;	// mask within set
  RegSet   *set;	// owner
  char     *name;	// symbolic name

  Reg(int i, unsigned m= 0, RegSet *s= 0, char *n= 0)
  {
    number= i;
    mask=   m;
    set=    s;
    name=   n;
  }

  inline bool operator ==(Reg &other)
  {
    return  (set == other.set) && (number == other.number);
  }

  inline bool operator !=(Reg &other) { return !operator ==(other); }

  int  print(void);
  void println(void);
};


class RegSet : public Bulk
{
public:
  Reg      *regs[33];	// up to 32 registers
  int       count;	// number of registers in set
  unsigned  mask;	// union of all register masks in set
  unsigned  used;	// union of all register masks allocated since ctor

  RegSet(void);

  typedef struct { int i; char *name; } Spec;

  RegSet *define(Heap *heap, int i, char *name= 0);
  RegSet *define(Heap *heap, Spec *specs);
  Reg    *use(int rno);
  Reg    *remove(int rno);

  inline Reg *at(int n) { return regs[n]; }
  Reg *atMask(unsigned mask);
  Reg *atNumber(int rno);
  Reg *allocateIn(unsigned mask);

  int  print(void);
  void println(void);
};


#endif // __vpu_Reg_h
