{ import: token.st }
{ import: scan.st }

C99Evaluator : Object ( cpp tokens stack file line )

C99Evaluator note: message
[
"
    StdOut nextPutAll: message.
    tokens notEmpty ifTrue: [StdOut tab; tab; putText: tokens].
    StdOut cr
"
]

C99Evaluator error: message
[
    super error: file, ':', line printString, ': ', message
]

C99Evaluator on: tokenList in: aPreprocessor
[
    self := self new.
    tokens := tokenList.
    cpp := aPreprocessor.
    stack := OrderedCollection new.
    file := tokens first file.
    line := tokens first line.
]

C99TokenList evaluateIn: cpp
[
    | result |
    self isEmpty ifTrue: [^0].
    result := (C99Evaluator on: self removeBlanks in: cpp) constantExpression; result.
    self isEmpty ifFalse: [self first error: 'junk after expression: ', self spelling].
    ^result
]

C99Evaluator warnings		[ ^cpp options warnings ]

C99Evaluator push: aNumber	[ ^stack addLast: aNumber ]
C99Evaluator pop		[ ^stack removeLast ]

C99Evaluator result
[
    stack size == 1 ifFalse: [self error: 'stack size is ', stack size printString, ' after evaluation'].
    ^stack removeLast
]

"   A.1.5 Constants 

    (6.4.4) constant: 
		integer-constant 
		floating-constant 
		enumeration-constant 
		character-constant "

C99Number constantValue
[
    | val dv base pos |
    val := 0.
    pos := 0.
    base := 10.
    (text beginsWith: '0x')
	ifTrue:  [base := 16.  pos := 2]
	ifFalse: [(text beginsWith: '0')
		      ifTrue:  [base := 8]].
    text from: pos do: [:c |
	(c == $L or: [c == $l or: [c == $U or: [c == $u]]])
	    ifFalse:
	       [dv := c digitValue.
		(dv isNil or: [dv >= base]) ifTrue: [self error: 'illegal digit ''', (String with: c), ''' in integer constant'].
		val := val * base + dv]].
    ^val
]

C99Character constantValue
[
    | val |
    val := 0.
    (C99Evaluator scanString: text) do: [:c | val := val * 256 + c].
    ^val
]

C99Evaluator scanString: aString
[
    | in out c |
    in := (aString copyFrom: 1 to: aString size - 2) readStream.
    out := WriteStream on: (aString new: aString size).
    [in atEnd]
	whileFalse:
	   [out nextPut: ((c := in next) == $\\ ifTrue: [self scanEscape: in] ifFalse: [c])].
    ^out contents
]

C99Evaluator scanEscape: aStream
[
    | c |
    aStream atEnd ifTrue: [^$\\].
    c := aStream peek.
    (c between: $0 and: $7)	ifTrue: [^self scanEscape: aStream radix: 8].
    (c == $x)			ifTrue: [^self scanEscape: (aStream next; yourself) radix: 16].
    ^c
]

C99Evaluator scanEscape: aStream radix: base
[
    | c val dv |
    val := 0.
    aStream atEnd ifTrue: [^self error: 'missing digits in numeric escape sequence'].
    [aStream atEnd or: [(dv := aStream peek digitValue) == -1 or: [dv >= base]]]
	whileFalse:
	   [val := val * base + dv.
	    aStream next].
    ^val
]

C99Evaluator constant
[
    | token |
    tokens isEmpty ifTrue: [self error: 'constant expression required'].
    self note: 'constant'.
    tokens removeFirst evalIn: self.
]

C99Token	evalIn: eval	[ self error: 'illegal expression' ]
C99Number	evalIn: eval	[ eval push: self constantValue ]
C99Character	evalIn: eval	[ eval push: self constantValue ]
C99Identifier	evalIn: eval	[ eval warnings ifTrue: [self warn: 'undefined name ''', text, ''' treated as 0'].  eval push: 0 ]

"   A.2.1 Expressions 

    (6.5.1) primary-expression: 
    		identifier 
    		constant 
    		string-literal 
    		( expression ) "

C99Evaluator primaryExpression
[
    self note: 'primary'.
    (tokens notEmpty and: [tokens first isOpenParen])
	ifTrue:
	   [tokens removeFirst.
	    self expression.
	    (tokens notEmpty and: [(tokens removeFirst) isCloseParen])
		ifFalse: [self error: ') expected']]
	ifFalse: [self constant]
]

"   (6.5.2) postfix-expression: 
    		primary-expression 
    		postfix-expression [ expression ] 
    		postfix-expression ( argument-expression-listopt ) 
    		postfix-expression . identifier 
    		postfix-expression -> identifier 
    		postfix-expression ++ 
    		postfix-expression -- 
    		( type-name ){ initializer-list } 
    		( type-name ){ initializer-list , } "

C99Evaluator postfixExpression
[
    self note: 'postfix'.
    self primaryExpression
]

"   (6.5.2) argument-expression-list: 
    		assignment-expression 
    		argument-expression-list , assignment-expression "

"   (6.5.3) unary-expression: 
    		postfix-expression 
    		++ unary-expression 
    		-- unary-expression 
    		unary-operator cast-expression 
    		sizeof unary-expression 
    		sizeof ( type-name ) "

"   (6.5.3) unary-operator: one of 
    		& * + - ~ ! "

C99Evaluator unaryExpression
[
    self note: 'unary'.
    (tokens notEmpty and: [tokens first isUnary])
	ifTrue:  [tokens removeFirst evalUnaryIn: self castExpression]
	ifFalse: [self postfixExpression]
]

C99Token	evalUnaryIn: eval	[ self error: 'illegal expression' ]
C99And		evalUnaryIn: eval	[ self error: 'illegal expression' ]
C99Multiply	evalUnaryIn: eval	[ self error: 'illegal expression' ]
C99Add		evalUnaryIn: eval	[ ]
C99Subtract	evalUnaryIn: eval	[ eval push: eval pop negated ]
C99Complement	evalUnaryIn: eval	[ eval push: -1 - eval pop ]
C99Not		evalUnaryIn: eval	[ eval push: (eval pop == 0 ifTrue: [1] ifFalse: [0]) ]

"   (6.5.4) cast-expression: 
    		unary-expression 
    		(type-name )cast-expression "

C99Evaluator castExpression
[
    self note: 'cast'.
    self unaryExpression
]

"   (6.5.5) multiplicative-expression: 
    		cast-expression 
    		multiplicative-expression * cast-expression 
    		multiplicative-expression / cast-expression 
    		multiplicative-expression % cast-expression "

C99Evaluator multiplicativeExpression
[
    self note: 'multiplicative'.
    self castExpression.
    [tokens notEmpty and: [tokens first isMultiplicative]]
	whileTrue:
	   [tokens removeFirst evalIn: self castExpression]
]

C99Multiply	evalIn: eval	[ eval push: eval pop * eval pop ]
C99Divide	evalIn: eval	[ | l r | r := eval pop.  l := eval pop.  eval push: l // r ]
C99Modulus	evalIn: eval	[ | l r | r := eval pop.  l := eval pop.  eval push: l \\ r ]

"   (6.5.6) additive-expression: 
    		multiplicative-expression 
    		additive-expression + multiplicative-expression 
    		additive-expression - multiplicative-expression "

C99Evaluator additiveExpression
[
    self note: 'additive'.
    self multiplicativeExpression.
    [tokens notEmpty and: [tokens first isAdditive]]
	whileTrue:
	   [tokens removeFirst evalIn: self multiplicativeExpression]
]

C99Add		evalIn: eval	[ | l r | r := eval pop.  l := eval pop.  eval push: l + r ]
C99Subtract	evalIn: eval	[ | l r | r := eval pop.  l := eval pop.  eval push: l - r ]

"   (6.5.7) shift-expression: 
    		additive-expression 
    		shift-expression << additive-expression 
    		shift-expression >> additive-expression "

C99Evaluator shiftExpression
[
    self note: 'shift'.
    self additiveExpression.
    [tokens notEmpty and: [tokens first isShift]]
	whileTrue:
	   [tokens removeFirst evalIn: self additiveExpression]
]

C99LeftShift	evalIn: eval	[ | l r | r := eval pop.  l := eval pop.  eval push: l << r ]
C99RightShift	evalIn: eval	[ | l r | r := eval pop.  l := eval pop.  eval push: l >> r ]

"   (6.5.8) relational-expression: 
    		shift-expression 
    		relational-expression < shift-expression 
    		relational-expression > shift-expression 
    		relational-expression <= shift-expression 
    		relational-expression >= shift-expression "

C99Evaluator relationalExpression
[
    self note: 'relational'.
    self shiftExpression.
    [tokens notEmpty and: [tokens first isRelational]]
	whileTrue:
	   [tokens removeFirst evalIn: self shiftExpression]
]

C99Less		evalIn: eval	[ | l r | r := eval pop.  l := eval pop.  eval push: (l <  r ifTrue: [1] ifFalse: [0]) ]
C99LessEqual	evalIn: eval	[ | l r | r := eval pop.  l := eval pop.  eval push: (l <= r ifTrue: [1] ifFalse: [0]) ]
C99Greater	evalIn: eval	[ | l r | r := eval pop.  l := eval pop.  eval push: (l >  r ifTrue: [1] ifFalse: [0]) ]
C99GreaterEqual	evalIn: eval	[ | l r | r := eval pop.  l := eval pop.  eval push: (l >= r ifTrue: [1] ifFalse: [0]) ]

"   (6.5.9) equality-expression: 
    		relational-expression 
    		equality-expression == relational-expression 
    		equality-expression != relational-expression "

C99Evaluator equalityExpression
[
    self note: 'equality'.
    self relationalExpression.
    [tokens notEmpty and: [tokens first isEquality]]
	whileTrue:
	   [tokens removeFirst evalIn: self relationalExpression]
]

C99Equal	evalIn: eval	[ | l r | r := eval pop.  l := eval pop.  eval push: (l == r ifTrue: [1] ifFalse: [0]) ]
C99NotEqual	evalIn: eval	[ | l r | r := eval pop.  l := eval pop.  eval push: (l ~~ r ifTrue: [1] ifFalse: [0]) ]

"   (6.5.10) AND-expression: 
    		equality-expression 
    		AND-expression & equality-expression "

C99Evaluator andExpression
[
    self note: 'and'.
    self equalityExpression.
    [tokens notEmpty and: [tokens first isAnd]]
	whileTrue:
	   [tokens removeFirst evalIn: self equalityExpression]
]

C99And	evalIn: eval	[ | l r | r := eval pop.  l := eval pop.  eval push: (l bitAnd: r) ]

"   (6.5.11) exclusive-OR-expression: 
    		AND-expression 
    		exclusive-OR-expression ^ AND-expression "

C99Evaluator exclusiveOrExpression
[
    self note: 'xor'.
    self andExpression.
    [tokens notEmpty and: [tokens first isXor]]
	whileTrue:
	   [tokens removeFirst evalIn: self andExpression]
]

C99Xor	evalIn: eval	[ | l r | r := eval pop.  l := eval pop.  eval push: (l bitXor: r) ]

"   (6.5.12) inclusive-OR-expression: 
    		exclusive-OR-expression 
    		inclusive-OR-expression | exclusive-OR-expression "

C99Evaluator inclusiveOrExpression
[
    self note: 'or'.
    self exclusiveOrExpression.
    [tokens notEmpty and: [tokens first isOr]]
	whileTrue:
	   [tokens removeFirst evalIn: self exclusiveOrExpression]
]

C99Or	evalIn: eval	[ | l r | r := eval pop.  l := eval pop.  eval push: (l bitOr: r) ]

"   (6.5.13) logical-AND-expression: 
    		inclusive-OR-expression 
    		logical-AND-expression && inclusive-OR-expression "

C99Evaluator logicalAndExpression
[
    self note: 'logand'.
    self inclusiveOrExpression.
    [tokens notEmpty and: [tokens first isLogAnd]]
	whileTrue:
	   [tokens removeFirst evalIn: self inclusiveOrExpression]
]

C99LogAnd evalIn: eval	[ | l r | r := eval pop.  l := eval pop.  eval push: ((l ~~ 0 and: [r ~~ 0]) ifTrue: [1] ifFalse: [0]) ]

"   (6.5.14) logical-OR-expression: 
    		logical-AND-expression 
    		logical-OR-expression || logical-AND-expression "

C99Evaluator logicalOrExpression
[
    self note: 'logor'.
    self logicalAndExpression.
    [tokens notEmpty and: [tokens first isLogOr]]
	whileTrue:
	   [tokens removeFirst evalIn: self logicalAndExpression]
]

C99LogOr evalIn: eval	[ | l r | r := eval pop.  l := eval pop.  eval push: ((l ~~ 0 or: [r ~~ 0]) ifTrue: [1] ifFalse: [0]) ]

"   (6.5.15) conditional-expression: 
    		logical-OR-expression 
    		logical-OR-expression ? expression : conditional-expression "

C99Evaluator conditionalExpression
[
    | op |
    self note: 'cond'.
    self logicalOrExpression.
    (tokens notEmpty and: [tokens first isQuery])
	ifTrue:
	   [op := tokens removeFirst.
	    self expression.
	    (tokens notEmpty and: [tokens removeFirst isColon]) ifFalse: [self error: ': missing in conditional expression'].
	    op evalIn: self conditionalExpression].
]

C99Query evalIn: eval	[ | c l r | r := eval pop.  l := eval pop.  c := eval pop.   eval push: (c ~~ 0 ifTrue: [l] ifFalse: [r]) ]

"   (6.5.16) assignment-expression: 
    		conditional-expression 
    		unary-expression assignment-operator assignment-expression "

C99Evaluator assignmentExpression
[
    self conditionalExpression
]

"   (6.5.16) assignment-operator: one of 
    		= *= /= %= += -= <<= >>= &= = |= "

"   (6.5.17) expression: 
    		assignment-expression 
    		expression , assignment-expression "

C99Evaluator expression
[
    | tok |
    self note: 'expression'.
    self assignmentExpression.
    [tokens notEmpty and: [tokens first isComma]]
	whileTrue:
	   [tok := tokens removeFirst.
	    self warnings ifTrue: [tok warn: 'comma operator in constant expression'].
	    self pop; assignmentExpression]
]

"   (6.6) constant-expression: 
    		conditional-expression "

C99Evaluator constantExpression
[
    self note: 'constantExpression'.
    ^self conditionalExpression
]
