{ import: scan }
{ import: eval }

C99PreprocessorOptions : Object ( userSearchPaths systemSearchPaths warnings )

C99PreprocessorOptions userSearchPaths		[ ^userSearchPaths   ]
C99PreprocessorOptions systemSearchPaths	[ ^systemSearchPaths ]

C99PreprocessorOptions new
[
    self := super new.
    userSearchPaths := OrderedCollection new.
    systemSearchPaths := OrderedCollection new.
    warnings := false.
]

C99PreprocessorOptions default
[
    self := self new.
    systemSearchPaths addAll: self defaultSystemSearchPaths.
    userSearchPaths   addAll: self defaultUserSearchPaths.
]

C99PreprocessorOptions warnings: flag	[  warnings := flag ]
C99PreprocessorOptions warnings		[ ^warnings ]

C99PreprocessorOptions defaultSystemSearchPaths	[ ^#('/usr/include' '/usr/include/X11' '/usr/X11R6/include' '/usr/local/include') ]
C99PreprocessorOptions defaultUserSearchPaths	[ ^#('.') ]

C99PreprocessorOptions addUserSearchPath: pathString		[ ^userSearchPaths   addFirst: pathString ]
C99PreprocessorOptions addSystemSearchPath: pathString		[ ^systemSearchPaths addLast:  pathString ]


C99Preprocessor : C99TokenList ( options macros )

C99Preprocessor withOptions: preprocessorOptions
[
    self := super new.
    options := preprocessorOptions.
    macros := FastIdentityDictionary new.
]

C99Preprocessor options	[ ^options ]

" 5.1.1.2  Translation phases

  4. Preprocessing directives are executed, macro invocations are expanded, and 
     _Pragma unary operator expressions are executed.  If a character sequence that 
     matches the syntax of a universal character name is produced by token 
     concatenation (6.10.3.3), the behavior is undefined.  A #include preprocessing 
     directive causes the named header or source file to be processed from phase 1 
     through phase 4, recursively.  All preprocessing directives are then deleted. "

C99Preprocessor nextToken
[
   | token |
   [ self isEmpty ifTrue: [^nil].
     (token := self removeFirst processBefore: self in: self) isNil ] whileTrue.
   ^token
]


C99Token processBefore: tokens in: cpp	[ ]


C99Untaint : C99Token ()

C99Untaint isUntaint	[ ^true ]
C99Token isUntaint	[ ^false ]

C99Untaint with: aMacro
[
    self := super new.
    text := '<UNTAINT>'.
    note := aMacro.
]

C99Untaint processUntaint	[ note untaint ]


C99Untaint processBefore: tokens in: cpp
[
    note untaint.
    ^nil
]


C99DirectiveInclude processBefore: tokens in: cpp
[
    body removeOuterBlanks first isIdentifier ifTrue: [(body := body expandedIn: cpp) removeOuterBlanks combineBlanks].
    body size > 2
	ifTrue:
	   [(body first isLess and: [body last isGreater])
		ifTrue:
		   [cpp sysInclude: (body removeFirst; removeLast; spelling) before: tokens.
		    ^nil].
	    (body first text == '"' and: [body last text == '"'])
		ifTrue:
		   [cpp usrInclude: (body removeFirst; removeLast; spelling) before: tokens.
		    ^nil]]
	ifFalse:
	   [(body notEmpty and: [body first isC99String])
		ifTrue:
		    [cpp usrInclude: body first withoutQuotes before: tokens.
		     ^nil]].
    (body isEmpty ifTrue: [self] ifFalse: [body first])
	error: '#include requires <path> or "path"'.
    ^nil
]

C99String withoutQuotes
[
    ^text copyFrom: 1 to: text size - 2
]

C99TokenList spelling
[
    | text |
    text := WriteStream on: (String new: 32).
    self do: [:token | text nextPutAll: token text asString].
    ^text contents
]

C99Preprocessor sysInclude: path before: tokens	[ self include: path before: tokens user: false ]
C99Preprocessor usrInclude: path before: tokens	[ self include: path before: tokens user: true ]

C99Preprocessor include: path before: tokens user: userFlag
[
    ((userFlag and: [self include: path in: options userSearchPaths before: tokens])
	    or: [self include: path in: options systemSearchPaths before: tokens])
	ifFalse: [self error: 'cannot find included file: ', path]
]

C99Preprocessor include: file in: dirs before: tokens
[
    | path |
    dirs do: [:dir | (self include: dir , '/', file before: tokens)ifTrue: [^true]].
    ^false
]

C99Preprocessor include: path before: tokens
[
    | file scanner fileTokens token |
    (file := File open: path ifAbsent: []) isNil ifTrue: [^false].
    scanner := C99Scanner on: file.
    fileTokens := C99TokenList new.
    [(token := scanner nextToken) isNil] whileFalse: [fileTokens addLast: token].
    file close.
    tokens addAllFirst: fileTokens.
    ^true
]

C99DirectiveLine    processBefore: tokens in: cpp	[ self warn: 'unimplemented directive: ', name text. ^nil ]

C99DirectiveError   processBefore: tokens in: cpp
[
    self error: body spelling
]

C99DirectivePragma  processBefore: tokens in: cpp	[ self warn: 'unimplemented directive: ', name text. ^nil ]

" conditional sections "

" 6.10.1  Conditional inclusion

  5 Each directive's condition is checked in order. If it evaluates to false (zero), the group 
    that it controls is skipped: directives are processed only through the name that determines 
    the directive in order to keep track of the level of nested conditionals; the rest of the 
    directives' preprocessing tokens are ignored, as are the other preprocessing tokens in the 
    group.  Only the first group whose control condition evaluates to true (nonzero) is 
    processed.  If none of the conditions evaluates to true, and there is a #else directive, the 
    group controlled by the #else is processed; lacking a #else directive, all the groups 
    until the #endif are skipped. "

C99DirectiveIf  processBefore: tokens in: cpp	[ cpp processIf: self.  ^nil ]	"inherited by dirIfdef, dirIfndef"

C99Preprocessor processIf: token
[
    [token isDirectiveEndif or: [token evaluateConditionIn: self]] whileFalse: [token := self skipTo: token nextGroup].
]


C99DirectiveElif processBefore: tokens in: cpp	[ cpp processElse: self.  ^nil ]
C99DirectiveElse processBefore: tokens in: cpp	[ cpp processElse: self.  ^nil ]

C99Preprocessor processElse: token
[
    [token isDirectiveEndif] whileFalse: [token := self skipTo: token nextGroup].
]

C99DirectiveEndif processBefore: tokens in: cpp	[ ^nil ]


C99Preprocessor skipTo: token
[
    | currentToken |
    [currentToken == token or: [self isEmpty]] whileFalse: [currentToken := self removeFirst].
    currentToken == token ifFalse: [token error: 'missing #endif'].
    ^currentToken		"token will be returned to client unless discarded"
]


C99DirectiveIf evaluateConditionIn: cpp
[
    | val |
    body removeOuterBlanks isEmpty ifTrue: [self error: '#if with no expression'].
    ^((body substituteDefinedIn: cpp; expandMacrosIn: cpp) substituteDefinedIn: cpp; evaluateIn: cpp) ~~ 0
]

C99DirectiveElif evaluateConditionIn: cpp
[
    | id |
    body removeOuterBlanks isEmpty ifTrue: [self error: '#elif with no expression'].
    ^((body substituteDefinedIn: cpp; expandMacrosIn: cpp) substituteDefinedIn: cpp; evaluateIn: cpp) ~~ 0
]

C99DirectiveIfdef evaluateConditionIn: cpp
[
    body removeBlanks.
    (body notEmpty and: [body first isIdentifier])	ifFalse: [self error: '#', text, ' requires an identifier'].
    (body size == 1)					ifFalse: [self warn: 'ignoring junk after #', text].
    ^(cpp lookup: body first text) notNil
]

C99DirectiveIfndef evaluateConditionIn: cpp	[ ^(super evaluateConditionIn: cpp) not ]
C99DirectiveElse   evaluateConditionIn: cpp	[ ^true ]

C99Preprocessor lookup: name	[ ^macros at: name ifAbsent: [] ]


C99TokenList substituteDefinedIn: cpp
[
    self removeBlanks.
    self do: [:tok | tok substituteDefinedIn: cpp].
]

C99Token substituteDefinedIn: cpp	[ ]

C99Identifier substituteDefinedIn: cpp
[
    | id |
    text == #defined ifFalse: [^self].
    nextLink isIdentifier
	ifFalse:
	   [(nextLink isOpenParen and: [nextLink nextLink isIdentifier and: [nextLink nextLink nextLink isCloseParen]])
		ifTrue:
		   [nextLink unlink.
		    nextLink nextLink unlink]
		ifFalse:
		   [self error: 'operator "defined" requires an identifier']].
    nextLink replaceWithNumber: ((cpp lookup: nextLink text) ifTrue: ['1'] ifFalse: ['0']).
    self unlink.
]

C99Identifier replaceWithNumber: numberText
[
    self replaceWith: (C99Number withText: numberText file: file line: line)
]

" macro un/definition "

C99DirectiveUndef processBefore: tokens in: cpp
[
    body removeOuterBlanks combineBlanks.
    (body notEmpty and: [body first isIdentifier])	ifFalse: [self error: '#undef requires an identifier'].
    cpp undefine: body removeFirst.
    body isEmpty					ifFalse: [self error: 'ignoring junk after #undef'].
    ^nil
]

C99Preprocessor undefine: id
[
    macros removeKey: id text ifAbsent: [options warnings ifTrue: [id warn: '#undef argument was not defined']]
]

C99DirectiveDefine processBefore: tokens in: cpp
[
    cpp define: (self createMacro: tokens).
    ^nil
]

C99Preprocessor define: macro
[
    | defn |
    defn := macros at: macro name text ifAbsent: [macros at: macro name text put: macro].
    macro = defn ifFalse: [macro name error: 'macro ', macro name text, ' redefined'].
    ^nil
]

C99DirectiveDefine createMacro: tokens
[
    | id |
    body removeOuterBlanks combineBlanks.
    (body notEmpty and: [body first isIdentifier])	ifFalse: [self error: '#define requires an identifier'].
    id := body removeFirst.
    ^((body notEmpty and: [body first isOpenParen]) ifTrue: [C99FunctionMacro] ifFalse: [C99ObjectMacro])
	withName:   id
	definition: body
]

C99Macro : Object ( tainted name replacementList )

C99Macro untaint		[ tainted := false ]
C99Macro name			[ ^name ]
C99Macro replacementList	[ ^replacementList ]

C99Macro withName: id definition: tokenList
[
    self := self new.
    tainted := false.
    name := id.
    replacementList := tokenList removeLeadingBlanks.
    replacementList isEmpty ifTrue: [replacementList addLast: (C99Blank withText: ' ' file: id file line: id line)].
]

C99ObjectMacro : C99Macro ()

C99ObjectMacro	isC99ObjectMacro [ ^true ]
C99Macro	isC99ObjectMacro [ ^false ]

C99ObjectMacro = aMacro
[
    ^self == aMacro or: [aMacro isC99ObjectMacro and: [replacementList = aMacro replacementList]]
]

C99ObjectMacro printOn: aStream
[
    aStream nextPutAll: 'ObjMac('; print: name; nextPutAll: ' -> '; print: replacementList; nextPut: $)
]


C99FunctionMacro : C99Macro ( argumentList )

C99FunctionMacro isC99FunctionMacro [ ^true ]
C99Macro	 isC99FunctionMacro [ ^false ]

C99FunctionMacro argumentList	[ ^argumentList ]

C99FunctionMacro withName: id definition: tokenList
[
    | args |
    args := C99MacroArgumentList fromTokens: tokenList in: id.
    args variadic ifTrue: [tokenList noteVariadic: args variadic].	"xxx GNU xxx"
    self := super withName: id definition: tokenList.
    argumentList := args
]

C99TokenList  noteVariadic: symbol	[ self do: [:tok | tok noteVariadic: symbol] ]	"xxx GNU xxx"
C99Token      noteVariadic: symbol	[ ]						"xxx GNU xxx"
C99Identifier noteVariadic: symbol	[ text == symbol ifTrue: [note := #variadic] ]	"xxx GNU xxx"


C99FunctionMacro = aMacro
[
    ^self == aMacro or:	[aMacro isC99FunctionMacro
			     and: [replacementList = aMacro replacementList
			     and: [argumentList = aMacro argumentList]]]
]

C99FunctionMacro printOn: aStream
[
    aStream nextPutAll: 'FunMac('; print: name; space; print: argumentList; nextPutAll: ' -> '; print: replacementList; nextPut: $)
]


C99MacroArgumentList : C99TokenList ( arity variadic )

C99MacroArgumentList variadic	[ ^variadic ]

" 6.1  Preprocessing directives

  control-line: 
	#define identifier replacement-list new-line 
	#define identifier lparen identifier-list-opt ) replacement-list new-line 
	#define identifier lparen ... ) replacement-list new-line 
	#define identifier lparen identifier-list , ... ) replacement-list new-line 

  identifier-list: 
	identifier 
	identifier-list , identifier "

C99MacroArgumentList fromTokens: tokenList in: id
[
    | token |
    self := self new.
    (tokenList notEmpty and: [tokenList removeFirst isOpenParen])	ifFalse: [self error: '( went AWOL'].
    self addFormalArgumentsFrom: tokenList in: id.
    (tokenList notEmpty and: [tokenList removeFirst isCloseParen])	ifFalse: [id expected: ') after formal arguments'].
]

C99MacroArgumentList addFormalArgumentsFrom: tokens in: id
[
    | token comma |
    arity := 0.
    variadic := nil.
    (tokens removeLeadingBlanks isEmpty or: [tokens first isCloseParen])
	ifTrue: [^self].
    [tokens notEmpty]
	whileTrue:
	   [token := tokens removeFirst.
	    tokens removeLeadingBlanks.
	    token isEllipsis ifTrue: [^variadic := #__VA_ARGS__].
	    token isIdentifier							ifFalse: [token expected: 'formal argument name'].
	    self addLast: token.
	    arity := arity + 1.
	    (tokens isEmpty or: [tokens first isCloseParen])
		ifTrue: [^self].
	    token := tokens removeFirst.
	    token isEllipsis					"xxx GNU xxx"
		ifTrue:
		   [arity := arity - 1.
		    ^variadic := self removeLast text].
	    (comma := token) isComma						ifFalse: [token expected: ', after formal argument'].
	    tokens removeLeadingBlanks]
]

C99MacroArgumentList printOn: aStream
[
    aStream nextPutAll: '( '.
    self do: [:tok | aStream print: tok text; space].
    variadic ifTrue: [aStream nextPutAll: '... '].
    aStream nextPutAll: ')'.
]

" macro expansion -- objects "

C99Identifier processBefore: tokens in: cpp
[
    | macro |
    (macro := cpp lookup: text) isNil ifTrue: [^self].
    ^(macro expand: self before: tokens in: cpp)
	ifTrue:  [nil]
	ifFalse: [self]
]

C99ObjectMacro expand: identifier before: tokens in: cpp
[
    | replacement |
    identifier note == #tainted ifTrue: [^false].
    tainted ifTrue: [identifier note: #tainted.  ^false].
    replacement := replacementList copyAllWithFile: identifier file line: identifier line.
    replacement pasteTokens.
    replacement removePlacemarkers.
    replacement combineBlanks.
    replacement addLast: (C99Untaint with: self).
    tokens addAllFirst: replacement.
    tainted := true.
    ^true
]

" 6.10.3.3 The ## operator 

  1 A ## preprocessing token shall not occur at the beginning or at the end of a replacement 
    list for either form of macro definition. "

C99TokenList pasteTokens	[ self do: [:token | token pasteTokens] ]

C99Token pasteTokens		[]

C99Paste pasteTokens
[
    | scanner token |
    [prevLink isBlank]	whileTrue: [prevLink unlink].
    prevLink isNullLink	ifTrue: [self error: '## cannot appear at the start of a replacement list'].
    prevLink isUntaint	ifTrue: [self error: 'untaint cannot appear before ##'].
    [nextLink isBlank]	whileTrue: [nextLink unlink].
    nextLink isNullLink	ifTrue: [self error: '## cannot appear at the start of a replacement list'].
    nextLink isUntaint	ifTrue: [self error: 'untaint cannot appear before ##'].
    (prevLink isPlacemarker)	ifTrue: [prevLink unlink.  ^self unlink].
    (nextLink isPlacemarker)	ifTrue: [nextLink unlink.  ^self unlink].
    note == #variadic		ifTrue: [^self unlink].				"xxx GNU xxx"
    scanner := C99Rescanner on: (prevLink spelling, nextLink spelling).
    token := scanner nextToken.
    (token notNil and: [scanner next isNil])
	ifFalse: [self error: 'pasting "', prevLink text, '" and "', nextLink text, '" does not yield a valid preprocessing token'].
    token line: line.
    prevLink unlink.
    nextLink unlink.
    (prevLink nextLink: token) prevLink: prevLink.
    (nextLink prevLink: token) nextLink: nextLink.
    self offList.
]

C99Placemarker : C99Token ()

C99Placemarker isPlacemarker	[ ^true ]
C99Token isPlacemarker		[ ^false ]

" macro expansion -- functions "

" 6.10.3.4  Rescanning and further replacement

  1 After all parameters in the replacement list have been substituted and # and ## 
    processing has taken place, all placemarker preprocessing tokens are removed. Then, the 
    resulting preprocessing token sequence is rescanned, along with all subsequent 
    preprocessing tokens of the source file, for more macro names to replace. "

C99FunctionMacro expand: identifier before: tokens in: cpp
[
    | arguments replacement |
    identifier note == #tainted ifTrue: [^false].
    tainted ifTrue: [identifier note: #tainted.  ^false].
    ((replacement := tokens firstNonBlank) notNil and: [replacement isOpenParen]) ifFalse: [^false].
    [self ignoreBlank: tokens] whileTrue.
    arguments := argumentList makeActualArguments: tokens in: cpp.
    replacement := replacementList copyAllWithFile: identifier file line: identifier line.
    replacement substituteArguments: arguments.
    replacement pasteTokens.
    replacement removePastes.
    replacement removePlacemarkers.
    replacement combineBlanks.
    replacement addLast: (C99Untaint with: self).
    tokens addAllFirst: replacement.
    tainted := true.
    ^true.
]

C99TokenList firstNonBlank
[
    self do: [:token | (token isBlank or: [token isUntaint]) ifFalse: [^token]].
    ^nil
]

C99FunctionMacro ignoreBlank: tokens
[
    tokens isEmpty ifTrue: [^false].
    tokens first isBlank ifTrue: [tokens removeFirst.  ^true].
    tokens first isUntaint ifTrue: [tokens removeFirst processUntaint.  ^true].
    ^false
]

C99TokenList	pasteOrStringifyTokens	[ self do: [:token | token pasteOrStringifyTokens] ]
C99Token	pasteOrStringifyTokens	[ ]
C99Paste	pasteOrStringifyTokens	[ self pasteTokens ]
C99Hash		pasteOrStringifyTokens	[ self stringifyTokens ]

C99TokenList	removePastes		[ self do: [:token | token removePastes] ]
C99Token	removePastes		[ ]
C99Paste	removePastes		[ self unlink ]

C99TokenList	removePlacemarkers	[ self do: [:token | token removePlacemarkers] ]
C99Token	removePlacemarkers	[ ]
C99Placemarker	removePlacemarkers	[ self unlink ]

File putText: obj [ obj putTextOn: self ]
SinkStream putText: obj [ obj putTextOn: self ]

" 6.10.3  Macro replacement

  11 The sequence of preprocessing tokens bounded by the outside-most matching parentheses 
     forms the list of arguments for the function-like macro. The individual arguments within 
     the list are separated by comma preprocessing tokens, but comma preprocessing tokens 
     between matching inner parentheses do not separate arguments.  If there are sequences of 
     preprocessing tokens within the list of arguments that would otherwise act as 
     preprocessing directives, the behavior is undefined. 

  12 If there is a ... in the identifier-list in the macro definition, then the trailing arguments, 
     including any separating comma preprocessing tokens, are merged to form a single item: 
     the variable arguments.  The number of arguments so combined is such that, following 
     merger, the number of arguments is one more than the number of parameters in the macro 
     definition (excluding the ...). "

C99MacroArgument : C99TokenList ( expanded placemarker )

C99MacroArgument expanded	[ ^expanded ]

C99MacroArgumentList makeActualArguments: tokenList in: cpp
[
    | actuals argument rest tok1 tok2 |
    actuals := IdentityDictionary new.
    (tokenList notEmpty and: [(tok1 := tokenList removeFirst) isOpenParen]) ifFalse: [self error: 'this cannot happen'].
    self do: [:token | (actuals at: (tok2 := token) text put: (self scanArgument: tokenList into: C99MacroArgument new)) expandIn: cpp]
	 separatedBy: [(tokenList notEmpty and: [tokenList removeFirst isComma])
			   ifFalse:
			      [tok2 warn: 'while scanning remaining arguments'.
			       tok1 error: 'too few arguments']].
    variadic
	ifTrue:
	   [rest := C99MacroArgument new.
	    (tokenList notEmpty and: [tokenList first isComma]) ifTrue: [tokenList removeFirst].
	    [tokenList isEmpty or: [tokenList first isCloseParen]]
		whileFalse:
		   [self scanArgument: tokenList into: rest.
		    (tokenList notEmpty and: [tokenList first isComma]) ifTrue: [rest addLast: tokenList removeFirst]].
	    (actuals at: variadic put: rest) expandIn: cpp].
    (tokenList notEmpty and: [tok1 := tokenList removeFirst isCloseParen]) ifFalse: [tok1 error: ') expected'].
    ^actuals
]

C99MacroArgumentList scanArgument: tokenList into: argument
[
    | token depth |
    depth := 0.
    [tokenList notEmpty]
	whileTrue:
	   [token := tokenList first.
	    (token isComma and: [depth == 0]) ifTrue: [^argument].
	    token isOpenParen
		ifTrue:  [depth := depth + 1]
		ifFalse: [(token isCloseParen and: [(depth := depth - 1) < 0])
			      ifTrue: [^argument]].
	    token isUntaint
		ifTrue:  [tokenList removeFirst processUntaint]
		ifFalse: [argument add: (tokenList removeFirst isEOL
					     ifTrue:  [C99Blank withText: ' ' file: token file line: token line]
					     ifFalse: [token])]].
    self error: ') expected'
]

C99MacroArgument printOn: aStream
[
    super printOn: aStream.
    aStream cr; tab; nextPutAll: 'EXP:'; print: expanded; cr.
]

C99TokenList substituteArguments: argumentDictionary
[
    | token |
    token := root nextLink.
    [token isNullLink] whileFalse: [token := token substituteArgument: argumentDictionary].
]

C99Token substituteArgument: argumentDictionary	[ ^nextLink ]

" 6.10.3.2  The # operator 

  1 Each # preprocessing token in the replacement list for a function-like macro shall be 
    followed by a parameter as the next preprocessing token in the replacement list.

  2 If, in the replacement list, a parameter is immediately preceded by a # preprocessing 
    token, both are replaced by a single character string literal preprocessing token that 
    contains the spelling of the preprocessing token sequence for the corresponding 
    argument.  Each occurrence of white space between the argument's preprocessing tokens 
    becomes a single space character in the character string literal.  White space before the 
    first preprocessing token and after the last preprocessing token composing the argument 
    is deleted.  Otherwise, the original spelling of each preprocessing token in the argument 
    is retained in the character string literal, except for special handling for producing the 
    spelling of string literals and character constants: a \ character is inserted before each ''
    and \ character of a character constant or string literal (including the delimiting ''
    characters), except that it is implementation-defined whether a \ character is inserted 
    before the \ character beginning a universal character name.  If the replacement that 
    results is not a valid character string literal, the behavior is undefined.  The character 
    string literal corresponding to an empty argument is "".  The order of evaluation of # and 
    ## operators is unspecified. "

C99Hash substituteArgument: argumentDictionary
[
    | id arg replacement |
    [nextLink isBlank] whileTrue: [nextLink unlink].
    (nextLink isIdentifier and: [(arg := argumentDictionary at: nextLink text ifAbsent: []) notNil])
	ifFalse: [self error: '# is not followed by a macro parameter: ', nextLink printString].
    nextLink unlink.
    replacement := arg isEmpty
	ifTrue:  ['""']
	ifFalse: [arg copyAll removeOuterBlanks combineBlanks spelling].
    self replaceWith: (C99String withText: replacement file: file line: line).
    ^nextLink
]

C99MacroArgument spelling
[
    | spelling |
    spelling := WriteStream on: (String new: 32).
    spelling nextPut: $".
    self do: [:token | spelling nextPutAll: token spelling].
    spelling nextPut: $".
    ^spelling contents
]

C99Token	spelling	[ ^text asString ]
C99Blank	spelling	[ ^' ' ]
C99Untaint	spelling	[ self error: 'this cannot happen' ]
C99String	spelling	[ ^text withEscapedQuoteAndBackslash ]
C99Character	spelling	[ ^text withEscapedQuoteAndBackslash ]

String withEscapedQuoteAndBackslash
[
    | result |
    result := WriteStream on: (String new: self size * 2).
    self do: [:c |
	(c == $" or: [c == $\\]) ifTrue: [result nextPut: $\\].
	result nextPut: c].
    ^result contents
]

" 6.10.3.1  Argument substitution

  1 After the arguments for the invocation of a function-like macro have been identified, 
    argument substitution takes place.  A parameter in the replacement list, unless preceded 
    by a # or ## preprocessing token or followed by a ## preprocessing token (see below), is 
    replaced by the corresponding argument after all macros contained therein have been 
    expanded.  Before being substituted, each argument's preprocessing tokens are 
    completely macro replaced as if they formed the rest of the preprocessing file; no other 
    preprocessing tokens are available. "

C99Identifier substituteArgument: argumentDictionary
[
    | arg replacement next paste prev |
    (arg := argumentDictionary at: text ifAbsent: []) isNil ifTrue: [^nextLink].
    replacement := ((nextLink nextNonBlank isPaste or: [paste := (prev := prevLink prevNonBlank) isPaste])
			ifTrue:  [arg tokensOrPlacemarker]
			ifFalse: [arg expanded])
	copyAllWithFile: file line: line.
    self replaceWithAll: replacement.
    (paste and: [note == #variadic]) ifTrue: [prev note: note].				"xxx GNU xxx"
    ^(replacement isEmpty ifTrue: [self] ifFalse: [replacement last]) nextLink.
]

C99Token nextNonBlank	[ [self isBlank] whileTrue: [self := self nextLink] ]
C99Token prevNonBlank	[ [self isBlank] whileTrue: [self := self prevLink] ]

C99MacroArgument expandIn: cpp
[
    expanded := self copyAll expandMacrosIn: cpp
]

C99TokenList expandMacrosIn: cpp
[
   | result token |
   result := C99TokenList new.
   [self isEmpty]
       whileFalse:
	  [((token := self removeFirst) expandBefore: self in: cpp) notNil
	       ifTrue: [result addLast: token]].
   ^result
]

C99Token	expandBefore: tokens in: cpp		[ ^self ]
C99Identifier	expandBefore: tokens in: cpp		[ ^self processBefore: tokens in: cpp ]
C99Untaint	expandBefore: tokens in: cpp		[  note untaint.  ^nil ]

C99MacroArgument tokensOrPlacemarker
[
    ^self isEmpty ifTrue: [self placemarker] ifFalse: [self]
]

C99MacroArgument placemarker
[
    ^placemarker ifNil: [placemarker := C99TokenList with: (C99Placemarker new)]
]

C99Token replaceWith: token
[
    (token prevLink: prevLink) nextLink: token.
    (token nextLink: nextLink) prevLink: token.
    token onList: onList.
    onList := nil
]

C99Token replaceWithAll: tokenList
[
    tokenList isEmpty
	ifTrue: [self unlink]
	ifFalse:
	   [tokenList do: [:token | token onList: onList].
	    prevLink nextLink: tokenList first.  tokenList first prevLink: prevLink.
	    nextLink prevLink: tokenList last.   tokenList last  nextLink: nextLink].
]
