{ import: Object-imp }
{ import: Object-lib }

Link : Object ( prevLink nextLink onList )

NullLink : Link ()

Link	 isNullLink	[ ^false ]	Link	 notNullLink	[ ^true ]
NullLink isNullLink	[ ^true ]	NullLink notNullLink	[ ^false ]

Link prevLink		[ ^prevLink ]
Link prevLink: aLink	[ ^prevLink := aLink ]
Link nextLink		[ ^nextLink ]
Link nextLink: aLink	[ ^nextLink := aLink ]
Link onList		[ ^onList ]
Link onList: aList	[ ^onList := aList ]

Link offList		[  onList := nil ]

Link unlink
[
    prevLink nextLink: nextLink.
    nextLink prevLink: prevLink.
    self offList.
]


LinkedList : Object ( root )

LinkedList isLinkedList	[ ^true ]
Object	   isLinkedList	[ ^false ]

LinkedList root	[ ^root ]

LinkedList new
[
    self := super new.
    root := self nullLink.
    root prevLink: root.
    root nextLink: root.
]

LinkedList nullLink	[ ^NullLink new ]

LinkedList with: aLink	[ ^self new addLast: aLink; yourself ]

LinkedList isEmpty	[ ^root nextLink == root ]
LinkedList notEmpty	[ ^root nextLink ~~ root ]

LinkedList emptyCheck	[ root nextLink == root ifTrue: [self errorEmptyCollection] ]

Link onListCheck	[ onList ifTrue: [self error: 'link is already on a list: ', self printString] ]

LinkedList addFirst: aLink
[
    | prev |
    aLink onListCheck.
    aLink prevLink: root.
    aLink nextLink: root nextLink.
    root nextLink prevLink: aLink.
    root nextLink: aLink.
    aLink onList: self.
    ^aLink.
]

LinkedList addLast: aLink
[
    | prev |
    aLink onListCheck.
    aLink nextLink: root.
    aLink prevLink: root prevLink.
    root prevLink nextLink: aLink.
    root prevLink: aLink.
    aLink onList: self.
    ^aLink.
]

LinkedList add: aLink	[ ^self addLast: aLink ]

LinkedList removeFirst
[
    | link |
    self emptyCheck.
    link := root nextLink.
    (root nextLink: link nextLink) prevLink: root.
    ^link offList
]

LinkedList removeLast
[
    | link |
    self emptyCheck.
    link := root prevLink.
    (root prevLink: link prevLink) nextLink: root.
    ^link offList
]

LinkedList first
[
    | link |
    (link := root nextLink) == root
	ifTrue: [self errorEmptyCollection].
    ^link
]

LinkedList second
[
    | link |
    ((link := root nextLink) == root or: [(link := link nextLink) == root])
	ifTrue: [self errorEmptyCollection].
    ^link
]

LinkedList third
[
    | link |
    ((link := root nextLink) == root or: [(link := link nextLink) == root or: [(link := link nextLink) == root]])
	ifTrue: [self errorEmptyCollection].
    ^link
]

LinkedList last
[
    self emptyCheck.
    ^root prevLink
]

LinkedList addAllFirst: aCollection	[ [aCollection isEmpty] whileFalse: [self addFirst: aCollection removeLast] ]
LinkedList addAllLast: aCollection	[ [aCollection isEmpty] whileFalse: [self addLast:  aCollection removeLast] ]

LinkedList size
[
    | count aLink |
    count := 0.
    aLink := root.
    [(aLink := aLink nextLink) == root] whileFalse: [count := 1 + count].
    ^count
]

LinkedList do: unaryBlock
[
    | aLink |
    aLink := root.
    [(aLink := aLink nextLink) == root] whileFalse: [unaryBlock value: aLink]
]

LinkedList reverseDo: unaryBlock
[
    | aLink |
    aLink := root.
    [(aLink := aLink prevLink) == root] whileFalse: [unaryBlock value: aLink]
]

LinkedList from: aLink do: unaryBlock
[
    [aLink == root]
	whileFalse:
	   [unaryBlock value: aLink.
	    aLink := aLink nextLink]
]

LinkedList do: unaryBlock separatedBy: aBlock
[
    | aLink |
    aLink := root.
    [(aLink := aLink nextLink) == root]
	whileFalse:
	   [aLink prevLink == root ifFalse: [aBlock value].
	    unaryBlock value: aLink]
]

LinkedList with: aList do: binaryBlock
[
    | link1 link2 |
    link1 := root.
    link2 := aList root.
    [link1 := link1 nextLink.
     link2 := link2 nextLink.
     (link1 isNullLink or: [link2 isNullLink])]
	whileFalse:
	    [binaryBlock value: link1 value: link2].
    (link1 isNullLink and: [link2 isNullLink])
	ifFalse: [self error: 'lists were of differing sizes'].
]

LinkedList = aList
[
    self == aList ifTrue: [^true].
    aList isLinkedList ifFalse: [^false].
    self size == aList size ifFalse: [^false].
    self with: aList do: [:x :y | x = y ifFalse: [^false]].
    ^true
]

"----------------------------------------------------------------"


"----------------------------------------------------------------"

C99Token : Link ( text note file line )

C99Token isC99Token	[ ^true ]
Object	 isC99Token	[ ^false ]

C99Token withText: textString file: fileString line: lineNumber note: aNote
[
    self := self withText: textString file: fileString line: lineNumber.
    note := aNote.
]

C99Token withText: textString file: fileString line: lineNumber
[
    self := self new.
    text := textString.
    file := fileString.
    line := lineNumber.
]

C99Token copy						[ ^self withText: text file: file	line: line	 note: note ]
C99Token copyWithFile: fileString line: lineNumber	[ ^self withText: text file: fileString line: lineNumber note: note ]

C99Token text		[ ^text ]
C99Token text: aString	[ ^text := aString ]
C99Token note		[ ^note ]
C99Token note: aNote	[ ^note := aNote ]
C99Token file		[ ^file ]
C99Token file: aString	[ ^file := aString ]
C99Token line		[ ^line ]
C99Token line: aNumber	[ ^line := aNumber ]

C99Token isEOL		[ ^note == #eol ]
C99Token notEOL		[ ^note ~~ #eol ]
C99Token isComment	[ ^note == #comment ]

C99Token = aToken	[ ^aToken isC99Token and: [self _vtable == aToken _vtable] ]

C99Token expected:	message	[ self error: 'expected ', message ]
C99Token error:		message	[ super error: file, ':', line printString, ': ERROR: ', message, ' near ''', text, '''' ]
C99Token warn:		message	[ StdErr nextPutAll: file, ':', line printString, ': WARNING: ', message, ' near ''', text, ''''; cr ]

C99Token printOn: aStream
[
    aStream nextPut: ${; nextPutAll: self debugName; space; print: note; space; print: text; nextPut: $}
]

C99Token putTextOn: aStream	[ aStream nextPutAll: text ]

"----------------------------------------------------------------"

C99TokenList : LinkedList ()

C99TokenList nullLink	[ ^C99NullLink new ]

C99NullLink : C99Token ()

C99NullLink isNullLink	[ ^true ]

C99TokenList removeLeadingBlanks
[
    [self notEmpty and: [root nextLink isBlank]] whileTrue: [self removeFirst].
]

C99TokenList removeTrailingBlanks
[
    [self notEmpty and: [root prevLink isBlank]] whileTrue: [self removeLast].
]

C99TokenList removeOuterBlanks
[
    self removeLeadingBlanks; removeTrailingBlanks
]

C99TokenList removeBlanks
[
    self do: [:tok | tok isBlank ifTrue: [tok unlink]].		"iteration is stable under deletion"
]

C99TokenList combineBlanks
[
    | prevBlank |
    prevBlank := false.
    self do: [:tok |
	tok isBlank
	    ifTrue:
	       [prevBlank
		    ifTrue:  [tok unlink]
		    ifFalse: [tok text: ' '].
		prevBlank := true]
	    ifFalse:
	       [prevBlank := false]]
]

C99TokenList copyAllWithFile: fileString line: lineNumber
[
    | copy |
    copy := self new.
    self do: [:token | copy addLast: (token copyWithFile: fileString line: lineNumber)].
    ^copy
]

C99TokenList copyAll
[
    | copy |
    copy := self new.
    self do: [:token | copy addLast: token copy].
    ^copy
]

C99TokenList error: message	[ super error: 'ERROR: ', message ]
C99TokenList warn: message	[ StdErr nextPutAll: 'WARNING: ', message ]

C99TokenList printOn: aStream
[
    aStream nextPutAll: '#{'.
    self do: [:tok | aStream space; print: tok].
    aStream nextPutAll: ' }'.
]

C99TokenList putTextOn: aStream
[
    self do: [:tok | tok putTextOn: aStream]
]

"----------------------------------------------------------------"

C99DivideInto	  : C99Token ()
C99Divide	  : C99Token ()
C99Paste	  : C99Token ()		C99Paste	isPaste		[ ^true ]	C99Token isPaste	[ ^false ]
C99Hash		  : C99Token ()
C99LeftShiftInto  : C99Token ()
C99LeftShift	  : C99Token ()
C99LessEqual	  : C99Token ()
C99Begin	  : C99Token ()
C99OpenParen	  : C99Token ()		C99OpenParen	isOpenParen	[ ^true ]	C99Token isOpenParen	[ ^false ]
C99OpenBracket	  : C99Token ()		C99OpenBracket	isOpenBracket	[ ^true ]	C99Token isOpenBracket	[ ^false ]
C99OpenBrace	  : C99Token ()		C99OpenBrace	isOpenBrace	[ ^true ]	C99Token isOpenBrace	[ ^false ]
C99Less		  : C99Token ()		C99Less		isLess		[ ^true ]	C99Token isLess		[ ^false ]
C99RightShiftInto : C99Token ()
C99RightShift	  : C99Token ()
C99GreaterEqual	  : C99Token ()
C99Greater	  : C99Token ()		C99Greater	isGreater	[ ^true ]	C99Token isGreater	[ ^false ]
C99LogAnd	  : C99Token ()		C99LogAnd	isLogAnd	[ ^true ]	C99Token isLogAnd	[ ^false ]
C99AndInto	  : C99Token ()
C99And		  : C99Token ()		C99And		isAnd		[ ^true ]	C99Token isAnd		[ ^false ]
C99LogOr	  : C99Token ()		C99LogOr	isLogOr		[ ^true ]	C99Token isLogOr	[ ^false ]
C99OrInto	  : C99Token ()
C99Or		  : C99Token ()		C99Or		isOr		[ ^true ]	C99Token isOr		[ ^false ]
C99MultiplyInto	  : C99Token ()
C99Multiply	  : C99Token ()		C99Multiply	isMultiply	[ ^true ]	C99Token isMultiply	[ ^false ]
C99Decrement	  : C99Token ()		C99Decrement	isDecrement	[ ^true ]	C99Token isDecrement	[ ^false ]
C99SubtractInto	  : C99Token ()
C99Member	  : C99Token ()		C99Member	isMember	[ ^true ]	C99Token isMember	[ ^false ]
C99MemberRef	  : C99Token ()		C99MemberRef	isMemberRef	[ ^true ]	C99Token isMemberRef	[ ^false ]
C99Subtract	  : C99Token ()
C99Increment	  : C99Token ()		C99Increment	isIncrement	[ ^true ]	C99Token isIncrement	[ ^false ]
C99AddInto	  : C99Token ()
C99Add		  : C99Token ()
C99NotEqual	  : C99Token ()
C99Not		  : C99Token ()
C99CloseParen	  : C99Token ()		C99CloseParen	isCloseParen	[ ^true ]	C99Token isCloseParen	[ ^false ]
C99CloseBracket	  : C99Token ()		C99CloseBracket	isCloseBracket	[ ^true ]	C99Token isCloseBracket	[ ^false ]
C99CloseBrace	  : C99Token ()		C99CloseBrace	isCloseBrace	[ ^true ]	C99Token isCloseBrace	[ ^false ]
C99Colon	  : C99Token ()		C99Colon	isColon		[ ^true ]	C99Token isColon	[ ^false ]
C99Equal	  : C99Token ()		C99Equal	isEqual		[ ^true ]	C99Token isEqual	[ ^false ]
C99Assign	  : C99Token ()
C99Paste	  : C99Token ()
C99Hash		  : C99Token ()
C99End		  : C99Token ()
C99ModulusInto	  : C99Token ()
C99Modulus	  : C99Token ()
C99XorInto	  : C99Token ()
C99Xor		  : C99Token ()		C99Xor		isXor		[ ^true ]	C99Token isXor		[ ^false ]
C99Ellipsis	  : C99Token ()		C99Ellipsis	isEllipsis	[ ^true ]	C99Token isEllipsis	[ ^false ]
C99Comma	  : C99Token ()		C99Comma	isComma		[ ^true ]	C99Token isComma	[ ^false ]
C99Semicolon	  : C99Token ()		C99Semicolon	isSemicolon	[ ^true ]	C99Token isSemicolon	[ ^false ]
C99Query	  : C99Token ()		C99Query	isQuery		[ ^true ]	C99Token isQuery	[ ^false ]
C99Complement	  : C99Token ()

"   (6.5.3) unary-operator: one of 
    		& * + -  ~ ! "

C99Token	isUnary	[ ^false ]
C99And		isUnary	[ ^true ]
C99Multiply	isUnary	[ ^true ]
C99Add		isUnary	[ ^true ]
C99Subtract	isUnary	[ ^true ]
C99Complement	isUnary	[ ^true ]
C99Not		isUnary	[ ^true ]

C99Token	isMultiplicative [ ^false ]
C99Multiply	isMultiplicative [ ^true ]
C99Divide	isMultiplicative [ ^true ]
C99Modulus	isMultiplicative [ ^true ]

C99Token	isAdditive [ ^false ]
C99Add		isAdditive [ ^true ]
C99Subtract	isAdditive [ ^true ]

C99Token	isShift [ ^false ]
C99LeftShift	isShift [ ^true ]
C99RightShift	isShift [ ^true ]

C99Token	isRelational [ ^false ]
C99Less		isRelational [ ^true ]
C99LessEqual	isRelational [ ^true ]
C99Greater	isRelational [ ^true ]
C99GreaterEqual	isRelational [ ^true ]

C99Token	isEquality [ ^false ]
C99Equal	isEquality [ ^true ]
C99NotEqual	isEquality [ ^true ]

C99Other	: C99Token ()
C99Number	: C99Token ()		C99Number	isNumber	[ ^true ]		C99Token isNumber	[ ^false ]
C99Character	: C99Token ()		C99Character	isCharacter	[ ^true ]		C99Token isCharacter	[ ^false ]
C99String	: C99Token ()		C99String	isC99String	[ ^true ]		C99Token isC99String	[ ^false ]
C99Identifier	: C99Token ()
C99Blank	: C99Token ()

C99Token isIdentifier	[ ^false ]		C99Identifier isIdentifier	[ ^true ]
C99Token notIdentifier	[ ^true ]		C99Identifier notIdentifier	[ ^false ]

C99Token isBlank	[ ^false ]		C99Blank isBlank	[ ^true ]

C99Directive : C99Token ( name body )

C99Directive new
[
    self := super new.
    body := C99TokenList new.
]

C99Directive name: aToken	[ ^name := aToken ]
C99Directive body		[ ^body ]
C99Directive body: aToken	[ ^body := aToken ]

C99Directive printOn: aStream
[
    super printOn: aStream.
    body do: [:tok | tok printOn: aStream].
]

C99Directive putTextOn: aStream
[
    aStream nextPut: $#; nextPutAll: name text.
    body do: [:tok | tok putTextOn: aStream].
    aStream cr.
]

C99GroupSection : C99Directive ( nextGroup )

C99GroupSection nextGroup		[ ^nextGroup ]
C99GroupSection nextGroup: aToken	[ ^nextGroup := aToken ]

C99GroupDirective : C99GroupSection ( enclosingGroup )

C99GroupDirective enclosingGroup		[ ^enclosingGroup ]
C99GroupDirective enclosingGroup: aToken	[ ^enclosingGroup := aToken ]

C99DirectiveIf  : C99GroupDirective ()
  C99DirectiveIfdef : C99DirectiveIf ()
    C99DirectiveIfndef : C99DirectiveIfdef ()

C99DirectiveElif : C99GroupDirective ()

C99DirectiveElse : C99GroupDirective ()

C99Directive     isDirectiveElse	[ ^false ]
C99DirectiveElse isDirectiveElse	[ ^true ]

C99DirectiveEndif : C99Directive ()

C99Directive      isDirectiveEndif	[ ^false ]
C99DirectiveEndif isDirectiveEndif	[ ^true ]

C99DirectiveInclude : C99Directive ()

C99DirectiveDefine : C99Directive ()

C99DirectiveUndef : C99Directive ()

C99DirectiveLine : C99Directive ()

C99DirectiveError : C99Directive ()

C99DirectivePragma : C99Directive ()
