{ import: Object }
{ include "tag.h" }

[ Object := _object ]	"alias"
[ Symbol := _selector ]	"alias"

SmallInteger value_: _value		{ return _O(v__value); }
SmallInteger _integerValue		{ return (oop)_I(v_self); }

ByteArray size_: _size value_: _value
[
    self := self new.
    size := SmallInteger value_: _size.
    _bytes := _value.
]

[ UndefinedObject _beNilType ]
[ SmallInteger    _beTagType ]

Object new			[ ^self _vtable _alloc: self _sizeof ]

Object yourself			[ "^self" ]

Object identityHash		{ return (oop)((long)self >> 3 | 1); }

Object = other			[ ^self == other ]

Object _palloc: size		{ return (oop)_palloc(sizeof(oop) * _I(v_size)); }
Object _balloc: size		{ return (oop)_balloc(              _I(v_size)); }

UndefinedObject _nil		{ return 0; }

[ nil   := UndefinedObject _nil ]		{ pragma: type nil   nil   }
[ true  := Object ]				{ pragma: type true  true  }
[ false := nil ]				{ pragma: type false false }

UndefinedObject not		[ ^true ]
Object not			[ ^false ]

UndefinedObject isNil		[ ^true ]
Object isNil			[ ^false ]

UndefinedObject notNil		[ ^false ]
Object notNil			[ ^true ]

Object debugName			[ ^String value_: self _debugName ]
Object printOn: aStream			[ ^aStream nextPutAll: self debugName ]
UndefinedObject printOn: aStream	[ ^aStream nextPutAll: 'nil' ]

Object printString
[
    | stream |
    stream := WriteStream on: (String new: 8).
    self printOn: stream.
    ^stream contents
]

Object ifNil: nilBlock			[ ^self ]
UndefinedObject ifNil: nilBlock		[ ^nilBlock value ]

Object ifNotNil: notNilBlock		[ ^notNilBlock value ]
UndefinedObject ifNotNil: notNilBlock	[ ^nil ]

Object ifTrue: aBlock			[ ^aBlock value ]
UndefinedObject ifTrue: aBlock		[ ^nil ]

Object ifFalse: aBlock			[ ^nil ]
UndefinedObject ifFalse: aBlock		[ ^aBlock value ]

Object abort		{ abort(); }
Object exit		{ exit(0); }
Object halt		[ StdErr cr; nextPutAll: 'halt'; cr.  { exit(1); } ]

Object perform: s
{
    struct __closure *c= _local_bind(v_s, v_self);
    return (c->method)((oop)c, v_self);
}

Object perform: s with: a
{
    struct __closure *c= _local_bind(v_s, v_self);
    return (c->method)((oop)c, v_self, v_a);
}

Object perform: s with: a with: b
{
    struct __closure *c= _local_bind(v_s, v_self);
    return (c->method)((oop)c, v_self, v_a, v_b);
}

Object perform: s with: a with: b with: c
{
    struct __closure *c= _local_bind(v_s, v_self);
    return (c->method)((oop)c, v_self, v_a, v_b, v_c);
}

Object perform: s with: a with: b with: c with: d
{
    struct __closure *c= _local_bind(v_s, v_self);
    return (c->method)((oop)c, v_self, v_a, v_b, v_c, v_d);
}

Object error: reason
[
    StdErr cr; nextPutAll: reason; cr.
    { exit(1); }.
]

Object shouldNotImplement: aSelector
[
    StdErr cr; nextPutAll: self debugName; nextPutAll: ' should not implement '; print: aSelector; cr.
    { exit(1); }.
]

Object errorOutOfBounds: anIndex
[
    StdErr cr;
	print: anIndex; nextPutAll: ' is out of bounds in ';
	nextPutAll: self debugName; nextPut: $(; print: self size; nextPut: $);
	cr.
    { exit(1); }.
]

"----------------------------------------------------------------"

Object	     isSmallInteger	[ ^false ]
SmallInteger isSmallInteger	[ ^true ]

SmallInteger negated		[ ^0 - self ]

"----------------------------------------------------------------"

SmallInteger identityHash	[ ^self * 17 ]

SmallInteger // arg
[
    { if ((long)v_self & (long)v_arg & 1) return _O(_I(v_self) / _I(v_arg)); }.
    ^self primitiveFailed
]

SmallInteger \\ arg
[
    { if ((long)v_self & (long)v_arg & 1) return _O(_I(v_self) % _I(v_arg)); }.
    ^self primitiveFailed
]

SmallInteger asDigit
[
    ^self + (self > 9 ifTrue: [$a - 10] ifFalse: [$0])
]

SmallInteger digitValue
[
    (self between: $A and: $Z) ifTrue: [^self - $A + 10].
    (self between: $a and: $z) ifTrue: [^self - $a + 10].
    (self between: $0 and: $9) ifTrue: [^self - $0     ].
    ^nil
]

SmallInteger max: anInteger
[
    ^self < anInteger ifTrue: [anInteger] ifFalse: [self]
]

SmallInteger between: low and: high
[
    ^low <= self and: [self <= high]
]

SmallInteger to: limit do: unaryBlock
[
    [self <= limit]
	whileTrue:
	   [unaryBlock value: self.
	    self := self + 1]
]

SmallInteger downTo: limit do: unaryBlock
[
    [self >= limit]
	whileTrue:
	   [unaryBlock value: self.
	    self := self - 1]
]

"----------------------------------------------------------------"

Object isString	[ ^false ]
String isString	[ ^true ]

String asString []

String new: capacity
[
    self := super new: capacity + 1.
    self at: size - 1 put: 0.
    size := size - 1.
]

String value_: _value
[
    self := self new.
    _bytes := _value.
    size := SmallInteger value_: self _size.
]

String _size
{
    return (oop)strlen((char *)self->v__bytes);
}

String _stringValue
{
    return (oop)self->v__bytes;
}

String beginsWith: aString
[
    self size < aString size ifTrue: [^false].
    aString doWithIndex: [:e :i | (self at: i) == e ifFalse: [^false]].
    ^true
]

String endsWith: aString
[
    | start |
    self size < aString size ifTrue: [^false].
    start := self size - aString size.
    aString doWithIndex: [:e :i | (self at: start + i) == e ifFalse: [^false]].
    ^true
]

String asSymbol		[ ^Symbol intern_: self _bytes size_: size _integerValue ]

String printOn: aStream	[ ^aStream nextPut: $'; nextPutAll: self; nextPut: $' ]


"----------------------------------------------------------------"

Object isSymbol	[ ^false ]
Symbol isSymbol	[ ^true ]

Symbol intern_: _string size_: _stringSize
{
    char *buf= (char *)alloca((int)v__stringSize + 1);
    memcpy((void *)buf, (void *)v__string, (int)v__stringSize);
    buf[(int)v__stringSize]= '\0';
    return _selector(buf);
}

Symbol intern: aString
[
    ^self intern_: aString _stringValue size_: aString size _integerValue
]

Symbol _bytes	[ ^_elements ]
Symbol size	[ ^SmallInteger value_: _size ]

Symbol _stringValue	[ ^_elements ]

Symbol asString	[ ^String value_: _elements ]

Symbol beginsWith: aString	[ ^self asString beginsWith: aString ]
Symbol endsWith: aString	[ ^self asString endsWith: aString ]

Symbol , aStringOrSymbol	[ ^(self asString , aStringOrSymbol asString) asSymbol ]
Symbol append: aStringOrSymbol	[ ^(self asString , aStringOrSymbol asString) asSymbol ]

Symbol at: offset	[ (offset between: 0 and: self size - 1) ifTrue: [^SmallInteger value_: (self _at_: offset _integerValue)] ]

Symbol _at_: _offset	{ return (oop)(long)(((char *)self->v__elements)[(long)v__offset]); }

Symbol = other		[ ^self == other ]

Symbol printOn: aStream	[ ^aStream nextPut: $#; nextPutAll: self asString ]

"----------------------------------------------------------------"

StaticBlockClosure function_: _implementationAddress arity_: _argumentCount
[
    { v_self= _alloc(v_self, sizeof(*self)); }.
    _arity    := _argumentCount.
    _function := _implementationAddress.
]

StaticBlockClosure arity	[ ^SmallInteger value_: _arity ]

"evaluating"

StaticBlockClosure value
[
    { if ((int)self->v__arity == 0) return ((_imp_t)(self->v__function))(v__closure, v_self); }.
    ^self errorArgumentCount: 0
]

StaticBlockClosure value: a
[
    { if ((int)self->v__arity == 1) return ((_imp_t)(self->v__function))(v__closure, v_self, v_a); }.
    ^self errorArgumentCount: 1
]

StaticBlockClosure value: a value: b
[
    { if ((int)self->v__arity == 2) return ((_imp_t)(self->v__function))(v__closure, v_self, v_a, v_b); }.
    ^self errorArgumentCount: 1
]

StaticBlockClosure value: a value: b value: c
[
    { if ((int)self->v__arity == 3) return ((_imp_t)(self->v__function))(v__closure, v_self, v_a, v_b, v_c); }.
    ^self errorArgumentCount: 1
]

StaticBlockClosure value: a value: b value: c value: d
[
    { if ((int)self->v__arity == 4) return ((_imp_t)(self->v__function))(v__closure, v_self, v_a, v_b, v_c, v_d); }.
    ^self errorArgumentCount: 1
]

StaticBlockClosure value: a value: b value: c value: d value: e
[
    { if ((int)self->v__arity == 5) return ((_imp_t)(self->v__function))(v__closure, v_self, v_a, v_b, v_c, v_d, v_e); }.
    ^self errorArgumentCount: 1
]

"controling"

StaticBlockClosure whileTrue
[
    [self value] whileTrue: []
]

StaticBlockClosure whileTrue: aBlock
[
    self value
	ifTrue:
	   [aBlock value.
	    self whileTrue: aBlock].
]

StaticBlockClosure whileFalse
[
    [self value] whileFalse: []
]

StaticBlockClosure whileFalse: aBlock
[
    self value
	ifFalse:
	   [aBlock value.
	    self whileFalse: aBlock].
]

StaticBlockClosure repeat
[
    [true] whileTrue: [self value]
]


BlockClosure function_: _implementationAddress arity_: _argumentCount
	     outer: outerBlock state: variableArray nlr_: _dynamicEnvironment
[
    { v_self= _alloc(v_self, sizeof(*self)); }.
    _function := _implementationAddress.
    _arity    := _argumentCount.
    outer     := outerBlock.
    state     := variableArray.
    _nlr      := _dynamicEnvironment.
]

"----------------------------------------------------------------"

"{ import: Collection }"

"{ import: Object }"

Collection isEmpty	[ ^self size == 0 ]
Collection notEmpty	[ ^self isEmpty not ]

Collection emptyCheck
[
    self isEmpty ifTrue: [self error: 'this collection is empty']
]

Collection addAll: aCollection		[ aCollection do: [:elt | self add: elt] ]
Collection removeAll: aCollection	[ aCollection do: [:elt | self remove: elt] ]

Collection printOn: aStream
[
    aStream nextPutAll: self debugName.
    self printContentsOn: aStream.
]

Collection printContentsOn: aStream
[
    aStream nextPut: $(.
    self do: [:element | element printOn: aStream] separatedBy: [aStream space].
    aStream nextPut: $).
]

Collection do: elementBlock separatedBy: betweenBlock
[
    | first |
    first := true.
    self do: [:element |
	first ifTrue: [first := false] ifFalse: [betweenBlock value].
	elementBlock value: element].
]

Collection includes: anObject
[
    self do: [:element | element == anObject ifTrue: [^true]].
    ^false
]

Collection readStream	[ ^ReadStream on: self ]
Collection writeStream	[ ^WriteStream on: self ]


SequenceableCollection isSequenceableCollection       [ ^true ]
Object isSequenceableCollection                       [ ^false ]

SequenceableCollection first		[ ^self at: 0			]
SequenceableCollection first: anObject	[ ^self at: 0 put: anObject	]
SequenceableCollection second		[ ^self at: 1			]
SequenceableCollection second: anObject	[ ^self at: 1 put: anObject	]
SequenceableCollection third		[ ^self at: 2			]
SequenceableCollection third: anObject	[ ^self at: 2 put: anObject	]
SequenceableCollection fourth		[ ^self at: 3			]
SequenceableCollection fourth: anObject	[ ^self at: 3 put: anObject	]
SequenceableCollection fifth		[ ^self at: 4			]
SequenceableCollection fifth: anObject	[ ^self at: 4 put: anObject	]
SequenceableCollection last		[ ^self at: self size - 1	]

SequenceableCollection asString		[ ^self as: String ]

SequenceableCollection as: newType
[
    ^(newType new: self size)
	replaceFrom: 0 to: self size - 1 with: self;
	yourself
]

SequenceableCollection withAll: aCollection
[
    ^(self new: aCollection size) replaceFrom: 0 to: aCollection size - 1 with: aCollection
]

SequenceableCollection new: size withAll: value
[
    ^(self new: size) atAllPut: value; yourself
]

SequenceableCollection new: size atAllPut: value
[
    ^(self new: size) atAllPut: value; yourself
]

SequenceableCollection atAllPut: value
[
    0 to: self size - 1 do: [:key | self at: key put: value]
]

SequenceableCollection atAll: keyCollection put: value
[
    keyCollection do: [:key | self at: key put: value]
]

SequenceableCollection addAllLast: aCollection
[
    aCollection do: [:elt | self addLast: elt]
]

SequenceableCollection = aCollection
[   
    self == aCollection ifTrue: [^true].
    (aCollection isSequenceableCollection and: [self size == aCollection size]) ifFalse: [^false].
    self with: aCollection do: [:a :b | a = b ifFalse: [^false]].
    ^true
]

SequenceableCollection from: firstKey to: lastKey put: value
[
    firstKey to: lastKey do: [:key | self at: key put: value]
]

SequenceableCollection collect: elementBlock
[
    | result |
    result := self new: self size.
    self doWithIndex: [:e :i | result at: i put: (elementBlock value: e)].
    ^result
]

SequenceableCollection do: elementBlock			[ self from: 0 do: elementBlock ]
SequenceableCollection reverseDo: elementBlock		[ self from: 0 reverseDo: elementBlock ]
SequenceableCollection doWithIndex: elementIndexBlock	[ self from: 0 doWithIndex: elementIndexBlock ]
SequenceableCollection reverseDoWithIndex: eiBlock	[ self from: 0 reverseDoWithIndex: eiBlock ]
SequenceableCollection do: aBlock separatedBy: bBlock	[ self from: 0 do: aBlock separatedBy: bBlock ]

SequenceableCollection from: start do: elementBlock
[
    start to: self size - 1 do: [:offset | elementBlock value: (self at: offset)].
]

SequenceableCollection from: start reverseDo: elementBlock
[
    self size - 1 downTo: start do: [:offset | elementBlock value: (self at: offset)].
]

SequenceableCollection from: start doWithIndex: elementIndexBlock
[
    start to: self size - 1 do: [:offset | elementIndexBlock value: (self at: offset) value: offset].
]

SequenceableCollection from: start reverseDoWithIndex: elementIndexBlock
[
    self size - 1 downTo: start do: [:offset | elementIndexBlock value: (self at: offset) value: offset].
]

SequenceableCollection from: start do: elementBlock separatedBy: betweenBlock
[
    | first |
    first := true.
    self from: start do: [:e |
	first ifTrue: [first := false] ifFalse: [betweenBlock value].
	elementBlock value: e].
]

SequenceableCollection inject: result into: binaryBlock
[
    self do: [:element | result := binaryBlock value: result value: element].
    ^result
]

SequenceableCollection detect: unaryBlock
[
    ^self detect: unaryBlock ifAbsent: []
]

SequenceableCollection detect: unaryBlock ifAbsent: defaultBlock
[
    self do: [:element || detected | (detected := unaryBlock value: element) notNil ifTrue: [^detected]].
    ^defaultBlock value
]

SequenceableCollection reverseDetect: unaryBlock ifAbsent: defaultBlock
[
    self reverseDo: [:element || detected | (detected := unaryBlock value: element) notNil ifTrue: [^detected]].
    ^defaultBlock value
]

SequenceableCollection with: aCollection do: elementBlock
[
    self size == aCollection size ifFalse: [self error: 'collection sizes differ'].
    0 to: self size - 1 do: [:offset | elementBlock value: (self at: offset) value: (aCollection at: offset)].
]

SequenceableCollection with: aCollection with: bCollection do: elementBlock
[
    (self size == aCollection size and: [self size == bCollection size]) ifFalse: [self error: 'collection sizes differ'].
    0 to: self size - 1 do: [:offset | elementBlock value: (self at: offset) value: (aCollection at: offset) value: (bCollection at: offset)].
]

SequenceableCollection with: aCollection doWithIndex: elementsIndexBlock
[
    self size == aCollection size ifFalse: [self error: 'collection sizes differ'].
    0 to: self size - 1 do: [:offset | elementsIndexBlock value: (self at: offset) value: (aCollection at: offset) value: offset].
]

SequenceableCollection with: aCollection replace: elementsBlock
[
    self with: aCollection doWithIndex: [:a :b :i | self at: i put: (elementsBlock value: a value: b)]
]

SequenceableCollection replaceFrom: first to: last with: aCollection
[
    ^self replaceFrom: first to: last with: aCollection startingAt: 0
]

SequenceableCollection replaceFrom: first to: last with: aCollection startingAt: collectionStart
[
    0 to: last - first do: [:offset | self at: first + offset put: (aCollection at: collectionStart + offset)].
]

SequenceableCollection copy
[
    ^(self new: self size) replaceFrom: 0 to: self size - 1 with: self startingAt: 0
]

SequenceableCollection copyFrom: first
[
    ^self copyFrom: first to: self size - 1
]

SequenceableCollection copyFrom: first to: last
[
    | size |
    size := 0 max: last - first + 1.
    ^(self new: size) replaceFrom: 0 to: size - 1 with: self startingAt: first
]

SequenceableCollection , aCollection
[
    ^(self new: self size + aCollection size)
	replaceFrom: 0 to: self size - 1 with: self;
	replaceFrom: self size to: self size + aCollection size - 1 with: aCollection
]

SequenceableCollection indexOf: anObject ifAbsent: errorBlock
[
    self doWithIndex: [:element :index | element == anObject ifTrue: [^index]].
    ^errorBlock value
]

SequenceableCollection concatenated
[
    | result |
    result := (self first new: 4) writeStream.
    self do: [:elt | result nextPutAll: elt].
    ^result contents
]

SequenceableCollection flattened
[
    | result |
    result := (self new: 4) writeStream.
    self flattenOn: result.
    ^result contents
]

Object flattenOn: aStream
[
    aStream nextPut: self
]

SequenceableCollection flattenOn: aStream
[
    aStream collection _vtable == self _vtable
	ifTrue:  [self do: [:elt | elt flattenOn: aStream]]
	ifFalse: [super flattenOn: aStream]
]

"----------------------------------------------------------------"

ArrayedCollection with: a
[
    ^(self new: 1)
	at: 0 put: a;
	yourself
]

ArrayedCollection with: a with: b
[
    ^(self new: 2)
	at: 0 put: a;
	at: 1 put: b;
	yourself
]

ArrayedCollection with: a with: b with: c
[
    ^(self new: 3)
	at: 0 put: a;
	at: 1 put: b;
	at: 2 put: c;
	yourself
]

ArrayedCollection with: a with: b with: c with: d
[
    ^(self new: 4)
	at: 0 put: a;
	at: 1 put: b;
	at: 2 put: c;
	at: 3 put: d;
	yourself
]

ArrayedCollection with: a with: b with: c with: d with: e
[
    ^(self new: 5)
	at: 0 put: a;
	at: 1 put: b;
	at: 2 put: c;
	at: 3 put: d;
	at: 4 put: e;
	yourself
]

ArrayedCollection new: capacity
[
    self := super new.
    size := capacity.
]

ArrayedCollection isEmpty
[
    ^size == 0.
]

ArrayedCollection size
[
    ^size.
]

"----------------------------------------------------------------"

_object isArray	[ ^false ]
Array isArray	[ ^true ]

Array size_: _size value_: _value
[
    self := self new: (SmallInteger value_: _size).
    _oops := _value.
    {
      int i;
      for (i= 0;  i < (int)v__size;  ++i)
        ((oop *)self->v__oops)[i]= *((oop **)v__value)[i];
    }.
]

Array new: capacity
[
    self := super new.
    size := capacity.
    _oops := self _palloc: size
]

Array _elements	[ ^_oops ]

Array at: offset
[
    ^(0 <= offset and: [offset < size])
	ifTrue:  [self at_: offset _integerValue]
	ifFalse: [self errorOutOfBounds: offset]
]

Array at_: _offset
{
    return ((oop *)self->v__oops)[(int)v__offset];
}

Array at: offset put: anObject
[
    ^(0 <= offset and: [offset < size])
	ifTrue:  [self at_: offset _integerValue put: anObject]
	ifFalse: [self errorOutOfBounds: offset]
]

Array at_: _offset put: anObject
{
    return ((oop *)self->v__oops)[(int)v__offset]= v_anObject;
}

Array  argumentVectorSize	[ ^SmallInteger value_: self _argumentVectorSize ]
Array _argumentVectorSize	{ return (oop)_param(0); }
Array  argumentVectorAt: offset	[ ^String value_: (self _argumentVectorAt_: offset _integerValue) ]
Array _argumentVectorAt_: _off	{ return (oop)(((char **)_param(1))[(int)v__off]); }

Array argumentVector
[
    self := self new: self argumentVectorSize.
    0 to: size - 1 do: [:offset | self at: offset put: (self argumentVectorAt: offset)].
]

Array printOn: aStream
[
    aStream nextPut: $#.
    self printContentsOn: aStream.
]

"----------------------------------------------------------------"

ByteArray : ArrayedCollection ( _bytes )

ByteArray isByteArray	[ ^true ]
Object	  isByteArray	[ ^false ]

ByteArray new: capacity
[
    self := super new: capacity.
    _bytes := self _balloc: capacity.
]

ByteArray _bytes	[ ^_bytes ]
ByteArray _elements	[ ^_bytes ]

ByteArray at: offset
[
    ^(0 <= offset and: [offset < size])
	ifTrue:  [SmallInteger value_: (self _at_: offset _integerValue)]
	ifFalse: [self errorOutOfBounds: offset]
]

ByteArray _at_: _offset
{
    return (oop)((long)(((unsigned char *)self->v__bytes)[(int)v__offset]));
}

ByteArray at: offset put: anInteger
[
    (0 <= offset and: [offset < size and: [0 <= anInteger and: [anInteger < 256]]])
	ifTrue:  [self at_: offset _integerValue put_: anInteger _integerValue]
	ifFalse: [self errorOutOfBounds: offset].
    ^anInteger.
]

ByteArray at_: _offset put_: _integer
{
    ((char *)self->v__bytes)[(int)v__offset]= (int)v__integer;
}

ByteArray replaceFrom: firstIndex to: lastIndex with: aCollection startingAt: startIndex
[
    | count _src |
    count := lastIndex - firstIndex + 1.
    (aCollection isByteArray
	   and: [startIndex + count <= aCollection size
	   and: [(firstIndex between: 0 and: lastIndex)
	   and: [lastIndex + count <= size]]])
	ifFalse: [^super replaceFrom: firstIndex to: lastIndex with: aCollection startingAt: startIndex].
    _src := aCollection _bytes.
    {
	char *dst= (char *)self->v__bytes + ((long)v_firstIndex >> 1);
	char *src= (char *)      v__src   + ((long)v_startIndex >> 1);
	memcpy(dst, src, (long)v_count >> 1);
    }.
]

"----------------------------------------------------------------"

IdentitySet : Collection ( tally lists )

IdentitySet lists
[
    ^lists
]

IdentitySet new
[
    ^self new: 1.
]

IdentitySet new: width
[
    self := super new.
    tally := 0.
    lists := Array new: width.
]

IdentitySet size	[ ^tally ]

IdSetDebug := [ 0 ]

IdentitySet add: anObject
[
    | listOffset list |
    listOffset := anObject identityHash \\ lists size.
    list := lists at: listOffset.
    list isNil
	ifTrue:
	   [list := lists at: listOffset put: self newList.
	    tally := tally + 1.
	    ^list at: 0 put: anObject].
    0 to: list size - 1 do: [:offset |
	| elt |
	elt := list at: offset.
	elt == nil
	    ifTrue:
	       [tally := tally + 1.
		^list at: offset put: anObject].
	(self compare: elt with: anObject)
	    ifTrue: [^elt]].
    (IdSetDebug := IdSetDebug + 1) = 20 ifTrue: [self error: 'recursion'].
    self
	grow: list at: listOffset;
	add: anObject.
    IdSetDebug := IdSetDebug - 1.
    ^anObject
]

IdentitySet remove: anObject	[ ^self remove: anObject ifAbsent: [self errorNotFound: anObject] ]

IdentitySet remove: anObject ifAbsent: errorBlock
[
    | listOffset list size |
    listOffset := anObject identityHash \\ lists size.
    list := lists at: listOffset.
    list isNil ifTrue: [^errorBlock value].
    0 to: (listOffset := list size - 1) do: [:offset |
	| elt |
	elt := list at: offset.
	elt == nil ifTrue: [^errorBlock value].
	(self compare: elt with: anObject)
	    ifTrue:
	       [offset to: listOffset - 1 do: [:index | list at: index put: (list at: index + 1)].
		list at: listOffset put: nil.
		^anObject]].
    ^errorBlock value
]

IdentitySet newList	[ ^Array new: 4 ]

IdentitySet grow: list at: listOffset
[
"
    list size * 4 > lists size
	ifTrue:  [self widen]
	ifFalse: [self deepen: list at: listOffset]
"
    self deepen: list at: listOffset
]

IdentitySet findElement: anObject
[
    | listOffset list |
    listOffset := anObject identityHash \\ lists size.
    list := lists at: listOffset.
    list isNil ifTrue: [^nil].
    0 to: list size - 1 do: [:offset |
	| elt |
	elt := list at: offset.
	elt == nil ifTrue: [^nil].
	(self compare: elt with: anObject) ifTrue: [^elt]].
    ^nil
]

IdentitySet compare: anEntry with: anObject
[
    ^anEntry == anObject
]


IdentitySet widen
[
    "StdOut nextPutAll: 'WIDEN '; print: self size; space; print: lists size + 1 * 2 - 1; cr."
    lists := (self new: lists size + 1 * 2 - 1) addAll: self; lists.
]

IdentitySet deepen: list at: listOffset
[
    | newList |
    "StdOut nextPutAll: 'DEEPEN '; print: self size; space; print: self lists size; space; print: list size * 2; cr."
    newList := list new: list size * 2.
    newList replaceFrom: 0 to: list size - 1 with: list startingAt: 0.
    lists at: listOffset put: newList.
]

IdentitySet do: unaryBlock
[
    lists do: [:list | list notNil ifTrue: [list do: [:elt | elt notNil ifTrue: [unaryBlock value: elt]]]]
]

IdentitySet includes: anObject
[
    ^(self findElement: anObject) notNil
]


FastIdentitySet : IdentitySet ()

FastIdentitySet newList	[ ^Array new: 1 ]

FastIdentitySet grow: list at: listOffset
[
"
    list size >= 4
	ifTrue:  [self widen]
	ifFalse: [self deepen: list at: listOffset]
"
    self deepen: list at: listOffset
]

"----------------------------------------------------------------"

Association : Object ( key value )

Association key		[ ^key ]
Association value	[ ^value ]

Object -> anObject
[
    ^Association withKey: self value: anObject
]

Association withKey: aKey value: aValue
[
    self := self new.
    key := aKey.
    value := aValue.
]

Association setValue: aValue
[
    ^value := aValue.
]

Association identityHash
[
    ^key identityHash
]

Association printOn: aStream
[
    aStream nextPut: $(; print: key; nextPutAll: ' -> '; print: value; nextPut: $)
]

"----------------------------------------------------------------"

IdentityDictionary compare: anEntry with: anObject
[
    ^anEntry key == anObject key
]

IdentityDictionary at: aKey
[
    ^self at: aKey ifAbsent: [self errorKeyNotFound]
]

IdentityDictionary at: aKey ifAbsent: failBlock
[
    | assoc |
    ^(assoc := self findElement: aKey -> nil) notNil
	ifTrue:  [assoc value]
	ifFalse: [failBlock value]
]

IdentityDictionary at: aKey put: anObject
[
    ^(self add: aKey -> anObject) setValue: anObject
]

IdentityDictionary associationAt: aKey ifAbsent: errorBlock
[
    | assoc |
    assoc := self findElement: aKey -> nil.
    ^assoc notNil ifTrue: [assoc] ifFalse: [errorBlock value]
]

IdentityDictionary includesKey: aKey
[
    ^(self associationAt: aKey ifAbsent: []) notNil
]

IdentityDictionary removeKey: aKey ifAbsent: errorBlock
[
    super remove: aKey -> nil ifAbsent: [^errorBlock value].
    ^aKey
]

"
IdentityDictionary do: unaryBlock
[
    super do: [:assoc | unaryBlock value: assoc value]
]

IdentityDictionary printContentsOn: aStream
[
    aStream nextPutAll: '( '.
    super do: [:assoc | aStream print: assoc; space].
    aStream nextPut: $).
]
"

FastIdentityDictionary : IdentityDictionary ()

FastIdentityDictionary newList	[ ^Array new: 1 ]

FastIdentityDictionary grow: list at: listOffset
[
    list size >= 4
	ifTrue:  [self widen]
	ifFalse: [self deepen: list at: listOffset]
]

"----------------------------------------------------------------"

SlotDictionary : Object ( _size _tally _keys _values default )

SlotDictionary new	[ ^self new: nil ]

SlotDictionary new: defaultElement
{
    struct t_SlotDictionary *this= (struct t_SlotDictionary *)_alloc(v_self, sizeof(struct t_SlotDictionary));
    this->v__size=    (oop)4;
    this->v__tally=   (oop)0;
    this->v__keys=    (oop)_palloc(sizeof(oop) * 4);
    this->v__values=  (oop)_palloc(sizeof(oop) * 4);
    this->v_default=  v_defaultElement;
    return (oop)this;
}

SlotDictionary at: aKey
{
    if (((oop *)self->v__keys)[0] == v_aKey)
      	return ((oop *)self->v__values)[0];
    else if (((oop *)self->v__keys)[1] == v_aKey)
      {
        register oop v1= ((oop *)self->v__values)[1];
			 ((oop *)self->v__keys  )[1]= ((oop *)self->v__keys  )[0];
			 ((oop *)self->v__values)[1]= ((oop *)self->v__values)[0];
			 ((oop *)self->v__keys  )[0]= v_aKey;
	return  	 ((oop *)self->v__values)[0]= v1;
      }
    else
      {
	register long index, lasti= (long)self->v__tally;
        for (index= 2;  index < lasti;  ++index)
	  {
	    if (((oop *)self->v__keys)[index] == v_aKey)
	      {
	        register oop v1= ((oop *)self->v__values)[index];
				 ((oop *)self->v__keys  )[index]= ((oop *)self->v__keys  )[lasti= index - 1];
				 ((oop *)self->v__values)[index]= ((oop *)self->v__values)[lasti];
				 ((oop *)self->v__keys  )[lasti]= v_aKey;
		return  	 ((oop *)self->v__values)[lasti]= v1;
	      }
	  }
      }
    return self->v_default;
}

SlotDictionary at: aKey put: aValue
{
    if (((oop *)self->v__keys)[0] == v_aKey)
      	return ((oop *)self->v__values)[0]= v_aValue;
    else if (((oop *)self->v__keys)[1] == v_aKey)
    {
		((oop *)self->v__keys  )[1]= ((oop *)self->v__keys  )[0];
		((oop *)self->v__values)[1]= ((oop *)self->v__values)[0];
		((oop *)self->v__keys  )[0]= v_aKey;
	return	((oop *)self->v__values)[0]= v_aValue;
    }
    else
    {
	register long index, lasti= (long)self->v__tally;
        for (index= 2;  index < lasti;  ++index)
	{
	    if (((oop *)self->v__keys)[index] == v_aKey)
	    {
			((oop *)self->v__keys  )[index]= ((oop *)self->v__keys  )[lasti= index - 1];
			((oop *)self->v__values)[index]= ((oop *)self->v__values)[lasti];
			((oop *)self->v__keys  )[lasti]= v_aKey;
		return	((oop *)self->v__values)[lasti]= v_aValue;
	    }
	}
    }
    {
        long tally= (long)self->v__tally;
    	if (tally == (long)self->v__size)
        {
            long oldSize= (long)self->v__size;
            long newSize= oldSize * 2;
	    oop *keys=   (oop *)_palloc(sizeof(oop) * newSize);
	    oop *values= (oop *)_palloc(sizeof(oop) * newSize);
	    memcpy(keys,   self->v__keys,   sizeof(oop) * oldSize);
	    memcpy(values, self->v__values, sizeof(oop) * oldSize);
	    self->v__keys=   (oop)keys;
	    self->v__values= (oop)values;
	    self->v__size=   (oop)newSize;
    	}
	((oop *)self->v__keys  )[tally]= v_aKey;
	((oop *)self->v__values)[tally]= v_aValue;
        self->v__tally= (oop)(tally + 1);
    }
    return v_aValue;
}

SlotDictionary at: aKey ifAbsent: errorBlock
[
    | value |
    ^(value := self at: aKey) ~~ default ifTrue: [value] ifFalse: [errorBlock value]
]

"----------------------------------------------------------------"

OrderedCollection : SequenceableCollection ( firstIndex lastIndex array )

OrderedCollection new	[ ^(self new: 4) removeAll ]

OrderedCollection new: size
[
    self := super new.
    array := Array new: size.
    firstIndex := 0.
    lastIndex := size.
]

OrderedCollection size			[ ^lastIndex - firstIndex ]

OrderedCollection add: anObject		[ ^self addLast: anObject ]

OrderedCollection addFirst: anObject
[
    firstIndex == 0 ifTrue: [self makeRoomAtFirst].
    firstIndex := firstIndex - 1.
    array at: firstIndex put: anObject.
    ^anObject
]

OrderedCollection addLast: anObject
[
    lastIndex == array size ifTrue: [self makeRoomAtLast].
    array at: lastIndex put: anObject.
    lastIndex := lastIndex + 1.
    ^anObject
]

OrderedCollection makeRoomAtFirst
[
    | delta newArray |
    delta := array size.
    (self size == 0 and: [delta > 0]) ifTrue: [^self resetToEnd].
    (self size * 2 < delta) ifTrue: [^self shiftToEnd].
    newArray := (array new: self size + delta) replaceFrom: firstIndex + delta to: lastIndex + delta - 1 with: array startingAt: firstIndex.
    array := newArray.
    firstIndex := firstIndex + delta.
    lastIndex := lastIndex + delta.
]

OrderedCollection makeRoomAtLast
[
    (self size == 0 and: [array size > 0]) ifTrue: [^self resetToBeginning].
    (self size * 2 < array size) ifTrue: [^self shiftToBeginning].
    array := (array new: self size + array size)
	replaceFrom: firstIndex to: lastIndex - 1 with: array startingAt: firstIndex;
	yourself.
]

OrderedCollection resetToBeginning	[ firstIndex := lastIndex := 0]
OrderedCollection resetToEnd		[ firstIndex := lastIndex := array size ]

OrderedCollection removeFirst
[
    | first |
    self emptyCheck.
    first := array at: firstIndex.
    firstIndex := firstIndex + 1.
    ^first
]

OrderedCollection removeLast
[
    | last |
    self emptyCheck.
    lastIndex := lastIndex - 1.
    last := array at: lastIndex.
    ^last
]

OrderedCollection removeAll
[
    firstIndex := lastIndex := 0.
]

OrderedCollection indexOf: anObject ifAbsent: errorBlock
[
    self doWithIndex: [:element :index | element == anObject ifTrue: [^index]].
    ^errorBlock value
]

OrderedCollection removeAll: aCollection ifAbsent: errorBlock
[
    aCollection do: [:element | self remove: element ifAbsent: [^errorBlock value]]
]

OrderedCollection remove: anObject ifAbsent: errorBlock
[
    ^self removeAtIndex: (self indexOf: anObject ifAbsent: [^errorBlock value])
]

OrderedCollection removeAtIndex: offset
[
    | arrayOffset |
    arrayOffset := firstIndex + offset.
    (offset >= 0 and: [arrayOffset < lastIndex]) ifFalse: [^self errorOutOfBounds: offset].
    arrayOffset to: lastIndex - 2 do: [:i | array at: i put: (array at: 1 + i)].
    lastIndex := lastIndex - 1.
]

OrderedCollection at: offset
[
    | arrayOffset |
    arrayOffset := firstIndex + offset.
    ^(offset >= 0 and: [arrayOffset < lastIndex])
	ifTrue:  [array at: arrayOffset]
	ifFalse: [self errorOutOfBounds: offset]
]

OrderedCollection at: offset put: anObject
[
    | arrayOffset |
    arrayOffset := firstIndex + offset.
    ^(offset >= 0 and: [arrayOffset < lastIndex])
	ifTrue:  [array at: arrayOffset put: anObject]
	ifFalse: [self errorOutOfBounds: offset]
]

OrderedCollection do: unaryBlock
[
    firstIndex to: lastIndex - 1 do: [:index | unaryBlock value: (array at: index)]
]


"----------------------------------------------------------------"

ConsoleOptions : Object (emacs)

ConsoleOptions emacs	[ ^emacs ]
ConsoleOptions emacs: flag	[ emacs := flag ]
ConsoleOptions new
[
    self := super new.
    emacs := true.
]

[ ConsoleOptions := ConsoleOptions new ]


File : Object ( _fd name )

File isFile	[ ^true ]
Object isFile	[ ^false ]

File withFd: fd
[
    self := self new.
    _fd := fd _integerValue.
    name := '<unknown>'.
]

File withFd: fd name: nameString
[
    self := self new.
    _fd := fd _integerValue.
    name := nameString.
]

File name			[ ^name ]

File print: anObject		[ anObject printOn: self ]
File print: aNumber base: base	[ aNumber printOn: self base: base ]
File cr				[ self nextPut: $\n ]
File cr: n			[ n timesRepeat: [self cr] ]
File space			[ self nextPut: $   ]
File space: n			[ n timesRepeat: [self space] ]
File tab			[ self nextPut: $\t ]
File tab: n			[ n timesRepeat: [self tab] ]

File nextPut: aByte
[
    { if ((long)v_aByte & 1) { char c= _I(v_aByte);  return _O(write((int)self->v__fd, (void *)&c, 1)); } }.
    ^self primitiveFailed
]

File nextPutAll: aString
[
    | _bytes _size |
    _bytes := aString _bytes.
    _size  := aString size _integerValue.
    { return _O(write((int)self->v__fd, (void *)v__bytes, (size_t)v__size)); }.
]

File  read: aCollection			[ ^self read: aCollection size: aCollection size ]
File  read: aCollection size: size	[ ^SmallInteger value_: (self _read_: aCollection _elements size_: size _integerValue) ]
File _read_: _bytes size_: _size	{ return (oop)read((int)self->v__fd, (void *)v__bytes, (size_t)v__size); }

File _isInteractive: _emacs		{ if ((int)v__emacs) {
						int tfd = ((int)(self->v__fd));
						return (oop)(0 <= tfd && tfd <= 2);
					  } else {
						return (oop)isatty((int)self->v__fd); 
					  }
				        }
File isInteractive			[^ self _isInteractive: ConsoleOptions emacs]

File open: pathString
[
    ^self open: pathString ifAbsent: [self error: pathString, ': No such file or directory'].
]

File open: pathString ifAbsent: errorBlock
[
    | descriptor |
    descriptor := self open_: pathString _stringValue.
    ^descriptor >= 0
	ifTrue:  [File withFd: descriptor name: pathString]
	ifFalse: [errorBlock value]
]

File open_: _path
{
    return (oop)(((long)open((char *)v__path, O_RDONLY, 0)) << 1 | 1);
}

File create: pathString
[
    ^self create: pathString ifAbsent: [self error: pathString, ': Cannot open for writing'].
]

File create: pathString ifAbsent: errorBlock
[
    | descriptor |
    descriptor := self create_: pathString _stringValue.
    ^descriptor >= 0
	ifTrue:  [File withFd: descriptor name: pathString]
	ifFalse: [errorBlock value]
]

File create_: _path
{
    return (oop)(((long)open((char *)v__path, O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH)) << 1 | 1);
}

File close
{
    close((long)self->v__fd);
}

File contents
[
    | buf out n |
    buf := ByteArray new: 8192.
    out := WriteStream on: (ByteArray new: 8192).
    [(n := self read: buf) > 0]
	whileTrue:
	   [out next: n putAll: buf].
    ^out contents
]

"----------------------------------------------------------------"

SinkStream : Object ()

SinkStream cr			[]
SinkStream tab			[]
SinkStream space		[]
SinkStream print: o		[]
SinkStream print: o base: b	[]
SinkStream nextPut: o		[]
SinkStream nextPutAll: o	[]

"----------------------------------------------------------------"

ReadStream : Object ( collection position readLimit )

ReadStream collection	[ ^collection ]

ReadStream on: aCollection
[
    self := self new.
    collection := aCollection.
    position := 0.
    readLimit := collection size.
]

ReadStream peek		[ ^self atEnd ifFalse: [collection at: position] ]
ReadStream peek: offset	[ ^position + offset < readLimit ifTrue: [collection at: position + offset] ]
ReadStream next		[ ^self atEnd ifFalse: [collection at: (position := position + 1) - 1] ]
ReadStream atEnd	[ ^position >= readLimit and: [self refill not] ]
ReadStream position	[ ^position ]
ReadStream position: i	[ ^position := i ]
ReadStream refill	[ ^false ]

ReadStream skip: count	[ count timesRepeat: [self next] ]

"----------------------------------------------------------------"

WriteStream : ReadStream ( writeLimit )

WriteStream isWriteStream	[ ^true ]
Object isWriteStream		[ ^false ]

WriteStream on: aCollection
[
    self := super on: aCollection.
    writeLimit := collection size.
    readLimit := 0.
]

WriteStream reset
[
    readLimit := position.
    position := 0.
]

WriteStream nextPut: anObject
[
    position >= writeLimit ifTrue: [self grow].
    collection at: position put: anObject.
    readLimit := (position := position + 1) max: readLimit.
    ^anObject
]

WriteStream nextPutAll: aCollection
[
    aCollection do: [:anObject | self nextPut: anObject].
    ^aCollection
]

WriteStream next: count putAll: seqCollection
[
    position + count >= writeLimit ifTrue: [self grow: (count max: collection size)].
    collection replaceFrom: position to: position + count - 1 with: seqCollection startingAt: 0.
    readLimit := (position := position + count) max: readLimit.
    ^seqCollection
]

WriteStream cr			[ self nextPut: $\n ]
WriteStream tab			[ self nextPut: $\t ]
WriteStream space		[ self nextPut: $   ]
WriteStream print: o		[ o printOn: self ]
WriteStream print: o base: b	[ o printOn: self base: b ]

WriteStream grow		[ self grow: collection size ]

WriteStream grow: count
[
    | newCollection |
    newCollection := collection new: (writeLimit := collection size + count).
    newCollection replaceFrom: 0 to: collection size - 1 with: collection startingAt: 0.
    collection := newCollection.
]

WriteStream contents
[
    ^collection copyFrom: 0 to: position - 1
]

WriteStream fullContents
[
    ^collection copyFrom: 0 to: readLimit - 1
]



WriteStream dump
[
    StdOut nextPutAll: self debugName;
	nextPutAll: '(collection='; nextPutAll: collection debugName; nextPut: $[; print: collection size; nextPut: $];
	nextPutAll: ' position=';   print: position;
	nextPutAll: ' readLimit=';  print: readLimit;
	nextPutAll: ' writeLimit='; print: writeLimit;
	nextPut: $); cr.
]


"----------------------------------------------------------------"

FileStream : ReadStream ( file )

FileStream refill
[
    | n |
    (n := file read: collection) > 0 ifFalse: [^false].
    position := 0.
    readLimit := n.
]

FileStream on: aFile
[
    self := super on: (ByteArray new: 8192).
    file := aFile.
    position := collection size.
]

FileStream close
[
    file close
]

"----------------------------------------------------------------"


ConsoleFileStream : FileStream ( _prompt )

Object isConsoleFileStream		[ ^false ]
ConsoleFileStream isConsoleFileStream	[ ^true ]

ConsoleFileStream on: aFile
[
    ^self on: aFile prompt: (aFile isInteractive ifTrue: ['.'] ifFalse: ['']).
]

ConsoleFileStream on: aFile prompt: aString
[
    self := super on: aFile.
    _prompt := aString _stringValue.
]

ConsoleFileStream prompt: aString
[
    _prompt := aString _stringValue.
]

{ include <readline/readline.h }
{ include <readline/history.h }

ConsoleFileStream _atexit
{
    stifle_history(1000);
    write_history(".jolt-history");
}

ConsoleFileStream refill
[
    | length _buffer _bytes |
    {
	static char *buffer= 0;
	if (buffer)
	{
	    free(buffer);
	    buffer= 0;
	}
	else
	{
	    using_history();
	    read_history(".jolt-history");
	    atexit((void (*)(void))ConsoleFileStream___5fatexit);
	}
	buffer= readline((char *)self->v__prompt);
	if (!buffer) return 0;
	if (*buffer) add_history(buffer);
	v_length= (oop)((strlen(buffer) + 1) << 1 | 1);
	v__buffer= (oop)buffer;
    }.
    collection size < length ifTrue: [collection := collection new: length].
    _bytes := collection _bytes.
    {
	strcpy((char *)v__bytes, (char *)v__buffer);
    }.
    collection at: length - 1 put: $\n.
    position := 0.
    readLimit := length.
]

"----------------------------------------------------------------"

SharedFile : File ( readStream )

SharedFile readStream
[
    readStream isNil ifTrue: [readStream := super readStream].
    ^readStream
]

File readStream
[
    ^(self isInteractive ifTrue: [ConsoleFileStream] ifFalse: [FileStream]) on: self
]

"----------------------------------------------------------------"

[ StdIn  := SharedFile	withFd: 0 name: '<stdin>'  ]
[ StdOut := File	withFd: 1 name: '<stdout>' ]
[ StdErr := File	withFd: 2 name: '<stderr>' ]

"----------------------------------------------------------------"

Link : Object ( prevLink nextLink )

Link prevLink		[ ^prevLink ]
Link prevLink: aLink	[ ^prevLink := aLink ]
Link nextLink		[ ^nextLink ]
Link nextLink: aLink	[ ^nextLink := aLink ]

Link unlink
[
    prevLink nextLink: nextLink.
    nextLink prevLink: prevLink.
]

LinkedList : Object ( firstLink lastLink )

LinkedList first	[ ^firstLink ]
LinkedList last		[ ^lastLink ]

LinkedList addFirst: aLink
[
    firstLink isNil ifTrue: [^firstLink := lastLink := aLink].
    firstLink prevLink: aLink.
    aLink     nextLink: firstLink.
    firstLink := aLink.
    ^aLink.
]

LinkedList addLast: aLink
[
    firstLink isNil ifTrue: [^firstLink := lastLink := aLink].
    lastLink nextLink: aLink.
    aLink    prevLink: lastLink.
    lastLink := aLink.
    ^aLink.
]

LinkedList remove: aLink
[
    aLink == firstLink ifTrue: [^self removeFirst].
    aLink == lastLink  ifTrue: [^self removeLast].
    aLink unlink.
    ^aLink
]

LinkedList do: unaryBlock
[
    | link |
    link := firstLink.
    [link notNil]
	whileTrue:
	   [unaryBlock value: link.
	    link := link nextLink]
]
