{ import: Object }

"----------------------------------------------------------------"

Point : Object ( x y )

Point x	[ ^x ]
Point y	[ ^y ]

Point withX: xValue y: yValue
[
    self := self new.
    x := xValue.
    y := yValue.
]

Number @ aNumber	[ ^Point withX: self y: aNumber ]

PointZero := [ 0@0 ]
PointUnit := [ 1@1 ]

Point zero		[ ^PointZero ]
Point unit		[ ^PointUnit ]

Point min: aPoint	[ ^(x min: aPoint x) @ (y min: aPoint y) ]
Point max: aPoint	[ ^(x max: aPoint x) @ (y max: aPoint y) ]

Point +  aPoint		[ ^self withX: x +  aPoint x y: y +  aPoint y ]
Point -  aPoint		[ ^self withX: x -  aPoint x y: y -  aPoint y ]
Point *  aPoint		[ ^self withX: x *  aPoint x y: y *  aPoint y ]
Point // aPoint		[ ^self withX: x // aPoint x y: y // aPoint y ]

Point <  aPoint		[ ^x <  aPoint x and: [y <  aPoint y] ]
Point <= aPoint		[ ^x <= aPoint x and: [y <= aPoint y] ]

Point r			[ ^(self dotProduct: self) sqrt ]

Point dotProduct: pt	[ ^(x * pt x) + (y * pt y) ]

Point printOn: aStream	[ aStream print: x; nextPut: $@; print: y ]

"----------------------------------------------------------------"

Rectangle : Object ( origin corner )

Rectangle origin	[ ^origin ]
Rectangle corner	[ ^corner ]
Rectangle extent	[ ^corner - origin ]

Rectangle withOrigin: originPoint corner: cornerPoint
[
    self := self new.
    origin := originPoint.
    corner := cornerPoint.
]

Point corner: aPoint	[ ^Rectangle withOrigin: self corner: aPoint ]
Point extent: aPoint	[ ^Rectangle withOrigin: self corner: self + aPoint ]

RectangleZero := [ Point zero corner: Point zero ]

Rectangle zero	[ ^RectangleZero ]

Rectangle includesPoint: aPoint
[
    ^origin <= aPoint and: [aPoint < corner]
]

Rectangle encompassing: listOfPoints
[
    | topLeft bottomRight |
    topLeft := bottomRight := nil.
    listOfPoints do: [:p |
	topLeft == nil
	    ifTrue:  [topLeft := bottomRight := p]
	    ifFalse: [topLeft := topLeft min: p.
		      bottomRight := bottomRight max: p]].
    ^topLeft corner: bottomRight
]

Rectangle insetBy: aPointOrNumber
[
    ^aPointOrNumber isNumber
	ifTrue:  [(origin x + aPointOrNumber) @ (origin y + aPointOrNumber) corner: (corner x - aPointOrNumber) @ (corner y - aPointOrNumber)]
	ifFalse: [self origin + aPointOrNumber corner: self corner - aPointOrNumber]
]

Rectangle translatedBy: aPoint
[
    aPoint isNumber ifTrue: [aPoint := aPoint @ aPoint].
    ^self withOrigin: origin + aPoint corner: corner + aPoint
]

Rectangle contains: aPoint
[
    ^origin <= aPoint and: [aPoint < corner]
]

Rectangle left		[ ^origin x ]
Rectangle top		[ ^origin y ]
Rectangle right		[ ^corner x ]
Rectangle bottom	[ ^corner y ]
Rectangle width		[ ^corner x - origin x ]
Rectangle height	[ ^corner y - origin y ]

Rectangle topLeft	[ ^origin ]
Rectangle bottomRight	[ ^corner ]
Rectangle bottomLeft	[ ^origin x @ corner y ]
Rectangle topRight	[ ^corner x @ origin y ]

Rectangle printOn: aStream	[ aStream nextPut: $(; print: origin; nextPutAll: ' corner: '; print: corner; nextPut: $) ]

"----------------------------------------------------------------"

Vector4 : Object ()
Matrix4x4 : Object ()

{ import: Matrix }
"
{ import: FloatArray }
"
