.SYNTAX Program

PrimaryExpression =
  PrimaryExpressionHead $PrimaryExpressionTail;

PrimaryExpressionHead =
  "(" Expression ")" /
  "function" .PRE["function"] (.ID .OUT(*) / .EMPTY .NIL) "(" Formals ")" "{" SourceElements "}" .CLOSE /
  "new" .PRE["new"] PrimaryExpression .CLOSE /
  .ID .PRE["identifier"] .OUT(*) .CLOSE /
  .NUMBER .PRE["number"] .OUT(*) .CLOSE /
  .STRING .PRE["string"] .OUT(*) .CLOSE;

PrimaryExpressionTail =
  "[" .POST["array-access"] Expression "]" .CLOSE /
  "." .POST["field-access"] .ID .OUT(*) .CLOSE /
  "(" .POST["call"] ArgumentList ")" .CLOSE;

Formals =
  .ID .POST .OUT(*) $("," .ID .OUT(*)) .CLOSE /
  .EMPTY .NIL;

ArgumentList =
  Expression .POST $("," Expression) .CLOSE /
  .EMPTY .NIL;

PostfixExpression =
  PrimaryExpression ("++" .POST["increment"] .CLOSE /
                     "--" .POST["decrement"] .CLOSE /
                     .EMPTY);

UnaryExpression =
  "+" UnaryExpression /
  "-" .PRE["negate"] UnaryExpression .CLOSE /
  "!" .PRE["not"]    UnaryExpression .CLOSE /
  PostfixExpression;

MultiplicativeExpression =
  UnaryExpression $("*" .POST["times"]   UnaryExpression .CLOSE /
                    "/" .POST["divide"]  UnaryExpression .CLOSE /
                    "%" .POST["modulus"] UnaryExpression .CLOSE);

AdditiveExpression =
  MultiplicativeExpression $("+" .POST["plus"] MultiplicativeExpression .CLOSE /
                             "-" .POST["minus"] MultiplicativeExpression .CLOSE);

RelationalExpression =
  AdditiveExpression $("<=" .POST["leq"] AdditiveExpression .CLOSE /
                       "<"  .POST["lt"]  AdditiveExpression .CLOSE / 
                       ">=" .POST["geq"] AdditiveExpression .CLOSE / 
                       ">"  .POST["gt"]  AdditiveExpression .CLOSE);

EqualityExpression =
  RelationalExpression $("==" .POST["equals"] RelationalExpression .CLOSE /
                         "!=" .POST["not-equals"] RelationalExpression .CLOSE);

LogicalANDExpression =
  EqualityExpression $("&&" .POST["logical-AND"] EqualityExpression .CLOSE);

LogicalORExpression =
  LogicalANDExpression $("||" .POST["logical-OR"] LogicalANDExpression .CLOSE);

Expression =
  LogicalORExpression ("?" .POST["cond-expr"] Expression ":" Expression .CLOSE /
                       "=" .POST["assignment"] Expression .CLOSE /
                       .EMPTY);

Statement =
  "{" StatementList "}" /
  "var" .PRE["var-decls"] VariableDeclarationList ";" .CLOSE /
  "if" .PRE["if"] "(" Expression ")" Statement ("else" Statement / .EMPTY .NIL) .CLOSE /
  "do" .PRE["do-loop"] Statement "while" "(" Expression ")" .CLOSE ";" /
  "for" .PRE["for-loop"] "(" "var" .PRE["var-decls"] VariableDeclarationList .CLOSE ";" (Expression / .EMPTY .NIL) ";"
    (Expression / .EMPTY .NIL) ")" Statement .CLOSE /
  "while" .PRE["while-loop"] "(" Expression ")" Statement .CLOSE /
  "continue" ";" .OUT("continue") /
  "break" ";" .OUT("break") /
  "return" (Expression .POST["return-val"] .CLOSE / .EMPTY .OUT("return")) ";" /
  Expression ";";

VariableDeclarationList =
  VariableDeclaration $("," VariableDeclaration);

VariableDeclaration =
  .ID .PRE .OUT(*) ("=" Expression / .EMPTY .NIL) .CLOSE;

StatementList =
  .PRE["block"] (Statement $Statement / .EMPTY) .CLOSE;

FunctionDeclaration =
 "function" .PRE["function"] .ID .OUT(*) "(" Formals ")" "{" SourceElements "}" .CLOSE;

SourceElement =
  FunctionDeclaration /
  Statement;

SourceElements =
  .PRE (SourceElement $SourceElement / .EMPTY) .CLOSE;

Program = SourceElements;

.END

