#[StdErr nextPutAll: '"Parsing and compiling JavaScript library... "];

#(sugar (makeJSType ct p)
   `(begin (define [,ct lookup: k]	[,p lookup: k])
	   (define [,ct set: k to: v]	[,p set: k to: v])
	   (define [,ct bind: k to: v]	[,p set: k to: v])));

#(define _ctxt		[ObjectProto withProto: 0]);
#(set    ObjectProto	[ObjectProto withProto: 0]);
#(define FunctionProto	[ObjectProto child]);

function Object() { }
Object.prototype = #ObjectProto;
Object.prototype.__array__ = #0;
Object.prototype.__class__ = "Object";
Object.prototype.toString = function() { return "[object " + this.__class__ + "]"; };

#(define true (import "_object"));
#(define TrueProto [ObjectProto child]);
#(makeJSType true TrueProto);
TrueProto = #TrueProto;
TrueProto.toString = function() { return "true"; };
true = #true;

#(define UndefinedObject (import "UndefinedObject"));
#(define NullProto [ObjectProto child]);
#(makeJSType UndefinedObject NullProto);
NullProto = #NullProto;
NullProto.toString = function() { return "null"; };
null = #0;
false = #0;

#(define Float (import "Float"));
function Number(n) { #[OS error: '"Number constructor not implemented!"]; }
#(define NumberProto (js-get (js-get _ctxt 'Number) 'prototype));
#(makeJSType SmallInteger NumberProto);
#(makeJSType Float NumberProto);
#(define [Float // f] [self / f]);
Number.prototype["+"] = #(lambda (this that)
			   (if [that isString]
			     (js-bsend '+ (js-send 'toString this) that)
			     [this + that]));
Number.prototype["-"] = #(lambda (this that) [this - that]);
Number.prototype["*"] = #(lambda (this that) [this * that]);
Number.prototype["/"] = #(lambda (this that) [this // that]);
Number.prototype["%"] = #(lambda (this that) [this \\ that]);
Number.prototype["<"] = #(lambda (this that) [this < that]);
Number.prototype["<="] = #(lambda (this that) [this <= that]);
Number.prototype[">"] = #(lambda (this that) [this > that]);
Number.prototype[">="] = #(lambda (this that) [this >= that]);
Number.prototype.toFixed = function(n) { return #[(js-get _ctxt 'this) toFixed: (js-get _ctxt 'n)]; };
Number.prototype.toString = function() { return #[(js-get _ctxt 'this) asString]; };

Math = new Object();
Math.PI = #[Float pi];
Math.floor = function(f) { return #[(js-get _ctxt 'f) asInteger]; };

function String(s) { this = s.toString(); }
#(define StringProto (js-get (js-get _ctxt 'String) 'prototype));
#(makeJSType String StringProto);
String.prototype.toString = function() { return this; };
String.prototype["+"] = #(lambda (this that) [this + (js-send 'toString that)]);

document = new Object();
document.write = function(x) {
  #[StdOut nextPutAll: (js-send 'toString (js-get _ctxt 'x))];
  #[StdOut cr];
};

function Array() {
  var l = arguments.length == 1 ? arguments[0] : 0;
  this.__array__ = #[Array new: (js-get _ctxt 'l)];
  this.length = l;
  for (var idx = 0; idx < l; idx++)
    this[idx] = null;
}
Array.prototype.toString = function() {
  var r = "[";
  for (var idx = 0; idx < this.length; idx++) {
    if (idx > 0)
      r += ",";
    r += this[idx].toString();
  }
  r += "]";
  return r;
};

#(define Date (import "Date"));
function Date() { this.__date__ = #[Date new]; }
Date.prototype.getSeconds = function() { return #[(js-get (js-get _ctxt 'this) '__date__) seconds]; };
Date.prototype.getMilliseconds = function() { return #[(js-get (js-get _ctxt 'this) '__date__) milliseconds]; };
Date.prototype.getMicroseconds = function() { return #[(js-get (js-get _ctxt 'this) '__date__) microseconds]; };
Date.prototype.toString = function() { return (this.getSeconds() * 1000 + this.getMilliseconds()).toString(); };

function eval(e) {
  e = e.toString();
  return #(let ((e (js-get _ctxt 'e))
		(_input [CheckpointStream on: e])
		(tree (APPLY SrcElem)))
	    (if tree
	      0
	      (begin
		(set _input [CheckpointStream on: [e + '";"]])
		(set tree (APPLY SrcElem))))
	    (if tree
	      [tree _eval]
	      (begin
		[StdOut nextPutAll: '"eval: "] [StdOut print: e]
		[OS error: '"... is not valid JavaScript"])));
}

#(define usleep (dlsym "usleep"));
function usleep(t) { return #(usleep [(js-get _ctxt 't) _integerValue]); }

//#(define sranddev (dlsym "sranddev"));
//function sranddev() { return #(sranddev); }
function sranddev() { return 0; }

#(define rand (dlsym "rand"));
function rand() { return #[SmallInteger value_: (rand)]; }

#[StdErr nextPutAll: '"done.\n"];
