[StdErr nextPutAll: '"Parsing and compiling JavaScript compiler... "]

with NoTokenizer {
  JSSpace	::= (' ' | '\t' | '\r' | '\n' |
		     "//" (~'\n' Char)* '\n' |
		     "/*" (~"*/" Char)* "*/")					=> '" ",
  JSNumber	::= Digit+:ids
		      ('.' Digit+:fds => `(number ,[Float fromString: [[[ids concatenated] + '"."] + [fds concatenated]]]) |
		       Empty	      => `(number ,(fixNumber [ids concatenated]))),
  JSSpecial	::= ('(' | ')' | '{' | '}' | '[' | ']' | ',' | ';' | '?' | ':' |
		     "!=" | "==" | '=' | ">=" | '>' | "<=" | '<' | "++" | "+=" |
		     '+' | "--" | "-=" | '-' | "*=" | '*' | "/=" | '/' | "%=" |
		     '%' | "&&=" | "&&" | "||=" | "||" | '.' | '!' | '#'):s	=> `(,[s asSymbol]),
  JSName	::= (Letter | '_'):first (Letter | Digit | '_')*:rest		=> (let ((s [[`(,first ,@rest) concatenated]
															 asSymbol]))
										     (if (or (== s 'break) (== s 'continue)
											     (== s 'do) (== s 'else) (== s 'for)
											     (== s 'function) (== s 'if) (== s 'new)
											     (== s 'return) (== s 'var)
											     (== s 'while))
										       `(,s)
										       `(name ,s))),
  JSTokenizer	::= JSSpace* (JSName | JSNumber | Str | JSSpecial)
}

with JSTokenizer {
  Expr		::= Empty,									; forward reference
  SrcElem	::= Empty,									; forward reference
  ArgList	::= Expr:first ("," Expr)*:rest							=> `(,first ,@rest)
		  | Empty									=> '(),
  Formals	::= #name:first ("," #name)*:rest						=> `(,first ,@rest)
		  | Empty									=> '(),
  FuncRest	::= "(" Formals:args ")" "{" SrcElem*:body "}"					=> `(js-func ,args (begin ,@body)),
  PrimExprHd	::= "(" Expr:e ")"								=> e
		  | #name:n									=> `(js-get _ctxt ',n)
		  | #number:n									=> `(quote ,n)
		  | #string:s									=> `(quote ,s)
		  | "function" FuncRest
		  | "new" #name:n "(" ArgList:a ")"						=> `(js-new ',n ,@a)
		  | "[" "]"									=> '(js-arr)
		  | "[" Expr:first ("," Expr)*:rest "]"						=> `(js-arr ,first ,@rest),
  PrimExpr	::= PrimExprHd:r
		      ("[" Expr:i "]" r<-`(js-arr-get ,r ,i) |
		       "." #name:m "(" ArgList:a ")" r<-`(js-send ',m ,r ,@a) |
		       "." #name:f r<-`(js-get ,r ',f) |
		       "(" ArgList:a ")" r<-`(js-call ,r ,@a))*					=> r,
  Postfix	::= PrimExpr:r ("++" r<-`(js-post-increment ,r) |
				"--" r<-`(js-post-decrement ,r) |
				Empty)								=> r,
  UnaryExpr	::= "-" Postfix:r								=> `[,r negated]
		  | "!" Postfix:r								=> `(js-not (js-bool ,r))
		  | ("+" | Empty) Postfix,
  MultExpr	::= UnaryExpr:r (("*" | "/" | "%"):o UnaryExpr:s r<-`(js-bsend ',o ,r ,s))*	=> r,
  AddExpr	::= MultExpr:r (("+" | "-"):o MultExpr:s r<-`(js-bsend ',o ,r ,s))*		=> r,
  RelExpr	::= AddExpr:r ((">" | ">=" | "<" | "<="):o AddExpr:s r<-`(js-bsend ',o ,r ,s))*	=> r,
  EqExpr	::= RelExpr:r (("==" | "!="):op RelExpr:s r<-`(if (,op ,r ,s) true 0))*		=> r,
  AndExpr	::= EqExpr:r ("&&" EqExpr:s r<-`(js-and (js-bool ,r) (js-bool ,s)))*		=> r,
  OrExpr	::= AndExpr:r ("||" AndExpr:s r<-`(js-or (js-bool ,r) (js-bool ,s)))*		=> r,
  Expr		::= "#" Coke
		  | OrExpr:r ("?" Expr:t ":" Expr:f r<-`(#"if" (js-bool ,r) ,t ,f) |
			      "=" Expr:rhs r<-`(js-set ,r ,rhs) |
			      "+=" Expr:rhs r<-`(js-set ,r (js-bsend '+ ,r ,rhs)) |
			      "-=" Expr:rhs r<-`(js-set ,r (js-bsend '- ,r ,rhs)) |
			      "*=" Expr:rhs r<-`(js-set ,r (js-bsend '* ,r ,rhs)) |
			      "/=" Expr:rhs r<-`(js-set ,r (js-bsend '/ ,r ,rhs)) |
			      "%=" Expr:rhs r<-`(js-set ,r (js-bsend '% ,r ,rhs)) |
			      "&&=" Expr:rhs r<-`(js-set ,r (and ,r ,rhs)) |
			      "||=" Expr:rhs r<-`(js-set ,r (or ,r ,rhs)) |
			      Empty)								=> r,
  Binding	::= #name:n ("=" Expr | Empty => '(js-get _ctxt 'null)):e			=> `(js-decl ',n ,e),
  Stmt		::= "{" Stmt*:s "}"								=> `(begin ,@s)
		  | "var" Binding:first ("," Binding)*:rest ";"					=> `(begin ,first ,@rest)
		  | "if" "(" Expr:c ")" Stmt:t ("else" Stmt |
						Empty => '0):f					=> `(if (js-bool ,c) ,t ,f)
		  | "while" "(" Expr:c ")" Stmt:s						=> `(while (js-bool ,c) ,s)
		  | "do" Stmt:stmt "while" "(" Expr:cond ")" ";"				=> `(while (begin ,stmt ,cond))
		  | "for" "(" ("var" Binding | Expr):init ";" Expr:cond ";" Expr:upd ")" Stmt:s	=> `(begin ,init
												      (while ,cond (begin ,s ,upd)))
		  | "break" ";"									=> '(break)
		  | "continue" ";"								=> '(continue)
		  | "return" (Expr:e => `(#return ,e) |
			      Empty => '(#return)):r ";"					=> r
		  | Expr:e ";"									=> e,
  SrcElem	::= "function" #name:n FuncRest:f						=> `(js-func-decl ',n ,f)
		  | Stmt
}

[StdErr nextPutAll: '"done.\nEntering JavaScript read-eval-print loop.\n"]
(while [[OS argumentFiles] notEmpty]
  (read-eval-print SrcElem [OS nextArgumentFile] 0 1 0))
