//sranddev();

Array.prototype.mirror = function() {
  var r = new Array(this.length);
  for (var i = 0; i < this.length; i++) {
    r[i] = new Array(this[i].length);
    for (var j = 0; j < this[i].length; j++)
      r[i][this[i].length - j - 1] = this[i][j];
  }
  return r;
};

Array.prototype.transpose = function() {
  var r = new Array(this[0].length);
  for (var j = 0; j < this[0].length; j++) {
    r[j] = new Array(this.length);
    for (var i = 0; i < this.length; i++)
      r[j][i] = this[i][j];
  }
  return r;
};

Array.prototype.flip = function() {
  var r = this.mirror().transpose();
  r.color = this.color;
  return r;
};

function colorBlock(shape, r, g, b) {
  shape.color = (r * 256 + g) * 256 + b;
  return shape;
}

var blocks = [colorBlock([[1, 1, 1, 1]],	255,   0,   0),

	      colorBlock([[1, 0, 0],
			  [1, 1, 1]],		255, 255,   0),

	      colorBlock([[0, 1, 1],
			  [1, 1, 0]],		  0, 255,   0),

	      colorBlock([[0, 1, 0],
			  [1, 1, 1]],		150, 150, 150),

	      colorBlock([[1, 1],
			  [1, 1]],		  0,   0, 255)];

function nextBlock() {
  var r = blocks[rand() % blocks.length];
  return (rand() % 2) == 0 ? r : r.flip();
}

var x = 5, y = -5, block = nextBlock();
var bl = 15, cols = 10, rows = 20;

var spaces = new Array(rows);
for (var i = 0; i < rows; i++) {
  spaces[i] = new Array(cols);
  for (var j = 0; j < cols; j++)
    spaces[i][j] = 0;
}

function collides(x, y, b) {
  for (var i = 0; i < b.length; i++)
    for (var j = 0; j < b[i].length; j++)
      if (b[i][j] != 0 && (x + j < 0 ||
			   x + j >= cols ||
			   y + i >= rows ||
			   y + i >= 0 && spaces[y + i][x + j] != 0))
	return true;
  return false;
}

function park(x, y, b) {
  for (var i = 0; i < b.length; i++)
    for (var j = 0; j < b[i].length; j++)
      if (b[i][j] != 0 &&
	  x + j >= 0 && x + j < cols &&
	  y + i >= 0 && y + i < rows)
	spaces[y + i][x + j] = b.color;
  var i = rows - 1;
  while (i >= 0) {
    var complete = true;
    for (var j = 0; j < cols; j++)
      if (spaces[i][j] == 0)
	complete = false;
    if (complete) {
      for (var ii = i; ii > 0; ii--)
	spaces[ii] = spaces[ii - 1];
      spaces[0] = new Array(cols);
      for (var j = 0; j < cols; j++)
	spaces[0][j] = 0;
    }
    else
      i--;
  }
}

function drawSquare(c, x, y) {
  pix.foreground(c);
  pix.fillRectangle(new Point(x * bl, y * bl).extent(new Point(bl, bl)));
  pix.foreground(c * 2 / 3);
  pix.drawLine(new Point(x * bl, (y + 1) * bl - 1), new Point((x + 1) * bl - 1, (y + 1) * bl - 1));
  pix.drawLine(new Point((x + 1) * bl - 1, y * bl), new Point((x + 1) * bl - 1, (y + 1) * bl - 1));
}

function draw() {
  pix.foreground(0); pix.fillRectangle(new Point(0, 0).extent(pix.extent()));
  pix.lineWidth(1);
  for (var i = 0; i < rows; i++)
    for (j = 0; j < cols; j++)
      if (spaces[i][j] != 0)
	drawSquare(spaces[i][j], j, i);
  for (var i = 0; i < block.length; i++)
    for (var j = 0; j < block[i].length; j++)
      if (block[i][j] != 0)
	drawSquare(block.color, x + j, y + i);
  pix.draw(win, new Point(0, 0));
}

Event.prototype.doIt = function() { };

ExposeEvent.prototype.doIt = function() { draw(); };

var done = false, UP = 30, DOWN = 31, LEFT = 17, RIGHT = 16, ESC = 27;

KeyPressEvent.prototype.doIt = function() {
  var k = this.keyCode();
  if (k == ESC)
    done = 1;
  else if (k == UP) {
    var midX = x + block[0].length / 2;
    var color = block.color;
    var tmpBlock = block.flip(), tmpX = midX - block[0].length / 2;
    if (!collides(tmpX, y, tmpBlock)) {
      x = tmpX;
      block = tmpBlock;
      draw();
    }
  }
  else if (k == DOWN && !collides(x, y + 1, block)) {
    y = y + 1;
    draw();
  }
  else if (k == LEFT && !collides(x - 1, y, block)) {
    x = x - 1;
    draw();
  }
  else if (k == RIGHT && !collides(x + 1, y, block)) {
    x = x + 1;
    draw();
  }
};

var win = new Surface(new Point(100, 100).extent(new Point(cols * bl, rows * bl))), pix = win.newPixmap(win.extent());

var ticks = 0;
while (!done) {
    usleep(2500);
  ticks = ticks + 1;
  if (ticks == 100) {
    if (collides(x, y + 1, block)) {
      park(x, y, block);
      block = nextBlock();
      y = -block.length;
      x = (cols - block[0].length) / 2;
    }
    else
      y = y + 1;
    draw();
    ticks = 0;
  }
  var e = win.pollEvent();
  if (e != null)
    e.doIt();
}
