#ifndef __cvpu_h
#define __cvpu_h

typedef struct _CVPU CVPU;
typedef struct _CLabel CLabel;

void  vpu_setMalloc(void *mallocFn);
void  vpu_setFree(void *freeFn);

CVPU *vpu_new(void);

void *vpu_newLocalLabel(CVPU *vpu);
CVPU *vpu_DefineLocalLabel(CVPU *vpu, void *l);

CVPU *vpu_Begin(CVPU *vpu, int n);
CVPU *vpu_Define(CVPU *vpu, int n);
CVPU *vpu_DefineLabel(CVPU *vpu, CLabel *l);
//CVPU *vpu_Define(CVPU *vpu, Label *l);
//CVPU *vpu_Define(CVPU *vpu, LocalLabel *l);
CVPU *vpu_End(CVPU *vpu, int n);

// function, args, temps

CVPU *vpu_Ienter(CVPU *vpu),		*vpu_Lenter(CVPU *vpu);
CVPU *vpu_Iarg(CVPU *vpu, int n),	*vpu_Larg(CVPU *vpu, int n);
CVPU *vpu_Itmp(CVPU *vpu, int n),	*vpu_Ltmp(CVPU *vpu, int n);
CVPU *vpu_LdArg(CVPU *vpu, int n);
CVPU *vpu_StArg(CVPU *vpu, int n);
CVPU *vpu_LdTmp(CVPU *vpu, int n);
CVPU *vpu_StTmp(CVPU *vpu, int n);
CVPU *vpu_DropTmp(CVPU *vpu, int n);
CVPU *vpu_LdFP(CVPU *vpu);
CVPU *vpu_AddrTmp(CVPU *vpu, int n);
CVPU *vpu_AddrArg(CVPU *vpu, int n);
//CVPU *vpu_Addr(CVPU *vpu, LocalLabel *l);
CVPU *vpu_SaveArgs(CVPU *vpu);
CVPU *vpu_Alloc(CVPU *vpu);

// stack

CVPU *vpu_LdInt(CVPU *vpu, int i);
//CVPU *vpu_Ld(CVPU *vpu, long long i);
CVPU *vpu_LdStr(CVPU *vpu, char *s);
CVPU *vpu_LdPtr(CVPU *vpu, void *p);
CVPU *vpu_Dup(CVPU *vpu, int n);
CVPU *vpu_Put(CVPU *vpu, int n);
CVPU *vpu_Drop(CVPU *vpu, int n);

// conditional

CVPU *vpu_Lt(CVPU *vpu);
CVPU *vpu_Le(CVPU *vpu);
CVPU *vpu_Eq(CVPU *vpu);
CVPU *vpu_Ne(CVPU *vpu);
CVPU *vpu_Ge(CVPU *vpu);
CVPU *vpu_Gt(CVPU *vpu);
CVPU *vpu_Not(CVPU *vpu);

CVPU *vpu_Bf(CVPU *vpu, int n);
CVPU *vpu_Bt(CVPU *vpu, int n);
CVPU *vpu_Br(CVPU *vpu, int n);
CVPU *vpu_BrLocal(CVPU *vpu, void *l);	//CVPU *vpu_Br(CVPU *vpu, LocalLabel *l);
CVPU *vpu_Bi(CVPU *vpu);

// call

CVPU *vpu_Icalli(CVPU *vpu, int n);
CVPU *vpu_Icall(CVPU *vpu, int n, void *fn);
CVPU *vpu_IcallLabel(CVPU *vpu, int n, CLabel *l);
CVPU *vpu_Ret(CVPU *vpu);

// arithmetic

CVPU *vpu_Icvt(CVPU *vpu);

CVPU *vpu_Neg(CVPU *vpu);

CVPU *vpu_Add(CVPU *vpu);
CVPU *vpu_Sub(CVPU *vpu);
CVPU *vpu_Div(CVPU *vpu);
CVPU *vpu_Mul(CVPU *vpu);
CVPU *vpu_Mod(CVPU *vpu);
CVPU *vpu_And(CVPU *vpu);
CVPU *vpu_Or (CVPU *vpu);
CVPU *vpu_Xor(CVPU *vpu);
CVPU *vpu_Asl(CVPU *vpu);
CVPU *vpu_Asr(CVPU *vpu);
CVPU *vpu_Lsl(CVPU *vpu);
CVPU *vpu_Lsr(CVPU *vpu);

// memory

CVPU *vpu_Rdb(CVPU *vpu);
CVPU *vpu_Rdh(CVPU *vpu);
CVPU *vpu_Rdw(CVPU *vpu);
CVPU *vpu_Wrb(CVPU *vpu);
CVPU *vpu_Wrh(CVPU *vpu);
CVPU *vpu_Wrw(CVPU *vpu);

CVPU *vpu_dump(CVPU *vpu);
void *vpu_compile(CVPU *vpu, unsigned flags, unsigned debug);
int   vpu_codeSize(CVPU *vpu);

void  vpu_iflush(void *first, void *last);

char *vpu_versionString(CVPU *vpu);

CLabel *label_new(void);

void *label_getAddress(CLabel *l);
void *label_setAddress(CLabel *l, void *a);
int   label_call0(CLabel *l);
int   label_call1(CLabel *l, int a);

#endif /* __cvpu_h */
