[StdOut nextPutAll: '"Parsing and compiling Monty compiler... "]

;;;
;;; lexical analyzer
;;;

MSpace		::= ' '								=> '" "
		  | '\t'							=> '"        ";

MName		::= (Letter | "_") (Letter | Digit | "_")*			=> (let ((s [[[_acc flattened] concatenated] asSymbol]))
										     (if (or (== s 'and)    (== s 'class)   (== s 'def)
										     	     (== s 'elif)   (== s 'else)    (== s 'from)
											     (== s 'if)	    (== s 'import)  (== s 'lambda)
											     (== s 'not)    (== s 'or)	    (== s 'pass)
											     (== s 'print)  (== s 'return)  (== s 'while))
										       `(,s)
										       `(name ,s)));

MNumber		::= Digit+							=> `(number ,(fixNumber [[_acc flattened] concatenated]));

MSpecial	::= ('(' | ')' | '{' | '}' | '[' | ']' | ',' | ':' | '.' |
		     "==" | '=' | ';' | '+' | '-' | '*' | '/' | "<=" | '<' |
		     ">=" | '>' | "!=")						=> `(,[[[_acc flattened] concatenated] asSymbol]);

iMToken		::= ?[[_input indent] < [[_input indentationStack] last]]	=> (begin [[_input indentationStack] removeLast]
											  [_input firstOfLine: 0]
											  '(dedent))
		  | ?(!= [_input indent] [[_input indentationStack] last])	=> [OS error: '"indentation error"]
		  | ?[_input firstOfLine]
		      (MSpace*):space
		      ![_input indent: (if [[[space first] size] > '0]
					 [[[space flattened] concatenated] size]
					 '0)]
		      !(define r 0)
		      (?[[_input indent] > [[_input indentationStack] last]]
			 ![[_input indentationStack] addLast: [_input indent]]
			 !(set r '(indent)) |
		       ?[[_input indent] < [[_input indentationStack] last]]
			 ![[_input indentationStack] removeLast]
			 !(set r '(dedent)))					=> (begin [_input firstOfLine: 0]
											  r)
		  | MSpace* !(define r 0)
		      (MName->r | MNumber->r | Str->r | MSpecial->r)		=> (begin [_input firstOfLine: 0]
											  r)
		  | MSpace* "\n"						=> (begin [_input firstOfLine: 1]
											  '(newline));

MToken		::= iMToken:r							=> (begin ;[StdOut print: r] [StdOut cr]
											  r);
;;;
;;; parser
;;;

(set Tokenizer	'MToken)

(define MExpr 0) ; forward reference

MKVPair		::= MExpr:k #":" MExpr:v					=> `(,k ,v);

MAtom		::= #name:r 							=> `(mp_lookup $ctxt ',r)
		  | #string:r							=> `',r
		  | #number:r							=> `',r
		  | #"(" MExpr:r #")"						=> r
		  | !(define r [WriteStream on: [Expression new: '8]]) #"{"
		      [MKVPair:kv ![r nextPut: kv]
		       (#"," MKVPair->kv ![r nextPut: kv])*] #"}"		=> `(mp_dictionary ,@[r contents])
		  | !(define r [WriteStream on: [Expression new: '8]]) #"["
		       MExpr:e ![r nextPut: e]
		      (#"," MExpr->e ![r nextPut: e])* #"]"			=> `(mp_list ,@[r contents]);

MExprList	::= !(define r [WriteStream on: [Expression new: '8]])
		      MExpr:e ![r nextPut: e]
		      (#"," MExpr->e ![r nextPut: e])*				=> [r contents];

MArgs		::= !(define r '()) #"(" [MExprList:args !(set r args)] #")"	=> r;

MPrimary	::= MAtom:r
		      (#"." #name:msg MArgs:args
			 !(set r `(mp_send ',msg ,r ,@args)) |
		       #"." #name:fd !(set r `(mp_lookup ,r ',fd)) |
		       #"[" MExpr:idx #"]" !(set r `(mp_subscript ,r ,idx)) |
		       MArgs:args !(set r `(mp_call ,r ,@args)))*		=> r;

MUnExpr		::= #"-" MUnExpr:r						=> `(mp_sub '0 ,r)
		  | #"+" MUnExpr:r						=> r
		  | MPrimary:r 							=> r;

MMulExpr	::= MUnExpr:r (#"*" MUnExpr:s !(set r `(mp_mul ,r ,s)) |
			       #"/" MUnExpr:s !(set r `(mp_div ,r ,s)))*	=> r;

MAddExpr	::= MMulExpr:r (#"+" MMulExpr:s !(set r `(mp_add ,r ,s)) |
				#"-" MMulExpr:s !(set r `(mp_sub ,r ,s)))*	=> r;

MCompExpr	::= MAddExpr:r [#"<" MAddExpr:s !(set r `(mp_lt ,r ,s)) |
				#"<=" MAddExpr:s !(set r `(mp_le ,r ,s)) |
				#">" MAddExpr:s !(set r `(mp_gt ,r ,s)) |
				#">=" MAddExpr:s !(set r `(mp_ge ,r ,s)) |
				#"==" MAddExpr:s !(set r `(mp_eq ,r ,s)) |
				#"!=" MAddExpr:s !(set r `(mp_ne ,r ,s))] 	=> r;

MNotExpr	::= #not MCompExpr:r						=> `(mp_not ,r)
		  | MCompExpr:r							=> r;

MAndExpr	::= MNotExpr:r (#and MNotExpr:s !(set r `(mp_and ,r ,s)))*	=> r;

MOrExpr		::= MAndExpr:r (#or MAndExpr:s !(set r `(mp_or ,r ,s)))*	=> r;

MFormals	::= !(define r [WriteStream on: [Expression new: '8]])
		      [#name:a ![r nextPut: a]
		        (#"," #name->a ![r nextPut: a])*]			=> [r contents];

MExpr		::= #lambda MFormals:args #":" MExpr:body			=> `(mp_lambda ,args ,body)
		  | MOrExpr:r							=> r;

MSimpleStmt	::= #pass							=> 1
		  | #name:t #"=" MExpr:e					=> `(mp_bind $ctxt ',t ,e)
		  | #print MExprList:args					=> `(mp_print ,@args)
		  | #return MExpr:arg						=> `(mp_return ,arg)
		  | #from #name #import #name					=> 1
		  | MExpr:r							=> r;

MStmtList	::= !(define r [WriteStream on: [Expression new: '8]])
		      MSimpleStmt:s ![r nextPut: s]
		      (#";" MSimpleStmt->s ![r nextPut: s])+ [#";"]		=> `(begin ,@[r contents])
		  | MSimpleStmt:r [#";"]					=> r;

(define MStmt 0) ; forward reference

MSuite		::= MStmtList:r	#newline					=> r
		  | #newline #indent
		      !(define r [WriteStream on: [Expression new: '8]])
		      (MStmt:s ![r nextPut: s])+ #dedent			=> `(begin ,@[r contents]);

MStmt		::= MStmtList:r #newline					=> r
		  | #if MExpr:c #":" MSuite:t
		      !(define r `(mp_if ,c ,t 0)) !(define curr r)
		      (#elif MExpr->c #":" MSuite->t
			 ![curr fourth: `(mp_if ,c ,t 0)]
			 !(set curr [curr fourth]))*
		      [#else #":" MSuite->t ![curr fourth: t]]			=> r
		  | #while MExpr:c #":" MSuite:b
		      !(define r `(mp_while ,c ,b 0))
		      [#else #":" MSuite->b ![r fourth: b]]			=> r
		  | #def #name:f #"(" MFormals:args #")" #":" MSuite:body	=> `(mp_bind $ctxt ',f (mp_lambda ,args ,body))
		  | #class #name:c !(define scs '())
		      [#"(" MFormals->scs #")"] #":" MSuite:body		=> `(mp_bind $ctxt ',c (mp_class ,scs ,body));

MTopLevel	::= #newline							=> 1
		  | MStmt:r							=> r;

[StdOut nextPutAll: '"done.\n"]

[StdErr nextPutAll: '"Entering Monty read-eval-print loop.\n"]

(while [[OS argumentFiles] notEmpty]
  (read-eval-print MTopLevel [OS nextArgumentFile] 0 1 0))
