{ import: Object-imp }
{ import: Object-lib }

String asPrototype
{
    return _local_import((char *)self->v__bytes);
}

UndefinedLoadMethod : Object ()

UndefinedLoadMethod loadFrom: aStream
[
    self error: 'undefined load method for type: ', (String with: aStream peek)
]

GenericLoadMethod : Object ()

GenericLoadMethod loadFrom: aStream
[
    | typeName |
    typeName := WriteStream on: (String new: 32).
    [aStream atEnd or: [aStream peek == $\n]] whileFalse: [typeName nextPut: aStream next].
    aStream atEnd ifTrue: [self error: 'object streaming protocol error (no newline after type name)'].
    aStream next.
    ^typeName contents asPrototype loadFrom: aStream
]

LoadMethodTable : Array ()
[
    LoadMethodTable := Array new: 256.
    LoadMethodTable
	atAllPut: UndefinedLoadMethod;
	atAll: '0123456789' put: SmallInteger;
	atAll: '#' put: Symbol;
	atAll: 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz_' put: GenericLoadMethod.
]

ReadStream nextObject			[ ^self atEnd ifFalse: [(LoadMethodTable at: self peek) perform: #loadFrom: with: self] ]

ReadStream contentsOfLine
[
    | line c |
    line := WriteStream on: (String new: 32).
    [self atEnd or: [(c := self next) == $\n]] whileFalse: [line nextPut: c].
    ^line contents
]

Object storeOn: aStream		[ self error: self debugName, ' should have overridden #storeOn:' ]

UndefinedObject storeOn: aStream	[ aStream print: self; cr ]
UndefinedObject loadFrom: aStream	[ ]

SmallInteger storeOn: aStream	[ aStream print: self; cr ]
SmallInteger loadFrom: aStream	[ ^aStream contentsOfLine inject: 0 into: [:val :c | val * 10 + c digitValue] ]

Symbol storeOn: aStream		[ aStream print: self; cr ]
Symbol loadFrom: aStream	[ ^(aStream next; contentsOfLine) asSymbol ]

SequenceableCollection storeOn: aStream
[
    aStream
	nextPutAll: self debugName;
	nextPut: $\n;
	print: self size;
	nextPut: $\n.
    self do: [:elt | elt storeOn: aStream]
]

SequenceableCollection loadFrom: aStream
[
    self := self new: aStream nextObject.
    0 to: self size - 1 do: [:index | self at: index put: aStream nextObject].
]

"----------------------------------------------------------------"

GenericLoadMethod test: original
[
    | stream string copy |
    StdOut nextPutAll: 'testing '; nextPutAll: original debugName; space; print: original; cr.
    stream   := (String new: 32) writeStream.
    original storeOn: stream.
    "StdOut print: stream contents; cr."
    copy := stream contents readStream nextObject.
    original _vtable == copy _vtable ifFalse: [StdErr nextPutAll: 'STORE-LOAD TYPE FAILURE FOR '; print: original; cr].
    original = copy ifFalse: [StdErr nextPutAll: 'STORE-LOAD CONTENT FAILURE FOR '; print: original; cr]
]

File store: anObject	[ anObject storeOn: self ]

{ import: Expression }

"
[
    | out in file |
    GenericLoadMethod
	test: nil;
	test: 42;
	test: 'hello, world';
	test: '\000\a\b\c\d\e\f\g\h\i\j\k\l\m\n\o\p\q\r\s\t\u\v\w\x\y\z\377';
	test: #symbol;
	test: #'wierd symbol with spaces and @ # $ % ^ & * ( ) characters';
	test: #(one two 'three' 4 5 (six 7 eight) 9 () () () 10);
	test: #(define store (lambda (x y z) (set (#'long@' (+ x (* 4 y))) z))) asExpression;
	test: #((dlsym 'printf') '%d' 42) asExpression;
	test: #[1 2 3 4];
	yourself.
    out := #('Mary had a little' (lambda (x y) (+ 3 4)) #'setting up' a stack).
    StdOut nextPutAll: 'testing file i/o with '; print: out; cr.
    file := File create: 'temp666.txt'.
    file store: out.
    file close.
    file := File open: 'temp666.txt'.
    in := file readStream nextObject.
    file close.
    in = out ifFalse: [StdErr nextOutAll: 'STORE-LOAD FILE TEST FAILURE'; cr].
    StdOut nextPutAll: 'tests done\n'
]
"
