{ import: Object-imp }
{ import: Object-lib }
{ import: Window-lib }
{ import: Window-x11 }

DesktopColour		:= [ 0xf0f0f0 ]
TextColour		:= [ 0x000000 ]
BorderColour		:= [ 0xaa00cc ]
BackgroundColour	:= [ 0xddddff ]
CursorColour		:= [ 0xff0000 ]
"
DesktopColour		:= [ 0xf0f0f0 ]
TextColour		:= [ 0x000000 ]
BorderColour		:= [ 0xff0000 ]
BackgroundColour	:= [ 0xccffcc ]
CursorColour		:= [ 0xff0000 ]
"

TweenText		:= [ false ]

"----------------------------------------------------------------"

Widget : Object ( parent children bounds borderWidth borderColour backgroundColour properties )

Widget new
[
    self := super new.
    borderWidth := 0.
]

Widget parent		[ ^parent ]
Widget parent: aWidget	[ ^parent := aWidget ]

Widget children		[ ^children notNil ifTrue: [children] ifFalse: [#()] ]

Widget add: aWidget
[
    children isNil ifTrue: [children := OrderedCollection new].
    aWidget parent: self.
    ^children add: aWidget.
]

Widget bounds: aRectangle		[ ^bounds := aRectangle ]
Widget origin: aPoint			[ ^bounds := aPoint corner: nil ]
Widget bounds				[ ^bounds ]

Widget borderWidth: anInteger		[ ^borderWidth := anInteger ]
Widget borderWidth			[ ^borderWidth ]

Widget borderColour: aColour		[ ^borderColour := aColour ]
Widget borderColour			[ ^borderColour ]

Widget backgroundColour: aColour	[ ^backgroundColour := aColour ]
Widget backgroundColour			[ ^backgroundColour ]

Widget propertyAt: aKey			[ ^properties notNil ifTrue: [properties at: aKey ifAbsent: []] ]
Widget propertyAt: aKey put: aValue 	[ ^(properties ifNil: [properties := FastIdentityDictionary new]) at: aKey put: aValue ]

Widget naturalSize			[ ^nil ]

Widget outermostWidgetAt: aPoint
[
    (self containsPoint: aPoint) ifTrue: [^self].
    ^self children detect: [:sub | sub outermostWidgetAt: aPoint].
]

Widget innermostWidgetAt: aPoint
[
    ^self children
	reverseDetect: [:sub | sub innermostWidgetAt: aPoint]
	ifAbsent:      [^(bounds contains: aPoint) ifTrue: [self]].
]

Widget moveBy: aPoint
[
    bounds := bounds translatedBy: aPoint.
    self children do: [:sub | sub moveBy: aPoint]
]

Widget resizeBy: aPoint
[
    bounds := bounds origin corner: bounds corner + aPoint.
    self children do: [:sub | sub resizeBy: aPoint]
]

Widget draw		[ self children do: [:sub | sub draw] ]
Widget redraw		[ parent notNil ifTrue: [parent redraw] ]

Widget drawOn: aSurface
[
    backgroundColour notNil ifTrue: [self drawBackgroundOn: aSurface].
    (borderColour notNil and: [borderWidth > 0]) ifTrue: [self drawBorderOn: aSurface].
    self drawForegroundOn: aSurface.
    self children do: [:sub | sub drawOn: aSurface]
]

Widget drawBackgroundOn: aSurface
[
    aSurface
	foreground:    backgroundColour;
	fillRectangle: bounds
]

Widget drawBorderOn: aSurface
[
    | tlInset brInset |
    tlInset := borderWidth // 2.
    brInset := borderWidth + 1 // 2.
    aSurface
	foreground:    borderColour;
	lineWidth:     borderWidth;
	"xxx AN X11 BUG OR DOES OpenGL SUFFER PROBLEM WITH WIDE LINES AND DRAW vs FILL ??? xxx"
	drawRectangle: (bounds origin + (tlInset @ tlInset) corner: bounds corner - (brInset @ brInset))
]

Widget drawForegroundOn: aSurface	[]

"----------------------------------------------------------------"

LabelWidget : Widget ( text )

LabelWidget new
[
    self := super new.
]

LabelWidget withText: aTextOrString
[
    self := self new.
    text := aTextOrString asText.
]

LabelWidget text: aTextOrString		[ ^text := aTextOrString asText ]

LabelWidget naturalSize			[ ^(borderWidth * 2 + (text size * 6)) @ (borderWidth * 2 + 13) ]

LabelWidget drawForegroundOn: aSurface
[
    text notNil ifTrue: [text drawOn: aSurface at: bounds origin + (borderWidth@borderWidth)]
]

"----------------------------------------------------------------"

TextWidget : LabelWidget ( cursorColour cursorIndex targetColumn )

TextWidget new
[
    self := super new.
    cursorIndex := 0.
    cursorColour := CursorColour.
    targetColumn := nil.
]

TextWidget keyPress: anEvent
[
    | keyCode |
    keyCode := anEvent keyCode.
    (keyCode == 30                     ) ifTrue: [^self moveCursorUp].
    (keyCode == 31                     ) ifTrue: [^self moveCursorDown].
    targetColumn := nil.
    (keyCode == 16                     ) ifTrue: [^self moveCursorRight].
    (keyCode == 17                     ) ifTrue: [^self moveCursorLeft].
    (keyCode ==  8 or: [keyCode == 127]) ifTrue: [^self deleteAtCursor].
    (keyCode == $\r)			 ifTrue: [keyCode := $\n].
    (keyCode >= 32 or: [keyCode == $\n]) ifTrue: [^self insertAtCursor: keyCode].
]

TextWidget moveCursorRight
[
    cursorIndex := cursorIndex + 1 min: text size.
    self redraw.
]

TextWidget moveCursorLeft
[
    cursorIndex := cursorIndex - 1 max: 0.
    self redraw.
]

TextWidget moveCursorUp		[ self moveCursorToLine: (self cursorLine - 1) column: (targetColumn ifNil: [targetColumn := self cursorColumn]) ]
TextWidget moveCursorDown	[ self moveCursorToLine: (self cursorLine + 1) column: (targetColumn ifNil: [targetColumn := self cursorColumn]) ]

TextWidget cursorLine		[ ^text lineAtIndex: cursorIndex ]
TextWidget cursorColumn		[ ^text columnAtIndex: cursorIndex ]

TextWidget moveCursorToLine: lineNumber column: columnNumber
[
    cursorIndex := lineNumber < 0 ifTrue: [0] ifFalse: [cursorIndex := text indexOfLine: lineNumber column: columnNumber].
    self redraw.
]

TextWidget insertAtCursor: aCharacter
[
    text insert: aCharacter at: cursorIndex.
    cursorIndex := cursorIndex + 1.
    self redraw.
]

TextWidget deleteAtCursor
[
    cursorIndex > 0
	ifTrue:
	   [cursorIndex := cursorIndex - 1.
	    text deleteAt: cursorIndex.
	    self redraw]
]

TextWidget drawForegroundOn: aSurface
[
    | hPosition vPosition hLimit vLimit index size bg char |
    (text isNil or: [bounds width < 6 or: [bounds height < 13]]) ifTrue: [^nil].
    vPosition := bounds origin + (borderWidth @ (borderWidth + 13)).
    index := 0.
    size := text size.
    hLimit := bounds right - 6 - borderWidth.
    vLimit := bounds bottom - borderWidth.
    [index < size and: [vPosition y < vLimit]]
	whileTrue:
	   [hPosition := vPosition.
	    [index < size and: [hPosition x < hLimit]]
		whileTrue:
		   [index == cursorIndex ifTrue: [self drawCursorOn: aSurface at: hPosition].
		    (char := text at: index) ~~ $\n
			ifTrue:
			   [(bg := text backgroundAt: index) notNil ifTrue: [aSurface background: bg].
			    aSurface
				foreground: (text foregroundAt: index);
				drawCharacter: char at: hPosition.
			    hPosition := hPosition + (6@0)]
			ifFalse:
			   [hPosition := hLimit @ hPosition y].
		    index := index + 1].
	    vPosition := vPosition + (0@13)].
    index == cursorIndex ifTrue: [self drawCursorOn: aSurface at: (hPosition ifNil: [vPosition])].
]

TextWidget drawCursorOn: aSurface at: aPoint
[
    | bPoint |
    aPoint := aPoint + (0@1).
    bPoint := aPoint - (1@12).
    aSurface
	foreground:    cursorColour;
	fillRectangle: (bPoint - (1@0) extent: 3@1);
	fillRectangle: (bPoint	       extent: 1@12);
	fillRectangle: (aPoint - (2@0) extent: 3@1)
]

"----------------------------------------------------------------"

Text : Object ( size capacity string foregrounds backgrounds )

String asText		[ ^Text withAll: self ]
Text   asText		[ ^self ]

Text withAll: aString
[
    self := self new.
    size := aString size.
    capacity := size.
    string := aString.
    foregrounds := Array new: capacity atAllPut: TextColour.
    backgrounds := Array new: capacity.
    TweenText
	ifTrue:
	   [0 to: capacity - 1 do: [:i |
	       backgrounds at: i put: (#(0xffffff 0xffff00 0xff00ff 0x00ffff 0xff0000 0x00ff00 0x0000ff 0x000000) at:     (i \\ 8)).
	       foregrounds at: i put: (#(0xffffff 0xffff00 0xff00ff 0x00ffff 0xff0000 0x00ff00 0x0000ff 0x000000) at: 7 - (i \\ 8))]].
]

Text size			[ ^string size ]
Text at: integerIndex		[ ^string at: integerIndex ]
Text backgroundAt: integerIndex	[ ^backgrounds at: integerIndex ]
Text foregroundAt: integerIndex	[ ^foregrounds at: integerIndex ]

Text lineAtIndex: integerIndex
[
    | line |
    line := 0.
    0 to: (integerIndex min: self size) - 1 do: [:i | (self at: i) == $\n ifTrue: [line := line + 1]].
    ^line
]

Text columnAtIndex: integerIndex
[
    | column |
    column := 0.
    0 to: (integerIndex min: self size) - 1 do: [:i | column := ((self at: i) == $\n ifTrue: [0] ifFalse: [column + 1])].
    ^column
]

Text indexOfLine: lineNumber column: columnNumber
[
    | index |
    index := self indexOfLine: lineNumber.
    [columnNumber > 0 and: [index < self size and: [(self at: index) ~~ $\n]]]
	whileTrue:
	   [columnNumber := columnNumber - 1.
	    index := index + 1].
    ^index
]

Text indexOfLine: lineNumber
[
    | index |
    index := 0.
    [lineNumber > 0 and: [index < self size]]
	whileTrue:
	    [(self at: index) == $\n ifTrue: [lineNumber := lineNumber - 1].
	     index := index + 1].
    ^index
]

Text insert: aCharacter at: integerIndex
[
    integerIndex >= string size
	ifTrue:
	   [string := string , (string with: aCharacter).
	    foregrounds := foregrounds , (foregrounds with: 0).
	    backgrounds := backgrounds , (backgrounds with: 0)]
	ifFalse:
	   [string := (string copyFrom: 0 to: integerIndex), (string copyFrom: integerIndex to: string size - 1).
	    foregrounds := (foregrounds copyFrom: 0 to: integerIndex), (foregrounds copyFrom: integerIndex to: foregrounds size - 1).
	    backgrounds := (backgrounds copyFrom: 0 to: integerIndex), (backgrounds copyFrom: integerIndex to: backgrounds size - 1).
	    string at: integerIndex put: aCharacter].
    foregrounds at: integerIndex put: TextColour.
    backgrounds at: integerIndex put: nil.
]

Text deleteAt: integerIndex
[
    string := (string copyFrom: 0 to: integerIndex - 1), (string copyFrom: integerIndex + 1 to: string size - 1).
    foregrounds := (foregrounds copyFrom: 0 to: integerIndex - 1), (foregrounds copyFrom: integerIndex + 1 to: foregrounds size - 1).
    backgrounds := (backgrounds copyFrom: 0 to: integerIndex - 1), (backgrounds copyFrom: integerIndex + 1 to: backgrounds size - 1).
]

"----------------------------------------------------------------"

World : Widget ( window layout eventHandler )

World new	[ ^self shouldNotImplement: #new ]

World withBounds: aRectangle
[
    self := super new.
    window := Surface new: aRectangle.
    window clear: DesktopColour.
    self bounds: (Point zero extent: aRectangle extent)
]

World window	[ ^window ]

World layout: aLayout			[ ^layout := aLayout ]
World eventHandler: anEventHandler	[ ^eventHandler := anEventHandler ]

World draw
[
    window beginDrawing.
    window clear.
    super drawOn: window.
    window endDrawing.
]

World redraw
[
    window beginDrawing.
    self drawOn: window.
    window endDrawing.
]

World outermostWidgetAt: aPoint
[
    ^self children detect: [:sub | sub outermostWidgetAt: aPoint].
]

World innermostWidgetAt: aPoint
[
    ^self children
	reverseDetect: [:sub | sub innermostWidgetAt: aPoint]
	ifAbsent:      []
]

World resize: newExtent
[
    bounds := bounds origin extent: newExtent.
    layout layout.
    self draw.
]

World layout	[ layout layout ]

World applicationLoop
[
    [eventHandler handle: window pollEvent] repeat.
]

"----------------------------------------------------------------"

EventHandler : Widget ( topWidget )

EventHandler on: aWidget
[
    self := self new.
    topWidget := aWidget.
]

EventHandler handle: anEvent	[]

"----------------------------------------------------------------"

ExampleEventHandler : EventHandler ( eventTable dragFocus dragPosition dragType )

ExampleEventHandler new
[
    self := super new.
    eventTable := IdentityDictionary new
	at: #ButtonPress	put: #buttonPress:;
	at: #ButtonRelease	put: #buttonRelease:;
	at: #Motion		put: #motion:;
	at: #KeyPress		put: #keyPress:;
	at: #KeyRelease		put: #keyRelease:;
	at: #Resize		put: #resize:;
	yourself.
]

ExampleEventHandler handle: anEvent
[
    anEvent notNil ifTrue: [self perform: (eventTable at: anEvent type ifAbsent: [#yourself]) with: anEvent]
]

ExampleEventHandler buttonPress: anEvent
[
    dragFocus := topWidget innermostWidgetAt: anEvent position.
    dragFocus isNil ifTrue: [^self].
    dragPosition := anEvent position.
    dragType := dragFocus bounds corner - anEvent position < (5@5) ifTrue:  [#resizeBy:] ifFalse: [#moveBy:].
    StdOut print: dragFocus; space; print: dragPosition; space; print: dragType; cr.
    dragFocus borderColour: BorderColour.
    topWidget draw.
]

ExampleEventHandler motion: anEvent
[
    dragFocus isNil ifTrue: [^self].
    dragFocus perform: dragType with: anEvent position - dragPosition.
    dragPosition := anEvent position.
    topWidget draw.
]

ExampleEventHandler buttonRelease: anEvent
[
    dragFocus isNil ifTrue: [^self].
    dragFocus borderColour: nil.
    dragFocus := nil.
    topWidget draw.
]

ExampleEventHandler keyPress: anEvent
[
    | target |
    (anEvent keyCode notNil and: [(target := topWidget innermostWidgetAt: anEvent position) notNil])
	ifTrue: [target keyPress: anEvent]
]

ExampleEventHandler keyRelease: anEvent
[
    anEvent keyCode == 27 ifTrue: [OS exit: 0].
]

ExampleEventHandler resize: anEvent
[
    topWidget resize: anEvent width @ anEvent height
]

Widget motion: anEvent		[]
Widget resize: anEvent		[]
Widget buttonPress: anEvent	[]
Widget buttonRelease: anEvent	[]
Widget keyPress: anEvent	[]
Widget keyRelease: anEvent	[]

"----------------------------------------------------------------"

Layout : Widget ( topWidget )

Layout on: aWidget
[
    self := self new.
    topWidget := aWidget.
]

Layout layout
[
    | parentBounds |
    parentBounds := topWidget bounds.
    topWidget children do: [:sub | self layout: sub in: parentBounds].
]

Layout layout: aWidget in: parentBounds
[
    | childOrigin childExtent |
    childOrigin := (aWidget bounds ifNil: [parentBounds]) origin.
    childExtent := self extentOf: aWidget at: childOrigin in: parentBounds.
    aWidget bounds: (parentBounds := childOrigin extent: childExtent).
    aWidget children do: [:sub | self layout: sub in: parentBounds].
]

Layout extentOf: aWidget at: originPoint in: parentBounds
[
    | minimumSize naturalSize maximumSize width height |
    minimumSize := aWidget propertyAt: #minimumExtent.
    naturalSize := aWidget naturalSize.
    maximumSize := aWidget propertyAt: #maximumExtent.
    width       := self resize:    (naturalSize ifNotNil: [naturalSize x])
			minimum:   (minimumSize ifNotNil: [minimumSize x])
			maximum:   (maximumSize ifNotNil: [maximumSize x])
			within:    parentBounds width - originPoint x
			expanding: (aWidget propertyAt: #hSizing).
    height      := self resize:    (naturalSize ifNotNil: [naturalSize y])
			minimum:   (minimumSize ifNotNil: [minimumSize y])
			maximum:   (maximumSize ifNotNil: [maximumSize y])
			within:    parentBounds height - originPoint y
			expanding: (aWidget propertyAt: #vSizing).
    ^width @ height
]

Layout resize: natural minimum: minimum maximum: maximum within: available expanding: expansion
[
    | size |
    size := expansion isNil ifTrue: [available] ifFalse: [self perform: expansion with: natural with: available].
    minimum notNil ifTrue: [size := size max: minimum].
    maximum notNil ifTrue: [size := size min: maximum].
    ^size min: available
]

Layout shrink :natural :available	[ ^natural isNil ifTrue: [available] ifFalse: [natural min: available] ]
Layout fill   :natural :available	[ ^natural isNil ifTrue: [available] ifFalse: [natural max: available] ]

Widget hShrink		[ self propertyAt: #hSizing put: #shrink ]
Widget hFill		[ self propertyAt: #hSizing put: #fill   ]
Widget vShrink		[ self propertyAt: #vSizing put: #shrink ]
Widget vFill		[ self propertyAt: #vSizing put: #fill   ]

"----------------------------------------------------------------"

Widget dumpTree		[ self dumpTree: 0 ]

Widget dumpTree: indent
[
    StdOut space: indent * 2; print: self; space; print: bounds; cr.
    self children do: [:sub | sub dumpTree: indent + 1].
]

ExampleText := [ 'Behind the ostensible government sits enthroned an invisible government owing no allegiance and acknowledging no responsibility to the people. To destroy this invisible government, to befoul the unholy alliance between corrupt business and corrupt politics is the first task of the statesmanship of today.  (President Theodore Roosevelt, 19 April 1906)' asText ]

[
    | world |
    world := World withBounds: (100@100 extent: 200@200).
    world add: (TextWidget new
		    borderWidth:	1;
		    borderColour:	nil;
		    backgroundColour:	BackgroundColour;
		    propertyAt:		#maximumExtent put: 400@400;
		    text:		ExampleText;
		    yourself).
    world layout: (Layout on: world).
    world eventHandler: (ExampleEventHandler on: world).
    world layout; draw.
    world applicationLoop.
]
