{ import: Time }
{ import: Dictionary }

FileStatus : Object (
    _mode		"format and permissions"
    userID		"file's owner"
    groupID		"file's group"
    accessTime		"last data access"
    modificationTime	"last data change"
    changeTime		"last inode change"
    size		"number of data bytes"
    attributes		"Dictionary of extended attributes, or nil"
    acl			"String encoding inode access control list, or nil"
)

FileStatus mode			{ return (oop)((long)self->v__mode << 1 | 1); }

FileStatus userID		[ ^userID ]
FileStatus groupID		[ ^groupID ]
FileStatus accessTime	 	[ ^accessTime ]
FileStatus modificationTime	[ ^modificationTime ]
FileStatus changeTime		[ ^changeTime ]
FileStatus size			[ ^size ]
FileStatus attributes		[ ^attributes ]
FileStatus acl			[ ^acl ]

FileStatus exists		[ ^size ~~ nil ]

FileStatus forPath: pathName
[
    ^self new readStatus: pathName
]

FileStatus initialize
[
    accessTime       := Time new.
    modificationTime := Time new.
    changeTime       := Time new.
    attributes	     := nil.
    acl		     := String new.
]

FileStatus readStatus: pathName
[
    size := nil.
    self readStatus_: pathName _stringValue.
]

FileStatus readStatus_: _path
[
    self loadStatus_: _path.
    self exists ifFalse: [^nil].
    self loadAttributes_: _path.
]

FileStatus writeStatus: pathName
[
    self writeStatus_: pathName _stringValue
]

FileStatus writeStatus_: _path
[
    ^(self storeStatus_: _path) and: [self storeAttributes_: _path]
]

{ include <fcntl.h> }		"filesec_t"
{ include <sys/acl.h> }		"acl_t"
{ include <errno.h> }		"ENOENT"

FileStatus loadStatus_: _path
{
    struct stat sb;
    filesec_t   fsec= filesec_init();
    acl_t       acl;
    char       *text;
    ssize_t     length;

    if (lstatx_np((char *)v__path, &sb, fsec))
      {
        if (ENOENT != errno) perror((char *)v__path);
        self->v_acl= 0;
      }
    else
      {
        self->v__mode=   (oop)(long)(sb.st_mode);
        self->v_size=	 (oop)(long)(sb.st_size << 1 | 1);
        self->v_userID=  (oop)(long)(sb.st_uid  << 1 | 1);
        self->v_groupID= (oop)(long)(sb.st_gid  << 1 | 1);
        ((struct t_Time *)self->v_accessTime      )->v__seconds=      (oop)sb.st_atimespec.tv_sec;
        ((struct t_Time *)self->v_accessTime      )->v__nanoseconds=  (oop)sb.st_atimespec.tv_nsec;
        ((struct t_Time *)self->v_modificationTime)->v__seconds=      (oop)sb.st_mtimespec.tv_sec;
        ((struct t_Time *)self->v_modificationTime)->v__nanoseconds=  (oop)sb.st_mtimespec.tv_nsec;
        ((struct t_Time *)self->v_changeTime      )->v__seconds=      (oop)sb.st_ctimespec.tv_sec;
        ((struct t_Time *)self->v_changeTime      )->v__nanoseconds=  (oop)sb.st_ctimespec.tv_nsec;
        if ((filesec_get_property(fsec, FILESEC_ACL, &acl) >= 0) && (text= acl_to_text(acl, &length)))
          {
    	    struct t_String *aclText= (struct t_String *)self->v_acl;
    	    aclText->v__bytes= _balloc(length);
    	    aclText->v_size= (oop)(length << 1 | 1);
    	    memcpy(aclText->v__bytes, text, length);
    	    acl_free(text);
          }
        else
          self->v_acl= 0;
      }

    filesec_free(fsec);
    return v_self;
}

FileStatus storeStatus_: _path
{
    struct stat sb;
    /*filesec_t fsec;*/
    struct timeval amtv[2];

    if (lstat((char *)v__path, &sb/*, fsec*/))
      {
	perror((char *)v__path);
	return 0;
      }

    amtv[0].tv_sec=  (unsigned long)((struct t_Time *)self->v_accessTime)->v__seconds;
    amtv[0].tv_usec= (unsigned long)((struct t_Time *)self->v_accessTime)->v__nanoseconds / 1000;
    amtv[1].tv_sec=  (unsigned long)((struct t_Time *)self->v_modificationTime)->v__seconds;
    amtv[1].tv_usec= (unsigned long)((struct t_Time *)self->v_modificationTime)->v__nanoseconds / 1000;
    utimes((char *)v__path, amtv);

    chown((char *)v__path, (uid_t)(long)self->v_userID >> 1, (gid_t)(long)self->v_groupID >> 1);
    chmod((char *)v__path, (mode_t)(long)self->v__mode);

    /*filesec_free(fsec);*/
    return v_self;
}

{ include <sys/xattr.h> }	"getxattr[list]"

FileStatus loadAttributes_: _path
[
    | names |
    names := self getAttributeNames_: _path into: String new.
    names isEmpty ifTrue: [^self].
    attributes := Dictionary new.
    [names isEmpty]
	whileFalse:
	   [| name |
	    name := String value_: names _stringValue.			"copies up to NUL separator"
	    names := names copyFrom: name size + 2 to: names size.	"remove name and NUL from list"
	    attributes
		at: name
		put: (self getAttribute_: name _stringValue of_: _path into: String new)].
]

FileStatus getAttributeNames_: _path into: aString
{
    struct t_String *string= (struct t_String *)v_aString;
    ssize_t count= listxattr((char *)v__path, 0, 0, XATTR_NOFOLLOW);
    if (count < 0) perror((char *)v__path);
    if (count > 0)
      {
        string->v__bytes= _balloc(count);
        string->v_size=   (oop)(count << 1 | 1);
	listxattr((char *)v__path, (char *)string->v__bytes, count, XATTR_NOFOLLOW);
      }
    return v_aString;
}

FileStatus getAttribute_: _name of_: _path into: aByteArray
{
    struct t_ByteArray *data= (struct t_ByteArray *)v_aByteArray;
    ssize_t count= getxattr((char *)v__path, (char *)v__name, 0, 0, 0, XATTR_NOFOLLOW);
    if (count)
      {
        data->v__bytes= _balloc(count);
        data->v_size=   (oop)(count << 1 | 1);
	getxattr((char *)v__path, (char *)v__name, data->v__bytes, count, 0, XATTR_NOFOLLOW);
      }
    return v_aByteArray;
}

FileStatus storeAttributes_: _path
[
    attributes notNil
	ifTrue: [attributes keysAndValuesDo: [:key :value |
	    self
		putAttribute_: key _stringValue
		of_:           _path
		with_:         value _bytes
		size_:         value _size]]
]

FileStatus putAttribute_: _name of_: _path with_: _bytes size_: _size
{
    setxattr((char *)v__path, (char *)v__name, (void *)v__bytes, (size_t)v__size, 0, XATTR_NOFOLLOW);
}

FileStatus isDirectory		{ return S_ISDIR((mode_t)(long)self->v__mode) ? v_self : 0; }
FileStatus isRegular		{ return S_ISREG((mode_t)(long)self->v__mode) ? v_self : 0; }
FileStatus isSymlink		{ return S_ISLNK((mode_t)(long)self->v__mode) ? v_self : 0; }

FileStatus type
[
    self isDirectory ifTrue: [^#directory].
    self isRegular   ifTrue: [^#regular].
    self isSymlink   ifTrue: [^#symlink].
    ^#unknown
]

FileStatus isOwnerReadable	{ return (S_IRUSR & (mode_t)(long)self->v__mode) ? v_self : 0; }
FileStatus isOwnerWritable	{ return (S_IWUSR & (mode_t)(long)self->v__mode) ? v_self : 0; }
FileStatus isOwnerExecutable	{ return (S_IXUSR & (mode_t)(long)self->v__mode) ? v_self : 0; }
FileStatus isGroupReadable	{ return (S_IRGRP & (mode_t)(long)self->v__mode) ? v_self : 0; }
FileStatus isGroupWritable	{ return (S_IWGRP & (mode_t)(long)self->v__mode) ? v_self : 0; }
FileStatus isGroupExecutable	{ return (S_IXGRP & (mode_t)(long)self->v__mode) ? v_self : 0; }
FileStatus isWorldReadable	{ return (S_IROTH & (mode_t)(long)self->v__mode) ? v_self : 0; }
FileStatus isWorldWritable	{ return (S_IWOTH & (mode_t)(long)self->v__mode) ? v_self : 0; }
FileStatus isWorldExecutable	{ return (S_IXOTH & (mode_t)(long)self->v__mode) ? v_self : 0; }

FileStatus dataMatches: other
[
    ^(self    type             == other type
	and: [size             == other size
	and: [modificationTime =  other modificationTime]])
]

FileStatus statusMatches: other
[
    ^(        userID       == other userID
	and: [groupID      == other groupID
	and: [self mode    == other mode
	and: [attributes   =  other attributes
	and: [acl          =  other acl]]]])
]

FileStatus printOn: aStream
[
    aStream
	nextPut: (self isDirectory	 ifTrue: [$d] ifFalse: [$-]);
	nextPut: (self isOwnerReadable	 ifTrue: [$r] ifFalse: [$-]);
	nextPut: (self isOwnerWritable	 ifTrue: [$w] ifFalse: [$-]);
	nextPut: (self isOwnerExecutable ifTrue: [$x] ifFalse: [$-]);
	nextPut: (self isGroupReadable	 ifTrue: [$r] ifFalse: [$-]);
	nextPut: (self isGroupWritable	 ifTrue: [$w] ifFalse: [$-]);
	nextPut: (self isGroupExecutable ifTrue: [$x] ifFalse: [$-]);
	nextPut: (self isWorldReadable	 ifTrue: [$r] ifFalse: [$-]);
	nextPut: (self isWorldWritable	 ifTrue: [$w] ifFalse: [$-]);
	nextPut: (self isWorldExecutable ifTrue: [$x] ifFalse: [$-]);
	space; print: userID;
	space; print: groupID;
	space; print: size;
	space; print: modificationTime.
    attributes notNil
	ifTrue: [attributes keys printOn: aStream]
]


