{ import: st80 }
{ import: File }
{ import: Time }
{ import: FileStatus }
{ import: FileStream }

TypeMask		:= [ 3 ]
HeaderTypeSizeAndClass	:= [ 0 ]
HeaderTypeClass		:= [ 1 ]
HeaderTypeFree		:= [ 2 ]
HeaderTypeShort		:= [ 3 ]

AllButTypeMask		:= [ 16r3FFFFFFF - TypeMask ]
SizeMask		:= [ 16r000000FC ]

SqueakImage : Object ( name specialObjects lastHash savedWindowSize fullScreenFlag baseAddress wordMemory byteMemory map )

SqueakImage info: aString		[ aString putln ]

SqueakImage byteAt: addr		[ ^byteMemory at: 1 +  addr		    ]
SqueakImage longAt: addr		[ ^wordMemory at: 1 + (addr // 4)	    ]
SqueakImage longAt: addr put: word	[ ^wordMemory at: 1 + (addr // 4) put: word ]

SqueakImage oopAt: addr
[
    | oop |
    oop := wordMemory at: 1 + (addr // 4).
    ^(oop bitAnd: 1) == 1 ifTrue: [oop >> 1] ifFalse: [oop]
]

SqueakImage mapAt: oop			[ ^map at: 1 + (oop // 4)		]
SqueakImage mapAt: oop put: anObject	[ ^map at: 1 + (oop // 4) put: anObject ]

SqueakImage objectAt: addr
[
    | oop |
    oop := wordMemory at: 1 + (addr // 4).
    ^(oop bitAnd: 1) == 1 ifTrue: [oop >> 1] ifFalse: [map at: oop]
]

SqueakImage sizeHeader: oop			[ ^self longAt: oop - 8 ]
SqueakImage classHeader: oop			[ ^self longAt: oop - 4 ]
SqueakImage classHeader: oop put: header	[ ^self longAt: oop - 4 put: header ]
SqueakImage baseHeader: oop			[ ^self longAt: oop ]
SqueakImage headerType: oop			[ ^(self longAt: oop) bitAnd: TypeMask ]
SqueakImage formatOfOop: oop			[ ^(self longAt: oop) >> 8 bitAnd: 15 ]
SqueakImage extraHeaderBytes: oopOrChunk	[ ^#(8 4 0 0) at: (1 + (self headerType: oopOrChunk)) ]
SqueakImage oopFromChunk: chunk			[ ^chunk + (self extraHeaderBytes: chunk) ]

SqueakImage objectAfter: oop
[
    | sz |
    sz := (self headerType: oop) == HeaderTypeFree
        ifTrue:  [self sizeOfFree: oop]
        ifFalse: [self sizeBitsOf: oop].
    ^self oopFromChunk: oop + sz
]

SqueakImage sizeOfFree: oop			[ ^(self longAt: oop) bitAnd: AllButTypeMask ]

SqueakImage sizeBitsOf: oop
[
    "Answer the number of bytes in the given object, including its
    base header, rounded up to an integral number of words."
    "Note: byte indexable objects need to have low bits subtracted
    from this size."
    | header |
    header := self baseHeader: oop.
    ^(header bitAnd: TypeMask) = HeaderTypeSizeAndClass
        ifTrue:  [(self sizeHeader: oop) bitAnd: AllButTypeMask]
    	ifFalse: [header bitAnd: SizeMask].
]

SqueakImage lastPointerOf: oop
[
    "Answer the byte offset of the last pointer field of the given object.
    Works with CompiledMethods, as well as ordinary objects."
    | fmt sz methodHeader header contextSize |
    header := self baseHeader: oop.
    fmt := header >> 8 bitAnd: 15.
    fmt <= 4
	ifTrue: [(fmt = 3 and: [self isContextHeader: header])
		     ifTrue: ["contexts end at the stack pointer"
			      contextSize := self fetchStackPointerOf: oop.
			      ^1 + 6 + contextSize * 4].
		 sz := self sizeBitsOf: oop.
		 ^sz - 4 "all pointers"].
    fmt < 12 ifTrue: [^0]. "no pointers"
    "CompiledMethod: contains both pointers and bytes:"
    methodHeader := self objectAt: oop + 4.
    ^((methodHeader >> 9) bitAnd: 255) * 4 + 4.
]

SqueakImage isContextHeader: aHeader
[
    | cc |
    cc := (aHeader >> 12) bitAnd: 16r1F.
    ^cc == 13 or: [cc == 14]
]

SqueakImage fetchStackPointerOf: aContext
[
    "Answer the stackPointer of a Context or BlockContext."
    | sp |
    sp := self longAt: aContext + 12.
    (sp bitAnd: 1) == 1 ifFalse: [^0].
    ^sp >> 1
]

SqueakImage fetchPointer: fieldIndex ofObject: oop
[
    ^self longAt: oop + 4 + (fieldIndex << 2)
]

ClassInteger	:= [  5 ]
CompactClasses	:= [ 28 ]

SqueakImage splObj: index
[
    ^specialObjects at: 1 + index
]

SqueakImage fetchClassOf: oop
[
    | ccIndex classOop |
    (oop bitAnd: 1) == 1 ifTrue: [^specialObjects at: 1 + ClassInteger].
    ccIndex := ((self baseHeader: oop) >> 12) bitAnd: 16r1F.
    ^ccIndex == 0
	ifTrue:  [self mapAt: ((self classHeader: oop) bitAnd: AllButTypeMask)]
	ifFalse: [(specialObjects at: 1 + CompactClasses) println at: ccIndex]
]

SqueakImage oopsDo: aBlock
[
    | oop endOfMemory |
    endOfMemory := byteMemory size - 4.
    oop := self oopFromChunk: 0.
    [oop < endOfMemory]
	whileTrue:
	   [(self headerType: oop) == HeaderTypeFree ifFalse: [aBlock value: oop].
	    oop := self objectAfter: oop].
]

SqueakImage open: imageName
[
    | stat file stream headerSize dataSize specialObjectsOop extraVMMemory tally |
    self := self new.
    (stat := (FileStatus forPath: imageName)) exists ifFalse: [self error: imageName, ' not found'].
    (file := File openForReading: imageName) isNil ifTrue: [self error: 'cannot open ', imageName].
    name := self absolutePathName: imageName.
    stream := FileStream on: file.
    stream nextInt32 == 6502
	ifFalse:
	   [stream := ReverseFileStream on: file.
	    stream nextInt32 == 6502 ifFalse: [self error: 'cannot read image file']].

    headerSize		:= stream nextInt32.
    dataSize		:= stream nextInt32.
    baseAddress		:= stream nextInt32.
    specialObjectsOop	:= stream nextInt32 - baseAddress.
    lastHash		:= stream nextInt32.
    savedWindowSize	:= stream nextInt32.
    fullScreenFlag	:= stream nextInt32.
    extraVMMemory	:= stream nextInt32.

    ('headerSize        ', headerSize printString) putln.
    ('dataSize          ', dataSize printString) putln.
    ('baseAddress       ', baseAddress printString) putln.
    ('specialObjectsOop ', specialObjects printString) putln.
    ('lastHash          ', lastHash printString) putln.
    ('savedWindowSize   ', savedWindowSize printString) putln.
    ('fullScreenFlag    ', fullScreenFlag printString) putln.
    ('extraVMMemory     ', extraVMMemory printString) putln.

    wordMemory := WordArray new: dataSize // 4 + 1.
    byteMemory := ByteArray size: dataSize + 4 value_: wordMemory _words.		"byte accessor"
    map := Array new: dataSize // 4.
    file seek: headerSize.
    (dataSize == (file read: byteMemory)) ifFalse: [self error: 'could not load image'].
    self longAt: dataSize put: 16r4 << 2 + HeaderTypeFree.	"install free chunk"
    stream isReversed
	ifTrue:
	   [self info: 'reverse bytes...'.
	    self reverseBytes].
    tally := 0.
    self info: 'relocate pointers...'.
    self oopsDo: [:oop | tally := tally + 1.  self relocate: oop].
    (tally printString, ' objects') putln.
    self info: 'instantiate fixed and pointer objects...'.
    self oopsDo: [:oop | self allocate: oop].
    specialObjects := self mapAt: specialObjectsOop.
    self info: 'instantiate remaining objects...'.
    self oopsDo: [:oop | self instantiate: oop].
    self info: 'release image...'.
    wordMemory := nil.
    byteMemory := nil.
    self info: 'release map...'.
    map atAllPut: nil.
    map := nil.
]

RelocationTable := [
    (Array new: 16)
	at:  1 put: #relocateEmpty:;	"    0 no fields"
	at:  2 put: #relocatePointers:;	"    1 fixed fields only (all containing pointers)"
	at:  3 put: #relocatePointers:;	"    2 indexable fields only (all containing pointers)"
	at:  4 put: #relocatePointers:;	"    3 both fixed and indexable fields (all containing pointers)"
	at:  5 put: #relocatePointers:;	"    4 both fixed and indexable weak fields (all containing pointers)"
	at:  6 put: #relocateUnused:;	"    5 unused"
	at:  7 put: #relocateEmpty:;	"    6 indexable word fields only (no pointers)"
	at:  8 put: #relocateUnused:;	"    7 unused"
	at:  9 put: #relocateEmpty:;	" 8-11 indexable byte fields only (no pointers)"
	at: 10 put: #relocateEmpty:;
	at: 11 put: #relocateEmpty:;
	at: 12 put: #relocateEmpty:;
	at: 13 put: #relocateMethod:;	"12-15 compiled methods"
	at: 14 put: #relocateMethod:;
	at: 15 put: #relocateMethod:;
	at: 16 put: #relocateMethod:;
	collect: [:s | s asSelector]
]

SqueakImage relocate: oop
[
    self _perform: (RelocationTable at: 1 + (self formatOfOop: oop)) w: oop
]

SqueakImage relocateEmpty: oop
[
    | header type |
    header := self baseHeader: oop.
    type := header bitAnd: TypeMask.
    (type == HeaderTypeClass or: [type == HeaderTypeSizeAndClass])
	ifTrue: [self classHeader: oop put: (self classHeader: oop) - baseAddress]
]

SqueakImage relocatePointers: oop
[
    | field elt |
    self relocateEmpty: oop.
    field := oop + (self lastPointerOf: oop).
    [field > oop]
	whileTrue:
	   [elt := self longAt: field.
	    (elt bitAnd: 1) == 0 ifTrue: [self longAt: field put: elt - baseAddress].
	    field := field - 4].
]

SqueakImage relocateMethod: oop
[
    | field elt |
    self relocateEmpty: oop.
    field := oop + (self lastPointerOf: oop).	"last literal"
    oop := oop + 4.
    [field > oop]
	whileTrue:
	   [elt := self longAt: field.
	    (elt bitAnd: 1) == 0 ifTrue: [self longAt: field put: elt - baseAddress].
	    field := field - 4].
]

{ import: SqueakObject }

"phase 1: allocate singletons (nil), fixed objects (classes), and pointer objects (compact class and special objects arrays)"

AllocationTable := [
    (Array new: 16)
	at:  1 put: #allocateEmpty:;		"    0 no fields"
	at:  2 put: #allocateFixed:;		"    1 fixed fields only (all containing pointers)"
	at:  3 put: #allocatePointers:;		"    2 indexable fields only (all containing pointers)"
	at:  4 put: #allocateLater:;		"    3 both fixed and indexable fields (all containing pointers)"
	at:  5 put: #allocateLater:;		"    4 both fixed and indexable weak fields (all containing pointers)"
	at:  6 put: #allocateUnused:;		"    5 unused"
	at:  7 put: #allocateLater:;		"    6 indexable word fields only (no pointers)"
	at:  8 put: #allocateUnused:;		"    7 unused"
	at:  9 put: #allocateLater:;	" 8-11 indexable byte fields only (no pointers)"
	at: 10 put: #allocateLater:;
	at: 11 put: #allocateLater:;
	at: 12 put: #allocateLater:;
	at: 13 put: #allocateLater:;	"12-15 compiled methods"
	at: 14 put: #allocateLater:;
	at: 15 put: #allocateLater:;
	at: 16 put: #allocateLater:;
	collect: [:s | s asSelector]
]

SqueakImage allocate: oop
[
    self mapAt: oop put: (self _perform: (AllocationTable at: 1 + (self formatOfOop: oop)) w: oop)
]

SqueakImage allocateUnused: oop		[ ^self error: 'illegal format: ', (self formatOfOop: oop) ]
SqueakImage allocateLater: oop		[ ^nil ]
SqueakImage allocateEmpty: oop		[ ^SqueakEmpty new ]
SqueakImage allocateFixed: oop		[ ^SqueakFixed  new: (self sizeBitsOf: oop) // 4 - 1 ]
SqueakImage allocatePointers: oop	[ ^SqueakPointers new: (self sizeBitsOf: oop) // 4 - 1 ]

InstantiationTable := [
    (Array new: 16)
	at:  1 put: #instantiateEmpty:class:;		"    0 no fields"
	at:  2 put: #instantiateFixed:class:;		"    1 fixed fields only (all containing pointers)"
	at:  3 put: #instantiatePointers:class:;	"    2 indexable fields only (all containing pointers)"
	at:  4 put: #instantiateFixedPointers:class:;	"    3 both fixed and indexable fields (all containing pointers)"
	at:  5 put: #instantiateFixedWeak:class:;	"    4 both fixed and indexable weak fields (all containing pointers)"
	at:  6 put: #instantiateUnused:class:;		"    5 unused"
	at:  7 put: #instantiateWords:class:;		"    6 indexable word fields only (no pointers)"
	at:  8 put: #instantiateUnused:class:;		"    7 unused"
	at:  9 put: #instantiateBytes:class:;		" 8-11 indexable byte fields only (no pointers)"
	at: 10 put: #instantiateBytes:class:;
	at: 11 put: #instantiateBytes:class:;
	at: 12 put: #instantiateBytes:class:;
	at: 13 put: #instantiateMethod:class:;		"12-15 compiled methods"
	at: 14 put: #instantiateMethod:class:;
	at: 15 put: #instantiateMethod:class:;
	at: 16 put: #instantiateMethod:class:;
	collect: [:s | s asSelector]
]

SqueakImage instantiate: oop
[
    self mapAt: oop put: (self _perform: (AllocationTable at: 1 + (self formatOfOop: oop)) w: oop w: (self fetchClassOf: oop))
]

"
SqueakImage instantiateMethod: type oop: oop
[
    | obj header field numLiterals literals |
    obj := self instantiateBytes: type oop: oop.
    header   := self oopAt: (field := oop + 4).
    literals := Array new: (numLiterals := ((header >> 9) bitAnd: 16rFF)).
    1 to: numLiterals do: [:index | literals at: index put: (self objectAt: (field := field + 4))].
    obj
	numArgs:   ((header >> 24) bitAnd: 16r0F)
	numTemps:  ((header >> 18) bitAnd: 16r3F)
	primitive: ((header >> 19) bitAnd: 16r200) + (header bitAnd: 16r1FF)
	literals:  literals.
    ^obj
]
"

SqueakImage absolutePathName: path
[
    "Answer the absolute path name to the file named by the receiver.
    The result starts with '/' and contains no '..' or '.'
    components."
    | cPath |
    cPath := path _stringValue.
    {{
      char *path= (char *)v_cPath;
      char buf[PATH_MAX];
#    if defined(WIN32)
      if (!_fullpath(buf, path, PATH_MAX)) return 0;
#    else
      if (!realpath(path, buf)) return 0;
#    endif
      v_cPath= (oop)buf;
    }}.
    ^String value_: cPath.
]
