{ import: SqueakVM }
{ import: perform }

BitBltSimulation : Object ( interpreterProxy
			    destForm sourceForm halftoneForm
			    combinationRule destX destY width height
			    sourceX sourceY clipX clipY clipWidth
			    clipHeight sourceWidth sourceHeight
			    sourceDepth sourcePitch sourceBits
			    sourcePPW sourceMSB destWidth destHeight
			    destDepth destPitch destBits destPPW
			    destMSB bitCount skew mask1 mask2 preload
			    nWords destMask hDir vDir sourceIndex
			    sourceDelta destIndex destDelta sx sy dx
			    dy bbW bbH halftoneHeight noSource
			    noHalftone halftoneBase sourceAlpha
			    srcBitShift dstBitShift bitBltOop
			    affectedL affectedR affectedT affectedB
			    maskTable ditherMatrix4x4
			    ditherThresholds16 ditherValues16
			    hasSurfaceLock warpSrcShift warpSrcMask
			    warpAlignShift warpAlignMask
			    warpBitShiftTable querySurfaceFn
			    lockSurfaceFn unlockSurfaceFn isWarping
			    cmFlags cmMask cmShiftTable cmMaskTable
			    cmLookupTable cmBitsPerColor dither8Lookup )

"Mask constants"
AllOnes		:= [ 16rFFFFFFFF ]
BinaryPoint	:= [ 14 ]
FixedPt1	:= [ 1 << BinaryPoint ]  "Value of 1.0 in Warp's fixed-point representation"

"Indices into stopConditions for scanning"
EndOfRun	:= [ 257 ]
CrossedX	:= [ 258 ]

"Form fields"
SqueakObject atFormBitsIndex		[ ^self instVarAt: 1 ]
SqueakObject atFormWidthIndex		[ ^self instVarAt: 2 ]
SqueakObject atFormHeightIndex		[ ^self instVarAt: 3 ]
SqueakObject atFormDepthIndex		[ ^self instVarAt: 4 ]

"BitBlt fields"
SqueakObject atBBDestFormIndex		[ ^self instVarAt:  1 ]
SqueakObject atBBSourceFormIndex	[ ^self instVarAt:  2 ]
SqueakObject atBBHalftoneFormIndex	[ ^self instVarAt:  3 ]
SqueakObject atBBRuleIndex		[ ^self instVarAt:  4 ]
SqueakObject atBBDestXIndex		[ ^self instVarAt:  5 ]
SqueakObject atBBDestYIndex		[ ^self instVarAt:  6 ]
SqueakObject atBBWidthIndex		[ ^self instVarAt:  7 ]
SqueakObject atBBHeightIndex		[ ^self instVarAt:  8 ]
SqueakObject atBBSourceXIndex		[ ^self instVarAt:  9 ]
SqueakObject atBBSourceYIndex		[ ^self instVarAt: 10 ]
SqueakObject atBBClipXIndex		[ ^self instVarAt: 11 ]
SqueakObject atBBClipYIndex		[ ^self instVarAt: 12 ]
SqueakObject atBBClipWidthIndex		[ ^self instVarAt: 13 ]
SqueakObject atBBClipHeightIndex	[ ^self instVarAt: 14 ]
SqueakObject atBBColorMapIndex		[ ^self instVarAt: 15 ]
SqueakObject atBBWarpBase		[ ^self instVarAt: 15 ]
SqueakObject atBBLastIndex		[ ^self instVarAt: 16 ]
SqueakObject atBBXTableIndex		[ ^self instVarAt: 17 ]

"RGBA indices"
RedIndex		:= [ 1 ]
GreenIndex		:= [ 2 ]
BlueIndex		:= [ 3 ]
AlphaIndex		:= [ 4 ]

"ColorMap indices"
SqueakObject atColorMapShiftsIndex	[ ^self instVarAt: 1 ]
SqueakObject atColorMapMasksIndex	[ ^self instVarAt: 2 ]
SqueakObject atColorMapColorsIndex	[ ^self instVarAt: 3 ]

"Color map flags"
ColorMapPresent		:= [ 1 ]	"do we have one?"
ColorMapFixedPart	:= [ 2 ]	"does it have a fixed part?"
ColorMapIndexedPart	:= [ 4 ]	"does it have an indexed part?"
ColorMapNewStyle	:= [ 8 ]	"new style color map"

OpTable := [ #( " 0" clearWord:with:
		" 1" bitAnd:with:
		" 2" bitAndInvert:with:
		" 3" sourceWord:with:
		" 4" bitInvertAnd:with:
		" 5" destinationWord:with:
		" 6" bitXor:with:
		" 7" bitOr:with:
		" 8" bitInvertAndInvert:with:
		" 9" bitInvertXor:with:
		"10" bitInvertDestination:with:
		"11" bitOrInvert:with:
		"12" bitInvertSource:with:
		"13" bitInvertOr:with:
		"14" bitInvertOrInvert:with:
		"15" destinationWord:with:
		"16" destinationWord:with: "unused - was old paint"
		"17" destinationWord:with: "unused - was old mask"
		"18" addWord:with:
		"19" subWord:with:
		"20" rgbAdd:with:
		"21" rgbSub:with:
		"22" OLDrgbDiff:with:
		"23" OLDtallyIntoMap:with:
		"24" alphaBlend:with:
		"25" pixPaint:with:
		"26" pixMask:with:
		"27" rgbMax:with:
		"28" rgbMin:with:
		"29" rgbMinInvert:with:
		"30" alphaBlendConst:with:
		"31" alphaPaintConst:with:
		"32" rgbDiff:with:
		"33" tallyIntoMap:with:
		"34" alphaBlendScaled:with:
		"35" alphaBlendScaled:with:	"unused here - only used by FXBlt"
		"36" alphaBlendScaled:with:	"unused here - only used by FXBlt"
		"37" rgbMul:with:
		"38" pixSwap:with:
		"39" pixClear:with:
		"40" fixAlpha:with: )
	    collect: [:op | SelectorTable at: op] ]

SmallInteger bitInvert32	[ ^-1 - self]

SqueakVM primitiveCopyBits
[
    "Invoke the copyBits primitive.  If the destination is the display, then copy it to the screen."
    BitBltSimulation
	setInterpreter: self;
	primitiveCopyBits
]

BitBltSimulation setInterpreter: vm
[
    interpreterProxy := vm
]

BitBltSimulation primitiveCopyBits
[
    (self loadBitBltFrom: (interpreterProxy stackValue: interpreterProxy methodArgumentCount))
	ifFalse: [^interpreterProxy primitiveFail].
    self copyBits.
    interpreterProxy failed ifTrue: [^nil].
    self showDisplayBits.
    interpreterProxy failed ifTrue: [^nil].
    interpreterProxy pop: interpreterProxy methodArgumentCount.
    (combinationRule == 22 or: [combinationRule == 32])
	ifTrue: [^interpreterProxy popThenPush: bitCount].
]

BitBltSimulation loadBitBltFrom: bbObj
[
    ^self loadBitBltFrom: bbObj warping: false.
]

BitBltSimulation loadBitBltFrom: bbObj warping: aBool
[
    "Load context from BitBlt instance.  Return false if anything is amiss."
    | ok |
    bitBltOop := bbObj.
    isWarping := aBool.
    combinationRule := bitBltOop atBBRuleIndex.
    (combinationRule isSmallInteger and: [combinationRule >= 0 and: [combinationRule < OpTable size]]) ifFalse: [$A put.  ^false].
    (combinationRule between: 16 and: 17) ifTrue: [^false].
    sourceForm := bitBltOop atBBSourceFormIndex.
    noSource := self ignoreSourceOrHalftone: sourceForm.
    halftoneForm := bitBltOop atBBHalftoneFormIndex.
    noHalftone := self ignoreSourceOrHalftone: halftoneForm.
    destForm := bbObj atBBDestFormIndex.
    (destForm isPointers and: [destForm fixedSize >= 4]) ifFalse: [^false].
    self loadBitBltDestForm ifFalse: [^false].
    destX  := self checkIntOrFloat: bitBltOop atBBDestXIndex  ifNil: 0.
    destY  := self checkIntOrFloat: bitBltOop atBBDestYIndex  ifNil: 0.
    width  := self checkIntOrFloat: bitBltOop atBBWidthIndex  ifNil: destWidth.
    height := self checkIntOrFloat: bitBltOop atBBHeightIndex ifNil: destHeight.
    interpreterProxy failed ifTrue: [^false].
    noSource
	ifTrue: [sourceX := sourceY := 0]
	ifFalse: 
	   [(sourceForm isPointers and: [sourceForm fixedSize >= 4 and: [self loadBitBltSourceForm and: [self loadColorMap]]])
		ifFalse: [^false].
	    (cmFlags bitAnd: ColorMapNewStyle) == 0 ifTrue: [self setupColorMasks].
	    sourceX := self checkIntOrFloat: bitBltOop atBBSourceXIndex ifNil: 0.
	    sourceY := self checkIntOrFloat: bitBltOop atBBSourceYIndex ifNil: 0].
    self loadHalftoneForm ifFalse: [^false].
    clipX      := self checkIntOrFloat: bitBltOop atBBClipXIndex      ifNil: 0.
    clipY      := self checkIntOrFloat: bitBltOop atBBClipYIndex      ifNil: 0.
    clipWidth  := self checkIntOrFloat: bitBltOop atBBClipWidthIndex  ifNil: destWidth.
    clipHeight := self checkIntOrFloat: bitBltOop atBBClipHeightIndex ifNil: destHeight.
    interpreterProxy failed ifTrue: [^false].
    clipX < 0 ifTrue: [clipWidth  := clipWidth  + clipX.  clipX := 0].
    clipY < 0 ifTrue: [clipHeight := clipHeight + clipY.  clipY := 0].
    clipX + clipWidth  > destWidth  ifTrue: [clipWidth  := destWidth  - clipX].
    clipY + clipHeight > destHeight ifTrue: [clipHeight := destHeight - clipY].
    ^true
]

BitBltSimulation ignoreSourceOrHalftone: formPointer
[
    ^formPointer == interpreterProxy nilObject
	or: [combinationRule == 0
	or: [combinationRule == 5
	or: [combinationRule == 10
	or: [combinationRule == 15]]]]
]

BitBltSimulation loadBitBltSourceForm
[
    "Load the source form for BitBlt. Return false if anything is wrong, true otherwise."
    | sourceBitsSize |
    sourceBits   := sourceForm atFormBitsIndex.
    sourceWidth  := sourceForm atFormWidthIndex.
    sourceHeight := sourceForm atFormHeightIndex.
    (sourceWidth >= 0 and: [sourceHeight >= 0])	ifFalse: [^false].
    sourceDepth  := sourceForm atFormDepthIndex.
    sourceMSB    := sourceDepth > 0.
    sourceDepth < 0 ifTrue: [sourceDepth := sourceDepth negated].
    "Ignore an integer bits handle for Display in which case the appropriate values will be obtained by calling ioLockSurfaceBits()."
    sourceBits isSmallInteger
	ifTrue: [self error: 'surfaces not supported']
	ifFalse:
	   [sourcePPW := 32 // sourceDepth.
	    sourcePitch := sourceWidth + (sourcePPW - 1) // sourcePPW * 4.
	    sourceBitsSize := sourceBits byteSize.
	    (sourceBits isWordsOrBytes and: [sourceBitsSize == (sourcePitch * sourceHeight)]) ifFalse: [^false].
	    "Skip header since external bits don't have one"
	    "sourceBits := sourceBits indexableFields"].
    ^true
]

BitBltSimulation loadBitBltDestForm
[
    "Load the dest form for BitBlt. Return false if anything is wrong, true otherwise."
    | destBitsSize |
    destBits   := destForm atFormBitsIndex.
    destWidth  := destForm atFormWidthIndex.
    destHeight := destForm atFormHeightIndex.
    (destWidth >= 0 and: [destHeight >= 0]) ifFalse: [^false].
    destDepth  := destForm atFormDepthIndex.
    destDepth isSmallInteger ifFalse: [^false].
    destMSB    := destDepth > 0.
    destDepth < 0 ifTrue: [destDepth := destDepth negated].
    "Ignore an integer bits handle for Display in which case the appropriate values will be obtained by calling ioLockSurfaceBits()."
    destBits isSmallInteger
	ifTrue: [self error: 'surfaces not supported']
	ifFalse:
	   [destPPW := 32 // destDepth.
	    destPitch := destWidth + (destPPW - 1) // destPPW * 4.
	    destBitsSize := destBits byteSize.
	    (destBits isWordsOrBytes and: [destBitsSize == (destPitch * destHeight)]) ifFalse: [^false].
	    "Skip  header since external bits don't have one"
	    "destBits := destBits indexableFields"].
    ^true
]

BitBltSimulation loadColorMap
[
    "ColorMap, if not nil, must be longWords, and 2^N long, where N =
    sourceDepth for 1, 2, 4, 8 bits, or N = 9, 12, or 15 (3, 4, 5 bits
    per color) for 16 or 32 bits."
    | cmSize oldStyle cmap cmOop |
    cmFlags := cmMask := cmBitsPerColor := 0.
    cmShiftTable := nil.
    cmMaskTable := nil.
    cmLookupTable := nil.
    cmOop := bitBltOop atBBColorMapIndex.
    cmOop == interpreterProxy nilObject ifTrue: [^true].
    cmFlags := ColorMapPresent. "even if identity or somesuch - may be cleared later"
    oldStyle := false.
    cmOop isWords
	ifTrue:		"This is an old-style color map (indexed only, with implicit RGBA conversion)"
	   [cmSize := cmOop indexableSize.
	    cmLookupTable := cmOop "firstIndexableField".
	    oldStyle := true]
	ifFalse:	"A new-style color map (fully qualified)"
	   [(cmOop isPointers and: [cmOop fixedSize >= 3]) ifFalse: [^false].
	    cmShiftTable := self loadColorMapShiftOrMaskFrom: cmOop atColorMapShiftsIndex.
	    cmMaskTable  := self loadColorMapShiftOrMaskFrom: cmOop atColorMapMasksIndex.
	    (cmap := cmOop atColorMapColorsIndex) == interpreterProxy nilObject
		ifTrue:  [cmSize := 0]
		ifFalse:
		    [cmap isWords ifFalse: [^false].
		     cmSize := cmap indexableSize.
		     cmLookupTable := cmap "firstIndexableField"].
	    cmFlags := cmFlags bitOr: ColorMapNewStyle].
    (cmSize bitAnd: cmSize - 1) == 0 ifFalse: [^false].
    cmMask := cmSize - 1.
    cmBitsPerColor := 0.
    cmSize ==   512 ifTrue: [cmBitsPerColor := 3].
    cmSize ==  4096 ifTrue: [cmBitsPerColor := 4].
    cmSize == 32768 ifTrue: [cmBitsPerColor := 5].
    cmSize == 0
	ifTrue:  [cmLookupTable := nil.  cmMask := 0]
	ifFalse: [cmFlags := cmFlags bitOr: ColorMapIndexedPart].
    oldStyle ifTrue: [self setupColorMasks].	 "needs implicit conversion"
    "Check if colorMap is just identity mapping for RGBA parts"
    (self isIdentityMap: cmShiftTable with: cmMaskTable)
	ifTrue:  [cmMaskTable := nil.  cmShiftTable := nil]
	ifFalse: [cmFlags := cmFlags bitOr: ColorMapFixedPart].
    ^true    
]

BitBltSimulation setupColorMasks
[
    "WARNING: For WarpBlt w/ smoothing the source depth is wrong here!"
    | bits targetBits |
    bits := targetBits := 0.
    sourceDepth <=  8 ifTrue: [^nil].
    sourceDepth == 16 ifTrue: [bits := 5].
    sourceDepth == 32 ifTrue: [bits := 8].
    cmBitsPerColor == 0
	ifTrue:		"Convert to destDepth"
	   [destDepth <=  8 ifTrue: [^nil].
	    destDepth == 16 ifTrue: [targetBits := 5].
	    destDepth == 32 ifTrue: [targetBits := 8]]
	ifFalse:
	   [targetBits := cmBitsPerColor].
    self setupColorMasksFrom: bits to: targetBits.
]

BitBltSimulation setupColorMasksFrom: srcBits to: targetBits
[
    "Setup color masks for converting an incoming RGB pixel value from srcBits to targetBits."
    | mask shifts masks deltaBits |
    shifts := WordArray new: 4.
    masks  := WordArray new: 4.
    deltaBits := targetBits - srcBits.
    deltaBits == 0 ifTrue: [^0].
    deltaBits <= 0
	ifTrue:
	   [mask := 1 << targetBits - 1.
	    "Mask for extracting a color part of the source"
	    masks at: RedIndex   put: mask << (srcBits * 2 - deltaBits).
	    masks at: GreenIndex put: mask << (srcBits - deltaBits).
	    masks at: BlueIndex  put: mask << (0 - deltaBits).
	    masks at: AlphaIndex put: 0]
	ifFalse:
	   [mask := 1 << srcBits - 1.
	    "Mask for extracting a color part of the source"
	    masks at: RedIndex   put: mask << (srcBits * 2).
	    masks at: GreenIndex put: mask << srcBits.
	    masks at: BlueIndex  put: mask].
    "Shifts for adjusting each value in a cm RGB value"
    shifts at: RedIndex   put: deltaBits * 3.
    shifts at: GreenIndex put: deltaBits * 2.
    shifts at: BlueIndex  put: deltaBits.
    shifts at: AlphaIndex put: 0.
    cmShiftTable := shifts.
    cmMaskTable  := masks.
    cmFlags      := cmFlags bitOr: (ColorMapPresent bitOr: ColorMapFixedPart).
]

BitBltSimulation isIdentityMap: shifts with: masks
[
    "Answer wether shiftTable/maskTable define an identity mapping."
    (shifts == nil or: [masks == nil]) ifTrue: [^true].
    ^         (shifts at: RedIndex)   == 0
	and: [(shifts at: GreenIndex) == 0
	and: [(shifts at: BlueIndex)  == 0
	and: [(shifts at: AlphaIndex) == 0
	and: [(masks  at: RedIndex)   == 16r00FF0000
	and: [(masks  at: GreenIndex) == 16r0000FF00
	and: [(masks  at: BlueIndex)  == 16r000000FF
	and: [(masks  at: AlphaIndex) == 16rFF000000]]]]]]]
]

BitBltSimulation checkSourceOverlap
[
    "check for possible overlap of source and destination"
    | t |
    (sourceForm == destForm and: [dy >= sy])
	ifTrue:
	   [dy > sy
		ifTrue:
		   ["have to start at bottom"
		    vDir := -1.
		    sy := sy + bbH - 1.
		    dy := dy + bbH - 1]
		ifFalse:
		   [(dy == sy and: [dx > sx])
			ifTrue:
			   ["y's are equal, but x's are backward"
			    hDir := -1.
			    sx := sx + bbW - 1.
			    "start at right"
			    dx := dx + bbW - 1.
			    "and fix up masks"
			    nWords > 1
				ifTrue: 
				   [t := mask1.
				    mask1 := mask2.
				    mask2 := t]]].
	    destIndex := destBits + (dy * destPitch) + ((dx // destPPW) * 4).
	    destDelta := (destPitch * vDir) - (4 * (nWords * hDir))]
]

BitBltSimulation checkIntOrFloat: anObject ifNil: defaultValue
[
    anObject == interpreterProxy nilObject ifTrue: [^defaultValue].
    (anObject isSmallInteger or: [anObject isFloat]) ifFalse: [^interpreterProxy primitiveFailed].
    ^anObject
]

BitBltSimulation loadHalftoneForm
[
    | halftoneBits |
    noHalftone
	ifTrue:
	   [halftoneBase := nil.
	    ^true].
    (halftoneForm isPointers and: [halftoneForm fixedSize >= 4])
	ifTrue:
	   ["Old-style 32xN monochrome halftone Forms"
	    halftoneBits := halftoneForm atFormBitsIndex.
	    halftoneHeight := halftoneForm atFormHeightIndex.
	    halftoneBits isWords ifFalse: [noHalftone := true]]
	ifFalse:
	   ["New-style WordArray"
	    halftoneForm isWords ifFalse: [^false].
	    halftoneBits := halftoneForm.
	    halftoneHeight := halftoneBits indexableSize].
    halftoneBase := halftoneBits "indexableFields".
    ^true
]

BitBltSimulation copyBits
[
    self clipRange.
    (bbW <= 0 or: [bbH <= 0])
	ifTrue:
	   [affectedL := affectedR := affectedT := affectedB := 0.
	    ^nil].
    self lockSurfaces ifFalse: [^interpreterProxy primitiveFail].
    self copyBitsLockedAndClipped.
    self unlockSurfaces.
]

BitBltSimulation clipRange
[
    "clip and adjust source origin and extent appropriately"
    "first in x"
    destX >= clipX
	ifTrue:  [sx := sourceX.                    bbW := width.                    dx := destX]
	ifFalse: [sx := sourceX + (clipX - destX).  bbW := width - (clipX - destX).  dx := clipX].
    (dx + bbW) > (clipX + clipWidth)
	ifTrue: [bbW := bbW - ((dx + bbW) - (clipX + clipWidth))].
    "then in y"
    destY >= clipY
	ifTrue:  [sy := sourceY.                  bbH := height.                    dy := destY]
	ifFalse: [sy := sourceY + clipY - destY.  bbH := height - (clipY - destY).  dy := clipY].
    (dy + bbH) > (clipY + clipHeight)
	ifTrue: [bbH := bbH - ((dy + bbH) - (clipY + clipHeight))].
    noSource ifTrue: [^nil].
    sx < 0 ifTrue: [dx := dx - sx.  bbW := bbW + sx.  sx := 0].
    sx + bbW > sourceWidth ifTrue: [bbW := bbW - (sx + bbW - sourceWidth)].
    sy < 0 ifTrue: [dy := dy - sy.  bbH := bbH + sy.  sy := 0].
    sy + bbH > sourceHeight ifTrue: [bbH := bbH - (sy + bbH - sourceHeight)].
]

BitBltSimulation lockSurfaces
[
    hasSurfaceLock := false.
    destBits == 0 ifTrue: [self error: 'surfaces not supported'].
    (sourceBits == 0 and: [noSource not]) ifTrue: [self error: 'surfaces not supported'].
    ^destBits ~~ 0 and: [sourceBits ~~ 0 or: [noSource]].
]

BitBltSimulation unlockSurfaces
[
]

BitBltSimulation copyBitsLockedAndClipped
[
    | done |
    "Try a shortcut for stuff that should be run as quickly as possible"
     self tryCopyingBitsQuickly ifTrue: [^nil].
    (combinationRule == 30 or: [combinationRule == 31])
	ifTrue:
	   ["Check and fetch source alpha parameter for alpha blend"
	    interpreterProxy methodArgumentCount == 1
		ifTrue:  [sourceAlpha := interpreterProxy stackIntegerValue: 0.
			  (interpreterProxy failed not and: [sourceAlpha >= 0 and: [sourceAlpha <= 255]])
			      ifFalse: [^interpreterProxy primitiveFail]]
		ifFalse: [^interpreterProxy primitiveFail]].
    bitCount := 0.
    self performCopyLoop.    "Choose and perform the actual copy loop."
    (combinationRule == 22 or: [combinationRule = 32])
	ifTrue:  [affectedL := affectedR := affectedT := affectedB := 0]. "zero width and height; return the count"
    hDir > 0
	ifTrue:  [affectedL := dx.            affectedR := dx + bbW]
	ifFalse: [affectedL := dx - bbW + 1.  affectedR := dx + 1].
    vDir > 0
	ifTrue:  [affectedT := dy.            affectedB := dy + bbH]
	ifFalse: [affectedT := dy - bbH + 1.  affectedB := dy + 1].
]

BitBltSimulation tryCopyingBitsQuickly
[
    "We need a source."
    noSource ifTrue: [^false].
    "We handle only combinationRule 34"
    combinationRule == 34 ifFalse: [^false].
    "We handle only sourceDepth 32"
    sourceDepth == 32 ifFalse: [^false].
    "We don't handle overlaps"
    sourceForm == destForm ifTrue: [^false].
    "We need at least 8bit deep dest forms"
    destDepth < 8 ifTrue: [^false].
    "If 8bit, then we want a color map"
    (destDepth == 8 and: [(cmFlags bitAnd: ColorMapPresent) == 0]) ifTrue: [^false].
    destDepth == 32 ifTrue: [self alphaSourceBlendBits32].
    destDepth == 16 ifTrue: [self alphaSourceBlendBits16].
    destDepth ==  8 ifTrue: [self alphaSourceBlendBits8].
    affectedL := dx.
    affectedR := dx + bbW.
    affectedT := dy.
    affectedB := dy + bbH.
    ^true
]

BitBltSimulation performCopyLoop
[
    self destMaskAndPointerInit.
    noSource
	ifTrue:  [self copyLoopNoSource] "Simple fill loop"
	ifFalse: "Loop using source and dest"
	   [self checkSourceOverlap.
	    (sourceDepth ~= destDepth or: [(cmFlags ~= 0) or: [sourceMSB ~= destMSB]])
		ifTrue:  [self copyLoopPixMap]
		ifFalse: [self sourceSkewAndPointerInit; copyLoop]]
]

BitBltSimulation destMaskAndPointerInit
[
    "Compute masks for left and right destination words"
    | startBits pixPerM1 endBits |
    pixPerM1 := destPPW - 1.  "A mask, assuming power of two"
    "how many pixels in first word"
    startBits := destPPW - (dx bitAnd: pixPerM1).
    destMSB
	ifTrue:  [mask1 := AllOnes >> (32 - (startBits * destDepth))]
	ifFalse: [mask1 := AllOnes << (32 - (startBits * destDepth))].
    "how many pixels in last word"
    endBits := ((dx + bbW - 1) bitAnd: pixPerM1) + 1.
    destMSB 
	ifTrue:  [mask2 := AllOnes << (32 - (endBits * destDepth))] 
	ifFalse: [mask2 := AllOnes >> (32 - (endBits * destDepth))].
    "determine number of words stored per line; merge masks if only 1"
    bbW < startBits
	ifTrue:  [mask1 := mask1 bitAnd: mask2.
		  mask2 := 0.
		  nWords := 1]
	ifFalse: [nWords := (bbW - startBits) + pixPerM1 // destPPW + 1].
    hDir := vDir := 1. "defaults for no overlap with source"
    "calculate byte addr and delta, based on first word of data"
    "Note pitch is bytes and nWords is longs, not bytes"
    destIndex := 0 "xxx destBits xxx" + (dy * destPitch) + ((dx // destPPW) * 4).
    destDelta := destPitch * vDir - (4 * (nWords * hDir)).  "byte addr delta"
]

BitBltSimulation sourceSkewAndPointerInit
[
    "This is only used when source and dest are same depth, i.e., when
    the barrel-shift copy loop is used."
    | dWid sxLowBits dxLowBits pixPerM1 |
    pixPerM1 := destPPW - 1.  "A mask, assuming power of two"
    sxLowBits := sx bitAnd: pixPerM1.
    dxLowBits := dx bitAnd: pixPerM1.
    "check if need to preload buffer (i.e., two words of source needed for first word of destination)"
    hDir > 0
	ifTrue:	"n Bits stored in 1st word of dest"
	   [dWid := bbW min: destPPW - dxLowBits.
	    preload := (sxLowBits + dWid) > pixPerM1]
	ifFalse:
	   [dWid := bbW min: dxLowBits + 1.
	    preload := (sxLowBits - dWid + 1) < 0].
    "calculate right-shift skew from source to dest"
    skew := sourceMSB
	ifTrue:  [(sxLowBits - dxLowBits) * destDepth] 
	ifFalse: [(dxLowBits - sxLowBits) * destDepth].  " -32..32 "
    preload ifTrue: [skew := skew + (skew < 0 ifTrue: [32] ifFalse: [-32])].
    "Calc byte addr and delta from longWord info"
    sourceIndex := "sourceBits +" (sy * sourcePitch) + ((sx // (32 // sourceDepth)) * 4).
    "calculate increments from end of 1 line to start of next"
    sourceDelta := (sourcePitch * vDir) - (4 * (nWords * hDir)).
    "Compensate for extra source word fetched"
    preload ifTrue: [sourceDelta := sourceDelta - (4 * hDir)].
]

BitBltSimulation copyLoopNoSource
[
    | halftoneWord mergeWord mergeFn destWord |
    mergeFn := OpTable at: 1 + combinationRule.
    1 to: bbH do: [:i |
	halftoneWord := noHalftone
	    ifTrue:  [AllOnes]
	    ifFalse: [self halftoneAt: dy + i - 1].
	"Note: the horizontal loop has been expanded into three parts for speed:"
	"This first section requires masking of the destination store..."
	destMask  := mask1.
	destWord  := self dstLongAt: destIndex.
	mergeWord := self _perform: mergeFn w: halftoneWord w: destWord.
	destWord  := (destMask bitAnd: mergeWord) bitOr: (destWord bitAnd: destMask bitInvert32).
	self dstLongAt: destIndex put: destWord.
	destIndex := destIndex + 4.
	"This central horizontal loop requires no store masking"
	destMask := AllOnes.
	combinationRule == 3
	    ifTrue:
	       ["Special inner loop for STORE"
		destWord := halftoneWord.
		2 to: nWords - 1 do: [:word |
		    self dstLongAt: destIndex put: destWord.
		    destIndex := destIndex + 4]]
	    ifFalse:
	       ["Normal inner loop does merge"
		2 to: nWords - 1 do: [:word |
		    destWord  := self dstLongAt: destIndex.
		    mergeWord := self _perform: mergeFn w: halftoneWord w: destWord.
		    self dstLongAt: destIndex put: mergeWord.
		    destIndex := destIndex + 4]].
	"This last section, if used, requires masking of the destination store..."
	nWords > 1
	    ifTrue:
	       [destMask  := mask2.
		destWord  := self dstLongAt: destIndex.
		mergeWord := self _perform: mergeFn w: halftoneWord w: destWord.
		destWord  := (destMask bitAnd: mergeWord) bitOr: (destWord bitAnd: destMask bitInvert32).
		self dstLongAt: destIndex put: destWord.
		destIndex := destIndex + 4].
	destIndex := destIndex + destDelta]
]

BitBltSimulation copyLoopPixMap
[
    | skewWord halftoneWord mergeWord scrStartBits nSourceIncs startBits endBits sourcePixMask destPixMask mergeFn nPix srcShift dstShift destWord words srcShiftInc dstShiftInc dstShiftLeft mapperFlags |
    mergeFn := OpTable at: 1 + combinationRule.
('copyLoopPixMap ', sourceDepth printString, ' ', destDepth printString, ' ', (String value_: mergeFn _string)) putln.
    "Additional inits peculiar to unequal source and dest pix size..."
    sourcePPW := 32 // sourceDepth.
    sourcePixMask := maskTable at: sourceDepth.
    destPixMask := maskTable at: destDepth.
    mapperFlags := cmFlags bitAnd: ColorMapNewStyle bitInvert32.
    sourceIndex := "sourceBits +" (sy * sourcePitch) + ((sx // sourcePPW) * 4).
    scrStartBits := sourcePPW - (sx bitAnd: sourcePPW - 1).
    nSourceIncs := bbW < scrStartBits
	ifTrue:  [0]
	ifFalse: [(bbW - scrStartBits) // sourcePPW + 1].
    sourceDelta := sourcePitch - (nSourceIncs * 4).
    "Note following two items were already calculated in destmask setup"
    startBits := destPPW - (dx bitAnd: destPPW - 1).
    endBits := ((dx + bbW - 1) bitAnd: destPPW - 1) + 1.
('sx ',			sx		printString
,' sy ',		sy		printString
,' bbW ',		bbW		printString
,' bbH ',		bbH		printString
) putln.
(' sourceIndex ',	sourceIndex	printString
,' scrStartBits ',	scrStartBits	printString
,' nSourceIncs ',	nSourceIncs	printString
,' sourceDelta ',	sourceDelta	printString
,' startBits ',		startBits	printString
,' endBits ',		endBits		printString
) putln.
    bbW < startBits ifTrue: [startBits := bbW].
    "Precomputed shifts for pickSourcePixels"
    srcShift := ((sx bitAnd: sourcePPW - 1) * sourceDepth).
    dstShift := ((dx bitAnd: destPPW - 1) * destDepth).
    srcShiftInc  := sourceDepth.
    dstShiftInc  := destDepth.
    dstShiftLeft := 0.
    sourceMSB
	ifTrue:
	   [srcShift := 32 - sourceDepth - srcShift.
	    srcShiftInc := 0 - srcShiftInc].
    destMSB
	ifTrue:
	   [dstShift := 32 - destDepth - dstShift.
	    dstShiftInc := 0 - dstShiftInc.
	    dstShiftLeft := 32 - destDepth].
(' srcShift ',		srcShift	printString
,' srcShiftInc ',	srcShiftInc	printString
,' dstShift ',		dstShift	printString
,' dstShiftInc ',	dstShiftInc	printString
,' dstShiftLeft ',	dstShiftLeft	printString
) putln.
    "here is the vertical loop"
    1 to: bbH do: [:i |
	halftoneWord := noHalftone
	    ifTrue:  [AllOnes]
	    ifFalse: [self halftoneAt: dy + i - 1].
	"setup first load"
	srcBitShift := srcShift.
	dstBitShift := dstShift.
	destMask    := mask1.
	nPix        := startBits.
	"Here is the horizontal loop..."
	words := nWords.
	[	"pick up the word"
		skewWord := self pickSourcePixels: nPix flags: mapperFlags 
				 srcMask: sourcePixMask destMask: destPixMask
				 srcShiftInc: srcShiftInc dstShiftInc: dstShiftInc.
		"align next word to leftmost pixel"
		dstBitShift := dstShiftLeft.
		destMask == AllOnes
		    ifTrue:	"avoid read-modify-write"
			[mergeWord := self _perform: mergeFn
					   w: (skewWord bitAnd: halftoneWord)
					   w: (self dstLongAt: destIndex).
			 self dstLongAt: destIndex put: (destMask bitAnd: mergeWord)]
		    ifFalse:	"General version using dest masking"
			[destWord := self dstLongAt: destIndex.
			 mergeWord := self _perform: mergeFn
					   w: (skewWord bitAnd: halftoneWord)
					   w: (destWord bitAnd: destMask).
			 destWord := (destMask bitAnd: mergeWord) bitOr: (destWord bitAnd: destMask bitInvert32).
			 self dstLongAt: destIndex put: destWord].
		destIndex := destIndex + 4.
		words == 2 "e.g., is the next word the last word?"
		    ifTrue:	"set mask for last word in this row"
		       [destMask := mask2.
			nPix := endBits]
		    ifFalse:	"use fullword mask for inner loop"
		       [destMask := AllOnes.
			nPix := destPPW].
		(words := words - 1) == 0]
	    whileFalse.
	"--- end of inner loop ---"
	sourceIndex := sourceIndex + sourceDelta.
	destIndex := destIndex + destDelta]
]

BitBltSimulation copyLoop
[
    "This version of the inner loop assumes noSource = false."
    | prevWord thisWord skewWord halftoneWord mergeWord hInc y unskew skewMask notSkewMask mergeFn destWord |
    mergeFn := OpTable at: 1 + combinationRule.
    hInc := hDir * 4.  "Byte delta"
    skew == -32
	ifTrue:
	   [skew := unskew := skewMask := 0]
	ifFalse:
	   [skew < 0
		ifTrue:
		   [unskew := skew + 32.
		    skewMask := AllOnes << (0 - skew)]
		ifFalse:
		   [skew == 0
			ifTrue:
			   [unskew := 0.
			    skewMask := AllOnes]
			ifFalse:
			   [unskew := skew - 32.
			    skewMask := AllOnes >> skew]]].
    notSkewMask := skewMask bitInvert32.
    noHalftone
	ifTrue:  [halftoneWord := AllOnes.  halftoneHeight := 0]
	ifFalse: [halftoneWord := self halftoneAt: 0].
    y := dy.
    "here is the vertical loop"
    1 to: bbH do: [:i |
	halftoneHeight > 1
	    ifTrue:	"Otherwise, its always the same"
	       [halftoneWord := self halftoneAt: y.
		y := y + vDir].
	preload
	    ifTrue:	"load the 64-bit shifter"
	       [prevWord := self srcLongAt: sourceIndex.
		sourceIndex := sourceIndex + hInc]
	    ifFalse:
	       [prevWord := 0].
	"Note: the horizontal loop has been expanded into three parts for speed:"
	"This first section requires masking of the destination store..."
	destMask := mask1.
	thisWord := self srcLongAt: sourceIndex.  "pick up next word"
	sourceIndex := sourceIndex + hInc.
	skewWord := ((prevWord bitAnd: notSkewMask) bitShift: unskew) bitOr: ((thisWord bitAnd: skewMask) bitShift: skew).
	prevWord := thisWord.
	destWord := self dstLongAt: destIndex.
	mergeWord := self _perform: mergeFn w: (skewWord bitAnd: halftoneWord) w: destWord.
	destWord := (destMask bitAnd: mergeWord) bitOr: (destWord bitAnd: destMask bitInvert32).
	self dstLongAt: destIndex put: destWord.
	destIndex := destIndex + hInc.
	"This central horizontal loop requires no store masking"
	destMask := AllOnes.
	combinationRule == 3
	    ifTrue:
	       [(skew == 0 and: [halftoneWord == AllOnes])
		    ifTrue: 
		       ["Very special inner loop for STORE mode with no skew -- just move words"
			hDir == -1
			    ifTrue:
			       ["Woeful patch: revert to older code for hDir = -1"
				2 to: nWords - 1 do: [:word |
				    thisWord := self srcLongAt: sourceIndex.
				    sourceIndex := sourceIndex + hInc.
				    self dstLongAt: destIndex put: thisWord.
				    destIndex := destIndex + hInc]]
			    ifFalse:
			       [2 to: nWords - 1 do: [:word |
				   "Note loop starts with prevWord loaded (due to preload)"
				   self dstLongAt: destIndex put: prevWord.
				   destIndex := destIndex + hInc.
				   prevWord := self srcLongAt: sourceIndex.
				   sourceIndex := sourceIndex + hInc]]]
		    ifFalse:
		       ["Special inner loop for STORE mode -- no need to call merge"
			2 to: nWords - 1 do: [:word |
			    thisWord := self srcLongAt: sourceIndex.
			    sourceIndex := sourceIndex + hInc.
			    skewWord := ((prevWord bitAnd: notSkewMask) bitShift: unskew)
				bitOr: ((thisWord bitAnd: skewMask) bitShift: skew).
			    prevWord := thisWord.
			    self dstLongAt: destIndex put: (skewWord bitAnd: halftoneWord).
			    destIndex := destIndex + hInc]]]
	    ifFalse:	 "Normal inner loop does merge:"
	       [2 to: nWords - 1 do: [:word |
		   thisWord := self srcLongAt: sourceIndex.  "pick up next word"
		   sourceIndex := sourceIndex + hInc.
		   skewWord := ((prevWord bitAnd: notSkewMask) bitShift: unskew) bitOr: ((thisWord bitAnd: skewMask) bitShift: skew).
		   prevWord := thisWord.
		   mergeWord := self _perform: mergeFn w: (skewWord bitAnd: halftoneWord) w: (self dstLongAt: destIndex).
		   self dstLongAt: destIndex put: mergeWord.
		   destIndex := destIndex + hInc]].
	"This last section, if used, requires masking of the destination store..."
	nWords > 1
	    ifTrue:
	       [destMask := mask2.
		thisWord := self srcLongAt: sourceIndex.  "pick up next word"
		sourceIndex := sourceIndex + hInc.
		skewWord := ((prevWord bitAnd: notSkewMask) bitShift: unskew) bitOr: ((thisWord bitAnd: skewMask) bitShift: skew).
		destWord := self dstLongAt: destIndex.
		mergeWord := self _perform: mergeFn w: (skewWord bitAnd: halftoneWord) w: destWord.
		destWord := (destMask bitAnd: mergeWord) bitOr: (destWord bitAnd: destMask bitInvert32).
		self dstLongAt: destIndex put: destWord.
		destIndex := destIndex + hInc].
	sourceIndex := sourceIndex + sourceDelta.
	destIndex := destIndex + destDelta].
]

BitBltSimulation srcLongAt: idx			[ ^sourceBits indexedAt: (1 + (idx // 4)) ]
BitBltSimulation srcLongAt: idx put: word	[ ^sourceBits indexedAt: (1 + (idx // 4)) put: word ]
BitBltSimulation dstLongAt: idx			[ ^destBits indexedAt: (1 + (idx // 4)) ]
BitBltSimulation dstLongAt: idx put: word	[ ^destBits indexedAt: (1 + (idx // 4)) put: word ]
BitBltSimulation halftoneAt: idx		[ ^halftoneBase indexedAt: 1 + (idx \\ halftoneHeight) ]

"combination rules"

BitBltSimulation clearWord: source with: destination			[ ^0 ]
BitBltSimulation bitAnd: sourceWord with: destinationWord		[ ^sourceWord bitAnd: destinationWord ]
BitBltSimulation bitAndInvert: sourceWord with: destinationWord		[^sourceWord bitAnd: destinationWord bitInvert32 ]
BitBltSimulation sourceWord: sourceWord with: destinationWord		[ ^sourceWord ]
BitBltSimulation bitInvertAnd: sourceWord with: destinationWord		[ ^sourceWord bitInvert32 bitAnd: destinationWord ]
BitBltSimulation destinationWord: sourceWord with: destinationWord	[ ^destinationWord ]
BitBltSimulation bitXor: sourceWord with: destinationWord		[ ^sourceWord bitXor: destinationWord ]
BitBltSimulation bitOr: sourceWord with: destinationWord		[ ^sourceWord bitOr: destinationWord ]
BitBltSimulation bitInvertAndInvert: sourceWord with: destinationWord	[ ^sourceWord bitInvert32 bitAnd: destinationWord bitInvert32 ]
BitBltSimulation bitInvertXor: sourceWord with: destinationWord		[ ^sourceWord bitInvert32 bitXor: destinationWord ]
BitBltSimulation bitInvertDestination: sourceWord with: destinationWord	[ ^destinationWord bitInvert32 ]
BitBltSimulation bitOrInvert: sourceWord with: destinationWord		[ ^sourceWord bitOr: destinationWord bitInvert32 ]
BitBltSimulation bitInvertSource: sourceWord with: destinationWord	[ ^sourceWord bitInvert32 ]
BitBltSimulation bitInvertOr: sourceWord with: destinationWord		[ ^sourceWord bitInvert32 bitOr: destinationWord ]
BitBltSimulation bitInvertOrInvert: sourceWord with: destinationWord	[ ^sourceWord bitInvert32 bitOr: destinationWord bitInvert32 ]
BitBltSimulation addWord: sourceWord with: destinationWord		[ ^sourceWord + destinationWord ]
BitBltSimulation subWord: sourceWord with: destinationWord		[ ^sourceWord - destinationWord ]

BitBltSimulation rgbAdd: sourceWord with: destinationWord
[
    destDepth < 16
	ifTrue:		"Add each pixel separately"
	   [^self partitionedAdd: sourceWord to: destinationWord nBits: destDepth nPartitions: destPPW].
    destDepth == 16
	ifTrue:		"Add RGB components of each pixel separately"
	    [^(self partitionedAdd: sourceWord to: destinationWord  nBits: 5 nPartitions: 3)
		 + ((self partitionedAdd: sourceWord >> 16 to: destinationWord >> 16 nBits: 5 nPartitions: 3) << 16)]
	ifFalse:	"Add RGB c	omponents of the pixel separately"
	    [^self partitionedAdd: sourceWord to: destinationWord nBits: 8 nPartitions: 3]
]

BitBltSimulation rgbSub: sourceWord with: destinationWord
[
    destDepth < 16
	ifTrue:		"Sub each pixel separately"
	   [^self partitionedSub: sourceWord from: destinationWord nBits: destDepth nPartitions: destPPW].
    destDepth == 16
	ifTrue:		"Sub RGB components of each pixel separately"
	   [^(self partitionedSub: sourceWord from: destinationWord nBits: 5 nPartitions: 3)
		+ ((self partitionedSub: sourceWord >> 16 from: destinationWord >> 16 nBits: 5 nPartitions: 3) << 16)]
	ifFalse:	"Sub RGB components of the pixel separately"
	    [^self partitionedSub: sourceWord from: destinationWord nBits: 8 nPartitions: 3]
]

BitBltSimulation OLDrgbDiff: sourceWord with: destinationWord
[
    "Subract the pixels in the source and destination, color by color,
    and return the sum of the absolute value of all the differences.
    For non-rgb, XOR the two and return the number of differing
    pixels.  Note that the region is not clipped to bit boundaries,
    but only to the nearest (enclosing) word.  This is because
    copyLoop does not do pre-merge masking.  For accurate results, you
    must subtract the values obtained from the left and right
    fringes."
    | diff pixMask |
    destDepth < 16
	ifTrue:		"Just xor and count differing bits if not RGB"
	   [diff := sourceWord bitXor: destinationWord.
	    pixMask := maskTable at: destDepth.
	    [diff == 0]
		whileFalse:
		   [(diff bitAnd: pixMask) ~~ 0 ifTrue: [bitCount := bitCount + 1].
		    diff := diff >> destDepth].
		^ destinationWord "for no effect"].
    destDepth == 16
	ifTrue:
	   [diff := (self partitionedSub: sourceWord from: destinationWord nBits: 5 nPartitions: 3).
	    bitCount := bitCount + (diff bitAnd: 16r1F)	+ (diff >> 5 bitAnd: 16r1F) + (diff >> 10 bitAnd: 16r1F).
	    diff := (self partitionedSub: sourceWord >> 16 from: destinationWord >> 16 nBits: 5 nPartitions: 3).
	    bitCount := bitCount + (diff bitAnd: 16r1F) + (diff >> 5 bitAnd: 16r1F) + (diff >> 10 bitAnd: 16r1F)]
	ifFalse:
	   [diff := (self partitionedSub: sourceWord from: destinationWord nBits: 8 nPartitions: 3).
	    bitCount := bitCount + (diff bitAnd: 16rFF) + (diff >> 8 bitAnd: 16rFF) + (diff >> 16 bitAnd: 16rFF)].
    ^destinationWord  "For no effect on dest"
]

BitBltSimulation OLDtallyIntoMap: sourceWord with: destinationWord
[
    "Tally pixels into the color map.  Note that the source should be
    specified = destination, in order for the proper color map checks
    to be performed at setup.  Note that the region is not clipped to
    bit boundaries, but only to the nearest (enclosing) word.  This is
    because copyLoop does not do pre-merge masking.  For accurate
    results, you must subtract the values obtained from the left and
    right fringes."
    (cmFlags bitAnd: (ColorMapPresent bitOr: ColorMapIndexedPart)) == (ColorMapPresent bitOr: ColorMapIndexedPart)
	ifFalse: [^destinationWord "no op"].
    destDepth < 16
	ifTrue:		"loop through all packed pixels."
	   [| pixMask shiftWord |
	    pixMask := (maskTable at: destDepth) bitAnd: cmMask.
	    shiftWord := destinationWord.
	    1 to: destPPW do: [:i |
		| mapIndex |
		mapIndex := shiftWord bitAnd: pixMask.
		self tallyMapAt: mapIndex put: (self tallyMapAt: mapIndex) + 1.
		shiftWord := shiftWord >> destDepth].
	    ^destinationWord].
    destDepth == 16
	ifTrue:		"Two pixels  Tally the right half..."
	   [| mapIndex |
	    mapIndex := self rgbMap: (destinationWord bitAnd: 16rFFFF) from: 5 to: cmBitsPerColor.
	      self tallyMapAt: mapIndex put: (self tallyMapAt: mapIndex) + 1.
	      "... and then left half"
	      mapIndex := self rgbMap: destinationWord >> 16 from: 5 to: cmBitsPerColor.
	      self tallyMapAt: mapIndex put: (self tallyMapAt: mapIndex) + 1]
	ifFalse:	"Just one pixel."
	   [| mapIndex |
	    mapIndex := self rgbMap: destinationWord from: 8 to: cmBitsPerColor.
	      self tallyMapAt: mapIndex put: (self tallyMapAt: mapIndex) + 1].
    ^destinationWord  "For no effect on dest"
]

BitBltSimulation alphaBlend: sourceWord with: destinationWord
[
    "Blend sourceWord with destinationWord, assuming both are 32-bit
    pixels.  The source is assumed to have 255*alpha in the high 8
    bits of each pixel, while the high 8 bits of the destinationWord
    will be ignored.  The blend produced is alpha*source +
    (1-alpha)*dest, with the computation being performed independently
    on each color component.  The high byte of the result will be 0."
    | alpha unAlpha colorMask result blend shift |
    alpha := sourceWord >> 24.  "High 8 bits of source pixel"
    alpha = 0 ifTrue: [^destinationWord].
    alpha = 255 ifTrue: [^sourceWord].
    unAlpha := 255 - alpha.
    colorMask := 16rFF.
    result := 0.

    "red"
    shift := 0.
    blend := ((sourceWord >> shift bitAnd: colorMask) * alpha)
	+ ((destinationWord >> shift bitAnd: colorMask) * unAlpha)
	+ 254 // 255 bitAnd: colorMask.
    result := result bitOr: blend << shift.
    "green"
    shift := 8.
    blend := ((sourceWord >> shift bitAnd: colorMask) * alpha)
	+ ((destinationWord >> shift bitAnd: colorMask) * unAlpha)
	+ 254 // 255 bitAnd: colorMask.
    result := result bitOr: blend << shift.
    "blue"
    shift := 16.
    blend := ((sourceWord >> shift bitAnd: colorMask) * alpha)
	+ ((destinationWord >> shift bitAnd: colorMask) * unAlpha)
	+ 254 // 255 bitAnd: colorMask.
    result := result bitOr: blend << shift.
    "alpha (pre-multiplied)"
    shift := 24.
    blend := (alpha * 255)
	+ ((destinationWord >> shift bitAnd: colorMask) * unAlpha)
	+ 254 // 255 bitAnd: colorMask.
    result := result bitOr: blend << shift.
    ^result
]

BitBltSimulation pixPaint: sourceWord with: destinationWord
[
    sourceWord == 0 ifTrue: [^destinationWord].
    ^sourceWord bitOr: (self partitionedAnd: sourceWord bitInvert32 to: destinationWord nBits: destDepth nPartitions: destPPW)
]

BitBltSimulation pixMask: sourceWord with: destinationWord
[
    ^self partitionedAnd: sourceWord bitInvert32 to: destinationWord nBits: destDepth nPartitions: destPPW
]

BitBltSimulation rgbMax: sourceWord with: destinationWord
[
    destDepth < 16
	ifTrue:		"Max each pixel separately"
	   [^self partitionedMax: sourceWord with: destinationWord nBits: destDepth nPartitions: destPPW].
    destDepth == 16
	ifTrue:		"Max RGB components of each pixel separately"
	   [^(self partitionedMax: sourceWord with: destinationWord nBits: 5 nPartitions: 3)
		+ ((self partitionedMax: sourceWord >> 16 with: destinationWord >> 16 nBits: 5 nPartitions: 3) << 16)]
	ifFalse:	"Max RGB components of the pixel separately"
	   [^self partitionedMax: sourceWord with: destinationWord nBits: 8 nPartitions: 3]
]

BitBltSimulation rgbMin: sourceWord with: destinationWord
[
    destDepth < 16
	ifTrue:		"Min each pixel separately"
	   [^self partitionedMin: sourceWord with: destinationWord nBits: destDepth nPartitions: destPPW].
    destDepth == 16
	ifTrue:		"Min RGB components of each pixel separately"
	   [^(self partitionedMin: sourceWord with: destinationWord nBits: 5 nPartitions: 3)
		+ ((self partitionedMin: sourceWord >> 16 with: destinationWord >> 16 nBits: 5 nPartitions: 3) << 16)]
	ifFalse:	"Min RGB components of the pixel separately"
	   [^self partitionedMin: sourceWord with: destinationWord nBits: 8 nPartitions: 3]
]

BitBltSimulation rgbMinInvert: wordToInvert with: destinationWord
[
    | sourceWord |
    sourceWord := wordToInvert bitInvert32.
    destDepth < 16
	ifTrue:		"Min each pixel separately"
	   [^self partitionedMin: sourceWord with: destinationWord nBits: destDepth nPartitions: destPPW].
    destDepth == 16
	ifTrue:		"Min RGB components of each pixel separately"
	   [^(self partitionedMin: sourceWord with: destinationWord nBits: 5 nPartitions: 3)
		+ ((self partitionedMin: sourceWord >> 16 with: destinationWord >> 16 nBits: 5 nPartitions: 3) << 16)]
	ifFalse:	"Min RGB components of the pixel separately"
	   [^self partitionedMin: sourceWord with: destinationWord nBits: 8 nPartitions: 3]
]

BitBltSimulation alphaBlendConst: sourceWord with: destinationWord
[
    ^self alphaBlendConst: sourceWord with: destinationWord paintMode: false
]

BitBltSimulation alphaBlendConst: sourceWord with: destinationWord paintMode: paintMode
[
    "Blend sourceWord with destinationWord using a constant alpha.
    Alpha is encoded as 0 meaning 0.0, and 255 meaning 1.0.  The blend
    produced is alpha*source + (1.0-alpha)*dest, with the computation
    being performed independently on each color component.  This
    function could eventually blend into any depth destination, using
    the same color averaging and mapping as warpBlt.  paintMode = true
    means do nothing if the source pixel value is zero."

    "This first implementation works with dest depths of 16 and 32
    bits only.  Normal color mapping will allow sources of lower
    depths in this case, and results can be mapped directly by
    truncation, so no extra color maps are needed.  To allow storing
    into any depth will require subsequent addition of two other
    colormaps, as is the case with WarpBlt."

    | pixMask destShifted sourceShifted destPixVal rgbMask sourcePixVal unAlpha result pixBlend shift blend maskShifted bitsPerColor |
    destDepth < 16 ifTrue: [^destinationWord "no-op"].
    unAlpha := 255 - sourceAlpha.
    pixMask := maskTable at: destDepth.
    bitsPerColor := destDepth == 16 ifTrue: [5] ifFalse: [8].
    rgbMask := (1 << bitsPerColor) - 1.
    maskShifted := destMask.
    destShifted := destinationWord.
    sourceShifted := sourceWord.
    result := destinationWord.
    destPPW == 1
	ifTrue:	"32bpp blends include alpha"
	   [(paintMode and: [sourceWord == 0])  "painting a transparent pixel"
		ifFalse:
		   [result := 0.
		    1 to: 4 do: [:i |
			shift := (i - 1) * 8.
			blend := (((sourceWord >> shift bitAnd: rgbMask) * sourceAlpha)
				      + ((destinationWord >> shift bitAnd: rgbMask) * unAlpha))
			    + 254 // 255 bitAnd: rgbMask.
			result := result bitOr: blend << shift]]]
	ifFalse:
	   [1 to: destPPW do: [:j |
	       sourcePixVal := sourceShifted bitAnd: pixMask.
	       ((maskShifted bitAnd: pixMask) == 0  "no effect if outside of dest rectangle"
			or: [paintMode and: [sourcePixVal == 0]])  "or painting a transparent pixel"
		   ifFalse:
		      [destPixVal := destShifted bitAnd: pixMask.
		       pixBlend := 0.
		       1 to: 3 do: [:i |
			   shift := (i - 1) * bitsPerColor.
			   blend := (((sourcePixVal >> shift bitAnd: rgbMask) * sourceAlpha)
					 + ((destPixVal >> shift bitAnd: rgbMask) * unAlpha))
			       + 254 // 255 bitAnd: rgbMask.
			   pixBlend := pixBlend bitOr: blend << shift].
		       result := destDepth == 16
			   ifTrue:  [(result bitAnd: (pixMask << (j - 1 * 16)) bitInvert32) bitOr: pixBlend << (j - 1 * 16)]
			   ifFalse: [pixBlend]].
	       maskShifted := maskShifted >> destDepth.
	       sourceShifted := sourceShifted >> destDepth.
	       destShifted := destShifted >> destDepth]].
    ^result
]

BitBltSimulation alphaPaintConst: sourceWord with: destinationWord
[
    sourceWord == 0 ifTrue: [^destinationWord  "opt for all-transparent source"].
    ^self alphaBlendConst: sourceWord with: destinationWord paintMode: true
]

BitBltSimulation rgbDiff: sourceWord with: destinationWord
[
    "Subract the pixels in the source and destination, color by color,
    and return the sum of the absolute value of all the differences.
    For non-rgb, return the number of differing pixels."
    | pixMask destShifted sourceShifted destPixVal bitsPerColor rgbMask sourcePixVal diff maskShifted |
    pixMask := maskTable at: destDepth.
    destDepth == 16
	ifTrue:  [bitsPerColor := 5.  rgbMask := 16r1F]
	ifFalse: [bitsPerColor := 8.  rgbMask := 16rFF].
    maskShifted := destMask.
    destShifted := destinationWord.
    sourceShifted := sourceWord.
    1 to: destPPW do: [:i |
	(maskShifted bitAnd: pixMask) > 0
	    ifTrue:		"Only tally pixels within the destination rectangle"
	       [destPixVal := destShifted bitAnd: pixMask.
		sourcePixVal := sourceShifted bitAnd: pixMask.
		destDepth < 16
		    ifTrue:  [sourcePixVal == destPixVal
				  ifTrue:  [diff := 0]
				  ifFalse: [diff := 1]]
		    ifFalse: [diff := (self partitionedSub: sourcePixVal from: destPixVal nBits: bitsPerColor nPartitions: 3).
			      diff := (diff bitAnd: rgbMask)
				  + (diff >> bitsPerColor bitAnd: rgbMask)
				  + ((diff >> bitsPerColor) >> bitsPerColor bitAnd: rgbMask)].
		bitCount := bitCount + diff].
	maskShifted := maskShifted >> destDepth.
	sourceShifted := sourceShifted >> destDepth.
	destShifted := destShifted >> destDepth].
    ^destinationWord  "For no effect on dest"
]

BitBltSimulation tallyIntoMap: sourceWord with: destinationWord
[
    "Tally pixels into the color map.  Those tallied are exactly those
    in the destination rectangle.  Note that the source should be
    specified == destination, in order for the proper color map checks
    to be performed at setup."
    | mapIndex pixMask destShifted maskShifted pixVal |
    (cmFlags bitAnd: (ColorMapPresent bitOr: ColorMapIndexedPart)) == (ColorMapPresent bitOr: ColorMapIndexedPart)
	ifFalse: [^destinationWord "no op"].
    pixMask := maskTable at: destDepth.
    destShifted := destinationWord.
    maskShifted := destMask.
    1 to: destPPW do: [:i |
	(maskShifted bitAnd: pixMask) == 0
	    ifFalse:	"Only tally pixels within the destination rectangle"
	       [pixVal := destShifted bitAnd: pixMask.
		destDepth < 16
		    ifTrue:  [mapIndex := pixVal]
		    ifFalse: [destDepth == 16
				  ifTrue:  [mapIndex := self rgbMap: pixVal from: 5 to: cmBitsPerColor]
				  ifFalse: [mapIndex := self rgbMap: pixVal from: 8 to: cmBitsPerColor]].
		self tallyMapAt: mapIndex put: (self tallyMapAt: mapIndex) + 1].
	maskShifted := maskShifted >> destDepth.
	destShifted := destShifted >> destDepth].
    ^destinationWord  "For no effect on dest"
]

BitBltSimulation alphaBlendScaled: sourceWord with: destinationWord
[
    "Blend sourceWord with destinationWord using the alpha value from
    sourceWord.  Alpha is encoded as 0 meaning 0.0, and 255 meaning
    1.0.  In contrast to alphaBlend:with: the color produced is
		srcColor + (1-srcAlpha) * dstColor
    i.e., it is assumed that the source color is already scaled."
    | unAlpha dstMask srcMask b g r a |
    unAlpha := 255 - (sourceWord >> 24).  "High 8 bits of source pixel"
    dstMask := destinationWord.
    srcMask := sourceWord.
    b := (dstMask bitAnd: 255) * unAlpha >> 8 + (srcMask bitAnd: 255).
    b > 255 ifTrue: [b := 255].
    dstMask := dstMask >> 8.
    srcMask := srcMask >> 8.
    g := (dstMask bitAnd: 255) * unAlpha >> 8 + (srcMask bitAnd: 255).
    g > 255 ifTrue: [g := 255].
    dstMask := dstMask >> 8.
    srcMask := srcMask >> 8.
    r := (dstMask bitAnd: 255) * unAlpha >> 8 + (srcMask bitAnd: 255).
    r > 255 ifTrue: [r := 255].
    dstMask := dstMask >> 8.
    srcMask := srcMask >> 8.
    a := (dstMask bitAnd: 255) * unAlpha >> 8 + (srcMask bitAnd: 255).
    a > 255 ifTrue: [a := 255].
    ^(((((a << 8) + r) << 8) + g) << 8) + b
]

BitBltSimulation rgbMul: sourceWord with: destinationWord
[
    destDepth < 16
	ifTrue:		"Mul each pixel separately"
	   [^self partitionedMul: sourceWord with: destinationWord nBits: destDepth nPartitions: destPPW].
    destDepth == 16
	ifTrue:		"Mul RGB components of each pixel separately"
	   [^(self partitionedMul: sourceWord with: destinationWord nBits: 5 nPartitions: 3)
		+ ((self partitionedMul: sourceWord >> 16 with: destinationWord >> 16 nBits: 5 nPartitions: 3) << 16)]
	ifFalse:	"Mul RGB components of the pixel separately"
	    [^self partitionedMul: sourceWord with: destinationWord nBits: 8 nPartitions: 3]
]

BitBltSimulation pixSwap: sourceWord with: destWord
[
    "Swap the pixels in destWord"
    | result shift lowMask highMask |
    destPPW == 1 ifTrue: [^destWord]. "a single pixel per word"
    result := 0.
    lowMask := (1 << destDepth) - 1. "mask low pixel"
    highMask := lowMask << (destPPW - 1 * destDepth). "mask high pixel"
    shift := 32 - destDepth.
    result := result bitOr: ((destWord bitAnd: lowMask) << shift bitOr: (destWord bitAnd: highMask) >> shift).
    destPPW <= 2 ifTrue: [^result].
    2 to: destPPW // 2 do: [:i |
	lowMask := lowMask << destDepth.
	highMask := highMask >> destDepth.
	shift := shift - (destDepth * 2).
	result := result bitOr: ((destWord bitAnd: lowMask) << shift bitOr: (destWord bitAnd: highMask) >> shift)].
    ^result
]

BitBltSimulation pixClear: sourceWord with: destinationWord
[
    "Clear all pixels in destinationWord for which the pixels of
    sourceWord have the same values. Used to clear areas of some
    constant color to zero."
    | mask result nBits |
    destDepth == 32 ifTrue: [^sourceWord == destinationWord ifTrue: [0] ifFalse: [destinationWord]].
    nBits := destDepth.
    mask := maskTable at: nBits.  "partition mask starts at the right"
    result := 0.
    1 to: destPPW do: [:i |
	| pv |
	pv := destinationWord bitAnd: mask.
	(sourceWord bitAnd: mask) == pv ifTrue: [pv := 0].
	result := result bitOr: pv.
	mask := mask << nBits "slide left to next partition"].
    ^result
]

BitBltSimulation fixAlpha: sourceWord with: destinationWord
[
    "For any non-zero pixel value in destinationWord with zero alpha
    channel take the alpha from sourceWord and fill it in. Intended
    for fixing alpha channels left at zero during 16->32 bpp
    conversions."
    destDepth == 32 ifFalse: [^destinationWord]. "no-op for non 32bpp"
    destinationWord == 0 ifTrue: [^0].
    (destinationWord bitAnd: 16rFF000000) == 0 ifFalse: [^destinationWord].
    ^destinationWord bitOr: (sourceWord bitAnd: 16rFF000000)
]

BitBltSimulation partitionedAnd: word1 to: word2 nBits: nBits nPartitions: nParts
[
    "AND word1 to word2 as nParts partitions of nBits each.  Any field
    of word1 not all-ones is treated as all-zeroes.  Used for erasing,
    eg, brush shapes prior to ORing in a color"
    | mask result |
    mask := maskTable at: nBits.  "partition mask starts at the right"
    result := 0.
    1 to: nParts do: [:i |
	(word1 bitAnd: mask) == mask ifTrue: [result := result bitOr: (word2 bitAnd: mask)].
	mask := mask << nBits  "slide left to next partition"].
    ^result
]

BitBltSimulation partitionedAdd: word1 to: word2 nBits: nBits nPartitions: nParts
[
    "Add word1 to word2 as nParts partitions of nBits each. This is
    useful for packed pixels, or packed colors"
    | mask result |
    mask := maskTable at: nBits.  "partition mask starts at the right"
    result := 0.
    1 to: nParts do: [:i |
	| sum |
	sum := (word1 bitAnd: mask) + (word2 bitAnd: mask).
	result := sum <= mask  "result must not carry out of partition"
	    ifTrue:  [result bitOr: sum]
	    ifFalse: [result bitOr: mask].
	mask := mask << nBits  "slide left to next partition"].
    ^result
]

BitBltSimulation partitionedMax: word1 with: word2 nBits: nBits nPartitions: nParts
[
    "Max word1 to word2 as nParts partitions of nBits each"
    | mask result |
    mask := maskTable at: nBits.  "partition mask starts at the right"
    result := 0.
    1 to: nParts do: [:i |
	result := result bitOr: ((word2 bitAnd: mask) max: (word1 bitAnd: mask)).
	mask := mask << nBits  "slide left to next partition"].
    ^result
]

BitBltSimulation partitionedMin: word1 with: word2 nBits: nBits nPartitions: nParts
[
    "Min word1 to word2 as nParts partitions of nBits each"
    | mask result |
    mask := maskTable at: nBits.  "partition mask starts at the right"
    result := 0.
    1 to: nParts do: [:i |
	result := result bitOr: ((word2 bitAnd: mask) min: (word1 bitAnd: mask)).
	mask := mask << nBits  "slide left to next partition"].
    ^result
]

BitBltSimulation partitionedMul: word1 with: word2 nBits: nBits nPartitions: nParts
[
    "Multiply word1 with word2 as nParts partitions of nBits each.
    This is useful for packed pixels, or packed colors.  Bug in loop
    version when non-white background"
    | sMask product result dMask |
    sMask := maskTable at: nBits.  "partition mask starts at the right"
    dMask := sMask << nBits.
    result := (((word1 bitAnd: sMask) + 1) * ((word2 bitAnd: sMask) + 1) - 1 bitAnd: dMask) >> nBits.	"optimized first step"
    product := (((word1 >> nBits bitAnd: sMask) + 1) * ((word2 >> nBits bitAnd: sMask) + 1) - 1 bitAnd: dMask).
    result := result bitOr: (product bitAnd: dMask).
    product := (((word1 >> (2 * nBits) bitAnd: sMask) + 1) * ((word2 >> (2 * nBits) bitAnd: sMask) + 1) - 1 bitAnd: dMask).
    result := result bitOr: (product bitAnd: dMask) << nBits.
    ^result
]

BitBltSimulation partitionedSub: word1 from: word2 nBits: nBits nPartitions: nParts
[
    "Subtract word1 from word2 as nParts partitions of nBits each.
    This is useful for packed pixels, or packed colors"
    | mask result |
    mask := maskTable at: nBits.  "partition mask starts at the right"
    result := 0.
    1 to: nParts do: [:i |
	| p1 p2 |
	p1 := word1 bitAnd: mask.
	p2 := word2 bitAnd: mask.
	result := p1 < p2  "result is really abs value of the difference"
	    ifTrue:  [result bitOr: p2 - p1]
	    ifFalse: [result bitOr: p1 - p2].
	mask := mask << nBits  "slide left to next partition"].
    ^result
]

BitBltSimulation pickSourcePixels: nPixels flags: mapperFlags srcMask: srcMask destMask: dstMask srcShiftInc: srcShiftInc dstShiftInc: dstShiftInc
[
    "Pick nPix pixels starting at srcBitIndex from the source, map by
    the color map, and justify them according to dstBitIndex in the
    resulting destWord."
    | sourceWord destWord srcShift dstShift nPix |
    sourceWord := self srcLongAt: sourceIndex.
    destWord := 0.
    srcShift := srcBitShift. "Hint: Keep in register"
    dstShift := dstBitShift. "Hint: Keep in register"
    nPix := nPixels. "always > 0 so we can use do { } while(--nPix);"
    (mapperFlags == (ColorMapPresent bitOr: ColorMapIndexedPart))
	ifTrue:
	   ["optimization for blits using indexed lookups only"
	    [	    | sourcePix destPix |
		    "grab, colormap and mix in pixel"
		    sourcePix := sourceWord >> srcShift bitAnd: srcMask.
		    destPix   := cmLookupTable indexedAt: 1 + (sourcePix bitAnd: cmMask).
		    destWord  := destWord bitOr: (destPix bitAnd: dstMask) << dstShift.
		    "adjust dest pix index"
		    dstShift  := dstShift + dstShiftInc.
		    "adjust source  pix index"
		    ((srcShift := srcShift + srcShiftInc) bitAnd: 16rFFFFFFE0) == 0
			ifFalse:
			    [srcShift := srcShift + (sourceMSB ifTrue: [32] ifFalse: [-32]).
			     sourceWord := self srcLongAt: (sourceIndex := sourceIndex + 4)].
		    (nPix := nPix - 1) == 0]
		whileFalse]
	ifFalse:
	   [[	    | sourcePix destPix |
		    "grab, colormap and mix in pixel"
		    sourcePix := sourceWord >> srcShift bitAnd: srcMask.
		    destPix   := self mapPixel: sourcePix flags: mapperFlags.
		    destWord  := destWord bitOr: (destPix bitAnd: dstMask) << dstShift.
		    "adjust dest pix index"
		    dstShift  := dstShift + dstShiftInc.
		    "adjust source pix index"
		    ((srcShift := srcShift + srcShiftInc) bitAnd: 16rFFFFFFE0) == 0
			ifFalse:
			    [srcShift := srcShift + (sourceMSB ifTrue:[32] ifFalse: [-32]).
			     sourceWord := self srcLongAt: (sourceIndex := sourceIndex + 4)].
		    (nPix := nPix - 1) == 0]
		whileFalse].
    srcBitShift := srcShift. "Store back"
    ^destWord
]

BitBltSimulation mapPixel: sourcePixel flags: mapperFlags
[
    "Color map the given source pixel."
    | pv |
    pv := sourcePixel.
    (mapperFlags bitAnd: ColorMapPresent) ~= 0
	ifTrue:
	   [(mapperFlags bitAnd: ColorMapFixedPart) ~= 0
		ifTrue:
		   [pv := self rgbMapPixel: sourcePixel flags: mapperFlags.
		    "avoid introducing transparency by color reduction"
		    (pv == 0 and: [sourcePixel ~= 0]) ifTrue: [pv := 1]].
	    (mapperFlags bitAnd: ColorMapIndexedPart) ~= 0
		ifTrue:
		   [pv := cmLookupTable indexedAt: 1 + (pv bitAnd: cmMask)]].
    ^pv
]

BitBltSimulation rgbMapPixel: sourcePixel flags: mapperFlags
[
    "Perform the RGBA conversion for the given source pixel"
    | val |
    val :=            ((sourcePixel bitAnd: (cmMaskTable at: RedIndex  )) bitShift: (cmShiftTable at: RedIndex  )).
    val := val bitOr: ((sourcePixel bitAnd: (cmMaskTable at: GreenIndex)) bitShift: (cmShiftTable at: GreenIndex)).
    val := val bitOr: ((sourcePixel bitAnd: (cmMaskTable at: BlueIndex )) bitShift: (cmShiftTable at: BlueIndex )).
    ^      val bitOr: ((sourcePixel bitAnd: (cmMaskTable at: AlphaIndex)) bitShift: (cmShiftTable at: AlphaIndex)).
]

BitBltSimulation showDisplayBits
[
    interpreterProxy 
	showDisplayBits: destForm
	left: affectedL
	top: affectedT
	right: affectedR
	bottom: affectedB
]

BitBltSimulation expensiveDither32To16: srcWord threshold: ditherValue
[
    "Dither the given 32bit word to 16 bit. Ignore alpha."
    | pv threshold value out |
    pv := srcWord bitAnd: 255.
    threshold := ditherThresholds16 at: 1 + (pv bitAnd: 7).
    value := ditherValues16 at: 1 + (pv bitShift: -3).
    out := ditherValue < threshold
	ifTrue:  [value + 1]
	ifFalse: [value].
    pv := (srcWord bitShift: -8) bitAnd: 255.
    threshold := ditherThresholds16 at: 1 + (pv bitAnd: 7).
    value := ditherValues16 at: 1 + (pv bitShift: -3).
    out := out bitOr: (ditherValue < threshold
			   ifTrue:  [value + 1 bitShift: 5]
			   ifFalse: [value bitShift: 5]).
    pv := (srcWord bitShift: -16) bitAnd: 255.
    threshold := ditherThresholds16 at: 1 + (pv bitAnd: 7).
    value := ditherValues16 at: 1 + (pv bitShift: -3).
    out := out bitOr: (ditherValue < threshold
			   ifTrue:  [value + 1 bitShift: 10]
			   ifFalse: [value bitShift: 10]).
    ^out
]

BitBltSimulation initialiseModule
[
    maskTable := Array new: 32.
    #(1 2 4 5 8 16 32) do: [:i | maskTable at: i put: (1 << i) - 1].
maskTable println.
    ditherMatrix4x4 :=	#(	0	8	2	10
				12	4	14	6
				3	11	1	9
				15	7	13	5 ).
    ditherThresholds16 := #(0 2 4 6 8 10 12 14 16).
    ditherValues16 :=	#(0 0 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 16 17 18 19 20 21 22 23 24 25 26 27 28 29 30).
    dither8Lookup := (Array new: 4096).
    0 to: 255 do: [ :b | 
	0 to: 15 do: [ :t |
	    | value |
	    value := self expensiveDither32To16: b threshold: t.
	    dither8Lookup at: 1 + ((t << 8) + b) put: value]].
    warpBitShiftTable := Array new: 32.
    ^true
]

[
    (BitBltSimulation := BitBltSimulation new) initialiseModule
]
