{ import: st80 }

{ import: SqueakObject }
{ import: SqueakImage }

Object printsp
[
    self print.
    ' ' put.
]


MaxPrimitiveIndex	:= [ 700 ]

MethodCacheSize		:= [ 16r1000 ]
MethodCacheMask		:= [ MethodCacheSize - 4 ]
MethodCacheSelector	:= [ 1 ]
MethodCacheClass	:= [ 2 ]
MethodCacheMethod	:= [ 3 ]
MethodCachePrim		:= [ 4 ]


SqueakVM : Object ( image nilObj trueObj falseObj lastHash
		    bytecodeTable primitiveTable activeContext
		    homeContext receiver method instructionPointer
		    stackPointer currentBytecode messageSelector
		    argumentCount lkupClass receiverClass methodClass
		    newMethod primitiveIndex successFlag
		    nextWakeupTick deferDisplayUpdates
		    interruptCheckCounter filesOpen methodCache )

SqueakVM open: fileName
[
    self := self new.
    image := SqueakImage open: fileName.
    self initializeBytecodeTable.
    self initializePrimitiveTable.
    self initializeInterpreter.
    self initializePrimitives.
    self initializePlatform.
    self interpret.
]

SqueakVM initializeBytecodeTable
[
    bytecodeTable := Array new: 256.
    self table: bytecodeTable from: self bytecodeTable
]

SqueakVM initializePrimitiveTable
[
    primitiveTable := Array new: 1 + MaxPrimitiveIndex.
    self table: primitiveTable from: self primitiveTable
]

SqueakVM table: anArray from: specArray
[
    "SpecArray is an array of either (index selector) or (index1 index2 selector)."
    | contiguous |
    contiguous := 0.
    specArray do: [ :spec |
        (spec at: 1) == contiguous ifFalse: [self error: 'Non-contiguous table entry'].
        spec size = 2
	    ifTrue:
	       [anArray at: ((spec at: 1) + 1) put: (SelectorTable at: (spec at: 2)).
		contiguous := contiguous + 1]
	    ifFalse:
	       [(spec at: 1) to: (spec at: 2) do: [:i | anArray at: (i + 1) put: (SelectorTable at: (spec at: 3))].
		contiguous := contiguous + ((spec at: 2) - (spec at: 1)) + 1]].
]

SqueakVM initializeInterpreter
[
    nilObj := image nilObject.
    trueObj := image trueObject.
    falseObj := image falseObject.
    lastHash := image lastHash.
"
    interpreterProxy := self sqGetInterpreterProxy.
    self dummyReferToProxy.
    self initializeObjectMemory: bytesToShift.
    self initCompilerHooks.
    self flushExternalPrimitives.
"
    activeContext			:= nilObj.
    homeContext				:= nilObj.
    method				:= nilObj.
    receiver				:= nilObj.
    messageSelector			:= nilObj.
    newMethod				:= nilObj.
    methodClass				:= nilObj.
    lkupClass				:= nilObj.
    receiverClass			:= nilObj.
    methodCache				:= Array new: MethodCacheSize.
    self flushMethodCache.
    self loadInitialContext.
    interruptCheckCounter		:= 0.
"
    interruptCheckCounterFeedBackReset	:= 1000.
    interruptChecksEveryNms		:= 3.
    nextPollTick			:= 0.
"
    nextWakeupTick			:= 0.
"
    lastTick				:= 0.
    interruptKeycode			:= 2094.  ""cmd-.""
    interruptPending			:= false.
    semaphoresUseBufferA		:= true.
    semaphoresToSignalCountA		:= 0.
    semaphoresToSignalCountB		:= 0.
"
    deferDisplayUpdates			:= false.
"
    pendingFinalizationSignals		:= 0.
"
]

SqueakVM loadInitialContext
[
    activeContext := image schedulerAssociation atValueIndex atActiveProcessIndex atSuspendedContextIndex.
    self fetchContextRegisters: activeContext.
]

SqueakVM fetchContextRegisters: activeCntx
[
    | tmp |
    homeContext := (activeCntx atMethodIndex isSmallInteger)
	ifTrue:  [activeCntx atHomeIndex]	"if the MethodIndex field is an integer, activeCntx is a block context"
	ifFalse: [activeCntx].			"otherwise, it is a method context and is its own home context"
    receiver := homeContext atReceiverIndex.
    method := homeContext atMethodIndex.
    instructionPointer := activeCntx atInstructionPointerIndex.
    stackPointer := activeCntx atStackPointerIndex.
"
    homeContext printsp.
    receiver printsp.
    activeCntx printsp.
    stackPointer printsp.
    method printsp.
    instructionPointer println.
"
]

SqueakVM storeContextRegisters: activeCntx
[
    activeCntx
	atInstructionPointerIndexPut: instructionPointer;
	atStackPointerIndexPut: stackPointer
]

SqueakVM fetchByte
[
    | byte |
    byte := method indexedAt: instructionPointer.
    instructionPointer := instructionPointer + 1.
    ^byte
]

SqueakVM fetchClassOf: anObject
[
    ^anObject isSmallInteger
	ifTrue:  [image classInteger]
	ifFalse: [anObject fetchClass]
]

SqueakVM jump: offset
[
    instructionPointer := instructionPointer + offset
]

SqueakVM jumpIfTrueBy: offset
[
    | boolean |
    (boolean := self stackTop) == trueObj
	ifTrue: [instructionPointer := instructionPointer + offset]
	ifFalse:
	   [boolean == falseObj
		ifFalse:
		   [messageSelector := image selectorMustBeBoolean.
		    argumentCount := 0.
		    ^self normalSend]].
    self pop.
]

SqueakVM jumpIfFalseBy: offset
[
    | boolean |
    (boolean := self stackTop) == falseObj
	ifTrue: [instructionPointer := instructionPointer + offset]
	ifFalse:
	   [boolean == trueObj
		ifFalse:
		   [messageSelector := image selectorMustBeBoolean.
		    argumentCount := 0.
		    ^self normalSend]].
    self pop.
]

SqueakVM stackTop		[ ^activeContext indexedAt: stackPointer ]
SqueakVM stackValue: n		[ ^activeContext indexedAt: stackPointer - n ]
SqueakVM pop			[  stackPointer := stackPointer - 1 ]
SqueakVM pop: count		[  stackPointer := stackPointer - count ]
SqueakVM popThenPush: anObject	[ ^activeContext indexedAt: stackPointer put: anObject ]
SqueakVM pop2thenPush: anObject	[ ^activeContext indexedAt: (stackPointer := stackPointer - 1) put: anObject ]
SqueakVM pop3thenPush: anObject	[ ^activeContext indexedAt: (stackPointer := stackPointer - 2) put: anObject ]
SqueakVM pop: n thenPush: o	[ ^activeContext indexedAt: (stackPointer := stackPointer - (n - 1)) put: o ]
SqueakVM push: anObject		[ ^activeContext indexedAt: (stackPointer := stackPointer + 1) put: anObject ]
SqueakVM pushBoolean: flag	[ ^self push: (flag ifTrue: [trueObj] ifFalse: [falseObj]) ]
SqueakVM unPop: count		[  stackPointer := stackPointer + count ]

SqueakVM stackPop
[
    | top |
    top := activeContext indexedAt: stackPointer.
    stackPointer := stackPointer - 1.
    ^top
]

SqueakVM popFloatOrInt
[
    | top |
    top := self stackPop.
    top isSmallInteger ifTrue: [^top].
    top fetchClass == image classFloat ifTrue: [^top _elements].
    successFlag := false.
    ^0
]

SqueakVM _popInteger
[
    | top |
    top := self stackPop.
    ^top isSmallInteger
	ifTrue:  [top _integerValue]
	ifFalse: [successFlag := false]
]

SqueakVM _popPos32BitInteger
[
    | top |
    top := self stackPop.
    ^top isSmallInteger
	ifTrue:  [top _integerValue]
	ifFalse:
	    [(top fetchClass == image classLargePositiveInteger and: [top indexableSize == 4])
		 ifTrue:  [top _largeIntegerValue]
		 ifFalse: [successFlag := false]]
]

SqueakVM signExtend16: int16
[
    "Convert a signed 16-bit integer into a signed 32-bit integer value."
    ^(int16 bitAnd: 16r8000) == 0
	ifTrue:  [int16]
	ifFalse: [int16 - 16r10000]
]

SqueakVM specialSelector: index
[
    ^image specialSelectors indexedAt: 1 + (index * 2)
]

SqueakVM specialSend: selectorIndex argumentCount: nArgs
[
    messageSelector := self specialSelector: selectorIndex.
    argumentCount := nArgs.
    self normalSend.
]

SqueakVM normalSend
[
    "Send a message, starting lookup with the receiver's class.
    Assume messageSelector and argumentCount have been set, and that
    the receiver and arguments have been pushed onto the stack."
    | rcvr |
    "('send ', messageSelector stringContents, ' with ', argumentCount printString, ' args') putln."
    rcvr := self stackValue: argumentCount.
    lkupClass := self fetchClassOf: rcvr.
    receiverClass := lkupClass.
    self commonSend.
]

SqueakVM superclassSend
[
    "Send a message to self, starting lookup with the superclass of
    the class containing the currently executing method. Assume:
    messageSelector and argumentCount have been set, and that the
    receiver and arguments have been pushed onto the stack."
    | rcvr |
    "('super ', messageSelector stringContents, ' with ', argumentCount printString, ' args') putln."
    lkupClass := method methodClass atSuperclassIndex.
    rcvr := self stackValue: argumentCount.
    receiverClass := self fetchClassOf: rcvr.
    self commonSend.
]

SqueakVM commonSend
[
    self findNewMethod; executeNewMethod.
]

SqueakVM findNewMethod
[
    "Find the compiled method to be run when the current
    messageSelector is sent to the class 'lkupClass', setting the
    values of 'newMethod' and 'primitiveIndex'."
    (self lookupInMethodCacheSel: messageSelector class: lkupClass)
	ifFalse:
	   [self lookupMethodInClass: lkupClass.
	    self addNewMethodToCache]
]

SqueakVM findNewMethodInClass: class
[
    "Find the compiled method to be run when the current
    messageSelector is sent to the given class, setting the values of
    'newMethod' and 'primitiveIndex'."
    (self lookupInMethodCacheSel: messageSelector class: class)
	ifFalse:
	   [self lookupMethodInClass: class.
	    lkupClass := class.
	    self addNewMethodToCache]
]

SqueakVM lookupInMethodCacheSel: selector class: class
[
    | hash probe |
    hash := selector identityHash bitXor: class identityHash.
    probe := hash bitAnd: MethodCacheMask.		"first probe"
    (((methodCache at: probe + MethodCacheSelector) == selector) and: [(methodCache at: probe + MethodCacheClass) == class])
	ifTrue:
	   [newMethod := methodCache at: probe + MethodCacheMethod.
	    primitiveIndex := methodCache at: probe + MethodCachePrim.
	    ^true].
    probe := (hash >> 1) bitAnd: MethodCacheMask.	"second probe"
    (((methodCache at: probe + MethodCacheSelector) == selector) and: [(methodCache at: probe + MethodCacheClass) == class])
	ifTrue:
	   [newMethod := methodCache at: probe + MethodCacheMethod.
	    primitiveIndex := methodCache at: probe + MethodCachePrim.
	    ^true].
    probe := (hash >> 2) bitAnd: MethodCacheMask.	"third probe"
    (((methodCache at: probe + MethodCacheSelector) == selector) and: [(methodCache at: probe + MethodCacheClass) == class])
	ifTrue:
	   [newMethod := methodCache at: probe + MethodCacheMethod.
	    primitiveIndex := methodCache at: probe + MethodCachePrim.
	    ^true].
    ^false
]

SqueakVM addNewMethodToCache
[
    | probe hash |
    hash := messageSelector identityHash bitXor: lkupClass identityHash.  "drop low-order zeros from addresses"
    0 to: 2 do: [:p |
	probe := (hash >> p) bitAnd: MethodCacheMask.
	(methodCache at: probe + MethodCacheSelector) isNil
	    ifTrue: [methodCache
			 at: probe + MethodCacheSelector put: messageSelector;
			 at: probe + MethodCacheClass	 put: lkupClass;
			 at: probe + MethodCacheMethod	 put: newMethod;
			 at: probe + MethodCachePrim	 put: primitiveIndex.
		     ^nil]].
    probe := hash bitAnd: MethodCacheMask.  "first probe"
    methodCache
	at: probe + MethodCacheSelector	put: messageSelector;
	at: probe + MethodCacheClass	put: lkupClass;
	at: probe + MethodCacheMethod	put: newMethod;
	at: probe + MethodCachePrim	put: primitiveIndex.
    "zap the following entries"
    1 to: 2 do: [:p |
	probe := (hash >> p) bitAnd: MethodCacheMask.
	methodCache at: probe + MethodCacheSelector put: nil].
]

SqueakVM flushMethodCache
[
    methodCache atAllPut: nil
]

SqueakVM lookupMethodInClass: class
[
    | currentClass dictionary found |
    currentClass := class.
    [currentClass ~~ nilObj]
	whileTrue:
	   [dictionary := currentClass atMessageDictionaryIndex.
	    dictionary == nilObj
		ifTrue:
		    ["MethodDict pointer is nil (hopefully due a swapped out stub) -- raise exception #cannotInterpret:."
		     self createActualMessageTo: class.
		     messageSelector := image selectorCannotInterpret.
		     ^self lookupMethodInClass: currentClass atSuperclassIndex].
	    found := self lookupMethodInDictionary: dictionary.
	    found ifTrue: [^methodClass := currentClass].
	    currentClass := currentClass atSuperclassIndex].
    (class nameOfClass, ' doesNotUnderstand ', messageSelector stringContents) putln.
    self printStack: activeContext.
    "Could not find #doesNotUnderstand: -- unrecoverable error."
    messageSelector == image selectorDoesNotUnderstand
	ifTrue: [self error: 'Recursive doesNotUnderstand error encountered'].
    "Cound not find a normal message -- raise exception #doesNotUnderstand:"
    self createActualMessageTo: class.
    messageSelector := image selectorDoesNotUnderstand.
    ^self lookupMethodInClass: class
]

SqueakVM lookupMethodInDictionary: dictionary
[
    "This method lookup tolerates integers as Dictionary keys to support
    execution of images in which Symbols have been compacted out"
    | length index mask wrapAround nextSelector methodArray |
    length := dictionary indexableSize.
    mask := length - 1.
    index := messageSelector isSmallInteger
	ifTrue:  [mask bitAnd: messageSelector]
	ifFalse: [mask bitAnd: messageSelector hashBits].
    "It is assumed that there are some nils in this dictionary, and search will
    stop when one is encountered.  However, if there are no nils, then wrapAround
    will be detected the second time the loop gets to the end of the table."
    index := index + 1.
    wrapAround := false.
    [true]
	whileTrue:
	   [nextSelector := dictionary indexedAt: index.
	    nextSelector == nilObj ifTrue: [^false].
	    nextSelector == messageSelector
		ifTrue:
		   [methodArray := dictionary atMethodArrayIndex.
		    newMethod := methodArray indexedAt: index.
		    "Check if newMethod is a CompiledMethod."
		    newMethod isSqueakMethod
			ifTrue:
			   [primitiveIndex := newMethod primitiveIndex.
			    "If primitiveIndex is out of range, set to zero before putting in cache.
    		            This is equiv to primFail, and avoids the need to check on every send."
			    primitiveIndex > MaxPrimitiveIndex ifTrue: [primitiveIndex := 0]]
			ifFalse:
			   ["indicate that this is not a CompiledMethod"
			    primitiveIndex := 248].
		    ^true].
	    (index := index + 1) > length
		ifTrue:
		   [wrapAround ifTrue: [^false].
		    wrapAround := true.
		    index := 1]].
]

SqueakVM executeNewMethod
[
    primitiveIndex > 0
	ifTrue:
	   [self primitiveResponse.
	    successFlag ifTrue: [^nil].
	    ('fail primitive ', primitiveIndex printString, ' ', messageSelector stringContents) putln.
	   ].
    self activateNewMethod.
    self quickCheckForInterrupts.
]

SqueakVM primitiveResponse
[
    | timerPending startTime delta primIdx nArgs |
    timerPending := nextWakeupTick ~~ 0.
    timerPending ifTrue: [startTime := self ioLowResMSecs].
    successFlag := true.
    self _perform: (primitiveTable at: 1 + primitiveIndex).
    (timerPending
	    and: [self ioLowResMSecs ~= startTime
	    and: [self ioMSecs >= nextWakeupTick]])
	ifTrue:
	   [successFlag
		ifTrue:  [self checkForInterrupts]
		ifFalse: [interruptCheckCounter := 0]].
    ^successFlag
]

SqueakVM activateNewMethod
[
    | newContext initialIP tempCount nilOop |
    (newContext := self allocateContext: newMethod needsLargeContext)
	atSenderIndexPut: activeContext;
	atInstructionPointerIndexPut: newMethod initialIP;
	atStackPointerIndexPut: newMethod numTemps;
	atMethodIndexPut: newMethod;
	atReceiverMapIndexPut: nilObj;
	atReceiverIndexPut: (self stackValue: argumentCount).
    "Copy the reciever and arguments..."
    1 to: argumentCount do: [:i | newContext indexedAt: i put: (self stackValue: argumentCount - i)].
    self pop: 1 + argumentCount.
    self newActiveContext: newContext.
]

SqueakVM newActiveContext: aContext
[
    self
	storeContextRegisters: activeContext;
	fetchContextRegisters: (activeContext := aContext)
]

SqueakVM allocateContext: largeFlag
[
    | cntxt |
    (cntxt := SqueakObject withClass: image classMethodContext hash: self newObjectHash)
	initFixed: 6;
	initPointers: (largeFlag ifTrue: [56] ifFalse: [16]) with: nilObj.
    ^cntxt
]

SqueakVM instantiateContext: class indexableSize: size
[
    | cntxt |
    (cntxt := SqueakObject withClass: class hash: self newObjectHash)
	initFixed: 6;
	initPointers: size with: nilObj.
    ^cntxt
]

SqueakVM instantiateClass: class indexableSize: size
[
    | object format fixedSize |
    object := SqueakObject withClass: class hash: self newObjectHash.
    format := class atInstanceSpecIndex >> 7 bitAnd: 16rF.
    "('instantiate format ', format printString) putln."
    format >= 8 ifTrue: [^object initBytes: size ].
    format == 6 ifTrue: [^object initWords: size ].
    format == 0	ifTrue: [^object].
    format == 1	ifTrue: [^object initFixed: class fixedSizeOfClass with: nilObj].
    format == 2	ifTrue: [^object initPointers: size with: nilObj].
    object initFixed: class fixedSizeOfClass with: nilObj.
    format == 3	ifTrue: [^object initPointers: size with: nilObj].
    format == 4	ifTrue: [^object initWeakPointers: size with: nilObj].
    ^object
]

SqueakVM positive32BitIntegerFor_: _word
[
    | result |
    {
      int word= (int)v__word;
      if (word >= 0 && word <= 0x3FFFFFFF)
        return (oop)(word << 1 | 1);
    }.
    result := self instantiateClass: image classLargePositiveInteger indexableSize: 4.
    result _largeIntegerValue_: _word.
    ^result
]

SqueakVM isFloat: obj
[
    ^(self fetchClassOf: obj) == image classFloat
]

SqueakVM newFloat
[
    ^self instantiateClass: image classFloat indexableSize: 2 "words"
]

SqueakVM transfer: count fromIndex: fromIndex ofObject: source toIndex: toIndex ofObject: destination
[
    count
	timesRepeat:
	   [destination indexedAt: toIndex put: (source indexedAt: fromIndex).
	    fromIndex := fromIndex + 1.
	    toIndex := toIndex + 1]
]

SqueakVM schedulerPointer
[
    ^image schedulerAssociation atValueIndex
]

SqueakVM synchronousSignal: aSemaphore
[
    (self isEmptyList: aSemaphore)		"no process is waiting on this Semaphore"
	ifTrue:  [aSemaphore atExcessSignalsIndexPut: aSemaphore atExcessSignalsIndex + 1]
	ifFalse: [self resume: (self removeFirstLinkOfList: aSemaphore)]
]

SqueakVM resume: aProcess
[
    | activeProc activePriority newPriority |
    activeProc := self schedulerPointer atActiveProcessIndex.
    activePriority := activeProc atPriorityIndex.
    newPriority := aProcess atPriorityIndex.
    newPriority > activePriority
	ifTrue:  [self putToSleep: activeProc; transferTo: aProcess]
	ifFalse: [self putToSleep: aProcess].
]

SqueakVM putToSleep: aProcess
[
    | priority processLists processList |
    priority := aProcess atPriorityIndex.
    processLists := self schedulerPointer atProcessListsIndex.
    processList := processLists indexedAt: priority.
    self addLastLink: aProcess toList: processList
]

SqueakVM addLastLink: proc toList: aList
[
    (self isEmptyList: aList)
	ifTrue:  [aList atFirstLinkIndexPut: proc]
	ifFalse:
	   [| lastLink |
	    lastLink := aList atLastLinkIndex.
	    lastLink atNextLinkIndexPut: proc].
    aList atLastLinkIndexPut: proc.
    proc atMyListIndexPut: aList.
]

SqueakVM removeFirstLinkOfList: aList
[
    | first |
    (first := aList atFirstLinkIndex) == aList atLastLinkIndex
	ifTrue:  [aList atFirstLinkIndexPut: nilObj; atLastLinkIndexPut: nilObj]
	ifFalse: [aList atFirstLinkIndexPut: first atNextLinkIndex].
    first atNextLinkIndexPut: nilObj.
    ^first
]

SqueakVM isEmptyList: aList
[
    ^aList atFirstLinkIndex == nilObj
]

SqueakVM transferTo: aProc
[
    | sched oldProc newProc |
    newProc := aProc.
    sched := self schedulerPointer.
    oldProc := sched atActiveProcessIndex.
    sched atActiveProcessIndexPut: newProc.
    oldProc atSuspendedContextIndexPut: activeContext.
    self newActiveContext: newProc atSuspendedContextIndex.
    newProc atSuspendedContextIndexPut: nilObj.
]

SqueakVM wakeHighestPriority
[
    | schedLists |
    schedLists := self schedulerPointer atProcessListsIndex.
    schedLists indexableSize downTo: 1 do: [:i |
	| processList |
	(self isEmptyList: (processList := schedLists indexedAt: i))
	    ifFalse: [^self removeFirstLinkOfList: processList]].
    self error: 'scheduler could not find a runnable process'
]

SqueakVM newObjectHash	[ ^lastHash := 13849 + (27181 * lastHash) bitAnd: 65535 ]

SqueakVM quickCheckForInterrupts
[
    (interruptCheckCounter := interruptCheckCounter - 1) <= 0
	ifTrue: [self checkForInterrupts].
]

SqueakVM checkForInterrupts
[
    "$. put."
]

SqueakVM return: returnValue to: returnContext
[
    | nilOop thisCntx localCntx localVal |
    nilOop := nilObj. "keep in a register"
    thisCntx := activeContext.
    localCntx := returnContext.
    localVal := returnValue.
    "make sure we can return to the given context"
    ((localCntx == nilOop) or: [localCntx atInstructionPointerIndex == nilOop])
	ifTrue: [^self cannotReturn: localVal].
    thisCntx := activeContext atSenderIndex.
    [thisCntx == localCntx]
	whileFalse:
	   [thisCntx == nilOop ifTrue: [^self cannotReturn: localVal].
	    (self isUnwindMarked: thisCntx) ifTrue: [^self aboutToReturn: localVal through: thisCntx].
	    thisCntx := thisCntx atSenderIndex].
    thisCntx := activeContext.
    [thisCntx == localCntx]
	whileFalse:
	   [| contextOfCaller |
	    contextOfCaller := thisCntx atSenderIndex.
	    thisCntx
		atSenderIndexPut: nilOop;
		atInstructionPointerIndexPut: nilOop.
	    thisCntx := contextOfCaller].
    self
	fetchContextRegisters: (activeContext := thisCntx);
	push: localVal.
]

SqueakVM isUnwindMarked: aContext
[
    ^aContext fetchClass == image classMethodContext and: [aContext atMethodIndex primitiveIndex == 198]
]

SqueakVM makePointwithxValue: x yValue: y
[
    | pointResult |
    (pointResult := SqueakObject withClass: image classPoint hash: self newObjectHash)
	initFixed: 2;
	instVarAt: 1 put: x;
	instVarAt: 2 put: y.
    ^pointResult
]


_object println	[ self print.  $\n put ]
_selector print	{ printf("%s", (char *)self->v__elements); }


SqueakVM printContext: context
[
    | home classAndSelector |
    home := (self fetchClassOf: context) == image classBlockContext
	ifTrue:  [context atHomeIndex]
	ifFalse: [context].
    classAndSelector := self findClassAndSelectorOfMethod: home atMethodIndex forReceiver: home atReceiverIndex.
    home atReceiverIndex print.  $\t put.
    context == home ifFalse: ['[] in ' put].
    classAndSelector first nameOfClass put.
    $> put.
    classAndSelector second isNil
	ifTrue:  [$? put]
	ifFalse: [classAndSelector second stringContents put].
]

SqueakVM printStack: context
[
    [context == nilObj]
	whileFalse:
	    [self printContext: context.
	     $\n put.
	     context := context atSenderIndex].
]

SqueakVM findClassAndSelectorOfMethod: meth forReceiver: rcvr
[
    | currClass |
    currClass := self fetchClassOf: rcvr.
    [currClass == nilObj]
	whileFalse:
	   [| classDict classDictSize methodArray |
	    classDict := currClass atMessageDictionaryIndex.
	    classDictSize := classDict indexableSize.
	    methodArray := classDict atMethodArrayIndex.
	    1 to: classDictSize do: [:i |
		meth == (methodArray indexedAt: i)
		    ifTrue: [^Array with: currClass with: (classDict indexedAt: i)]].
	    currClass := currClass atSuperclassIndex].
    ^Array with: (self fetchClassOf: rcvr) with: nil    "method not found in superclass chain"
]

SqueakVM image			[ ^image ]
SqueakVM methodArgumentCount	[ ^argumentCount ]
SqueakVM nilObject		[ ^nilObj ]
SqueakVM failed			[ ^successFlag not ]

SqueakVM showDisplayBits: aForm left: l top: t right: r bottom: b
[
    "Repaint the portion of the Smalltalk screen bounded by the
    affected rectangle. Used to synchronize the screen after a Bitblt
    to the Smalltalk Display object."
    deferDisplayUpdates ifTrue: [^nil].
    self displayBitsOf: aForm Left: l Top: t Right: r Bottom: b
]

SqueakVM displayBitsOf: aForm Left: l Top: t Right: r Bottom: b
[
    "Repaint the portion of the Smalltalk screen bounded by the
    affected rectangle. Used to synchronize the screen after a Bitblt
    to the Smalltalk Display object."
    | displayObj dispBits w h dispBitsIndex d left right top bottom surfaceHandle |
    displayObj := image theDisplay.
    aForm == displayObj ifFalse: [^nil].
    (displayObj isPointers and: [displayObj fixedSize >= 4]) ifFalse: [^successFlag := false].
    dispBits := displayObj atFormBitsIndex.
    w := displayObj atFormWidthIndex.
    h := displayObj atFormHeightIndex.
    d := displayObj atFormDepthIndex.
    left   := l < 0 ifTrue: [0] ifFalse: [l].
    right  := r > w ifTrue: [w] ifFalse: [r].
    top    := t < 0 ifTrue: [0] ifFalse: [t].
    bottom := b > h ifTrue: [h] ifFalse: [b].
    ((left <= right) and: [top <= bottom]) ifFalse: [^nil].
    dispBits isSmallInteger ifTrue: [self error: 'surfaces not supported'].
    dispBitsIndex := dispBits _words.  "index in memory byte array"
    self ioShowDisplay_: dispBitsIndex w: w h: h d: d left: left right: right top: top bottom: bottom
]

SqueakVM trace: bytecode
[
    self printContext: activeContext.  $\t put.
    instructionPointer print.  $\t put.  bytecode print.  $\t put.
    $[ put.  stackPointer print.  $  put.  stackPointer == 0 ifTrue: [$- put] ifFalse: [self stackTop print].  ']\t' put.
    (bytecodeTable at: 1 + bytecode) println.
]

SqueakVM interpret
[
    | count |
    count := 0.
    [true]
	whileTrue:
	   [currentBytecode := method indexedAt: instructionPointer.
	    "count \\ 1 == 0 ifTrue: [self trace: currentBytecode]."
	    count := count + 1.
	    instructionPointer := instructionPointer + 1.
	    self _perform: (bytecodeTable at: 1 + currentBytecode)]
]


{ import: SqueakVM-bytecodes }
{ import: SqueakVM-primitives }
{ import: SqueakVM-io }
