{ import: st80 }

{ include "surface-imp.h" }
{ include "surface-imp.c" }

Window : Object ( _window )
Pixmap : Object ( _pixmap )

Event  : Object ( x y )

Event x		[ ^ x ]
Event y		[ ^ y ]

Event withEvent_: _evt		[ ^self new ]

KeyEvent : Event ( key )

KeyEvent key	[ ^key ]

KeyEvent withEvent_: _evt
[
    self := super withEvent_: _evt.
    {
      CWindowKeyEvent *evt= (CWindowKeyEvent *)v__evt;
      self->v_x=   (oop)((evt->x  ) << 1 | 1);
      self->v_y=   (oop)((evt->y  ) << 1 | 1);
      self->v_key= (oop)((evt->key) << 1 | 1);
    }
]

KeyPressEvent   : KeyEvent ()

Event isKeyPress		[ ^false ]
KeyPressEvent isKeyPress	[ ^true ]

KeyReleaseEvent : KeyEvent ()

Event isKeyRelease		[ ^false ]
KeyPressEvent isKeyRelease	[ ^true ]

ButtonEvent : Event ( button )

ButtonEvent button	[ ^button ]

ButtonEvent withEvent_: _evt
[
    self := super withEvent_: _evt.
    {
      CWindowButtonEvent *evt= (CWindowButtonEvent *)v__evt;
      self->v_x=      (oop)((evt->x     ) << 1 | 1);
      self->v_y=      (oop)((evt->y     ) << 1 | 1);
      self->v_button= (oop)((evt->button) << 1 | 1);
    }
]

ButtonPressEvent   : ButtonEvent ()

Event isButtonPress		[ ^false ]
ButtonPressEvent isButtonPress	[ ^true ]

ButtonReleaseEvent : ButtonEvent ()

Event isButtonRelease			[ ^false ]
ButtonReleaseEvent isButtonRelease	[ ^true ]

MotionEvent : Event ()

Event isMotion		[ ^false ]
MotionEvent isMotion	[ ^true ]

MotionEvent withEvent_: _evt
[
    self := super withEvent_: _evt.
    {
      CWindowMotionEvent *evt= (CWindowMotionEvent *)v__evt;
      self->v_x=   (oop)((evt->x) << 1 | 1);
      self->v_y=   (oop)((evt->y) << 1 | 1);
    }
]

EventMap := [
    (Array new: 5)
	at: 1 put: KeyPressEvent;
	at: 2 put: KeyReleaseEvent;
	at: 3 put: ButtonPressEvent;
	at: 4 put: ButtonReleaseEvent;
	at: 5 put: MotionEvent;
	yourself
]

Window _window	 [ ^_window ]

Window withWidth: w height: h
[
    self := self new.
    _window := self _newWidth_: w _integerValue height_: h _integerValue.
]

Window _newWidth_: _w height_: _h
{
    return (oop)newCWindow((int)v__w, (int)v__h);
}

Window close
{
    deleteCWindow((CWindow *)self->v__window);
}

Window _nextEvent: block
{
    return (oop)nextCWindowEvent((CWindow *)self->v__window, (int)v_block);
}

Window nextEvent: block
[
    | _evt type |
    _evt := self _nextEvent: block.
    nil == _evt ifTrue: [^nil].
    type := Event typeOf_: _evt.
    ^(EventMap at: type) withEvent_: _evt
]
    
Pixmap withWindow: window width: w height: h depth: d
[
    self := self new.
    _pixmap := self _newWindow_: window _window w_: w _integerValue h_: h _integerValue d_: d _integerValue.
    { printf("new Pixmap %p -> %p\n", self, self->v__pixmap); }.
]

Pixmap _newWindow_: _window w_: _w h_: _h d_: _d
{
    return (oop)newCPixmap((CWindow *)v__window, (int)v__w, (int)v__h, (int)v__d, 0);
}

Pixmap withWindow: window width: w height: h depth: d bits_: _b
[
    self := self new.
    _pixmap := self _newWindow_: window _window w_: w _integerValue h_: h _integerValue d_: d _integerValue bits_: _b.
    { printf("new Pixmap %p -> %p\n", self, self->v__pixmap); }.
]

Pixmap _newWindow_: _window w_: _w h_: _h d_: _d bits_: _b
{
    printf("CPixmap %d %d %d %p\n", (int)v__w, (int)v__h, (int)v__d, (void *)v__b);
    return (oop)newCPixmap((CWindow *)v__window, (int)v__w, (int)v__h, (int)v__d, (void *)v__b);
}

Pixmap shortAt: index put: pixel
[
    self shortAt_: index _integerValue put_: pixel _integerValue
]

Pixmap shortAt_: _index put_: _pixel
{
    CPixmap *pix= (CPixmap *)self->v__pixmap;
    unsigned short *bits= (unsigned short *)pix->bits;
    bits[(int)v__index - 1]= (unsigned short)(long)v__pixel;
}

Pixmap displayOn: aWindow x: x y: y
[
    self blitWindow_: aWindow _window x_: x _integerValue y_: y _integerValue
]

Pixmap blitWindow_: _window x_: _x y_: _y
{
    CPixmap *pix= (CPixmap *)self->v__pixmap;
    blitCWindow((CWindow *)v__window, (int)v__x, (int)v__y, pix, 0, 0, pix->width, pix->height);
}

Pixmap x: x y: y w: w h: h displayOn: aWindow x: dx y: dy
[
    self x_: x _integerValue y_: y _integerValue w_: w _integerValue h_: h _integerValue
	 blitWindow_: aWindow _window x_: dx _integerValue y_: dy _integerValue
]

Pixmap x_: _x y_: _y w_: _w h_: _h blitWindow_: _window x_: _dx y_: _dy
{
    CPixmap *pix= (CPixmap *)self->v__pixmap;
    blitCWindow((CWindow *)v__window, (int)v__dx, (int)v__dy, pix, (int)v__x, (int)v__y, (int)v__w, (int)v__h);
}

Pixmap  width	{ return (oop)((((CPixmap *)self->v__pixmap)->width)  << 1 | 1); }
Pixmap  height	{ return (oop)((((CPixmap *)self->v__pixmap)->height) << 1 | 1); }
Pixmap  depth	{ return (oop)((((CPixmap *)self->v__pixmap)->depth)  << 1 | 1); }
Pixmap  _bits	{ return (oop)((((CPixmap *)self->v__pixmap)->bits)           ); }

Pixmap close
{
    printf("close Pixmap %p -> %p\n", self, self->v__pixmap);
    deleteCPixmap((CPixmap *)(self->v__pixmap));
}

Event  typeOf_: _evt	[ ^SmallInteger value_: (self _typeOf_: _evt) ]
Event _typeOf_: _evt	{  return (oop)(((CWindowEvent *)v__evt)->type); }
