{ import: Object }

BitSet : ByteArray ( msb )

BitSet new: capacity
[
    self := super new: capacity + 7 // 8.
    msb := capacity - 1.
    self clearAll
]

BitSet copy
[
    ^(super new: size) replaceFrom: 0 to: size - 1 with: self startingAt: 0.
]

BitSet clearAll	[ self atAllPut:   0 ]
BitSet setAll	[ self atAllPut: 255 ]

BitSet set:   firstIndex through: lastIndex	[ firstIndex to: lastIndex do: [:i | self set:   i]]
BitSet clear: firstIndex through: lastIndex	[ firstIndex to: lastIndex do: [:i | self clear: i]]

BitSet set: bitIndex
[
    | byteIndex |
    byteIndex := bitIndex // 8.
    bitIndex  := bitIndex \\ 8.
    self at: byteIndex put: ((self at: byteIndex) bitOr: (1 << bitIndex)).
]

BitSet setAll: aBitSet	[ self with: aBitSet replace: [:b1 :b2 | b1 bitOr: b2] ]

BitSet clear: bitIndex
[
    | byteIndex |
    byteIndex := bitIndex // 8.
    bitIndex  := bitIndex \\ 8.
    self at: byteIndex put: ((self at: byteIndex) bitAnd: (255 bitXor: 1 << bitIndex)).
]

BitSet clearAll: aBitSet	[ self with: aBitSet replace: [:b1 :b2 | b1 bitAnd: (255 bitXor: b2)] ]

BitSet test: bitIndex
[
    | byteIndex |
    byteIndex := bitIndex // 8.
    bitIndex  := bitIndex \\ 8.
    ^((self at: byteIndex) bitAnd: 1 << bitIndex) ~~ 0.
]

BitSet testAny: aBitSet
[
    self with: aBitSet do: [:a :b | (a bitAnd: b) ~~ 0 ifTrue: [^true]].
    ^false
]

BitSet testAll: aBitSet
[
    self with: aBitSet do: [:a :b | (a == b) ifFalse: [^false]].
    ^true
]

BitSet first	[ 0   to:     msb do: [:i | (self test: i) ifTrue: [^i]].  ^nil ]
BitSet last	[ msb downTo: 0   do: [:i | (self test: i) ifTrue: [^i]].  ^nil ]

BitSet removeFirst	[ | element | (element := self first) notNil ifTrue: [self clear: element].  ^element ]
BitSet removeLast	[ | element | (element := self last ) notNil ifTrue: [self clear: element].  ^element ]

BitSet printContentsOn: aStream
[
    aStream nextPut: $(.
    self reverseDo: [:b | aStream nextPut: (b // 16) asDigit; nextPut: (b \\ 16) asDigit].
    aStream nextPut: $).
]
