{ import: GrammarProduction }

ReductionGrammar : Object ( startSets listing )

ReductionGrammar new
[
    self := super new.
    startSets := IdentityDictionary new.
    listing := StdOut.
]

ReductionGrammar listing: l	[ listing := l ]

ReductionGrammar at: startSymbol
[
    ^startSets at: startSymbol ifAbsent: [startSets at: startSymbol put: OrderedCollection new]
]

ReductionGrammar at: startSymbol add: pattern
[
    self at: startSymbol add: pattern do: [:x|]
]

ReductionGrammar at: startSymbol add: pattern do: actionBlock
[
    self at: startSymbol add: pattern if: [:x | true] do: actionBlock
]

ReductionGrammar at: startSymbol add: pattern if: predicateBlock do: actionBlock
[
    (self at: startSymbol)
	addFirst: (GrammarProduction withSymbol: startSymbol
				     pattern: pattern
				     predicate: predicateBlock
				     action: actionBlock).
]

ReductionGrammar reduce: aTree to: startSymbol
[
    | startSet |
    ""listing nextPutAll: '  reduce: '; print: aTree; nextPutAll: ' to: '; print: startSymbol; cr.""
    aTree isArray ifFalse: [self error: 'ReductionGrammar cannot reduce a non-tree'].
    aTree first == #par ifTrue: [^self reducePar: aTree to: startSymbol].
    (aTree first beginsWith: 'call') ifTrue: [self reduceArguments: aTree].
    startSet := startSets at: startSymbol ifAbsent: [self error: 'start symbol #', startSymbol asString, ' not in rule set'].
    startSet do: [:rule |
	(self match: aTree with: rule pattern)
	    ifTrue:
	       [listing print: aTree; nextPutAll: ' --> '; print: rule pattern; nextPutAll: ' -> '; print: rule symbol; cr.
		rule action value: aTree.
		^startSymbol]].
    ^self cannotReduce: aTree to: startSymbol
]

ReductionGrammar reducePar: aTree to: startSymbol
[
    ""listing nextPutAll: '  reducePar: '; print: aTree; nextPutAll: ' to: '; print: startSymbol; cr.""
    aTree from: 1 do: [:e | self reduce: e to: (e type isVoidType ifTrue: [#stmt] ifFalse: [startSymbol])].
    ^startSymbol
]

ReductionGrammar reduceArguments: aTree
[
    ""listing nextPutAll: '  reduceCall: '; print: aTree; cr.""
    aTree from: 2 do: [:e | self reduce: e to: #stmt].
]

ReductionGrammar match: aList with: pattern
[
    ""listing nextPutAll: '    ? '; print: aList; nextPutAll: ' = '; print: pattern; cr.""
    pattern isSymbol ifTrue: [^(self reduce: aList to: pattern) notNil].
    aList put: #gen value: nil.
    aList first == pattern first ifFalse: [^false].
    1 to: pattern size - 1 do: [:index |
	(self match: (aList at: index) with: (pattern at: index))
	    ifFalse: [listing nextPutAll: 'fail\n'.  ^false]].
    ^true
]

ReductionGrammar cannotReduce: aTree to: startSymbol
[
    StdErr nextPutAll: '\ncannot reduce: '; print: aTree; nextPutAll: ' to: '; print: startSymbol; cr; abort.
]

ReductionGrammar printOn: aStream
[
    super printOn: aStream.
    aStream nextPut: $(.
    startSets do: [:startSet |
	aStream cr; nextPutAll: '  '; print: startSet key.
	startSet value do: [:rule | aStream nextPutAll: '\n    ->\t'; print: rule pattern]].
    aStream cr; nextPut: $)
]

