/* asm-common.h -- dynamic assembler support
 *
 * @copyright@
 *
 * Last edited: 2005-10-17 09:08:14 by piumarta on emilia.local
 */

#ifndef __ccg_asm_common_h
#define __ccg_asm_common_h_

#ifdef EMBEDDED
# include "osdep/stdio.h"
# include "osdep/stdlib.h"
#else
# include <stdio.h>
# include <stdlib.h>
# define sys_fprintf fprintf
# define sys_abort   abort
#endif

#define ASMWARN(MSG) asmWarn(MSG, __FILE__, __LINE__, __FUNCTION__)
#define ASMFAIL(MSG) asmFail(MSG, __FILE__, __LINE__, __FUNCTION__)

static int asmWarn(const char *msg, const char *file, int line, const char *function)
{
  sys_fprintf(stderr, "%s: In function `%s':\n", file, function);
  sys_fprintf(stderr, "%s:%d: %s\n", file, line, msg);
  return 0;
}

static int asmFail(const char *msg, const char *file, int line, const char *function)
{
  asmWarn(msg, file, line, function);
  sys_abort();
  return 0;
}

#include "asm-cache.h"

#ifndef _ASM_LOCALPC
  static int   asm_pass;	/* 0 (single pass) or 1/2 (two pass) */
  static insn *asm_pc;
#endif

#ifndef  _ASM_APP_1
# define _ASM_APP_1	{ asm_pass= 0; {
#endif
#ifndef  _ASM_NOAPP_1
# define _ASM_NOAPP_1	}}
#endif
#ifndef  _ASM_APP_2
# define _ASM_APP_2	{ insn *asm_lwm= asm_pc; for (asm_pass= 1; ((asm_pass < 3) && ((asm_pc= asm_lwm), 1)); ++asm_pass) {
#endif
#ifndef  _ASM_NOAPP_2
# define _ASM_NOAPP_2	}}
#endif

#ifndef  _ASM_ORG
# define _ASM_ORG(O)	(asm_pc= (insn *)(O))
#endif
#ifndef  _ASM_LBL
# define _ASM_LBL(V)	static insn *V= 0
#endif
#ifndef  _ASM_DEF
# define _ASM_DEF(V)	(V= (((asm_pass==2)&&(asm_pc!=(V))) ? (insn *)(long)ASMFAIL("phase error") : asm_pc))
#endif
#ifndef  _ASM_ORG_
# define _ASM_ORG_(O)	_ASM_ORG(O)
#endif

/* integer predicates */

#define _MASK(N)	((u_int32_t)((1<<(N)))-1)
#define _siP(N,I)	((((int32_t)(I)) >= (-1<<((N)-1))) && (((int)(I)) < (1<<((N)-1))))
#define _uiP(N,I)	(((u_int32_t)(I)) < (1UL<<(N)))

#endif /* __ccg_asm_common_h */
