{ import: Object }
{ import: Expression }
{ import: PowerPC }

"----------------------------------------------------------------"

CokeTransform : Object ()

CokeTransform begin :expr
[
    1 to: expr size - 1 do: [:i | expr at: i put: (self transform: (expr at: i))].
    ^expr
]

CokeTransform note: name
[
    StdOut nextPutAll: self debugName; nextPutAll: ' defaulting '; nextPutAll: name; cr.
]

CokeTransform cnsti4	:expr	[ "self note: 'cnsti4'."   ^expr ]
CokeTransform cnstp4	:expr	[ "self note: 'cnstp4'."   ^expr ]
CokeTransform addrgp4	:expr	[ "self note: 'addrgp4'."  ^expr ]
CokeTransform addi4	:expr	[ "self note: 'addi4'."    ^self begin :expr ]
CokeTransform calli4	:expr	[ "self note: 'calli4'."   ^self begin :expr ]
CokeTransform enteri4	:expr	[ "self note: 'enteri4'."  ^expr ]
CokeTransform reti4	:expr	[ "self note: 'reti4'."    ^self begin :expr ]

"----------------------------------------------------------------"

CokeSymbolTable := [ IdentityDictionary new ]

CokeCompiler : CokeTransform ()

CokeCompiler new
[
    self := super new.
]

CokeCompiler cons:  head :cdar			[ ^Expression with: head with: cdar ]
CokeCompiler cons2: head :cdar :cddar		[ ^Expression with: head with: cdar with: cddar ]

CokeCompiler transform: expr			[ ^expr compileIn: self ]

CokeCompiler define :expr
[
    | name storage |
    storage := CokeSymbolTable at: name ifAbsent: [CokeSymbolTable at: name put: (ByteArray new: 4)].
    ^self cons2: #asgni4 :(self cons: #cnstp4 :storage) :(expr third)
]

"----------------------------------------------------------------"

CokeGenerator : CokeTransform ( asm )

CokeGenerator new
[
    self := super new.
    asm := PowerPC new.
]

CokeGenerator execute	[ ^asm execute ]

CokeEncoder : CokeTransform ( generator depth )

CokeEncoder withGenerator: aGenerator
[
    self := self new.
    generator := aGenerator.
    depth := 0.
]

Object evaluate
[
    ^SmallInteger value_: self compile execute.
]

Object compile
[
    | generator |
    generator := CokeGenerator new.
    self compileIn:  CokeCompiler new.				"StdOut print: self; cr."
    self encodeIn:  (CokeEncoder withGenerator: generator).	"StdOut print: self; cr."
    self generateIn: generator.					"StdOut print: self; cr."
    ^generator.
]

Object compileIn:  aCompiler	[ self shouldNotImplement: #compileIn:  ]
Object encodeIn:   anEncoder	[ self shouldNotImplement: #encodeIn:   ]
Object generateIn: aGenerator	[ self shouldNotImplement: #generateIn: ]

Expression compileIn:  aCompiler	[ ^aCompiler  apply: self first to: self ]
Expression encodeIn:   anEncoder	[ ^anEncoder  apply: self first to: self ]
Expression generateIn: aGenerator	[ ^aGenerator apply: self first to: self ]

CokeCompiler  apply: aSelector to: args	[ ^self perform: aSelector with: args ]
CokeEncoder   apply: aSelector to: args	[ ^self perform: aSelector with: args ]
CokeGenerator apply: aSelector to: args	[ ^self perform: aSelector with: args ]

"----------------------------------------------------------------"

CokeEncoder _lookupSymbol_: _sym
{
    void *addr= dlsym(RTLD_DEFAULT, (char *)v__sym);
    if (!addr) fprintf(stderr, "%s: %s\n", (char *)v__sym, dlerror());
    return (oop)addr;
}

CokeEncoder transform: expr	[ ^expr encodeIn: self ]

CokeEncoder pushLocation	[ ^depth := depth + 1 ]
CokeEncoder popLocation		[ ^depth := depth - 1 ]
CokeEncoder popLocations: n	[ ^depth := depth - n ]

CokeEncoder cnsti4 :expr	[ expr put: #location value: self pushLocation.  ^expr ]
CokeEncoder cnstp4 :expr	[ expr put: #location value: self pushLocation.  ^expr ]

CokeEncoder addrgp4 :expr
[
    | _addr |
    _addr := self _lookupSymbol_: expr second _cString.
    _addr == nil ifTrue: [StdErr nextPutAll: 'undefined global symbol: '; print: expr second; cr; abort].
    expr
	first:  #cnstp4;
	second: (ByteArray size_: 4 _integerValue value_: _addr);
	put:    #location value: self pushLocation.
    ^expr
]

CokeEncoder asgni4 :expr
[
    expr third:  (expr third  encodeIn: self).
    expr second: (expr second encodeIn: self).
    self popLocation.
    expr put: #location value: (expr third get: #location).
    ^expr
]

CokeEncoder addi4 :expr
[
    expr second: (expr second encodeIn: self).
    expr third:  (expr third  encodeIn: self).
    self popLocations: 2.
    expr put: #location value: self pushLocation.
    ^expr
]

CokeEncoder calli4 :expr
[
    | indir |
    expr size - 1 downTo: 1 do: [:i | expr at: i put: ((expr at: i) encodeIn: self)].
    expr second: (expr second encodeIn: self).
    self popLocations: expr size - 1.
    expr put: #location value: self pushLocation.
    ^expr
]

CokeEncoder enteri4 :expr
[
    ^expr
]

CokeEncoder reti4 :expr
[
    expr second: (expr second encodeIn: self).
    self popLocation.
    ^expr
]

"----------------------------------------------------------------"

CokeGenerator hi_: _ptr		{ return (oop)((((unsigned long)v__ptr) >> 16   ) << 1 | 1); }
CokeGenerator lo_: _ptr		{ return (oop)((((unsigned long)v__ptr) & 0xffff) << 1 | 1); }

CokeGenerator offsetOf: expr	[ ^(expr get: #location) * 4 + 8 ]

CokeGenerator transform: expr	[ ^expr generateIn: self ]

CokeGenerator cnstp4 :expr
[
    asm note: 'CNSTP4' print: (SmallInteger value_: expr second);
	lis	:11     :(self hi_: expr second _elements);
	ori	:11 :11 :(self lo_: expr second _elements);
	stw	:11 :(self offsetOf: expr) :1.
]

CokeGenerator cnsti4 :expr
[
    | hi lo |
    asm note: 'CNSTI4' print: expr second.
    hi := expr second >> 16.
    lo := expr second bitAnd: 0xffff.
    hi == 0
	ifTrue:  [asm li  :11 :lo]
	ifFalse: [asm lis :11 :hi;  ori :11 :11 :lo].
    asm stw :11 :(self offsetOf: expr) :1.
    ^expr
]

CokeGenerator addrgp4 :expr
[
    self error: 'global symbol not encoded'
]

CokeGenerator asgni4 :expr
[
    expr third  generateIn: self.
    expr second generateIn: self.
    asm note: 'ASGNI4';
	lwz	:11 :(self offsetOf: expr third ) :1;
	lwz	:12 :(self offsetOf: expr second) :1;
	stw	:11 :0 :12.
    ^expr
]

CokeGenerator addi4 :expr
[
    expr second generateIn: self.
    expr third  generateIn: self.
    asm note: 'ADDI4';
	lwz	:11 :(self offsetOf: expr second) :1;
	lwz	:12 :(self offsetOf: expr third ) :1;
	add	:11 :11 :12;
	stw	:11 :(self offsetOf: expr) :1.
    ^expr
]

CokeGenerator calli4 :expr
[
    expr size - 1 downTo: 1 do: [:i | (expr at: i) generateIn: self].
    asm note: 'CALLI4';
	lwz	:11 :(self offsetOf: expr second) :1;
	mtctr	:11.
    2 to: expr size - 1 do: [:i | asm lwz :(1 + i) :(self offsetOf: (expr at: i)) :1].
    asm bctrl;
	stw :3 :(self offsetOf: expr) :1.
    ^expr
]

CokeGenerator enteri4 :expr
[
    asm note: 'ENTERI4';
	mflr	:0;
	stw	:0 :8 :1;
	stwu	:1 :-96 :1.
    ^expr
]

CokeGenerator reti4 :expr
[
    expr second generateIn: self.
    asm note: 'RETI4';
	lwz	:3 :(self offsetOf: expr second) :1;
	addi	:1 :1 :96;
	lwz	:0 :8 :1;
	mtlr	:0;
	blr.
    ^expr
]
