{ import: st80 }
{ import: File }
{ import: FileStatus }
{ import: SortedCollection }

PathName : String ( status )

PathName species	[ ^PathName ]

String pathName		[ ^PathName size_: self _size value_: _bytes ]

{ include <libgen.h> }

PathName directoryName		[ ^self value_: (self _dirname_: self _stringValue) ]
PathName _dirname_: _path	{ return (oop)dirname((char *)v__path); }

PathName baseName		[ ^self value_: (self _basename_: self _stringValue) ]
PathName _basename_: _path	{ return (oop)basename((char *)v__path); }

PathName status
[
    status isNil ifTrue: [status := FileStatus forPath: self].
    ^status.
]

PathName reset			[ status := nil ]

PathName exists			[ ^self status notNil ]
PathName fileType		[ ^self status notNil ifTrue: [status type] ]
PathName isDirectory		[ ^self status notNil and: [status isDirectory] ]

PathName pathTo: fileName	[ ^self, '/', fileName ]

{ include <dirent.h> }

PathName directoryContents
[
    | contents _dir _ent |
    contents := SortedCollection new.
    (_dir := self _opendir_: self _stringValue) == nil
	ifFalse: [[(_ent := self _readdir_: _dir) == nil] whileFalse: [contents add: (String value_: _ent)].
		  self closedir_: _dir].
    ^contents asArray.
]

PathName _opendir_: _name	{ return (oop)opendir((char *)v__name); }
PathName closedir_: _dir	{ closedir((DIR *)v__dir); }

PathName _readdir_: _dir
{
    struct dirent *entry= readdir((DIR *)v__dir);
    return entry ? (oop)entry->d_name : 0;
}

PathName removeFile
[
    self removeFileIfError:
	[self isDirectory
	     ifTrue:  [self error: 'removeFile: ', self, ': Is a directory']
	     ifFalse: [Smalltalk osError: self]]
]

PathName removeFileIfError: aBlock
[
    (self isDirectory or: [(self removeFile_: self _stringValue) ~= 0])
	ifTrue: [aBlock value]
]

PathName removeFile_: _path		{ return (oop)(unlink((char *)v__path) << 1 | 1); }


PathName removeDirectory
[
    self removeDirectoryIfError:
	[self isDirectory
	     ifTrue:  [Smalltalk osError: self]
	     ifFalse: [self error: 'removeDirectory: ', self, ': Is not a directory']]
]

PathName removeDirectoryIfError: aBlock
[
    (self isDirectory not or: [(self removeDirectory_: self _stringValue) ~= 0])
	ifTrue: [aBlock value]
]

PathName removeDirectory_: _path	{ return (oop)(rmdir((char *)v__path) << 1 | 1); }


PathName removeRecursively		[ self removeRecursivelyIfError: [self error: 'cannot remove directory: ', self] ]

PathName removeRecursivelyIfError: aBlock
[
    self isDirectory ifFalse: [^self removeFileIfError: aBlock].
    self directoryContents do: [ :name |
	(name ~= '.' and: [name ~= '..'])
	    ifTrue: [(self, '/', name) removeRecursivelyIfError: aBlock]].
    self removeDirectoryIfError: aBlock.
]


PathName rename: newName	[ ^0 ==  (self rename_: self _stringValue _: newName _stringValue) ]
PathName rename_: _old _: _new	{ return (oop)(rename((char *)v__old, (char *)v__new) << 1 | 1); }

PathName createForWriting	[ ^File createForWriting: self ]
PathName openForReading		[ ^File openForReading: self ]

DataBuffer : ByteArray ()	[ self := self new: 8*1024*1024 ]

PathName copySafelyFrom: sourceName
[
    | sourceFile destinationFile tempName tempFile |
    tempName        := self, '.+'.
    tempName exists ifTrue: [tempName delete].
    (             (tempFile   := tempName createForWriting) notNil
	    and: [(sourceFile := sourceName openForReading) notNil
	    and: [(tempFile copyDataFrom: sourceFile withBuffer: DataBuffer)]])
	ifTrue:
	   [tempFile close.
	    sourceFile close.
	    sourceName := tempName.
	    (tempName := self, '.-') exists ifTrue: [tempName delete].
	    self exists ifTrue: [self rename: tempName].
	    (sourceName rename: self)
		ifTrue:
		   [tempName exists ifTrue: [tempName delete].
		    self reset.
		    ^true]].
    tempFile isNil ifFalse: [tempFile close].
    sourceFile isNil ifFalse: [sourceFile close].
    ^false
]

PathName copyStatusFrom: sourceName
[
    self println.
    self exists println.
    sourceName exists println.
    ^(self exists and: [sourceName exists and: [sourceName status writeStatus: self]])
]



FileHierarchy : Object ( root files )

FileHierarchy root	[ ^root ]
FileHierarchy files	[ ^files ]

FileHierarchy withRoot: rootPath
[
    rootPath exists ifFalse: [self error: '\'', rootPath, '\' not found'].
    rootPath isDirectory ifFalse: [self error: '\'', rootPath, '\' is not a directory'].
    self := self new.
    root := rootPath.
]

FileHierarchy initialize
[
    super initialize.
    files := OrderedCollection new.
]

FileHierarchy addEntry: file
[
    | path fileType |
    path := root pathTo: file.
    fileType := path fileType.
    fileType isNil ifTrue: [^('cannot read: ', path) putln].
    fileType == #unknown ifTrue: [^('ignoring: ', path) putln].
    (files size \\ 1"000" = 0) ifTrue: [(files size printString, '\t', root, ' @ ', file) putln].
    files add: file.
    fileType == #directory ifTrue: [self addDirectory: file].
]

FileHierarchy addDirectory: directory
[
    | path contents |
    path := root pathTo: directory.
    contents := path directoryContents.
    contents do: [:file | (file ~= '.' and: [file ~= '..']) ifTrue: [self addEntry: (directory pathTo: file)]]
]

FileHierarchy size	[ ^files size ]

FileHierarchy next	[ ^files isEmpty ifFalse: [files removeFirst] ]


FileHierarchy copyIfChanged: aFile from: sourceHierarchy
[
    ('? ', (sourceHierarchy root, ' @ ', aFile), ' -> ', (root, ' @ ', aFile)) putln.
]

FileHierarchy copy: aFile from: sourceHierarchy
[
    | sourceName destinationName |
    sourceName := sourceHierarchy root pathTo: aFile.
    destinationName := root pathTo: aFile.
    ('> ', sourceName, ' -> ', destinationName) putln.
    ((destinationName copySafelyFrom: sourceName) and: [destinationName copyStatusFrom: sourceName])
	ifFalse: ['copy failed' putln].
]

FileHierarchy delete: aFile
[
    ('- ', (root, ' @ ', aFile)) putln.
]



xcp : Object (
    sourceFiles		"FileHierarchy"
    destinationFiles	"FileHierarchy"
    sourceFile		"PathName at current source position"
    destinationFile	"PathName at current destination position"
)

xcp main
[
    | args |
    'temp' pathName removeRecursively.
    Smalltalk quit.
    args := Smalltalk arguments.
    args size = 2 ifFalse: [^self usage].
    self new copyAll: args first pathName to: args second pathName.
]

xcp usage
[
    'usage: xcp source destination' putln.
]

xcp copyAll: sourcePath to: destinationPath
[
    sourceFiles      := FileHierarchy withRoot: sourcePath.
    destinationFiles := FileHierarchy withRoot: destinationPath.
    sourceFiles      addEntry: '.' pathName.  (sourceFiles      size printString, ' files in ', sourcePath) putln.
    destinationFiles addEntry: '.' pathName.  (destinationFiles size printString, ' files in ', destinationPath) putln.
    sourceFile := sourceFiles next.
    destinationFile := destinationFiles next.
    "traverse source and destination hierarchies until one off them runs out "
    [sourceFile notNil and: [destinationFile notNil]]	whileTrue: [self compareSourceAndDestination].
    [sourceFile notNil                              ]	whileTrue: [self copySource].
    [                       destinationFile notNil  ]	whileTrue: [self deleteDestination].
]

xcp compareSourceAndDestination
[
    sourceFile < destinationFile ifTrue: [^self copySource].
    sourceFile > destinationFile ifTrue: [^self deleteDestination].
    "(sourceFile = destinationFile) assert"
    destinationFiles copyIfChanged: sourceFile from: sourceFiles.
    sourceFile := sourceFiles next.
    destinationFile := destinationFiles next.
]

xcp copySource
[
    destinationFiles copy: sourceFile from: sourceFiles.
    sourceFile := sourceFiles next.
]

xcp deleteDestination
[
    destinationFiles delete: destinationFile.
    destinationFile := destinationFiles next.
]

xcp updateStatus: src to: dst
[
    (' > ', src, ' ', dst) putln halt.
]



[ xcp main ]
